/*****************************************************************************
**
**      Filename:   nivxint.c
**
**      Purpose:    The package emulates the NiVXI functionalities
**                  used by the SCTDAQ software.
**                  
**      Version 2.1                   20/04/2001   
**
**      Authors:    P.Bruckman,  J.Olszowska,  A.Spyrzewska   
**                        INP  Cracow 
**
*****************************************************************************/


#include        <stdio.h>
#include        <stdlib.h>
#include        <string.h>
#include        <ctype.h>
#include        "btapi.h"
#include        "sct_hardware.h"
#include        "nivxi.h"

#define DEF_UNIT        0    /* default unit number */
#define SLOG_BUFFER     SLOG_SPACING
#define MUS_BUFFER      MUS_SPACING
#define CLOAC_BUFFER    0x100
#define BILED_BUFFER    BILED_SPACING
#define SCTHV_BUFFER    SCTHV_SPACING
#define SCTLV_BUFFER    0xf0000


//#include        <basetsd.h>
//typedef    short   NIVXI_STATUS;    // try to mimick the NI ret type...

   /* Define global variables to hold info about opened devices */
int            windows[SCT_MAX_MODULES+1];
bt_desc_t      descriptors[SCT_MAX_MODULES+1];
unsigned long  addresses[SCT_MAX_MODULES+1];
void          *pointers[SCT_MAX_MODULES+1];
int            buffers[SCT_MAX_MODULES+1];
int            n_modules = 0;
unsigned long  def_buffer_length = 0;

/*****************************************************************************
**
**      Function    MapVXIAddress
**
**      Purpose:    Maps NiVXI MapVXIAddress functionality on the BIT3
**                  API. 
**                  
**
**
*****************************************************************************/
void* MapVXIAddress(     UINT32 accessparms,
                         UINT32 address,
                         INT32  timo,
                         UINT32 *window,
                         short  *ret)
{

    int             unit = DEF_UNIT;          /* Which board to use */
    char            devname[BT_MAX_DEV_NAME]; /* Device name string */
    bt_error_t      status;                   /* Mirror API return value */
    bt_dev_t        type = BT_DEV_A24;        /* device type for bt_open */
    unsigned long   remote_address;           /* copy of the remote address */
    int             length;                   /* Device buffer length */
    void            *remote_p = NULL;         /* holds pointer to a memory region */
//    volatile char   *addr_p = NULL;           /* current location being accessed */
    int             address_mod;              /* Address Modifier for mmap(...) */
    static int             n_slog=0, n_mus=0, n_cloac=0, n_biled=0;
	static int             n_sctlv=0, n_scthv;
	static unsigned long   new_slog =SLOG_ADDRESS;
	static unsigned long   new_mus  =MUS_ADDRESS;
	static unsigned long   new_cloac=CLOAC_ADDRESS;
	static unsigned long   new_biled=BILED_ADDRESS;
	static unsigned long   new_scthv=SCTHV_ADDRESS;
	static unsigned long   new_sctlv=SCTLV_CRATE_ADDRESS;

    *ret = 0;

	/* Map the necessary parameters */

	/* VMI base address */
    remote_address = address;
    /* Now identify the module according to its base address,
	   this compared to the one declared in sct_hardware.h */
    if      ( remote_address == new_slog ) {
		n_slog++;
        new_slog += SLOG_SPACING;
		if (n_slog > SLOG_NUMBER) {
          printf( "Declared number of SLOG modules exceeded: %d\n", n_slog);
		  *ret = -3;
		  return NULL;
		}
		length = SLOG_BUFFER;
	}
	else if ( remote_address == new_mus ) {
		n_mus++;
        new_mus += MUS_SPACING;
		if (n_mus > MUS_NUMBER) {
          printf( "Declared number of MuSTARD modules exceeded: %d\n", n_mus);
		  *ret = -3;
		  return NULL;
		}
		length = MUS_BUFFER;
	}
	else if ( remote_address == new_cloac ) {
		n_cloac++;
//        new_cloac += CLOAC_SPACING;
//		if (n_cloac > CLOAC_NUMBER) {
//          printf( "Declared number of CLOAC modules exceeded: %d\n", n_cloac);
//		  *ret = -3;
//		  return NULL;
//		};
		length = CLOAC_BUFFER;
	}
	else if ( remote_address == new_biled ) {
		n_biled++;
        new_biled += BILED_SPACING;
		if (n_biled > BILED_NUMBER) {
          printf( "Declared number of BILED modules exceeded: %d\n", n_biled);
		  *ret = -3;
		  return NULL;
		};
		length = BILED_BUFFER;
	}
	else if ( remote_address == new_scthv ) {
		n_scthv++;
        new_scthv += SCTHV_SPACING;
		if (n_scthv > SCTHV_NUMBER) {
          printf( "Declared number of SCTHV modules exceeded: %d\n", n_scthv);
		  *ret = -3;
	  	  return NULL;
		};
		length = SCTHV_BUFFER;
	}
	else if ( remote_address == new_sctlv ) {
		n_sctlv++;
        new_sctlv += SCTLV_SPACING;
		if (n_sctlv > SCTLV_NUMBER) {
          printf( "Declared number of SCTLV modules exceeded: %d\n", n_sctlv);
  		*ret = -3;
	    	return NULL;
		};
		length = SCTLV_BUFFER;
	}
	else {
        printf( "Unrecognized module base address: %x\n", remote_address);
		*ret = -3;
		return NULL;
	}


	/* decide what to use as the actual buffer length for mmap */
#ifdef MAPSIZE
	length = def_buffer_length;
#endif


	/* Address Modifier */
    switch (accessparms) {

	case 1:           /* A16 Non priv data */
		address_mod = 0x29;
        type = BT_DEV_A16;     /* in practice dummy par. */
		break;

	case 2:           /* A24 non_priv data */
		address_mod = 0x39;
        type = BT_DEV_A24;     /* in practice dummy par. */
		break;

	case 3:           /* A32 non_priv data */
		address_mod = 0x09;
        type = BT_DEV_A32;     /* in practice dummy par. */
		break;

	case 5:           /* A16 Supervisory data */
		address_mod = 0x2D;
        type = BT_DEV_A16;     /* in practice dummy par. */
		break;

	case 6:           /* A24 Supervisory data */
		address_mod = 0x3D;
        type = BT_DEV_A24;     /* in practice dummy par. */
		break;

	case 7:           /* A32 Supervisory data */
		address_mod = 0x0D;
        type = BT_DEV_A32;     /* in practice dummy par. */
		break;

    default :
        printf( "Unsupported data modifier: %s\n", accessparms);
		*ret = -2;
		return NULL;
	};





    /* Open the unit */
	n_modules++;
    status = bt_open(&descriptors[n_modules],
        bt_gen_name(unit, type, &devname[0], BT_MAX_DEV_NAME),
        BT_RD | BT_WR);
    if (BT_SUCCESS != status) {
        bt_perror(descriptors[n_modules], status, "Could not open the device");
		n_modules--;
		*ret = -9;
        return NULL ;
    }

    /* Clear any outstanding errors */
    status = bt_clrerr(descriptors[n_modules]);
    if (BT_SUCCESS != status) {
        bt_perror(descriptors[n_modules], status, "Could not clear errors from the device");
        (void) bt_close(descriptors[n_modules]);
		n_modules--;
		*ret = -9;
        return NULL;
    }

	/* Set specific parameters of device descriptor */
    status = bt_set_info(descriptors[n_modules], BT_INFO_BLOCK, FALSE);
    if (BT_SUCCESS != status) {
        bt_perror(descriptors[n_modules], status, "Could not modify BT_INFO_BLOCK");
        (void) bt_close(descriptors[n_modules]);
		n_modules--;
		*ret = -9;
        return NULL;
    }
    status = bt_set_info(descriptors[n_modules], BT_INFO_MMAP_AMOD, address_mod);
    if (BT_SUCCESS != status) {
        bt_perror(descriptors[n_modules], status, "Could not modify BT_INFO_MMAP_AMOD");
        (void) bt_close(descriptors[n_modules]);
		n_modules--;
		*ret = -9;
        return NULL;
    }
    status = bt_set_info(descriptors[n_modules], BT_INFO_SWAP, BT_SWAP_DEFAULT);
    if (BT_SUCCESS != status) {
        bt_perror(descriptors[n_modules], status, "Could not modify BT_INFO_SWAP");
        (void) bt_close(descriptors[n_modules]);
		n_modules--;
		*ret = -9;
        return NULL;
    }

    /* Do the actual mmap */
    status = bt_mmap(descriptors[n_modules], &remote_p, remote_address, length, 
        BT_RD | BT_WR, BT_SWAP_DEFAULT);
    if (BT_SUCCESS != status) {
        bt_perror(descriptors[n_modules], status, "Could not memory map the device");
        (void) bt_close(descriptors[n_modules]);
		n_modules--;
		*ret = -9;
        return NULL;
    }


    /* Lock the adaptor */
#ifdef  BTLOCK
    status = bt_lock(descriptors[n_modules], BT_FOREVER);
    if (BT_SUCCESS != status) {
        bt_perror(descriptors[n_modules], status, "Could not lock the device");
        (void) bt_unmmap(descriptors[n_modules], &remote_p, length);
        (void) bt_close(descriptors[n_modules]);
        n_modules--;
		*ret = -9;
        return NULL;
    }
#endif

     /* if all is Ok return proper values */
	 windows[n_modules] = n_modules;
	 addresses[n_modules] = remote_address;
	 pointers[n_modules] = remote_p;
	 buffers[n_modules] = length;
     *window = n_modules;
     return (void *) remote_p;

}

/*****************************************************************************
**
**      Function    UnMapVXIAddress
**
**      Purpose:    Maps NiVXI UnMapVXIAddress functionality on the BIT3
**                  API. 
**                  
**
**
*****************************************************************************/
int   UnMapVXIAddress(     UINT32 window )
{

    bt_error_t      status;                   /* Mirror API return value */

	if ( (int) window > n_modules) {
        printf( "Invalid window number: %d\n", window);
		return -1;
    }

    /* Unlock the adaptor  */
#ifdef  BTLOCK
    status = bt_unlock(descriptors[window]);
    if (BT_SUCCESS != status) {
        bt_perror(descriptors[window], status, "Could not unlock the device");
        (void) bt_close(descriptors[window]);
        return -9;
    }
#endif

    /* Release the mmap resources */
    status = bt_unmmap(descriptors[window], pointers[window], buffers[window]);
    if (BT_SUCCESS != status) {
        bt_perror(descriptors[window], status, "Could not release the memory map");
        (void) bt_close(descriptors[window]);
        return 1;
    }

    /* Check for Adaptor status errors */
    status = bt_chkerr(descriptors[window]);
    if (BT_SUCCESS != status) {
        bt_perror(descriptors[window], status, "Memory mapped read failed");
        (void) bt_close(descriptors[window]);
        return -9;
    }

    /* Close the device */
    status = bt_close(descriptors[window]);
    if (BT_SUCCESS != status) {
        bt_perror(descriptors[window], status, "Could not close the device");
        return -9;
    }

    return 0 ;
}

/*****************************************************************************
**
**      Function    MapVXIAddressSize
**
**      Purpose:    Maps NiVXI MapVXIAddressSize functionality on the BIT3
**                  API. 
**                  
**
**
*****************************************************************************/
int     MapVXIAddressSize(     UINT32 size )
{
     def_buffer_length = size;
	 return 0;
}




/*****************************************************************************
**
**      Function    GetWindowRange
**
**      Purpose:    Maps NiVXI GetWindowRange functionality on the BIT3
**                  API. 
**                  
**
**
*****************************************************************************/
int   GetWindowRange(      int     window,
					       UINT32 *wstart,
						   UINT32 *wend )
{
	if (window > n_modules) {
        printf( "Invalid window number: %d\n", window);
		return -1;
    }
    
	*wstart = addresses[window];
	*wend   = addresses[window] + buffers[window];

	return 0;
}

/*****************************************************************************
**
**      Function    InitVXIlibrary
**
**      Purpose:    Maps NiVXI InitVXIlibrary functionality on the BIT3
**                  API. 
**                  
**
**
*****************************************************************************/
int     InitVXIlibrary( void )
{
	 return 0;
}

/*****************************************************************************
**
**      Function    CloseVXIlibrary
**
**      Purpose:    Maps NiVXI CloseVXIlibrary functionality on the BIT3
**                  API. 
**                  
**
**
*****************************************************************************/
int     CloseVXIlibrary( void )
{
	 return 0;
}

/*****************************************************************************
**
**      Function    VXIin
**
**      Purpose:    Maps NiVXI VXIin functionality on the BIT3
**                  API.
**                   
**                  
**
**
*****************************************************************************/
NIVXI_STATUS VXIin   (   unsigned int accessparms,
                         UINT32 address,
                         unsigned int accwidth,
                         void   *value )

{

    int             unit = DEF_UNIT;          /* Which board to use */
    char            devname[BT_MAX_DEV_NAME]; /* Device name string */
    bt_error_t      status;                   /* Mirror API return value */
    bt_dev_t        type = BT_DEV_A24;        /* device type for bt_open */
    unsigned long   remote_address;           /* copy of the remote address */
	unsigned        data_depth;               /* copy of the buffer size */
 	unsigned        act_data_depth;           /* buffer size actually written */
    bt_desc_t       descriptor;               /* local device descriptor */
    int             address_mod;              /* Address Modifier for mmap(...) */

	/* Map the necessary parameters */

	/* VMI base address */
    remote_address = address;
	data_depth     = accwidth;


	/* Address Modifier */
    switch (accessparms) {

	case 1:           /* A16 Non priv data */
		address_mod = 0x29;
        type = BT_DEV_A16;     /* in practice dummy par. */
		break;

	case 2:           /* A24 non_priv data */
		address_mod = 0x39;
        type = BT_DEV_A24;     /* in practice dummy par. */
		break;

	case 3:           /* A32 non_priv data */
		address_mod = 0x09;
        type = BT_DEV_A32;     /* in practice dummy par. */
		break;

	case 5:           /* A16 Supervisory data */
		address_mod = 0x2D;
        type = BT_DEV_A16;     /* in practice dummy par. */
		break;

	case 6:           /* A24 Supervisory data */
		address_mod = 0x3D;
        type = BT_DEV_A24;     /* in practice dummy par. */
		break;

	case 7:           /* A32 Supervisory data */
		address_mod = 0x0D;
        type = BT_DEV_A32;     /* in practice dummy par. */
		break;

    default :
        printf( "Unsupported data modifier: %s\n", accessparms);
		return -2;
	};


    /* Open the unit */
    status = bt_open(&descriptor,
        bt_gen_name(unit, type, &devname[0], BT_MAX_DEV_NAME),
        BT_RD);
    if (BT_SUCCESS != status) {
        bt_perror(descriptor, status, "Could not open the device");
        return -9;
    }

    /* Clear any outstanding errors */
    status = bt_clrerr(descriptor);
    if (BT_SUCCESS != status) {
        bt_perror(descriptor, status, "Could not clear errors from the device");
        (void) bt_close(descriptor);
        return -9;
    }

	/* Set specific parameters of device descriptor */
    status = bt_set_info(descriptor, BT_INFO_BLOCK, FALSE);
    if (BT_SUCCESS != status) {
        bt_perror(descriptor, status, "Could not modify BT_INFO_BLOCK");
        (void) bt_close(descriptor);
        return -9;
    }
    status = bt_set_info(descriptor, BT_INFO_DMA_THRESHOLD, 0x1000);
    if (BT_SUCCESS != status) {
        bt_perror(descriptor, status, "Could not modify BT_INFO_DMA_THRESHOLD");
        (void) bt_close(descriptor);
        return -9;
    }
    status = bt_set_info(descriptor, BT_INFO_PIO_AMOD, address_mod);
    if (BT_SUCCESS != status) {
        bt_perror(descriptor, status, "Could not modify BT_INFO_PIO_AMOD");
        (void) bt_close(descriptor);
        return -9;
    }
    status = bt_set_info(descriptor, BT_INFO_SWAP, BT_SWAP_DEFAULT);
    if (BT_SUCCESS != status) {
        bt_perror(descriptor, status, "Could not modify BT_INFO_SWAP");
        (void) bt_close(descriptor);
        return -9;
    }

    /* Do the actual read */
    status = bt_read(descriptor, &value, remote_address,
		     data_depth, &act_data_depth );
    if (BT_SUCCESS != status) {
        bt_perror(descriptor, status, "Could not read from the device");
        (void) bt_close(descriptor);
        return -9;
    }


    /* Close the device */
    status = bt_close(descriptor);
    if (BT_SUCCESS != status) {
        bt_perror(descriptor, status, "Could not close the device");
        return -9;
    }


     /* if all is Ok return zero */
     return 0;

}
/*****************************************************************************
**
**      Function    VXIout
**
**      Purpose:    Maps NiVXI VXIout functionality on the BIT3
**                  API.
**                   
**                  
**
**
*****************************************************************************/
NIVXI_STATUS VXIout  (   unsigned int accessparms,
                         UINT32 address,
                         unsigned int accwidth,
                         UINT32 value )

{

    int             unit = DEF_UNIT;          /* Which board to use */
    char            devname[BT_MAX_DEV_NAME]; /* Device name string */
    bt_error_t      status;                   /* Mirror API return value */
    bt_dev_t        type = BT_DEV_A24;        /* device type for bt_open */
    unsigned long   remote_address;           /* copy of the remote address */
	unsigned        data_depth;               /* copy of the buffer size */
 	unsigned        act_data_depth;           /* buffer size actually written */
    bt_desc_t       descriptor;               /* local device descriptor */
    int             address_mod;              /* Address Modifier for mmap(...) */

	/* Map the necessary parameters */

	/* VMI base address */
    remote_address = address;
	data_depth     = accwidth;


	/* Address Modifier */
    switch (accessparms) {

	case 1:           /* A16 Non priv data */
		address_mod = 0x29;
        type = BT_DEV_A16;     /* in practice dummy par. */
		break;

	case 2:           /* A24 non_priv data */
		address_mod = 0x39;
        type = BT_DEV_A24;     /* in practice dummy par. */
		break;

	case 3:           /* A32 non_priv data */
		address_mod = 0x09;
        type = BT_DEV_A32;     /* in practice dummy par. */
		break;

	case 5:           /* A16 Supervisory data */
		address_mod = 0x2D;
        type = BT_DEV_A16;     /* in practice dummy par. */
		break;

	case 6:           /* A24 Supervisory data */
		address_mod = 0x3D;
        type = BT_DEV_A24;     /* in practice dummy par. */
		break;

	case 7:           /* A32 Supervisory data */
		address_mod = 0x0D;
        type = BT_DEV_A32;     /* in practice dummy par. */
		break;

    default :
        printf( "Unsupported data modifier: %s\n", accessparms);
		return -2;
	};


    /* Open the unit */
    status = bt_open(&descriptor,
        bt_gen_name(unit, type, &devname[0], BT_MAX_DEV_NAME),
        BT_WR);
    if (BT_SUCCESS != status) {
        bt_perror(descriptor, status, "Could not open the device");
        return -9;
    }

    /* Clear any outstanding errors */
    status = bt_clrerr(descriptor);
    if (BT_SUCCESS != status) {
        bt_perror(descriptor, status, "Could not clear errors from the device");
        (void) bt_close(descriptor);
        return -9;
    }

	/* Set specific parameters of device descriptor */
    status = bt_set_info(descriptor, BT_INFO_BLOCK, FALSE);
    if (BT_SUCCESS != status) {
        bt_perror(descriptor, status, "Could not modify BT_INFO_BLOCK");
        (void) bt_close(descriptor);
        return -9;
    }
    status = bt_set_info(descriptor, BT_INFO_DMA_THRESHOLD, 0x1000);
    if (BT_SUCCESS != status) {
        bt_perror(descriptor, status, "Could not modify BT_INFO_DMA_THRESHOLD");
        (void) bt_close(descriptor);
        return -9;
    }
    status = bt_set_info(descriptor, BT_INFO_PIO_AMOD, address_mod);
    if (BT_SUCCESS != status) {
        bt_perror(descriptor, status, "Could not modify BT_INFO_PIO_AMOD");
        (void) bt_close(descriptor);
        return -9;
    }
    status = bt_set_info(descriptor, BT_INFO_SWAP, BT_SWAP_DEFAULT);
    if (BT_SUCCESS != status) {
        bt_perror(descriptor, status, "Could not modify BT_INFO_SWAP");
        (void) bt_close(descriptor);
        return -9;
    }

    /* Do the actual write */
    status = bt_write(descriptor, &value, remote_address,
		     data_depth, &act_data_depth );
    if (BT_SUCCESS != status) {
        bt_perror(descriptor, status, "Could not write on the device");
        (void) bt_close(descriptor);
        return -9;
    }


    /* Close the device */
    status = bt_close(descriptor);
    if (BT_SUCCESS != status) {
        bt_perror(descriptor, status, "Could not close the device");
        return -9;
    }


     /* if all is Ok return zero */
     return 0;

}

void NI_pause()  {
//    FUDGE: pause to match PCI->VME timing
      int i, j;
      for (i=0; i<PAUSE_LENGTH; i++) j=i;
}

