/* 
 * sctlv_lib.h 
 *
 * JCH 24.06.02 version 0.54 - changes for RCC
 * PWP 30.08.01 version 0.51 - minor bugfixes
 * PWP 04.07.01 version 0.50 include new register map, limits
 *                           and scales to suit SCTLV3
 *
 * PWP 22.03.01 version 0.42 revisisons to RESTART functionality
 * PWP 17.01.01 version 0.41 include mods for multiple frames
 * PWP 17.01.01 version 0.40 introducing info struct
 *              also some internal reorganisation
 * PWP 15.05.00 add function sctlv_check_trip
 *
 * PWP 08.12.99 version 0.30  include support for VPIN on SCTLV2
 *                            also some corrections to register list
 *
 * PWP 28.05.99 version 0.21 substantial rewrite
 */

#include <time.h>

/*
 * These definitions are used within sctlv_lib
 */

#define VccOut 0 /* SCTLV3 only */
#define Vcc    1
#define Vi1    2 /* SCTLV1 and SCTLV2 only */
#define VddOut 3 /* SCTLV3 only */
#define Vdd    4
#define Vld0   5
#define Vld1   6
#define Vpin   7
#define Icc    8
#define Ivi1   9 /* SCTLV1 and SCTLV2 only */
#define Ipin   9 /* SCTLV3 only */
#define Idd   10
#define Ild0  11 /* SCTLV1 and SCTLV2 only */
#define T0    11 /* SCTLV3 only */
#define Ild1  12 /* SCTLV1 and SCTLV2 only */
#define T1    12 /* SCTLV3 only */

#define LV_NMON 13 /* number of registers for monitoring */
#define LV_NSET 15 /* number of registers for status and control */

/*
 * Here we define the offsets corresponding to each SCTLV register.
 * The list is defined in terms of data type char 
 * (was defined in terms of short before 19.01.01)
 */

/* SCTLV1 and SCTLV2 registers */

#define LV_M_VCC_HI     0x00
#define LV_M_VCC_LO     0x02
#define LV_M_VI1        0x04
#define LV_M_VDD_HI     0x06
#define LV_M_VDD_LO     0x08
#define LV_M_VLD0       0x0A
#define LV_M_VLD1       0x0C
#define LV_M_VPIN       0x0E
#define LV_M_ICC        0x10
#define LV_M_IVI1       0x12
#define LV_M_IDD        0x14
#define LV_M_ILD0       0x16
#define LV_M_ILD1       0x18
#define LV_S_VCC_HI     0x1C
#define LV_S_VCC_LO     0x1E
#define LV_S_VI1        0x20
#define LV_S_VDD_HI     0x22
#define LV_S_VDD_LO     0x24
#define LV_S_VLD0       0x26
#define LV_S_VLD1       0x28
#define LV_S_VPIN       0x1A /* SCTLV2 only */
#define LV_S_ICC        0x2A
#define LV_S_IVI1       0x2C
#define LV_S_IDD        0x2E
#define LV_S_ILD0       0x30
#define LV_S_ILD1       0x32

/* SCTLV3 registers */

#define LV3_M_VCCOUT     0x00
#define LV3_M_VCC        0x02
#define LV3_M_VDDOUT     0x06
#define LV3_M_VDD        0x08
#define LV3_M_VLD0       0x0A
#define LV3_M_VLD1       0x0C
#define LV3_M_VPIN       0x0E
#define LV3_M_ICC        0x10
#define LV3_M_IPIN       0x12
#define LV3_M_IDD        0x14
#define LV3_M_T0         0x16
#define LV3_M_T1         0x18
#define LV3_S_VCC        0x1E
#define LV3_S_VDD        0x24
#define LV3_S_VLD0       0x26
#define LV3_S_VLD1       0x28
#define LV3_S_VPIN       0x1A
#define LV3_S_ICC        0x2A
#define LV3_S_IPIN       0x2C
#define LV3_S_IDD        0x2E
#define LV3_S_TL0        0x30
#define LV3_S_TL1        0x32

/* Registers common to all SCTLV boards */

#define LV_SELECTS      0x34 /* PWRON, SELECT, RESET */
#define LV_FLAG         0x36
#define LV_0STATUS      0x70 
#define LV_0DIAGNOSTIC  0x72
#define LV_1STATUS      0x74 
#define LV_1DIAGNOSTIC  0x76
#define LV_VERSION      0x78 /*
                              * version 3 = SCTLV1 
                              * version 4 = SCTLV1 with upgrade
                              * version 5 = SCTLV2
                              * version 6 = SCTLV3
                              */
#define LV_TEMP         0x7A
#define LV_BOARD_STATUS 0x7C /* PWP 22.03.00        */
#define LV_RESTART      0x7E /* after SJMP 21.03.00 */

#define LV_CHAN_OFFSET  0x38

/*
 * definitions of number of attempts to make whilst setting flags
 * and various software delays, all in mS.  These values OK at RAL.
 */
 
#define LV_NTRIES_FLAG       10   // Was 10 changed to 50 tbMay00 GFM
                                  // and back to 10 PWP 08.02.01
#define LV_NTRIES_RESTART    10
#define LV_WAIT_FLAG         10   // Was 50 changed to 10 PWP 08.02.01
#define LV_WAIT_RAMP        250
#define LV_WAIT_RESET      1000
#define LV_WAIT_RESTART     100


/*
 * Names, scaling factors and range limits for voltage and current channels.
 * Currents are in mA, voltages are in V, temperatures are in degrees Celcius.
 */

/* SCTLV1 and SCTLV2 */

static char vi2_name[LV_NMON][6]={ "    ",     "Vcc ",     "Vi1 ",     "    ",
                                   "Vdd ",     "Vld0",     "Vld1",     "Vpin",
                                   "Icc ",     "Ivi1",     "Idd ",     "Ild0",     "Ild1"};

static char vi2_unit[LV_NMON][3]={ "  ",       " V",       " V",       "  ",
                                   " V",       " V",       " V",       " V", 
                                   "mA",       "mA",       "mA",       "mA",       "mA"};

static float vi2_def[LV_NMON] =  { (float)0.0, (float)3.5, (float)0.0, (float)0.0,
                                   (float)4.0, (float)6.0, (float)6.0, (float)6.0,
                                   (float)1000,(float)8.0, (float)600, (float)8.0, (float)8.0};

static double vi2_scale[LV_NMON]={ 0.000,      0.010,      0.020,      0.000,
                                   0.010,      0.040,      0.040,      0.040,
                                   5.0,        0.1,        5.0,        0.1,        0.1};

static float vi2_min[LV_NMON] =  { (float)0.0, (float)3.2, (float)0.0, (float)0.0,
                                   (float)3.2, (float)0.0, (float)0.0, (float)0.0,
                                   (float)0.0, (float)0.0, (float)0.0, (float)0.0, (float)0.0};

static float vi2_max[LV_NMON] = {  (float)0.0, (float)4.5, (float)4.0, (float)0.0,  
                                   (float)5.5, (float)6.6, (float)6.6, (float)10.0,
                                   (float)1300,(float)8.0, (float)1600,(float)8.0, (float)8.0};


/* SCTLV3 */

static char vi3_name[LV_NMON][7]={ "VccOut",   "Vcc   ",   "      ",   "VddOut",
                                   "Vdd   ",   "Vld0  ",   "Vld1  ",   "Vpin  ",
                                   "Icc   ",   "Ipin  ",   "Idd   ",   "T0    ",   "T1    "};

static char vi3_unit[LV_NMON][3]={ " V",       " V",       "  ",       " V",
                                   " V",       " V",       " V",       " V", 
                                   "mA",       "mA",       "mA",       " C",       " C"};

static float vi3_def[LV_NMON] =  { (float)0.0, (float)3.5, (float)0.0, (float)0.0,
                                   (float)4.0, (float)6.0, (float)6.0, (float)6.0,
                                   (float)1000,(float)8.0, (float)600, (float)60,  (float)60};

static double vi3_scale[LV_NMON]={ 0.040,      0.020,      0.000,      0.040,
                                   0.020,      0.040,      0.040,      0.050,
                                   10.0,       0.010,      10.0,       1.0,        1.0};

static float vi3_min[LV_NMON] =  { (float)0.0, (float)2.5, (float)0.0, (float)0.0,
                                   (float)2.5, (float)0.0, (float)0.0, (float)0.0,
                                   (float)0.0, (float)0.0, (float)0.0, (float)0.0, (float)0.0};

static float vi3_max[LV_NMON] = {  (float)0.0, (float)4.5, (float)0.0, (float)0.0,  
                                   (float)5.5, (float)6.6, (float)6.6, (float)10.0,
                                   (float)1400,(float)8.0, (float)1400,(float)127.0, (float)127.0};



/*
 * And now, the function prototypes....
 * The lowest level routines are first.
 */

int sctlv_vmemap(unsigned long address, short *card);
int sctlv_exit();

char sctlv_rd(short id, unsigned long reg);
void sctlv_wt(short id, unsigned long reg, char dta);

int sctlv_test_rd_flag(short id);
int sctlv_test_wt_flag(short id);

void sctlv_wait_ms(clock_t wait);
int sctlv_init(short id);
int sctlv_board_restart(short id);  /* NB was sctlv_do_board_reset */

/*
 * Get routines
 */

int sctlv_get_power(short id,  short *power);
int sctlv_get_reset(short id,  short *reset);
int sctlv_get_select(short id, short *select);
int sctlv_get_selects(short id, short *power, short *reset, short *select);

int sctlv_get_diagnostic(short id, char *diagnostic);
int sctlv_get_status(short id, char *status);
int sctlv_get_board_status(short id, char *status);
int sctlv_get_t0(short id, char *t0);
int sctlv_get_t1(short id, char *t1);
int sctlv_get_temp(short id, char *temp);
int sctlv_get_temps(short id, char *t0, char *t1);
int sctlv_get_version(short id, char *version);

/*
 * Set routines
 */

int sctlv_set_power(short id, short power);
int sctlv_set_reset(short id, short reset);
int sctlv_set_select(short id, short select);
int sctlv_set_selects(short id, short power, short reset, short select);

int sctlv_set_vcc(short id, float vcc, float icc);
int sctlv_set_vvi1(short id, float vvi1, float ivi1);
int sctlv_set_vdd(short id, float vdd, float idd);
int sctlv_set_vled0(short id, float vled0, float iled0);
int sctlv_set_vled1(short id, float vled1, float iled1);
int sctlv_set_vpin(short id, float vpin);

/* 
 *  Low level routines to get/set control/monitored values 
 */

int sctlv_get_monval(short id, short ch, float *val);
int sctlv_get_setval(short id, short ch, float *val);
int sctlv_setval(short id, short ch, float val);

/* 
 * Detector module reset
 */

int sctlv_do_reset(short id);
int sctlv_do_reset_all();
int sctlv_do_reset_many(short lo, short hi);

/*
 * Higher level routines, mainly for diagnostics
 */

int sctlv_set_defaults(short id);
int sctlv_mon_all(short id, float *mon);
int sctlv_print_status(short id);
int sctlv_brief_status(short id); 
int sctlv_ramp_v(short id, short ch, float start, float stop, float step);
int sctlv_check_HW_on(short id);
int sctlv_check_pwron(short id);
int sctlv_check_trip(short id);

int sctlv_monitor(short id, char* status, char* temp, float * mon);

short sctlv_n_cards();
short sctlv_n_channels();

void sctlv_show_error(int rcode);

/* Useful internal structure for SCTLV info*/

struct sctlv_info_structure
{
  unsigned long address;
#ifdef NIVXI
  unsigned short accessparms;
#endif 
#ifdef RCC
  int window;
#endif
  char version;
  char missing;
};

#define sctlv_info_size sizeof(struct sctlv_info_structure)
