/* 
 * sctlv_lib.c  SCT LV Power Supply
 *
 * Version 0.21 PWP 28.05.99
 * Version 0.22 PWP/JCH 29.07.99 - RAID compatibility
 * Version 0.30 PWP 08.12.99 - updated for SCTLV v5
 *              changes should be backward compatible
 *
 * NB. requires version 0.3 of sctlv_lib.h
 * version 0.40 PWP 30.01.01 - addition of structure to hold SCTLV info;
 *                             and protection against out of range useage;
 *                             addition of code to read module temps t0+t1;
 *                             some internal reorganisation;
 * version 0.41 PWP 22.02.01 - inclusion of mods to facilitate running SCTLV
 *                             across multiple VME frames with NIVXI, in A16 or A24
 * version 0.42 PWP 22.03.01 - bugfixes in sctlv_board_restart, after Simon Peeters
 * version 0.42 PWP 12.06.01 - trivial changes
 */

#include <stdio.h>
#include <time.h>

#ifdef VXINT
#include "../include/nivxi.h"
#include <stdlib.h>
#include <string.h>
#endif

#ifdef OS9
#include <stdlib.h>
#include <string.h>
#include <process.h>
#include <signal.h>
#include <types.h>
#endif

#ifdef RAID
#include <posix/time.h>
#include <stdlib.h>
#include <string.h>
#define CLOCKS_PER_SEC 100000
#endif

#include "sctlv_lib.h"
/* #define LV_DEBUG */


/* Global variables */
short  sctlv_card_total = 0;               /* Total number of SCTLV cards    */
short  sctlv_chan_total = 0;               /* Total number of SCTLV channels */
struct sctlv_info_structure *sctlv_info;   /* structure to hold SCTLV info   */

int do_test_rd_flag = 1;


/*
 * Low level functions
 */

/*
 * VME mapping and low level communications
 */

int sctlv_vmemap(unsigned long address, short* card)
{
#ifdef VXINT
  char dta;
  NIVXI_STATUS ret;
#else
  char *addr;
  int status;
#endif
#ifdef RAID
  int am;
#endif
  

  *card = sctlv_card_total++;
  sctlv_chan_total +=2;

/* Use malloc/realloc to get space for the structure */
  if(sctlv_card_total == 1)
    sctlv_info = (struct sctlv_info_structure *) malloc(sctlv_info_size);
  else
    sctlv_info = (struct sctlv_info_structure *) realloc((void *)sctlv_info,sctlv_card_total*sctlv_info_size);

  sctlv_info[*card].address = 0;
  sctlv_info[*card].version = 0;
  sctlv_info[*card].missing = 0;
  sctlv_info[*card].icc_range = 0;
  sctlv_info[*card].idd_range = 0;
#ifdef VXINT
  
#endif

/*
 * In the case of OS9 or RAID, one window is opened
 * of size 0x8000 into which all cards must be placed
 */
  
#ifdef OS9
  status = _os_permit((void *)address, 0x8000,2,0);
  if(status != 0) {
     printf("VME mapping failed, status = %d\n",status);
     return status;
  }
  sctlv_info[*card].address = address;
  sctlv_init((short)(2*(*card)));
  sctlv_check_current_ranges((short)(2*(*card)));
#endif

#ifdef RAID
  /* Establish from identifier whether whole crate needs mapping */
  am = 0x39; /* A24 standard address mode */
  if((address&0xFF000000) != 0) am=0x09;

    status = RD13_VmeMap(address,am,0x8000,&addr,0);
    if(status != 0) {
       printf("VME mapping failed, status = %d\n",status);
       return status;
    }
  sctlv_info[*card].address = (unsigned long) addr;
  sctlv_init((short)(2*(*card)));
  sctlv_check_current_ranges((short)(2*(*card)));
#endif

#ifdef VXINT
  /*
   * For VXINT, use higher level VXI/VME access such that the window 
   * can be opened and closed "on the fly", allowing SCTLV cards 
   * to be distributed amongst VME crates.
   * Some restrictions:
   *   A16 - can have cards with base addresses 0x0000 - 0xB000.
   *         (0xC000-0xFFFF is reserved).  The number of cards
   *         allowed in one crate must be a power of 2 (1,2,4,8,16).
   *         If placing only 3 cards (0-2) in one crate, the next
   *         crate must start with card 4.
   *   A24 - cannot be split across crates given the present address
   *         range of SCTLV.
   * PWP 08.02.01
   */
  
  if(sctlv_card_total == 1){
    printf("sctlv_lib: initialising VXI library\n");
    InitVXIlibrary();
  }
  
  sctlv_info[*card].address = address;

  if(address>0xffff){  /* assume A24 */
    sctlv_info[*card].accessparms = 2;
    printf ("sctlv_vmemap: card %2d in A24 at address 0x%lx",
          *card, sctlv_info[*card].address);
  }else{               /* assume A16 */
    sctlv_info[*card].accessparms = 1;
    printf ("sctlv_vmemap: card %2d in A16 at address 0x%lx",
          *card, sctlv_info[*card].address);
  }

  ret = VXIin(sctlv_info[*card].accessparms,address+0x37,1,&dta);

  if(ret<0){
    printf (" - rcode %d => MISSING\n",ret);
    sctlv_info[*card].missing = 1;
    return ret;
  } else {
    printf (" - OK\n");
    sctlv_init((short)(2*(*card)));
    sctlv_check_current_ranges((short)(2*(*card)));
  }

#endif

  return 0;
}

char sctlv_rd(short id, unsigned short reg){

  /*
   * Read byte from SCTLV channel id register reg
   *
   * New version PWP 17.01.01
   *
   * The LV system comprises sctlv_card_total cards or sctlv_chan_total channels.
   * The base address of each card is given by sctlv_info[card].address.
   * Odd numbered channels are (basically) offset by LV_CHAN_OFFSET.
   */

  int card;
  char dta;
#ifdef VXINT
  NIVXI_STATUS ret;
#else
  char *ip;
#endif

  if(id%2){
    if (reg < LV_CHAN_OFFSET)  {(reg += LV_CHAN_OFFSET);}
    if (reg == LV_0STATUS)     {(reg = LV_1STATUS);}
    if (reg == LV_0DIAGNOSTIC) {(reg = LV_1DIAGNOSTIC);}
  } 
  reg ++;
  card = id/2;

#ifdef VXINT

  ret = VXIin(sctlv_info[card].accessparms,(sctlv_info[card].address + reg),1,&dta);
  if(ret<0){
    printf ("VXIin rcode %d for address 0x%lx\n",ret,(sctlv_info[card].address + reg));
  }
    
#else

  ip = (unsigned char*)(sctlv_info[card].address + (unsigned long)reg);
  dta = *ip;

  #ifdef LV_DEBUG
    printf ("\ndebug card 0x%lx base 0x%lx ip 0x%lx data %d", card, sctlv_info[card].address, &ip, dta);
  #endif

#endif

  return dta;
}

void sctlv_wt(short id, unsigned short reg, char dta){

  /*
   * Write byte to SCTLV channel id register reg
   *
   * New version PWP 17.01.01
   *
   * The LV system comprises sctlv_card_total cards or sctlv_chan_total channels.
   * The base address of each card is given by sctlv_info[card].address.
   * Odd numbered channels are (basically) offset by LV_CHAN_OFFSET.
   */

  int card;
#ifdef VXINT
  NIVXI_STATUS ret;
#else
  char *ip;
#endif

  if(id%2){
    if (reg < LV_CHAN_OFFSET)  {(reg += LV_CHAN_OFFSET);}
    if (reg == LV_0STATUS)     {(reg = LV_1STATUS);}
    if (reg == LV_0DIAGNOSTIC) {(reg = LV_1DIAGNOSTIC);}
  } 
  reg ++;
  card = id/2;
  

#ifdef VXINT

  ret = VXIout(sctlv_info[card].accessparms,(sctlv_info[card].address + reg),1,dta);

  if(ret<0){
    printf ("VXIout rcode %d for address 0x%lx\n",ret,(sctlv_info[card].address + reg));
  }

#else

  ip = (unsigned char*)(sctlv_info[card].address + (unsigned long)reg);
  *ip = dta;

#endif

#ifdef LV_DEBUG
  printf ("\ndebug card 0x%lx base 0x%lx ip 0x%lx data %d", card, sctlv_info[card].address, &ip, dta);
#endif

 
}

int sctlv_test_rd_flag(short id){

  /* flag must return 0 before data can be read */

  int i = 0;
  unsigned short flag = 1;

  sctlv_wt(id, LV_FLAG, 1);
  while((flag)&(i<LV_NTRIES_FLAG)){
    flag = sctlv_rd(id,LV_FLAG);
    if(flag) sctlv_wait_ms(LV_WAIT_FLAG);
    i++;
  }
  if (flag){
    printf("SCTLV ch %d test_rd_flag failed after %d attempts\n",id, i);
  }
  return (int)(flag);
}

int sctlv_test_wt_flag(short id){

  /* flag must return 1 before data can be written */

  int i = 0;
  unsigned short flag;

  flag = sctlv_rd(id,LV_FLAG);
  if(!flag){
    sctlv_wt(id, LV_FLAG, 1);
    while((!flag)&(i<LV_NTRIES_FLAG)){
      flag = sctlv_rd(id,LV_FLAG);
      if(!flag) sctlv_wait_ms(LV_WAIT_FLAG);
      i++;
    }
  }
  if (!flag){
    printf("SCTLV ch %d test_wt_flag failed after %d attempts\n",id, i);
  }
  return (int)(flag);
}

int sctlv_board_restart(short id){

  /* restart both channels!!
   * Channel must be HW OFF in order for restart to work 
   * Flag will return 0 once the board has been restarted
   * At the time of writing this routine may not work correctly 
   * PWP 18.05.99 revised PWP 13.12.00 
   * REVISED PWP 22.03.01 following bug report from Simon Peeters
   *  - should work (more) correctly now!!
   */

  char status = 0;
  char board_status = 0;
  int i = 0;
  int flag = 1;

  /* Read Channel and Board Status bytes - if possible!! */

  do_test_rd_flag = 0;
  flag = sctlv_test_rd_flag(id);
  if(flag==0){ /* Communication established */
    sctlv_get_status(id,&status);
    sctlv_get_board_status(id,&board_status);
  }
  do_test_rd_flag = 1;

  if(status&0x80){
    printf("SCTLV channel %x has status HW ON: it cannot be restarted\n",id);
    return(1);
  }

  /* The following section is probably temporary, for diagnostics... */

  if(board_status&0x01) printf(" bit 0x01 set - POWER FAIL\n");
  if(board_status&0x02) printf(" bit 0x02 set - OVERHEAT\n");
  if(board_status&0x04) printf(" bit 0x04 set - HW Interlock\n");
  if(board_status&0x08) printf(" bit 0x08 set - ???\n");
  if(board_status&0x00) printf(" bit 0x10 set - Power Fail Latch\n");
  if(board_status&0x00) printf(" bit 0x20 set - Overheat Latch\n");
  if(board_status&0x00) printf(" bit 0x40 set - HW Interlock Latch\n");
  if(board_status&0x80) printf(" bit 0x80 set - HW Interlock Enabled\n");

  /* Actually do the restart */

  flag = 1;
  sctlv_wt(id, LV_RESTART, 1);

  while((flag)&&(i<LV_NTRIES_RESTART)){
    flag =  sctlv_rd(id,LV_RESTART);
    if(flag) sctlv_wait_ms(LV_WAIT_RESTART);
    i++;
  }

  if (flag){
    printf("SCTLV ch %d do_board_restart failed after %d attempts\n",id, i);
  }
  return(flag);
}

void sctlv_wait_ms(clock_t wait){

   long t_zero,ldelay;

   ldelay = (long)wait*CLOCKS_PER_SEC/1000;    /*get it in msec*/
   for(t_zero=(long)clock();(long)clock()-t_zero<=ldelay;);

}

int sctlv_init(short id){

  /*
   * Restart crashed cards and establish 
   * which version(s) of SCTLV are in use
   * returns 1 if successful (SHOULD CHANGE THIS??)
   * Return -1 if chan invalid
   * Return -2 if card is marked as missing
   */

  int rcode;
  
  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  rcode=sctlv_check_HW_on(id);
  if(rcode==1){
    /* SCTLV HW on */
    sctlv_test_rd_flag(id);
    sctlv_info[(id/2)].version = sctlv_rd(id, LV_VERSION);
  }
  return rcode;
}

int sctlv_check_current_ranges(short id){

  /*
   * Cards used for irradiation studies have increased current ranges.
   * This can be detected by reading the contents of the Vcc, Vdd high
   * bytes when the power is OFF.  That is the purpose of this routine.
   * returns 1 if successful (SHOULD CHANGE THIS??)
   * Return -1 if chan invalid
   * Return -2 if card is marked as missing
   * PWP 18.04.01
   */

  int rcode;
  char selects  = 0;
  char vcc_byte = 0;
  char vdd_byte = 0;

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  rcode=sctlv_check_HW_on(id);
  if(rcode==1){
    /* SCTLV HW on */
    sctlv_test_rd_flag(id);
    selects  = sctlv_rd(id,LV_SELECTS);
    vcc_byte = sctlv_rd(id,LV_M_VCC_HI);
    vdd_byte = sctlv_rd(id,LV_M_VDD_HI);
  }

  if((selects & 0x01)==0){ /* power is off, therefore the current
                              range flags actually have meaning  */

    if(vcc_byte&0x80) sctlv_info[(id/2)].icc_range = 2;
    else              sctlv_info[(id/2)].icc_range = 1;

    if(vdd_byte&0x80) sctlv_info[(id/2)].idd_range = 2;
    else              sctlv_info[(id/2)].idd_range = 1;
  }

  return rcode;
}


/*
 * Get Routines
 */


int sctlv_get_power(short id,  short *power){

  /*
   * Get the value of the power bit only.
   * Return -1 if chan invalid
   * Return -2 if card is marked as missing
   */

  short l_reset, l_select;
  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  sctlv_get_selects(id, power, &l_reset, &l_select);
  return 0;
}

int sctlv_get_reset(short id,  short *reset){

  /*
   * Get the value of the reset bit only.
   * Return -1 if chan invalid
   * Return -2 if card is marked as missing
   */

  short l_power, l_select;
  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  sctlv_get_selects(id, &l_power, reset, &l_select);
  return 0;
}

int sctlv_get_select(short id, short *select){

  /*
   * Get the value of the select bit only.
   * Return -1 if chan invalid
   * Return -2 if card is marked as missing
   */

  short l_power, l_reset;
  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  sctlv_get_selects(id, &l_power, &l_reset, select);
  return 0;
}

int sctlv_get_selects(short id, short *power, short *reset, short *select){
  
  /*
   * Get the values of the power, status and select bits.
   * Return -1 if chan invalid.
   * Return -2 if card is marked as missing.
   */

  char dta;
  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  /* An introduction to the bits of the LV_SELECTS register...
   * 0 (0x1) - PWRON  - active HIGH
   * 1 (0x2) - SELECT - active LOW
   * 2 (0x4) - RESET  - active HIGH */

  if(do_test_rd_flag) sctlv_test_rd_flag(id);
  dta = sctlv_rd(id,LV_SELECTS);
  * power  = (short)(dta & 0x01);
  * select = (short)((dta & 0x02)>>1);
  * reset  = (short)((dta & 0x04)>>2);

  return 0;
}

int sctlv_get_diagnostic(short id, char *diagnostic){

  /*
   * Read SCTLV diagnostic register.
   * Return -1 if chan invalid.
   * Return -2 if card is marked as missing.
   */

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  if(do_test_rd_flag) sctlv_test_rd_flag(id);
  *diagnostic = sctlv_rd(id, LV_0DIAGNOSTIC);

  return 0;
}

int sctlv_get_status(short id, char *status){

   /*
   * Read SCTLV status byte.
   * Return -1 if chan invalid.
   * Return -2 if card is marked as missing.
   */

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;
  
  if(do_test_rd_flag) sctlv_test_rd_flag(id);
  *status = sctlv_rd(id, LV_0STATUS);

  return 0;
}

int sctlv_get_board_status(short id, char *status){

   /*
   * Read SCTLV board status byte.
   * Return -1 if chan invalid.
   * Return -2 if card is marked as missing.
   */

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;
  
  if(do_test_rd_flag) sctlv_test_rd_flag(id);
  *status = sctlv_rd(id, LV_BOARD_STATUS);

  return 0;
}

int sctlv_get_t0(short id, char *t0){

  /*
   * Read module temperature sensor T0 - SCTLV3 only.
   * Resolution 1 degree.
   * Return -1 if chan invalid.
   * Return -2 if card is marked as missing.
   * Return -4 if SCTLV version 5 or older.
   */

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;
  if(sctlv_info[id/2].version<6) return -4;

  if(do_test_rd_flag) sctlv_test_rd_flag(id);
  *t0 = sctlv_rd(id,LV_M_T0);

  return 0;
}
 
int sctlv_get_t1(short id, char *t1){

  /*
   * Read module temperature sensor T1 - SCTLV3 only.
   * Resolution 1 degree.
   * Return -1 if chan invalid.
   * Return -2 if card is marked as missing.
   * Return -4 if SCTLV version 5 or older.
   */

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;
  if(sctlv_info[(id/2)].version<6) return -4;

  if(do_test_rd_flag) sctlv_test_rd_flag(id);
  *t1 = sctlv_rd(id,LV_M_T1);

  return 0;
} 

int sctlv_get_temp(short id, char *temp){

  /*
   * Read SCTLV board temp.  Resolution 1 degree.
   * Return -1 if chan invalid.
   * Return -2 if card is marked as missing.
   */

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  if(do_test_rd_flag) sctlv_test_rd_flag(id);
  *temp =  sctlv_rd(id,LV_TEMP);
  
  return 0;
}
 
int sctlv_get_temps(short id, char *t0, char *t1){

  /*
   * Read both module temperature sensors - SCTLV3 only.
   * Resolution 1 degree.
   * Return -1 if chan invalid.
   * Return -2 if card is marked as missing.
   * Return -4 if SCTLV version 5 or older.
   */

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;
  if(sctlv_info[id/2].version<6) return -4;

  if(do_test_rd_flag) sctlv_test_rd_flag(id);
  *t0 = sctlv_rd(id,LV_M_T0);
  *t1 = sctlv_rd(id,LV_M_T1);

  return 0;
} 


int sctlv_get_version(short id, char *version){

  /*
   * Return previously read content of SCTLV version register.
   * Return -1 if chan invalid.
   * Return -2 if card is marked as missing.
   */

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  *version = sctlv_info[(id/2)].version;

  return 0;
}


/*
 * Set Routines
 */


int sctlv_set_power(short id, short power){

  /*
   * Set the value of the power bit only.
   * Return -1 if chan invalid.
   * Return -2 if card is marked as missing.
   */

  short l_power, l_reset, l_select;
  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  sctlv_get_selects(id, &l_power, &l_reset, &l_select);
  sctlv_set_selects(id, power, l_reset, l_select);

  return 0;
}

int sctlv_set_reset(short id, short reset){

  /*
   * Set the value of the reset bit only.
   * Return -1 if chan invalid.
   * Return -2 if card is marked as missing.
   */

  short l_power, l_reset, l_select;
  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  sctlv_get_selects(id, &l_power, &l_reset, &l_select);
  sctlv_set_selects(id, l_power, reset, l_select);

  return 0;
}

int sctlv_set_select(short id, short select){

  /*
   * Set the value of the select bit only.
   * Return -1 if chan invalid.
   * Return -2 if card is marked as missing.
   */

  short l_power, l_reset, l_select;
  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  sctlv_get_selects(id, &l_power, &l_reset, &l_select);
  sctlv_set_selects(id, l_power, l_reset, select);

  return 0;
}

int sctlv_set_selects(short id, short power, short reset, short select){

  /*
   * Set the values of the power, status and select bits.
   * Return -1 if chan invalid.
   * Return -2 if card is marked as missing.
   */

  char dta = 0;
  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  if(power)  dta+= 0x1;
  if(select) dta+= 0x2;   
  if(reset)  dta+= 0x4;   

  sctlv_test_wt_flag(id);
  sctlv_wt(id,LV_SELECTS,dta);

  return 0;
}

int sctlv_set_vcc(short id, float vcc, float icc){

  /*
   * Set Vcc and Icc limit of the requested LV channel
   * Return -1 if chan invalid
   * Return -2 if card is marked as missing.
   */

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  sctlv_setval(id,Vcc,vcc);
  sctlv_setval(id,Icc,icc);
  return 0;
}

int sctlv_set_vvi1(short id, float vvi1, float ivi1){

  /*
   * Set Vvi1 and Ivi1 limit of the requested LV channel
   * Return -1 if chan invalid
   * Return -2 if card is marked as missing.
   * Return -4 if SCTLV3 - Vi1 does not exist
   */

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;
  if(sctlv_info[(id/2)].version>=6) return -4;

  sctlv_setval(id,Vi1,vvi1);
  sctlv_setval(id,Ivi1,ivi1);
  return 0;
}

int sctlv_set_vdd(short id, float vdd, float idd){

  /*
   * Set Vdd and Idd limit of the requested LV channel
   * Return -1 if chan invalid
   * Return -2 if card is marked as missing.
   */

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  sctlv_setval(id,Vdd,vdd);
  sctlv_setval(id,Idd,idd);
  return 0;
}

int sctlv_set_vled0(short id, float vled0, float iled0){

  /*
   * Set Vled0 and Iled0 limit of the requested LV channel
   * Return -1 if chan invalid
   * Return -2 if card is marked as missing.
   */

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  sctlv_setval(id,Vld0,vled0);
  sctlv_setval(id,Ild0,iled0);
  return 0;
}

int sctlv_set_vled1(short id, float vled1, float iled1){

  /*
   * Set Vled1 and Iled1 limit of the requested LV channel
   * Return -1 if chan invalid
   * Return -2 if card is marked as missing.
   */

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  sctlv_setval(id,Vld1,vled1);
  sctlv_setval(id,Ild1,iled1);
  return 0;
}

int sctlv_set_vpin(short id, float vpin){

  /*
   * Set Vpin of the requested LV channel
   * Return -1 if chan invalid
   * Return -2 if card is marked as missing.
   */

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  sctlv_setval(id,Vpin,vpin);
  return 0;
}


/*
 * Low level Get/Set routines
 */


int sctlv_get_monval(short id, short ch, float *val){

  /*
   * Read the monitored value of any ADC in "real" units.
   * Return -1 if chan invalid.
   * Return -2 if card is marked as missing.
   * Return -3 if ADC  invalid for any SCTLV.
   * Return -4 if ADC  invalid for SCTLV version 6 or newer.
   */

  int bits;

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;
  if((ch<Vcc)||(ch>Ild1))  return -3;

  if(sctlv_info[(id/2)].version>=6){
    if((ch==Vi1)||(ch==Ivi1))  return -4;
  }

  if(do_test_rd_flag) sctlv_test_rd_flag(id);

  switch(ch){
    /*
     * Vcc and Vdd use 16 bit ADCs, other channels are just 8 bit
     */
    case Vcc:
      bits  = 0x7f00 & ( ((unsigned char) sctlv_rd(id,LV_M_VCC_HI)) <<8);
      bits += 0x00ff & (unsigned char) sctlv_rd(id,LV_M_VCC_LO);
      break;
    case Vi1:
      bits  = 0x00ff & (unsigned char) sctlv_rd(id,LV_M_VI1);
      break;
    case Vdd:
      bits  = 0x7f00 & ( ((unsigned char) sctlv_rd(id,LV_M_VDD_HI)) <<8);
      bits += 0x00ff & (unsigned char) sctlv_rd(id,LV_M_VDD_LO);
      break;
    case Vld0:
      bits  = 0x00ff & (unsigned char) sctlv_rd(id,LV_M_VLD0);
      break;
    case Vld1:
      bits  = 0x00ff & (unsigned char) sctlv_rd(id,LV_M_VLD1);
      break;
    case Vpin:
      bits  = 0x00ff & (unsigned char) sctlv_rd(id,LV_M_VPIN);
      break;
    case Icc:
      bits  = 0x00ff & (unsigned char) sctlv_rd(id,LV_M_ICC);
      if(sctlv_info[id/2].icc_range==2) bits *= 2; /* double range */
      break;
    case Ivi1:
      bits  = 0x00ff & (unsigned char) sctlv_rd(id,LV_M_IVI1);
      break;
    case Idd:
      bits  = 0x00ff & (unsigned char) sctlv_rd(id,LV_M_IDD);
      if(sctlv_info[id/2].idd_range==2) bits *= 2; /* double range */
      break;
    case Ild0:
      bits  = 0x00ff & (unsigned char) sctlv_rd(id,LV_M_ILD0);
      break;
    case Ild1:
      bits  = 0x00ff & (unsigned char) sctlv_rd(id,LV_M_ILD1);
      break;
    default:
      bits=0;
      break;
  }

  * val = (float) ( ((double)bits) * vi_scale[ch] );
  return 0;
}

int sctlv_get_setval(short id, short ch, float *val){

  /*
   * Read back the set value of any DAC in "real" units.
   * Return -1 if chan invalid
   * Return -2 if card is marked as missing.
   * Return -3 if DAC  invalid
   * Return -4 if attempting to read Vpin DAC on SCTLV1
   * Return -5 if attempting to read Vi1  DAC on SCTLV3
   * Return -6 if attempting to read Ivi1 limit on SCTLV3 
   */

  int bits;

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;
  if((ch<Vcc)||(ch>Ild1))  return -3;

  if(sctlv_info[(id/2)].version<5){
    if(ch==Vpin){
      * val = -1;
      return -4;
    }
  }

  if(sctlv_info[(id/2)].version>=6){
    if(ch==Vi1){
      * val = -1;
      return -5;
    }

    if(ch==Ivi1){
      * val = -1;
      return -6;
    }
  }

  if(do_test_rd_flag) sctlv_test_rd_flag(id);

  switch(ch){
    /*
     * Vcc and Vdd use 16 bit DACs, other channels are just 8 bit
     */
    case Vcc:
      bits  = 0x7f00 & ( ((unsigned char) sctlv_rd(id,LV_S_VCC_HI)) <<8);
      bits += 0x00ff & (unsigned char) sctlv_rd(id,LV_S_VCC_LO);
      break;
    case Vi1:
      bits  = 0x00ff & (unsigned char) sctlv_rd(id,LV_S_VI1);
      break;
    case Vdd:
      bits  = 0x7f00 & ( ((unsigned char) sctlv_rd(id,LV_S_VDD_HI)) <<8);
      bits += 0x00ff & (unsigned char) sctlv_rd(id,LV_S_VDD_LO);
      break;
    case Vld0:
      bits  = 0x00ff & (unsigned char) sctlv_rd(id,LV_S_VLD0);
      break;
    case Vld1:
      bits  = 0x00ff & (unsigned char) sctlv_rd(id,LV_S_VLD1);
      break;
    case Vpin: /* added PWP 08.12.99 */
      bits  = 0x00ff & (unsigned char) sctlv_rd(id,LV_S_VPIN);
      break;
    case Icc:
      bits  = 0x00ff & (unsigned char) sctlv_rd(id,LV_S_ICC);
      if(sctlv_info[id/2].icc_range==2) bits *= 2; /* double range */
      break;
    case Ivi1:
      bits  = 0x00ff & (unsigned char) sctlv_rd(id,LV_S_IVI1);
      break;
    case Idd:
      bits  = 0x00ff & (unsigned char) sctlv_rd(id,LV_S_IDD);
      if(sctlv_info[id/2].idd_range==2) bits *= 2; /* double range */
      break;
    case Ild0:
      bits  = 0x00ff & (unsigned char) sctlv_rd(id,LV_S_ILD0);
      break;
    case Ild1:
      bits  = 0x00ff & (unsigned char) sctlv_rd(id,LV_S_ILD1);
      break;
    default:
      bits=0;
      break;
  }

  * val = (float) ( ((double)bits) * vi_scale[ch] );
  return 0;
}

int sctlv_setval(short id, short ch, float val){

  /*
   * Set any DAC to a value given in "real" units.
   * Return -1 if chan invalid.
   * Return -2 if card is marked as missing.
   * Return -3 if DAC  invalid.
   * Return -4 if attempting to set Vpin DAC on SCTLV1
   * Return -5 if attempting to set Vi1  DAC on SCTLV3
   * Return -6 if attempting to set Ivi1 limit on SCTLV3 
   * Return 2 if value set to lower limit
   * Return 3 if value set to upper limit
   */

  int rcode = 0;
  unsigned short bits;

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;
  if((ch<Vcc)||(ch>Ild1))  return -3;

  if(sctlv_info[(id/2)].version<5){
    if(ch==Vpin){
      val = -1;
      return -4;
    }
  }

  if(sctlv_info[(id/2)].version>=6){
    if(ch==Vi1){
      val = -1;
      return -5;
    }

    if(ch==Ivi1){
      val = -1;
      return -6;
    }
  }

  if(val < vi_min[ch]){   /* trim to minimum value */
    printf("sctlv_setval called for board %d ch %d with value %6.1f\n", id, ch, val);
    printf("             setting to limit of %6.1f instead\n", vi_min[ch]);
    val = vi_min[ch];
    rcode = 2;
  }

  if(val > vi_max[ch]){   /* trim to maximum value */
    printf("sctlv_setval called for board %d ch %d with value %6.1f\n", id, ch, val);
    printf("             setting to limit of %6.1f instead\n", vi_max[ch]);
    val = vi_max[ch];
    rcode = 3;
  }
    /* fix for rounding errors - add half step - PWP 13.07.99 */

    bits = (unsigned short) ((double)(val+(vi_scale[ch]/2)) / vi_scale[ch]);

#ifdef LV_DEBUG
    printf("  setting LV ch %x %s to %6.2f %s (= %3d bits)\n",
            id,vi_name[ch],val,vi_unit[ch],bits);
#endif

  sctlv_test_wt_flag(id);

  switch(ch){
    /*
     * Vcc and Vdd use 16 bit ADCs, other channels are just 8 bit
     */
    case Vcc:
      sctlv_wt(id,LV_S_VCC_HI, ((unsigned char)((bits&0xff00)>>8)));
      sctlv_wt(id,LV_S_VCC_LO, ((unsigned char) (bits&0x00ff)    ));
      break;
    case Vdd:
      sctlv_wt(id,LV_S_VDD_HI, ((char)((bits&0xff00)>>8)));
      sctlv_wt(id,LV_S_VDD_LO, ((char) (bits&0x00ff)    ));
      break;
    case Vi1:
      sctlv_wt(id,LV_S_VI1,((char)(bits&0xff)));
      break;
    case Vld0:
      sctlv_wt(id,LV_S_VLD0,((char)(bits&0xff)));
      break;
    case Vld1:
      sctlv_wt(id,LV_S_VLD1,((char)(bits&0xff)));
      break;
    case Vpin: /* added PWP 08.12.99 */
      sctlv_wt(id,LV_S_VPIN,((char)(bits&0xff)));
      break;
    case Icc:
      if(sctlv_info[id/2].icc_range==2) bits /=2; /* double range */
      sctlv_wt(id,LV_S_ICC,((char)(bits&0xff)));
      break;
    case Ivi1:
      sctlv_wt(id,LV_S_IVI1,((char)(bits&0xff)));
      break;
    case Idd:
      if(sctlv_info[id/2].idd_range==2) bits /=2; /* double range */
      sctlv_wt(id,LV_S_IDD,((char)(bits&0xff)));
      break;
    case Ild0:
      sctlv_wt(id,LV_S_ILD0,((char)(bits&0xff)));
      break;
    case Ild1:
      sctlv_wt(id,LV_S_ILD1,((char)(bits&0xff)));
      break;
    default:
      break;
  }
  return rcode;
}


/* 
 * Detector module reset
 */

int sctlv_do_reset(short id){

  /* 
   * Issue HARD_RESET to a single detector module.
   * Assumes that reset is initially LOW, ie. RESETB is HIGH
   * Return -1 if chan invalid.
   * Return -2 if card is marked as missing.
   */

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  sctlv_set_reset(id, 1);
  sctlv_wait_ms(LV_WAIT_RESET);
  sctlv_set_reset(id, 0);

  return 0;
}

int sctlv_do_reset_all(){

  /* 
   * Issue HARD_RESET to all detector modules, i.e. toggle
   * reset on each SCTLV channel.
   * Assumes that reset is initially LOW, ie. RESETB is HIGH
   */

  sctlv_do_reset_many((short)0,(short)(sctlv_chan_total-1));

  return 0;
}

int sctlv_do_reset_many(short lo, short hi){

  /* 
   * Issue HARD_RESET to all detector modules from lo to hi.
   * Assumes that reset is initially LOW, ie. RESETB is HIGH
   * Return -1 if lo >= sctlv_chan_total
   * coerce hi to (sctlv_chan_total-1) if value too high
   */

  short id;

  if(lo>=sctlv_chan_total) return -1;
  if(hi>=sctlv_chan_total) hi = (short) (sctlv_chan_total-1);

  for (id=lo; id<=hi; id++){sctlv_set_reset(id,1);}
  sctlv_wait_ms(LV_WAIT_RESET);
  for (id=lo; id<=hi; id++){sctlv_set_reset(id,0);}

  return 0;
}  




/*
 New High level routines, mainly for diagnostic use, PWP 04.03.99
 */

int sctlv_set_defaults(short id){

  /*
   * Simple routine to set hardcoded default values.
   * Return -1 if chan invalid.
   * Return -2 if card is marked as missing.
   */

  int i;
  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  for(i=Vcc; i<=Ild1; i++){
    sctlv_setval(id, (short)i, vi_def[i]);
  } 

  sctlv_set_selects(id, 0, 0, 0); /* power off,
                                     reset low (=> RESETB HIGH)
                                     select low */

  return 0;
}

int sctlv_mon_all(short id, float * mon){

  /* Read all monitored and set values
   * Return -1 if chan invalid.
   * Return -2 if card is marked as missing.
   */

  int i;
  int flag;

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  flag = do_test_rd_flag;
  
  if(flag){
    sctlv_test_rd_flag(id);
    do_test_rd_flag=0;
  }

  for(i=Vcc; i<=Ild1; i++){
    sctlv_get_monval(id, (short)i, &mon[i]);
    sctlv_get_setval(id, (short)i, &mon[i+11]);
  } 

  do_test_rd_flag = flag;
  return (0); 
}

int sctlv_print_status(short id){

  /*
   * Print out channel status to the screen for diagnostic purposes
   * Return -1 if chan invalid.
   * Return -2 if card is marked as missing.
   */

  char diagnostic, status, temp, t0, t1;
  float mon[22];
  int i;
  short power, reset, select;

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  
  do_test_rd_flag = 0;
  sctlv_test_rd_flag(id);
  
  sctlv_get_status(id, &status);
  sctlv_mon_all(id, mon);
  sctlv_get_selects(id, &power, &reset, &select);
  sctlv_get_temp(id, &temp);

  if(sctlv_info[(id/2)].version>=6){
    sctlv_get_temps(id,&t0,&t1);
  }
  sctlv_get_diagnostic(id, &diagnostic);

  do_test_rd_flag = 1;

  printf("\nSCTLV channel %d (0x%x) status:0x%2x \n",id,id,(status&0xff));

  if(status&0x01)printf("** Analogue Subch. Parity Error\n");
  if(status&0x02)printf("** Digital Subch. Parity Error\n");
  if(status&0x04)printf("** Analogue Subch. Communication Error\n");
  if(status&0x08)printf("** Digital Subch. Communication Error\n");
  if(status&0x10)printf("** Current Limit\n");
  if(status&0x20)printf("** Current Trip\n");
  if(status&0x40)printf("** Channel SW ON\n");
  if(status&0x80)printf("** Channel HW ON\n");

    printf("          Set    Mon         Limit    Mon\n");
  for(i=Vcc;i<Vpin;i++){
    printf("  %s  %1.3f  %1.3f   %s  %5.1f  %5.1f\n",
            vi_name[i], mon[11+i],mon[i],
            vi_name[Icc+i],mon[11+Icc+i], mon[Icc+i]);
  }
    printf("  Vpin         %1.3f\n", mon[Vpin]);
    printf("          V      V             mA     mA\n\n");

    if(sctlv_info[(id/2)].version>=6){
      printf("  Detector Module temperatures T0 %2d and T1 %2d\n",t0,t1);
    }      
    printf("  Board Diagnostic 0x%2x\n", diagnostic); 
    printf("  Board Temperature %d celcius\n",temp);
    printf("  Board Version is %d\n",sctlv_info[(id/2)].version);
    printf("  Icc range is %d\n",sctlv_info[(id/2)].icc_range);
    printf("  Idd range is %d\n",sctlv_info[(id/2)].idd_range);
    printf("  Power is ");
      if(power){printf("ON, ");}
      if(!power){printf("OFF,");}
    printf("            Select is ");
      if(select){printf("HIGH\n");}
      if(!select){printf(" LOW\n");}
    printf("  Reset is ");
      if(reset){printf("HIGH,");}
      if(!reset){printf("LOW, ");}
    printf(" therefore RESETB is ");
      if(!reset){printf("HIGH\n");}
      if(reset){printf(" LOW\n");}

 
  return 0;
}

int sctlv_ramp_v(short id, short ch, float start, float stop, float step){

  /*
   * Problems at RAL with hardware current trips due to transients
   *  at switch on.  This is a crude way to get around the problem
   * for the time being... PWP 04.02.99 revised 17.05.99
   * NB - problem seems to have gone away 19.05.99
   * Return -1 if chan invalid.
   * Return -2 if card is marked as missing.
   */

  char status=0;
  float v = start;

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  while( (v<stop) & (!(status&0x3F)) ){
    sctlv_setval(id,ch,v);
    if (v==start){
      sctlv_wait_ms(LV_WAIT_RAMP);
      sctlv_set_power(id,1);
      printf("  Power is now ON\n");
    }
    sctlv_wait_ms(250);
    sctlv_get_status(id, &status);
    v += step;
  }

  if(!(status&0x3F)){
    v = stop;
    sctlv_setval(id,ch,v);
    sctlv_wait_ms(LV_WAIT_RAMP);
  }

  if((status&0x3F)){
    printf("\nsctlv_ramp: LV ch %x %s tripped at %f %s\n",id,vi_name[ch],v,vi_unit[ch]);
    return(1);
  }
  else{
    printf("\nsctlv_ramp: LV ch %x %s reached %f %s\n",id,vi_name[ch],v,vi_unit[ch]);
  }

  return 0;
}

int sctlv_check_HW_on(short id){

  /*
   *Check that channel has HW status ON,
   * if not, restart the channel
   * It is up to the user to switch power on as appropriate
   * Return -1 if chan invalid
   * Return -2 if card is marked as missing.
   * Return 0 if chan has HW OFF
   * Return 1 if chan has HW ON
   */

  char status;

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  sctlv_get_status(id, &status);
  if(status&0x80){
    return 1;
  }
  printf("SCTLV channel %x has initial status HW OFF!!\n",id);

  sctlv_board_restart(id);

  sctlv_get_status(id, &status);
  if(status&0x80){
  printf("SCTLV channel %x has been restarted\n",id);
  return 1;
  }
  else{
    printf("SCTLV channel %x restart FAILED!!\n",id);
    return 0;
  }
}

int sctlv_brief_status(short id){

  /*
   * Print out channel status to the screen for diagnostic purposes
   * Return -1 if chan invalid
   * Return -2 if card is marked as missing.
   */

  char status;
  float mon[22];
  short ilim, trip, swon, hwon;
  short power, reset, select;

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  do_test_rd_flag = 0;
  sctlv_test_rd_flag(id);
  
  sctlv_get_status(id, &status);
  sctlv_mon_all(id, mon);
  sctlv_get_selects(id, &power, &reset, &select);

  do_test_rd_flag = 1;

  ilim = (status&0x10) ? 1 : 0;
  trip = (status&0x20) ? 1 : 0;
  swon = (status&0x40) ? 1 : 0;
  hwon = (status&0x80) ? 1 : 0;

  printf("ch ilim trip SW HW Pwr Rst Sel   Vcc   mon  Icc  mon   Vdd   mon  Idd  mon\n");
  printf("%2d %4d %4d %2d %2d %3d %3d %3d %1.3f %1.3f %4.0f %4.0f %1.3f %1.3f %4.0f %4.0f\n",
         id, ilim, trip, swon, hwon, power, reset, select,
         mon[Vcc+11], mon[Vcc], mon[Icc+11], mon[Icc],
         mon[Vdd+11], mon[Vdd], mon[Idd+11], mon[Idd]);
      
  return 0;
}

int sctlv_check_pwron(short id){

  /*
   * 1) Check HW ON
   *      - restart if HW OFF, then turn power on
   * 2) Check that channel has not tripped
   *      - toggle power to recover
   * Return -1 if chan invalid
   * Return -2 if card is marked as missing.
   * Return 1 if channel has tripped
   * Return 2 if card has crashed
   */

  char status;

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  sctlv_get_status(id, &status);

  if((status&0x80)!= 0x80){
    /* HW off - try restart */
    printf("SCTLV channel %x has initial status HW OFF!!\n",id);

    sctlv_board_restart(id);
    sctlv_get_status(id, &status);
    if((status&0x80)!= 0x80){
      printf("SCTLV channel %x restart FAILED!!\n",id);
      return 2;
    }
    printf("SCTLV channel %x has been restarted!!\n",id);

    sctlv_set_power(id, 1);
    sctlv_wait_ms(500);
    sctlv_get_status(id, &status);
  }
  
  if(status&0x20){
    /* channel has tripped - toggle power */
    sctlv_set_power(id, 0);
    sctlv_wait_ms(500);
    sctlv_set_power(id, 1);

    return 1;
  }
  else {
    sctlv_set_power(id,1);
    return 0;
  }
}

int sctlv_check_trip(short id){

  /*
   * Check for a trip condition on channel id
   * Return -1 if chan invalid
   * Return -2 if card is marked as missing.
   * Return 1 if channel has tripped
   * Return 2 if card has crashed
   */

  char status;
  int rcode = 0;

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  do_test_rd_flag = 0;

  if(sctlv_test_rd_flag(id)==0){
    /* Communication established */

    sctlv_get_status(id,&status);

    if((status&0x80)!= 0x80){
      /* Channel  has status hardware off - probably crashed */
      rcode = 2;
    } else if(status&0x20){
      /* channel has tripped */
      sctlv_brief_status(id);
      rcode = 1;
    }
  } else {
    /* timeout: again, the card has probably crashed */
    rcode = 2;
  }
 
  do_test_rd_flag = 1;
  return rcode;
}

int sctlv_monitor(short id, char* status, char* temp, float * mon){

  /* Read all monitored values 
   * Return -1 if chan invalid
   * Return -2 if card is marked as missing.
   * Return 1 if channel has tripped
   * Return 2 if card has crashed
   * PWP 12.12.00
   */

  int i;
  int rcode = 0;

  if(id>=sctlv_chan_total) return -1;
  if(sctlv_info[id/2].missing) return -2;

  do_test_rd_flag = 0;

  if(sctlv_test_rd_flag(id)==0){
    /* Communication established */

    sctlv_get_status(id,status);
    sctlv_get_temp(id,temp);

    for(i=Vcc; i<=Ild1; i++){
      sctlv_get_monval(id, (short)i, &mon[i]);
    }

    if(*status&0x20){
      /* channel has tripped */
      rcode = 1;
    }
  } else {
    /* timeout: probably the card has crashed */
    rcode = 2;
  } 

  do_test_rd_flag = 1;
  return rcode; 
}

short sctlv_n_cards(){
  return sctlv_card_total;
}

short sctlv_n_channels(){
  return sctlv_chan_total;
}

void sctlv_show_error(int rcode){

  /* 
   * Simple display of error codes
   */

  switch(rcode){
    case -1:
      printf("sctlv_show_error: code %d - invalid channel\n",rcode);
      break;
    case -2:
      printf("sctlv_show_error: code %d - card MISSING\n",rcode);
      break;
    case -3:
      printf("sctlv_show_error: code %d - invalid DAC or ADC\n",rcode);
      break;
    case -4:
    case -5:
    case -6:
      printf("sctlv_show_error: code %d - inappropriate VERSION\n",rcode);
      break;
    default:
      break;
  }
  return;
}


