// LiveScan.cpp
// Version 0.1 PWP SJP 01.08.00
//
// Macro to continuosly take scans and keep track of the sigma and mean of 
///the fitted results. Current set-up shows three channels per chip: 
// first, middle and last-3 (choosen to fit CAL0).
//
// Example of usage:
// e->burst_ntrigs = 500;        // set number of triggers per burst
// e->ConfigureScan(1,0,250,10); // set-up to scan the threshold (scan nr 1)
//                               // from 0 mV to 250 mV with a 10 mV step
// .x LiveScan.cpp(1000,0)       // perform 1000 scans for module 0
//
// The performance at the CERN forward system was the time taken per scan 
// is approximately 5 seconds.

#define LS_NCHAN 36
#define LS_NBINS 64

TH2F    * hr[LS_NCHAN];

// Channels chosen to suit CAL0 :
static const int 
chan[LS_NCHAN]={   0,  64, 124,  128, 192, 252,  256, 320, 380,
		 384, 448, 508,  512, 576, 636,  640, 704, 764,
		 768, 832, 892,  896, 960,1020, 1024,1088,1148,
		1152,1216,1276, 1280,1344,1404, 1408,1472,1532}; 
float sigma     [LS_NCHAN][LS_NBINS];
float mean      [LS_NCHAN][LS_NBINS];
float time_axis [LS_NBINS];
int   loop = 0;

void LiveScan(int nscans, int moi){

  int stop_loop = 0;
  int old_calloop = e->do_cal_loop;

  ls_create_display(moi);
  st_start_stopper();  // start stop dialog

  e->do_cal_loop=0;
  e->ConfigureVariable(10,0); // set ST_CAL_MODE = 0

  while(loop<nscans){
    if ( e->scan_count >= 127 ) e->scan_count = 0;
    e->InitScan(); // will open file for ascii ntuple if ascii_nt !=0

    while(e->scan_running){ 

      e->ScanPoint();
      stop_loop=st_do_stop();
      if(stop_loop==1){
	e->scan_running=0;
	loop=nscans;
      }
    }
    e->EndScan(); // will close file for ascii ntuple if ascii_nt !=0
    ls_update_display(moi);
    //st_display_scan();
    //st_save_scan();
    ++loop;
  }
  st_stop_stopper();  // stop stop dialog
  e->do_cal_loop = old_calloop;

}

void ls_create_display(int moi){
  char h_name[32], h_title[32];
  char s[256];
  float range_lo, range_hi;
  int i;

  // Load other files needed by this one
  gROOT->LoadMacro("stan_util.cpp");

  // Create new canvas for scurves
  TCanvas * csc = new TCanvas ("csc",s,1100,750);
  sprintf(s,"%s - scurves Module %d", e->scan_name, moi);
  TPad * pad_t0 = new TPad("pad_t0","title",    0.00,0.95,1.00,1.000,10);
  pad_t0->Draw();
  TPad * pad_c0 = new TPad("pad_c0","SCurves",  0.00,0.30,1.00,0.950,10);
  pad_c0->Draw();
  TPad * pad_c1 = new TPad("pad_c1","LiveScan", 0.00,0.00,1.00,0.300,10);
  pad_c1->Draw();
  pad_c1->cd();
  pad_c1->Divide(2,1,0,0,10);
  pad_c0->cd();
  pad_c0->Divide(9,4,0,0,10);

  Float_t lo_y, hi_y;
  if (e->burst_trtype==5) {
    lo_y = -0.1;
    hi_y =  1.1;
  } else {
    lo_y = 0.1*e->burst_ntrigs;
    hi_y = 1.1*e->burst_ntrigs;
  }

  for(i=0;i<LS_NCHAN;i++){

    if(e->scan_stop>e->scan_start){
      range_lo = e->scan_start;
      range_hi = e->scan_stop;
    } else {
      range_lo = e->scan_stop;
      range_hi = e->scan_start;
    }

    pad_c0->cd((i%LS_NCHAN)+1);

    sprintf(h_name,"h_sc%d",i);
    sprintf(h_title,"Channel %d",chan[i]);
    hr[i] = new TH2F(h_name,h_title,2,range_lo,range_hi,2,lo_y,hi_y);
    hr[i]->Draw();
    hr[i]->SetXTitle(e->scan_name);  
    hr[i]->SetYTitle("occupancy");  
  }

  for ( i = 0 ; i < LS_NBINS ; ++i ) {
    time_axis [i] = i;
  }

  //  ls_update_display(moi);
}

void ls_update_display(int moi){
  char h_name[32], h_title[32];
  char s[256];
  int  i, j, n;
  float err2, n2, range_lo, range_hi;

  float x[128];
  float xe[128];
  float y[LS_NCHAN][128];
  float ye[LS_NCHAN][128];

  float chan_p0[LS_NCHAN];
  float chan_p1[LS_NCHAN];
  float chan_p2[LS_NCHAN];

  pad_t0->cd();
  pad_t0->Clear();

  TPaveText * ttitle = new TPaveText(0.01,0.05,0.99,0.95);
  ttitle->SetFillColor(10);  
  ttitle->SetTextAlign(13);
  ttitle->SetTextSize(0.4);
  ttitle->SetTextColor(4);

  sprintf(s,"ATLAS SCT Module Test - Run %d Scan %d Module %d", 
	  e->runnum,e->scan_count,moi);
  ttitle->AddText(s);
  sprintf(s,"%s", e->scan_description_string);
  ttitle->AddText(s);
  ttitle->Draw();

  //Select Appropriate Fit Function
  switch (e->scan_type){
  case 1: // ST_VTHR
    TF1 * f1 = new TF1 ("f1",erfc,  0, 1000, 3);
    f1->SetParNames("Constant","Mean","Sigma");
    break;
  case 2: // ST_VCAL
    TF1 * f1 = new TF1 ("f1", erf,  0, 1000, 3);
    f1->SetParNames("Constant","Mean","Sigma");
    break;
  case 3: // ST_STROBE_DELAY
    // fit the back edge - > front edge of charge injection pulse
    TF1 * f1 = new TF1 ("f1", erfc,  0, 1000, 3);
    f1->SetParNames("Constant","Mean","Sigma");
    break;
  case 41: // ST_QTHR
    TF1 * f1 = new TF1 ("f1",erfc,  0, 1000, 3);
    f1->SetParNames("Constant","Mean","Sigma");
    break;
  case 42: // ST_QCAL
    TF1 * f1 = new TF1 ("f1", erf,  0, 1000, 3);
    f1->SetParNames("Constant","Mean","Sigma");
    break;
  default: // use gaussian for now 
    TF1 * f1 = new TF1 ("f1", gaus,  0, 1000, 3);
    f1->SetParNames("Constant","Mean","Sigma");
    break;
  }


  // Determine maximum occupancy

  if ( (e->m[moi]->h_scan0->GetMaximum()<1.0)
       && (e->m[moi]->h_scan1->GetMaximum()<1.0)) n=1.0;
  else n=e->burst_ntrigs;
  n2 = e->burst_ntrigs;
  if (n2 ==0) n2 = 1;

  // Project arrays needed to form graphs

  TAxis * axis = e->m[moi]->h_scan0->GetYaxis();

  for(i=0;i<e->scan_max_points; i++){
    x[i] = axis->GetBinCenter(i+1);
    xe[i] = 1;
  }

  for(i=0;i<LS_NCHAN;i++){
    for(j=0;j<e->scan_max_points;j++){
      if(chan[i]>767)
	y[i][j] = e->m[moi]->h_scan1->GetBinContent((770*(j+1))+(chan[i]-768));
      else
	y[i][j] = e->m[moi]->h_scan0->GetBinContent((770*(j+1))+chan[i]);

      if(n>1){
	err2 = (1+(y[i][j])) * (1+(n2-y[i][j])) / n2;
	ye[i][j]=sqrt(err2);
      } else {
	err2 = (1+(n2*y[i][j])) * (1+(n2-(n2*y[i][j]))) / n2;
	ye[i][j]=sqrt(err2)/n2;
      }
    }
  }

  // Project graphs

  TGraphErrors * gr      [LS_NCHAN];
  TGraph * sigma_gr[LS_NCHAN];
  TGraph * mean_gr [LS_NCHAN];
  float min_sigma;
  float max_sigma;
  float min_mean;
  float max_mean;

  for(i=0;i<LS_NCHAN;i++){
    
    int channel_color = (i%49)+1;
    if ( channel_color > 9 ) ++channel_color;
    if(e->scan_stop>e->scan_start){
      range_lo = e->scan_start;
      range_hi = e->scan_stop;
    } else {
      range_lo = e->scan_stop;
      range_hi = e->scan_start;
    }
    Float_t xtpos = 0.4*fabs(range_hi-range_lo);

    pad_c0->cd((i%LS_NCHAN)+1);

    hr[i]->Draw();

    gr[i] = new TGraphErrors(e->scan_max_points,x,y[i],xe,ye[i]);
    gr[i]->SetTitle(h_title);
    gr[i]->Draw("P");

    Float_t max = 0.0;
    for(j=0;j<e->scan_max_points;j++){
      if(y[i][j]>max) max = y[i][j];
    }

    if(max>0.0){

      Float_t p0_guess = max;
      Float_t p1_guess = range_lo+range_hi;   // OVERLY SIMPLISTIC!!
      Float_t p2_guess = 10;                  // OVERLY SIMPLISTIC!!
      if(p1_guess>0) p1_guess=p1_guess/2;
      else           p1_guess = 1;

      if((e->scan_type==1)||(e->scan_type==41)||(e->scan_type==3)){
	// VTHR, QTHR or STROBE DELAY
	// search from high end down
	j=e->scan_max_points-1;
	while((j>=0)&&(y[i][j]<(0.99*max))){
          range_lo = x[j];                         // last bin <99%
          if (y[i][j]<(0.82*max)) p2_guess = x[j]; // last bin <82%
          if (y[i][j]<(0.50*max)) p1_guess = x[j]; // last bin <50%
          j--;
	}
	p2_guess = p2_guess - p1_guess;

      } else if((e->scan_type==2)||(e->scan_type==42)){
	// VCAL or QCAL 
	// search from low end up
	j=0;
	while((j<e->scan_max_points)&&(y[i][j]<(0.99*max))){
          if (y[i][j]<(0.50*max)) p1_guess = x[j]; // first bin <50%
          if (y[i][j]<(0.82*max)) p2_guess = x[j]; // first bin <82%
          range_hi = x[j];                         // first bin <99%
          j++;
	}
	p2_guess = p2_guess - p1_guess;
      }
      if(p2_guess<0) p2_guess = p2_guess*-1.0;
      if(p2_guess==0) p2_guess =1;  // emergency!!
  
      f1->SetParameters(p0_guess,p1_guess,p2_guess);
      f1->SetParLimits(0,p0_guess,p0_guess);
      f1->SetRange(range_lo, range_hi);
      f1->SetLineColor(channel_color);

      //printf("chan %d p0 %f p1 %f p2 %f lo %d hi %d\n",
      //        chan[i],p0_guess,p1_guess,p2_guess,range_lo,range_hi); 
      gr[i]->SetLineColor(channel_color);
      gr[i]->SetMarkerColor(channel_color);
      gr[i]->Fit("f1","QR"); // Quiet Mode
      chan_p0[i] = f1->GetParameter(0); 
      chan_p1[i] = f1->GetParameter(1); 
      chan_p2[i] = f1->GetParameter(2);

    } else {
      chan_p0[i] = -1;
      chan_p1[i] = -1;
      chan_p2[i] = -1;
    }

    TText * t = new TText();
    t->SetTextSize(.08);
    sprintf(h_title,"Mean = %03.2f",    chan_p1[i]);
    t->DrawText(xtpos,(0.9*n),h_title);
    sprintf(h_title,"Sigma = %03.2f",   chan_p2[i]);
    t->DrawText(xtpos,(0.8*n),h_title);
    sprintf(h_title,"Max Eff. = %03.2f",chan_p0[i]);
    t->DrawText(xtpos,(0.7*n),h_title);
    //
    // Update graphs of sigma and mean
    //
    if ( loop < LS_NBINS ) {
      mean [i][loop] = chan_p1[i];
      sigma[i][loop] = chan_p2[i];
      for ( j = 0 ; j < loop ; ++j ) {
	if ( i == 0 && j == 0 ) {
          min_sigma = sigma[i][j];
          max_sigma = sigma[i][j];
          min_mean  = mean[i][j];
          max_mean  = mean[i][j];
	}
	else {
          if ( min_sigma > sigma[i][j] ) min_sigma = sigma[i][j];
          if ( max_sigma < sigma[i][j] ) max_sigma = sigma[i][j];
          if ( min_mean > mean[i][j] )   min_mean  = mean[i][j];
          if ( max_mean < mean[i][j] )   max_mean  = mean[i][j];
	}
      }
      sigma_gr[i] = new TGraph(loop+1,time_axis,sigma[i]);
      mean_gr[i] = new TGraph(loop+1,time_axis,mean[i]);
    }
    else {
      for ( j = 0 ; j < LS_NBINS-1 ; ++j ) {
	mean [i][j] = mean [i][j+1];
	sigma[i][j] = sigma[i][j+1];
	if ( i == 0 ) { 
          time_axis[j] = loop-LS_NBINS+j+1;
	}
	if ( i == 0 && j == 0 ) {
          min_sigma = sigma[i][j];
          max_sigma = sigma[i][j];
          min_mean  = mean[i][j];
          max_mean  = mean[i][j];
	}
	else {
          if ( min_sigma > sigma[i][j] ) min_sigma = sigma[i][j];
          if ( max_sigma < sigma[i][j] ) max_sigma = sigma[i][j];
          if ( min_mean > mean[i][j] )   min_mean  = mean[i][j];
          if ( max_mean < mean[i][j] )   max_mean  = mean[i][j];
	}
      }
      mean [i][LS_NBINS-1] = chan_p1[i];
      sigma[i][LS_NBINS-1] = chan_p2[i];
      if ( i == 0 ) time_axis[LS_NBINS-1] = loop;
      if ( min_sigma > sigma[i][LS_NBINS-1] ) min_sigma = sigma[i][LS_NBINS-1];
      if ( max_sigma < sigma[i][LS_NBINS-1] ) max_sigma = sigma[i][LS_NBINS-1];
      if ( min_mean > mean[i][LS_NBINS-1] )   min_mean  = mean[i][LS_NBINS-1];
      if ( max_mean < mean[i][LS_NBINS-1] )   max_mean  = mean[i][LS_NBINS-1];
      sigma_gr[i] = new TGraph(LS_NBINS,time_axis,sigma[i]);
      mean_gr[i] = new TGraph(LS_NBINS,time_axis,mean[i]);
    } 
    sigma_gr[i]->SetMarkerColor(channel_color);
    sigma_gr[i]->SetLineColor(channel_color);
    mean_gr[i]->SetMarkerColor(channel_color);
    mean_gr[i]->SetLineColor(channel_color);
  }

  pad_c1->cd(1);
  if ( 0.9*min_sigma == 1.1*max_sigma ) max_sigma = min_sigma+1;
  TH2F * sigma_frame = new TH2F("sigma_frame","sigma (mV)",
				2,time_axis[0],time_axis[LS_NBINS-1],
				2,0.9*min_sigma,1.1*max_sigma);
  sigma_frame->SetYTitle("sigma (mV)");
  sigma_frame->SetXTitle("scannumber");
  sigma_frame->Draw();
  for ( i = 0 ; i < LS_NCHAN ; ++i ) sigma_gr[i]->Draw("LP");
  pad_c1->cd(2);
  if ( 0.9*min_mean == 1.1*max_mean ) max_mean = min_mean+1;
  TH2F * mean_frame = new TH2F("mean_frame","50% point (mV)",
			       2,time_axis[0],time_axis[LS_NBINS-1],
			       2,0.9*min_mean,1.1*max_mean);
  mean_frame->SetYTitle("50% point (mV)");
  mean_frame->SetXTitle("scannumber");
  mean_frame->Draw();
  for ( i = 0 ; i < LS_NCHAN ; ++i ) mean_gr[i]->Draw("LP");
  csc->Update();
}

void LiveScan(int moi){
  LiveScan(1,moi);
}


void LiveScan(){

  printf("Possible Uses of LiveScan:\n\n");
  printf(".x LiveScan.cpp(moi)\n");
  printf(".x LiveScan.cpp(nscans,moi)\n");
}



