void PVAnalysis(Int_t run, Int_t scan, Int_t moi, Float_t startVdd, Float_t stopVdd, Float_t stepVdd) {

  // analysis of PVTest
  // version 0.1 PWP 14.11.00
  // version 0.2 LE  19.02.00 - including BYPASS 13 and 14
  //                          - stricter definition of a good chip
  // version 0.3 BJG 30.04.02 - fix some memory leaks
  //
  // Analysis of data:
  // Every channel is classified as 
  //   - Good (98-102% occupancy)
  //   - Dead (0% occupancy)
  //   - Unknown (otherwise)
  // Chips with more than 120 good channels are classified as good, 
  // chips with more than 120 dead channels are classified as dead,
  // otherwise, chip is in unknown state.


  #ifdef DATA_DIRECTORY
    // do not reload parameters.h
  #else
    // must load parameters.h
  #include <parameters.h>
  #endif
  
  #define MAX_CONFIGS 20
  #define N_TRIGS     50

  Int_t module, link, chip, chan, i;
  char f_name[128], h_name[64];
  char mname[MAX_MODULES][32];

  Short_t chip_dead[MAX_LINKS][MAX_CHIPS];
  Short_t chip_good[MAX_LINKS][MAX_CHIPS];
  Short_t chip_error[MAX_LINKS][MAX_CHIPS];
  Short_t chip_status[MAX_CONFIGS][MAX_LINKS][MAX_CHIPS];

  Short_t desired_result[MAX_LINKS][MAX_CHIPS] = {1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1};

  
  Short_t module_status[MAX_CONFIGS];
  Short_t n_bad_chips, n_good_chips, n_missing_chips, overall_status;

  // Load other files needed by this one
  gROOT->LoadMacro("stan_util.cpp");

  // 
  // Ensure that the module name array is cleared to start
  //

  for (module=0; module<MAX_MODULES; module++){
    sprintf(mname[module]," ");
  }

  Float_t step;
  Float_t minVdd;
  Float_t maxVdd;
  Int_t n_configs;

  if(startVdd<stopVdd) {
    step = TMath::Abs(stepVdd);
    minVdd  = startVdd;
    maxVdd  = stopVdd;
    n_configs=(int)(TMath::Abs((startVdd-stopVdd)/stepVdd)+1);
  } else {
    step = TMath::Abs(stepVdd);
    minVdd  = stopVdd;
    maxVdd  = startVdd;
    n_configs=(int)(TMath::Abs((startVdd-stopVdd)/stepVdd)+1);
  }    

  // 
  // Load and process the data
  //

  // Find today's date - needed for file output

  TDatime *dt = new TDatime;
  Int_t date = dt->GetDate();
  Int_t the_time = dt->GetTime();
    
  // Load file

  sprintf(f_name,"%sstrun%d_%d.root",DATA_DIRECTORY,run,scan);
  TFile *strun = new TFile(f_name);

  if(strun==0){
    printf("Failed to open file %s - EXITING!!\n",f_name);
    return;
  }
  
  TObjString* time  = (TObjString*)strun->Get("Time");

  //
  //  loop over modules
  //

  Int_t start = 0;
  Int_t end   = 0;

  if(moi==-1){
    start = 0;
    end   = MAX_MODULES;

  } else {
    start = moi;
    end   = moi+1;
  }
  
  for (module=start; module<end; module++){

    for(link=0;link<MAX_LINKS;link++){
      for(chip=0;chip<MAX_CHIPS;chip++){
        chip_dead[link][chip]=0;
        chip_good[link][chip]=0;
        for(i=0;i<MAX_CONFIGS;i++){
          chip_status[i][link][chip]=-1;
        }
      }
    }

    for(link=0;link<MAX_LINKS;link++){

      // find histograms

      sprintf(h_name,"ModuleName;%d",module+1);
      TObjString* name = (TObjString*)strun->Get(h_name);
      if(name) {
        sprintf(mname[module],"%s",name->GetName());
        delete name;
      }

      sprintf(h_name,"h_scan%d;%d",link,module+1);
      TH2F *h_scan= (TH2F*)strun->Get(h_name);

      if(h_scan){
        if(h_scan->GetMean()>0){
          printf("Analysing data for module %d link %d\n",
          module, link);

          // Loop over each tested configurations
          for(i=0;i<n_configs;i++){

            for(chip=0;chip<MAX_CHIPS;chip++){
              chip_dead[link][chip]=0;
              chip_good[link][chip]=0;
            }

            // Categorise each channel
            for(chan=0;chan<MAX_CHANS;chan++){
              chip = chan/128;

              Float_t occ = h_scan->GetBinContent((770*(i+1))+chan+1);

              if( (occ>-3)&&(occ<3) ){
		// allow +/-2 hits failing per channel
                chip_dead[link][chip]++;
              }else if( (occ>N_TRIGS-3)&&(occ<N_TRIGS+3) ) {
		// allow +/-2 hits failing per channel
                chip_good[link][chip]++;
              }else{
		chip_error[link][chip]++;
	      }
            } // end for(chan)


            // How many chips gave data?
            for(chip=0;chip<MAX_CHIPS;chip++){
              if(chip_good[link][chip]>=120){

                chip_status[i][link][chip]=1; // CHIP GAVE CORRECT DATA
		
              }else if(chip_dead[link][chip]>=120){

                chip_status[i][link][chip]=0; // CHIP DEAD
	      
	      }else{
		
		chip_status[i][link][chip]=2; // CHIP STATUS UNDEFINED
              }
            }
          } // end for(i)

        } // end if(mean>0) [histogram is not empty]
      } // end if(h_scan) [histogram exists]
    } // end for(link)

    // does each configuration work as expected?

    // Loop over configurations
    for(i=0;i<n_configs;i++){
      n_bad_chips=0;
      n_good_chips=0;
      n_missing_chips=0;
      for(link=0;link<MAX_LINKS;link++){
        for(chip=0;chip<MAX_CHIPS;chip++){
          if(chip_status[i][link][chip]==desired_result[link][chip]){
            n_good_chips++;
          }else if(chip_status[i][link][chip]==-1){
            n_missing_chips++;
          }else{
            n_bad_chips++;
          }
        }
      }

      if((n_bad_chips+n_good_chips)==0){
        module_status[i] = -1;  // NO DATA

      }else if(n_good_chips==12){
        module_status[i] =  1; // PASS

      }else{
        module_status[i] =  0; // FAIL
      }

    }
    // And finally, has the module passed?
  
    overall_status=0;
    for(i=0;i<n_configs;i++){
      if(module_status[i]==1){
        overall_status++;
      }
    } 

    if(module_status[0]>-1){ // Do not display or save results if no data!!


      //
      // Results to screen
      //

      printf("\n  Key to read the table.\n");
      printf(" '1' > 120 channels/chip returns 98 - 102%% occupancy\n");
      printf(" '0' > 120 channels/chip returns -2 - 2%% occupancy\n");
      printf(" '2' otherwise\n\n");

      if(overall_status==n_configs){
        printf("  Module %d (%s) has PASSED\n",module,mname[module]);
      }else{
        printf("  Module %d (%s) has FAILED\n",module,mname[module]);
      }

      printf("            pass  M0  S1  S2  S3  S4  E5  M8  S9 S10 S11 S12 E13\n");

      for(i=0;i<n_configs;i++){

        if(module_status[i]){
          printf("  Vdd %1.1f PASS  ", minVdd+(i*step));
        }else{
          printf("  Vdd %1.1f FAIL  ", minVdd+(i*step));
        }

        for(link=0;link<MAX_LINKS;link++){
          for(chip=0;chip<MAX_CHIPS;chip++){
            printf("%2d  ",chip_status[i][link][chip]);

          } // end for(chip)
        } // end for(link)
        printf("\n");
      } // end for(config)


      //
      // Results to file
      //

      Short_t Pass;
      if(overall_status==n_configs){
        Pass = 1;
      }else{
        Pass = 0;
      }

      FILE* outfile = OpenResults("PVTest", strun, module, run, scan, Pass, 0);
      
      if(outfile==NULL){
        printf("Could not open result file %s\n",f_name);
        return;
      } else {

        fprintf(outfile,"#\n");
        fprintf(outfile,"%%NConfigs  %d\n",n_configs);
        fprintf(outfile,"#\n");

        fprintf(outfile,"#\n");
        fprintf(outfile,"#cfg  pass  M0  S1  S2  S3  S4  E5  M8  S9 S10 S11 S12 E13\n");

        for(i=0;i<n_configs;i++){
          fprintf(outfile,"%1.1f     %1d", minVdd+(step*i), module_status[i]);

          for(link=0;link<MAX_LINKS;link++){
            for(chip=0;chip<MAX_CHIPS;chip++){
              fprintf(outfile,"  %2d",chip_status[i][link][chip]);
            } // end for(chip)
          } // end for(link)
          fprintf(outfile,"\n");
        } // end for(config)
        fclose(outfile);

      } // end if(outfile)


      Float_t vdd = 4.0;
      for(i=n_configs-1;i>=0;i--){
        if(module_status[i]==1){
          vdd = minVdd+(step*i);
        }
      }
      // Don't set Vdd smaller than 4.0
      if(vdd<4.0) vdd = 4.0;

      printf("module %s minimum Vdd %1.1f\n",mname[module],vdd);
      e->ConfigureVariable(module,33,vdd);


    }
  } // end(module)

  if(time) delete time;

  strun->Close();
  delete strun;
}


void PVAnalysis(Int_t run, Int_t scan, Int_t module){

  PVAnalysis(run,scan,module,4.5,3.8,-0.1);
}

void PVAnalysis(Int_t run, Int_t scan){

  PVAnalysis(run,scan,-1,4.5,3.8,-0.1); 
}
