/*
  RD13 Map/Unmap Emulation Library for the National Instruments VME-VXI System

  Gareth Moorhead  970409
*/

#define  DEBUG

#define	 VXINT
#include <nivxi.h>

#include <stdio.h>
#include <time.h>
#include <math.h>
#include <memory.h>

#include "../include/rd13lib.h"
#include "../include/rd13liberrs.h"

#define Em_RD13_MAX 256
unsigned int Em_RD13_BASES[Em_RD13_MAX];

int Em_RD13_ACTIVE = 0;
void Em_RD13_DumpMaps();

/***********************************************************/
/*                                                         */
/*    RD13_VmeMap(phys, am, size, base, cage)              */
/*    ---------------------------------------              */
/*                                                         */
/*                   - T C / I X -                         */
/*                                                         */
/*    Maps a VME device in the RAID Master Interface.      */
/*                                                         */
/*       phys  = external VME address to map.              */
/*       am    = Address Modifier.                         */
/*       size = Dimension of the addressing space mapped  */
/*               on the VME.                               */
/*       base  = Base contains the pointer to the internal */
/*               base address of the space addressed in    */
/*               VME.                                      */
/*       cage  = dummy parameter to be compatible with SUN */
/*                                                         */
/*    It returns:                                          */ 
/*       RD13_SUCCESS   - Success                          */
/*       RD13_TOO_PAGES - Too many pages has been mapped.  */
/*                        There are no available pages.    */
/*       RD13_INVAM     - Bad address modifier.            */
/*       RD13_AM_MISM   - too much space required          */
/*                        according to the am              */
/*                                                         */
/***********************************************************/

int RD13_VmeMap(char *phys, int am, int size, char **base, int cage) {

	uint16	access;     /* Access parameters */
	uint32	address;    /* VME address */
	int32	timo;       /* Timeout period (ms) */
	uint32	window;     /* MapVXI memory window id */
	int16	ret;		/* Return status */
	void*	addr;		/* Unix address */


#ifdef DEBUG
	printf("Em_RD13_VmeMap: Em_RD13_ACTIVE %8x\n", Em_RD13_ACTIVE);
#endif

	if (Em_RD13_ACTIVE != 0x1111) {

#ifdef DEBUG
		printf("Em_RD13_VmeMap: calling InitVXIlibrary\n");
#endif
		Em_RD13_ACTIVE = 0x1111;
		if (ret = InitVXIlibrary() < 0) {
			printf("Em_RD13VmeMap: ERROR %d from InitVXIlibrary\n", ret);
			return(ret);
		}
	}

 
	if ((am & 0x30) == 0x30) { access = 2; }        /* Standard A24 */
	else if ((am & 0x30) == 0x20) { access = 1; }   /* Short A16 */
	else if ((am & 0x30) == 0x00) { access = 3; }   /* Extended A32 */
	else access = 2;
	
	address = (uint32) phys;
	
	timo = 5000L;

#ifdef DEBUG
	printf("Em_RD13_VmeMap: calling MapVXIAddress for address %8x access %2x\n",
		address, access);
#endif

	addr = MapVXIAddress( access, address, timo, &window, &ret);

#ifdef DEBUG
	printf("Em_RD13_VmeMap: MapVXIAddress ret %d win %d addr %8x for vme %8x\n", 
		ret, window, addr, address);
#endif

	*base = addr;

	if (window < Em_RD13_MAX) {
		Em_RD13_BASES[window] = (unsigned int) addr;
	}

#ifdef DEBUG
	printf("Em_RD13_VmeMap: ending with return code %x\n", ret);
#endif

	return(ret);  /* ### should translate error codes correctly */

}


/***********************************************************/
/*                                                         */
/*    RD13_VmeUnmap(base)                                  */
/*    -------------------                                  */
/*                                                         */
/*    Unmap the VME space in to a user application         */
/*                                                         */
/*       base = Pointer to the base of the internal space  */
/*              mapped.                                    */
/*                                                         */
/*    RETURN:                                              */
/*    RD13_SUCCESS    - Success                            */
/*    RD13_NOT_MAPPED - No mapped Space with base          */
/*                                                         */
/***********************************************************/

int RD13_VmeUnmap(char *base) {

	uint32	window = 0;  /* find correct VXI window from base */
	int16	ret;

	while ((window<Em_RD13_MAX) && ((unsigned int) base != Em_RD13_BASES[window])) 
		window++;

#ifdef DEBUG
	printf("Em_RD13_VmeUnmap: base %8x returns window %d\n", (unsigned int) base, window);
#endif

	if (window < Em_RD13_MAX) {
		ret = UnMapVXIAddress(window);
		return(ret);  /* ### should translate error codes correctly */
	} else {
#ifdef DEBUG
		printf("Em_RD13_VmeUnmap: ERROR: Unable to find address in window table\n");
		Em_RD13_DumpMaps();
#endif 	
		return(-1);
	}

}

void Em_RD13_DumpMaps(){
	int n = 0;
	printf("Em_RD13_DumpMaps: dumping all non-zero entries\n");
	while (n < Em_RD13_MAX) {
		if (Em_RD13_BASES[n] != 0) printf("%3x %8x\n", n, Em_RD13_BASES[n]);
		n++;
	}
	printf("Em_RD13_DumpMaps: finished\n");
}
