/*
 * scthv_lib.c.
 * Library of routines to handle SCTHV 
 *
 * Some assumptions:
 * The sysmap contains two variables regarding HV cards:
 * short hv_id  =>   card number
 * short hv_ch  =>   channel number, 0-3
 * 
 * in the context of VME based HV cards, this seems to be
 * a more sensible use of these variables their original
 * usage (crate number (=0), channel number 0-16)
 *
 * Limitation: cards must be numbered consecutively
 *
 * Version 0.20 PWP 27.07.01 
 * Version 0.21 PWP 08.12.00 - suit revised cards
 * Version 0.22 PWP 12.06.01 - detect presence of cards
 *
 * Version 0.30 PWP 12.07.01 - Many revisions, including:
 *                             o track which cards are switched on and suppress
 *                                 ramp / staus commands to cards which are off
 *                             o detect "hung" state during status request
 *                             o wait for correct first byte during status request
 *                             o sanitised error codes
 *                             o new routines sctlv_print_error and sctlv_ison
 *                             This version is much more reliable!!
 * Version 0.40 PWP 31.01.02 - o Keep record of expected state of each channel
 *                             o Keep record of requested vset, ilim and ramp rate
 *                             o Use this information, together with monitor
 *                               data, to identify certain error conditions
 *                             o Detect "STUCK" condition if monitored values
 *                               remain unchanged for more than 5 minutes
 *                             o Enforce minimum time intervals between monitor
 *                               and ramp requestsramp and bn
 *                             o add routine to resend last requested ramp command
 * Version 0.41 PWP 27.03.02 - bugfix affecting monitoring multiple SCTHV channels
 *                             (put back short delay that had been commented out)
 * Version 0.42 PWP 28.05.02 - merge revisions by Carlos for operation with UNIX
 * Version 0.50 PWP 10.06.02 - move the low level functionality from scthv_monitor 
 *                             to scthv_get_status.  The low level code may now be 
 *                             called by scthv_monitor more than once, and indeed it 
 *                             is if a trip has been reported.  Between the two trials 
 *                             the ramp command is resent to reset the communication 
 *                             between the master and channel processors on SCTHV.  
 *                             This is in response to reports of "phantom" HV trips 
 *                             occuring during module long term tests.  Note that 
 *                             if the trip is real, the HV channel must be switched 
 *                             OFF to recover the trip.
 * Version 0.51 JCH 24.06.02 - additions for ATLAS RCC library support (for SBCs)              
 */

/* includes and defines */

#include "scthv_lib.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifdef NIVXI
#include <nivxi.h>
#endif

#ifdef OS9
#include <process.h>
#include <signal.h>
#include <types.h>
#endif

#ifdef RAID
#define CLOCKS_PER_SEC 100000
#endif

#ifdef QB_linux
#define WND_SIZE 0x1000
#include "../QB/InitVME.h"
#endif

#ifdef RCC
#include <rcc_error.h>
#include <vme_rcc.h>
#endif

/* Global variables */
short  scthv_total = 0;                    /* Total number of SCTHV cards  */
struct scthv_info_structure *scthv_info;   /* structure to hold SCTHV info */

#ifdef unix
/*#include <sys/time.h>*/
#include <unistd.h>
#undef CLOCKS_PER_SEC
#define CLOCKS_PER_SEC 1000000
static clock_t _clock(void) {
   static unsigned long seconds=0;
   struct timeval now;
   clock_t ret;
   gettimeofday (&now, NULL);
   if ( seconds == 0 )
      seconds = now.tv_sec;
   ret = (now.tv_sec-seconds)*CLOCKS_PER_SEC + now.tv_usec;
   return ret;
}
#else
#define _clock clock
#endif

/* Functions */

int scthv_exit() {
  /*
   * scthv_exit.
   * Exit routine for SCTHV library. 
   * Should be called once only to tidy things up.
   * Unmaps the modules and deletes the malloced memory.
   */

  int i;
#ifdef RCC
  VME_ErrorCode_t status;
#else
  int status;
#endif

/* loop over the modules */
  for(i=0;i<scthv_total;i++) {
#ifdef RAID
    status = RD13_VmeUnmap(scthv_info[i].address);
#endif
#ifdef NIVXI
    status = UnMapVXIAddress(scthv_info[i].window);
#endif
#ifdef QB_linux
    munmap((void *)(scthv_info[i].address+1),WND_SIZE);
#endif 
#ifdef RCC
    status = VME_MasterUnmap(scthv_info[i].window);
#endif

  }  /*end of loop*/

#ifdef NIVXI
    CloseVXIlibrary();
#endif
#ifdef RCC
    status = VME_Close();
#endif

/* release memory */
  free((void *)scthv_info);
  return 0;
}


int scthv_ison(short card, short chan) {
  /*
   * Establish if the requested HV channel has been turned on
   * Return -1 if card invalid
   * Return -2 if card missing
   * Return -3 if chan invalid
   * Return 0 if channel OFF
   * Return 1 if channel ON
   */

  if((card<0)||(card>=scthv_total)) return -1;
  if(scthv_info[card].missing) return -2;
  if((chan<0)||(chan>=4)) return -3;

  if(scthv_info[card].state[chan]==CH_OFF){
    return 0;
  }else{
    return 1;
  }
}


int scthv_off(short card, short chan) {
  /*
   * Turn off the requested HV channel
   * Return -1 if card invalid
   * Return -2 if card missing
   * Return -3 if chan invalid
   */

  volatile unsigned char *ip;

  if((card<0)||(card>=scthv_total)) return -1;
  if(scthv_info[card].missing) return -2;
  if((chan<0)||(chan>=4)) return -3;

#ifdef HV_DEBUG
  printf("SCTHV: turning OFF card %d chan %d\n",card,chan);
#endif

  /* Enforce minimum time interval of one second between commands */
  while(_clock() < (scthv_info[card].last_update[chan] + CLOCKS_PER_SEC) );
  scthv_info[card].last_update[chan]= _clock();

  ip = (unsigned char *)(scthv_info[card].address+0x000D);

  *ip = (unsigned char) 0x01;
  *ip = (unsigned char) chan;

  /*
   * Keep track that this channel has been switched off:
   * ramp and status requests must not be sent to channels
   * which have not been switched on since this will mess up
   * the communication between the five on board processors
   */

  scthv_info[card].state[chan]=CH_OFF;

  /* Print summary of detected error conditions */

  printf("SCTHV: card %d chan %d nerror %d nrepeat %d nsame %d nstuck %d\n", card, chan,
    scthv_info[card].nerror[chan], scthv_info[card].nrepeat[chan], 
    scthv_info[card].nsame[chan], scthv_info[card].nstuck[chan]);

  return 0;
}


int scthv_on(short card, short chan) {
  /*
   * Turn on the requested HV channel
   * Return -1 if card invalid
   * Return -2 if card missing
   * Return -3 if chan invalid
   */

  volatile unsigned char *ip;

  if((card<0)||(card>=scthv_total)) return -1;
  if(scthv_info[card].missing) return -2;
  if((chan<0)||(chan>=4)) return -3;

#ifdef HV_DEBUG
  printf("SCTHV: turning ON card %d chan %d\n",card,chan);
#endif

  /* Enforce minimum time interval of one second between commands */
  while(_clock() < (scthv_info[card].last_update[chan] + CLOCKS_PER_SEC) );
  scthv_info[card].last_update[chan]= _clock();

  ip = (unsigned char *)(scthv_info[card].address+0x000D);

  *ip = (unsigned char)  0x01;
  *ip = (unsigned char) (0x10 + chan);

  /*
   * Keep track that this channel has been switched on:
   * ramp and status requests must not be sent to channels
   * which have not been switched on since this will mess up
   * the communication between the five on board processors
   */

  if(scthv_info[card].state[chan]==CH_OFF){
    scthv_info[card].state[chan]=CH_ON;
  }
  return 0;
}


int scthv_get_status(short card, short chan, 
                    float* vmon, float* imon, short* status) {
  /*
   * Monitor voltage, current and status of the requested HV channel
   * - low level portion including VME access and data conversion -
   * Return -1 if card invalid
   * Return -2 if card missing
   * Return -3 if chan invalid
   * Return -4 if chan not switched on
   * Return -8 if timeout waiting to read status
   * Return -9 if minimum interval between commands has not elapsed
   */

  int i,j, nloop, vmon_int, imon_int;
  volatile unsigned char *ip;
  unsigned char nbyte;
  unsigned char sbyte;
  unsigned char vbyte[2];
  unsigned char ibyte[2];
  unsigned char tbyte[4];

  *vmon = (float)0;
  *imon = (float)0;
  *status = (short)0;

  if((card<0)||(card>=scthv_total)) return -1;
  if(scthv_info[card].missing) return -2;
  if((chan<0)||(chan>=4)) return -3;
  if(scthv_info[card].state[chan]==CH_OFF) return -4;

  /*
   * Only pass the status query if more than one second has elapsed
   * since the last attempt, otherwise return the previous data
   */
  if(_clock() < (scthv_info[card].last_update[chan] + CLOCKS_PER_SEC) ){

    i = (int)(scthv_info[card].status[chan] &0x03);
    *vmon = (float)(scthv_info[card].vmon[chan] * V_STEP);
    *imon = (float)(scthv_info[card].imon[chan] / i_step[i]);
    *status = (short) scthv_info[card].status[chan];
    return -9;
  }


  /* request data */

  ip = (unsigned char *)(scthv_info[card].address+(2*chan)+1);
  *ip = 0x01;

  ip = (unsigned char *)(scthv_info[card].address+0x000D);
  *ip = (unsigned char)(0x30 + (unsigned char) chan);


  /*
   * read back the data:
   * During tests with a single module, performance seemed to be 
   * more reliable without the delay loop - but to run with multiple
   * modules the delay is required.  PWP 27.03.02
   */

  j=0;
  nbyte=0;
  nloop=0;
  while((nbyte!=9)&&(nloop<=HV_MAX_LOOPS)){
    for (i=0;i<2000;i++){
      j++;
    }
    nbyte = *ip;
    nloop++;
  }

  sbyte = *ip;
  vbyte[HI] = *ip;
  vbyte[LO] = *ip;
  ibyte[HI] = *ip;
  ibyte[LO] = *ip;

  /* four test bytes - meaning unspecified */

  tbyte[0]  = *ip;
  tbyte[1]  = *ip;
  tbyte[2]  = *ip;
  tbyte[3]  = *ip;
	
  /* convert to "real" units */

  vmon_int = ((vbyte[HI]<<8)+vbyte[LO]);
  imon_int = ((ibyte[HI]<<8)+ibyte[LO]);

  i = (int)(sbyte &0x03);

  *vmon = (float)(vmon_int * V_STEP);
  *imon = (float)(imon_int / i_step[i]);
  *status = (short) sbyte;

  /*
   * We record the time at which this update was requested. If new data
   * was returned, we also update the time of the last new data event
   * and some aspects of the returned data itself.
   */

  scthv_info[card].last_update[chan]= _clock();

  if((scthv_info[card].vmon[chan] != vmon_int) ||
     (scthv_info[card].imon[chan] != imon_int) ||
     (scthv_info[card].status[chan] != sbyte)){

    scthv_info[card].diff_update[chan]= scthv_info[card].last_update[chan];

    scthv_info[card].vmon[chan] = vmon_int;
    scthv_info[card].imon[chan] = imon_int;
    scthv_info[card].status[chan] = sbyte;

    scthv_info[card].nsame[chan]=0;
  }else{
    scthv_info[card].nsame[chan]++;
  }

#ifdef HV_DEBUG
  printf("nloops %d 0x%2x %2x %2x %2x %2x %2x %2x %2x %2x %2x\n",
  nloop,nbyte,sbyte,vbyte[HI],vbyte[LO],ibyte[HI],ibyte[LO],tbyte[0],tbyte[1],tbyte[2],tbyte[3]);
#endif

  if(nloop>HV_MAX_LOOPS)   return -8;
  return 0;
}


int scthv_monitor(short card, short chan, 
                  float* vmon, float* imon, short* status) {
  /*
   * Monitor voltage, current and status of the requested HV channel
   * - high level portion including error detection and checks on trip validity -
   * Return -1 if card invalid
   * Return -2 if card missing
   * Return -3 if chan invalid
   * Return -4 if chan not switched on
   * Return -8 if timeout waiting to read status
   * Return -9 if minimum interval between commands has not elapsed
   * Return  1 if channel tripped due to overcurrent
   * Return  2 if channel tripped due to overvoltage
   * Return  3 if channel has frozen during ramping
   * Return  4 if channel is ramping
   * Return  5 if a channel expected to be STEADY, reports a strange voltage
   */

  int error, rcode;

  rcode = scthv_get_status(card, chan, vmon, imon, status);
  if(rcode) return rcode;


  /*
   * Sometimes the communication between SCTHV's main processor
   * and the processor responsible for a given channel can fail.
   * Dependent upon the expected state of the channel, we make
   * some simple checks to spot error conditions. If an error
   * condition has been detected, recovery may usually be
   * effected by resending the last ramp request.
   */


  /* Identify tripped channels */
  if((*status&HV_OVERCURRENT)||(*status&HV_OVERVOLTAGE)){
    if(scthv_info[card].state[chan]!=CH_TRIPPED){
      scthv_info[card].state[chan]=CH_TRIP_PENDING;
    }
  }

  /* Take further action appropriate to the expected channel state */
  error=0;

  switch(scthv_info[card].state[chan]){
    case CH_OFF:
      break;

    case CH_ON:
      /*
       * The channel is known to have been switched on, but
       * otherwise its expected status is not known.  This can only
       * be rectified by sending a ramp command - otherwise the
       * following error detection logic will get rather confused!
       */
      break;

    case CH_RAMPING:
      /*
       * The channel is expected to be ramping.
       * If the time of the last update which returned new data is not the
       * same as the time of the last update, then the channel is stuck.
       */

      if(*status&HV_RAMPING){
        if(scthv_info[card].last_update[chan] != scthv_info[card].diff_update[chan]){
          error=1;
        }
      }else{
        scthv_info[card].state[chan]=CH_STEADY;
      }
      break;

    case CH_STEADY:
      /*
       * The channel is expected to be giving steady output.
       * If vmon is more than 5V different from vset we assume 
       * that an error has occurred.
       */

      if( (*vmon<(scthv_info[card].vset[chan] - 5.0)) ||
          (*vmon>(scthv_info[card].vset[chan] + 5.0)) ){
        error=2;
        scthv_info[card].nerror[chan]++;
      }

      /*
       * The intended validity check of the imon data was actually
       * a null test - so I removed the code. PWP 10.06.02
       */

      break;

    case CH_TRIP_PENDING:
      /*
       * The channel has reported a trip, but this may not be real. We issue
       * a repeat ramp command in an attempt to establish the truth...
       */

      printf("SCTHV: card %d chan %d trip pending. Verification in progress...\n",card,chan);
      scthv_repeat_request(card,chan); 

      /* Enforce 2 second time interval for SCTHV to process our request */
      scthv_info[card].last_update[chan]=_clock();
      while(_clock() < (scthv_info[card].last_update[chan] + (2*CLOCKS_PER_SEC)) );

      rcode = scthv_get_status(card, chan, vmon, imon, status);
      if(rcode) return rcode;

      if((*status&HV_OVERCURRENT)||(*status&HV_OVERVOLTAGE)){
        printf("SCTHV: card %d chan %d TRIP CONFIRMED\n",card,chan);
        scthv_info[card].state[chan]=CH_TRIPPED;
      }else{
        printf("SCTHV: card %d chan %d false alarm...\n",card,chan);
        scthv_info[card].state[chan]=CH_RAMPING;
      }
      break;

    case CH_TRIPPED:
      /*
       * The channel has repeatedly reported a trip.
       * Time to give up.
       */
       break;
  }


  /*
   * We now make two tests for the "stuck" condition: has the
   * returned data remained constant over more than HV_NSAME
   * samples spaced over more than HV_TSTUCK seconds?
   */

  if( (scthv_info[card].nsame[chan]>HV_NSAME) && (scthv_info[card].last_update[chan] >
      (scthv_info[card].diff_update[chan] + (HV_TSTUCK*CLOCKS_PER_SEC)) ) ){
    error=1;
    scthv_info[card].nstuck[chan]++;
  }


#ifdef HV_DEBUG
  printf("SCTHV: status for card %d chan %d\n",card,chan);
  printf("       ints:   V %d,    I %d,    status 0x%2x, state %d error %d\n", 
          vmon_int, imon_int, status, scthv_info[card].state[chan], error);
  printf("       floats: V %3.2f,   I %4.2f\n", *vmon, *imon);
#endif

  /* return codes */



  if(*status&HV_OVERCURRENT) return 1;
  if(*status&HV_OVERVOLTAGE) return 2;
  if(error==1)               return 3;
  if(*status&HV_RAMPING)     return 4;
  if(error==2)               return 5;
  
  return 0;
}


void scthv_print_error(int rcode, short card, short chan){
  /*
   * scthv_print_error
   * Print explanation of the error code to the console window
   */

  switch(rcode){
    case  5: /* wrong volts - corrupt monitor data */
      printf("SCTHV ERROR: card %d chan %d wrong volts - corrupt monitored data\n",card,chan);
      break;
    case  4: /* channel is ramping */
      printf("SCTHV card %d chan %d is ramping\n",card,chan);
      break;
    case  3: /* channel seems to be stuck during ramping */
      printf("SCTHV ERROR: card %d chan %d is stuck\n",card,chan);
      break;
    case  2: /* channel tripped due to overvoltage */
      printf("SCTHV ERROR: card %d chan %d tripped due to overvoltage\n",card,chan);
      break;
    case  1: /* channel tripped due to overcurrent */
      printf("SCTHV ERROR: card %d chan %d tripped due to overcurrent\n",card,chan);
      break;
    case 0:  /* no error */
      break;
    case -1: /* card invalid */
      printf("SCTHV ERROR: card %d invalid\n",card);
      break;
    case -2: /* card missing */
      printf("SCTHV ERROR: card %d missing\n",card);
      break;
    case -3: /* chan invalid */
      printf("SCTHV ERROR: card %d chan %d invalid\n",card,chan);
      break;
    case -4: /* chan not switched on */
      printf("SCTHV ERROR: card %d chan %d not switched on\n",card,chan);
      break;
    case -5: /* vset out of range */
      printf("SCTHV ERROR: card %d chan %d requested vset out of range",card,chan);
      break;
    case -6: /* ilim out of range */
      printf("SCTHV ERROR: card %d chan %d requested ilim out of range",card,chan);
      break;
    case -7: /* ramp out of range */
      printf("SCTHV ERROR: card %d chan %d requested ramp out of range",card,chan);
      break;
    case -8: /* timeout waiting to read status */
      printf("SCTHV ERROR: card %d chan %d timeout waiting to read status",card,chan);
      break;
    case -9: /* monitor interval not elapsed */
      printf("SCTHV ERROR: card %d chan %d monitor interval not elapsed",card,chan);
      break;
    default:
      printf("SCTHV ERROR: unknown error %d",rcode);
      break;
  }

}


int scthv_print_status(short card, short chan){

  /* Request and print V, Imon and status */

  float vmon, imon;
  short status;
  unsigned short probe;
  unsigned short ramp = 0;
  unsigned short itrip = 0;
  unsigned short vtrip = 0;
  int rcode; 

  rcode = scthv_monitor(card, chan, &vmon, &imon, &status);

  if(rcode ==-1){
    printf("scthv_status: card %d invalid\n",card);
    return -1;
  }
  if(rcode ==-2){
    printf("scthv_status: card %d is missing\n",chan);
    return -2;
  }
  if(rcode ==-3){
    printf("scthv_status: channel %d invalid\n",chan);
    return -3;
  }

  probe = (status&0x03);
  if(status&HV_RAMPING) ramp =1;
  if(status&HV_OVERCURRENT) itrip =1;
  if(status&HV_OVERVOLTAGE) vtrip =1;

  printf("HV id ch prob ramp itrip  vtrip  vmon    imon  stat rc vset   state\n");
  printf("   %2d %2d  %2d    %1d    %1d      %1d    %3.2f  %4.2f  0x%02x %2d %3.2f %s\n",
          card, chan, probe, ramp, itrip, vtrip, vmon, imon, status, rcode,
          scthv_info[card].vset[chan], states[(scthv_info[card].state[chan])]);

  return rcode;
}


int scthv_repeat_request(short card, short chan) {
  /*
   * Repeat the last ramp request.
   * Useful to recover from communication errors
   * between the processors on the SCTHV card.
   * added PWP 23.01.02
   */

  int rcode;

  rcode = scthv_request(card, chan, scthv_info[card].vset[chan],
                                    scthv_info[card].ilim[chan],
                                    scthv_info[card].ramp[chan]);

  scthv_info[card].nrepeat[chan]++;
  return rcode;
}


int scthv_request(short card, short chan, 
                  float vset, float ilim, short ramp) {
  /*
   * Program desired voltage, current limit and ramp rate of the requested HV channel
   * Return -1 if card invalid
   * Return -2 if card missing
   * Return -3 if chan invalid
   * Return -4 if chan not switched on
   * Return -5 if vset out of range
   * Return -6 if ilim out of range
   * Return -7 if ramp out of range
   */

  int i, vset_int;
  int ilim_int[4];
  volatile unsigned char *ip;

  unsigned char vbyte[2];
  unsigned char ibyte[4][2];
  unsigned char rbyte;

#ifdef HV_DEBUG
  printf("SCTHV: request for card %d chan %d V %f I %f r %d\n",
	  card, chan, vset,ilim,ramp);
#endif

  if((card<0)||(card>=scthv_total)) return -1;
  if(scthv_info[card].missing) return -2;
  if((chan<0)||(chan>=4)) return -3;
  if(scthv_info[card].state[chan]==CH_OFF) return -4;

  if((vset<0)||(vset>500.0)) return -5;
  if((ilim<0)||(ilim>5000)) return -6;
  if((ramp<0)||(ramp>4)) return -7;

  /*
   * Having got this far, we store the requested values
   */

  scthv_info[card].vset[chan] = vset;
  scthv_info[card].ilim[chan] = ilim;
  scthv_info[card].ramp[chan] = ramp;

  /*
   * Conversion of "real" units to DAC and ADC bits.
   * The current trip limit is calculated four times,
   * once for each probe resistor.
   */

  vset_int = (int)((vset+(V_STEP/2)) / V_STEP);
  if(vset<20.0) vset_int++;  /* why? */

  if(vset_int>4000) vset_int=4000; /* extra protection */ 
  vbyte[HI] = (unsigned char) (vset_int/256);
  vbyte[LO] = (unsigned char) (vset_int%256);

#ifdef HV_DEBUG
  printf("SCTHV: new request to card %d chan %d\n",card,chan);
  printf("       Vfloat %3.2f,   Ilim %4.2f, Vint %d \n", vset, ilim, vset_int);
#endif

  for(i=0; i<4; i++){
    ilim_int[i] = (int)(ilim*i_step[i]);

    if(ilim_int[i]>0x0fff) ilim_int[i]=0x0fff; /* 12 bit ADC */ 
    ibyte[i][HI] = (unsigned char) (ilim_int[i]/256);
    ibyte[i][LO] = (unsigned char) (ilim_int[i]%256);

#ifdef HV_DEBUG
  printf("Iint[%d] %d  \n", i, ilim_int[i]);
#endif
  }

  rbyte = (unsigned char) (ramp&0x07);

#ifdef HV_DEBUG
  printf(" ramp %d %d\n", ramp, rbyte);
#endif


  /* Enforce minimum time interval of one second between commands */
  while(_clock() < (scthv_info[card].last_update[chan] + CLOCKS_PER_SEC) );
  scthv_info[card].last_update[chan]=_clock();


  /* Send request */

  ip = (unsigned char *)(scthv_info[card].address+(2*chan)+1);
  *ip = 0x0C;

  ip = (unsigned char *)(scthv_info[card].address+0x000D);

  *ip = (unsigned char)(0x20 + chan);
  *ip = vbyte[HI]; 
  *ip = vbyte[LO]; 
  *ip = ibyte[0][HI]; 
  *ip = ibyte[0][LO]; 
  *ip = ibyte[1][HI]; 
  *ip = ibyte[1][LO]; 
  *ip = ibyte[2][HI]; 
  *ip = ibyte[2][LO]; 
  *ip = ibyte[3][HI]; 
  *ip = ibyte[3][LO]; 
  *ip = rbyte; 
 
  scthv_info[card].state[chan]=CH_RAMPING;
  scthv_info[card].nsame[chan]=0;

  return 0; 
}


int scthv_vmemap(unsigned long address, short *card) {
  /*
   * Map SCTHV to the CPU address space. The input address is the true
   * VME address for UNIX and WNT implementations, mapped address for OS9 (this
   * is because the mapping in OS9 is CPU-specific and so it is not possible to
   * provide a generalised mapping within this routine). The mapping in UNIX/WNT
   * is A24.
   */

#if defined(NIVXI)
  char dta;
  NIVXI_STATUS ret;
  UINT32 w,wstart, wend;
#elif defined(RCC)
  u_char data;
  VME_ErrorCode_t status;
  VME_MasterMap_t master_map;
  u_int virtual_address;
  int handle;
#else
  int status;
#endif
  int am;
  char *addr;

  short chan;

  am = 0x9;
  *card = scthv_total++;		/*Simplest approach? */

/* Use malloc/realloc to get space for the structure */
  if(scthv_total == 1)
    scthv_info = (struct scthv_info_structure *) malloc(scthv_info_size);
  else
    scthv_info = (struct scthv_info_structure *) realloc((void *)scthv_info,scthv_total*scthv_info_size);

  scthv_info[*card].address = address;
  scthv_info[*card].missing = 0;

  for(chan=0;chan<4;chan++){  
    scthv_info[*card].vmon[chan] = 0;
    scthv_info[*card].imon[chan] = 0;
    scthv_info[*card].status[chan] = 0;

    scthv_info[*card].vset[chan] = 0;
    scthv_info[*card].ilim[chan] = 0;
    scthv_info[*card].ramp[chan] = 4;

    scthv_info[*card].state[chan] = CH_OFF;

    scthv_info[*card].nerror[chan]=0;
    scthv_info[*card].nrepeat[chan]=0;
    scthv_info[*card].nsame[chan]=0;
    scthv_info[*card].nstuck[chan]=0;

    scthv_info[*card].last_update[chan]=0;
    scthv_info[*card].diff_update[chan]=0;
  }

#ifdef NIVXI
  scthv_info[*card].window = 0;
#endif

/* Now do the mapping - CPU-dependent - allow a 0x1000 window */

#ifdef OS9
  status = _os_permit((void *)address, 0x1000,2,0);
  if(status != 0) {
     printf("VME mapping failed, status = %d\n",status);
     return status;
  }
#endif

#ifdef RAID
  am = 0x39;		/* A24 standard access */
  status = RD13_VmeMap(address,am,0x1000,&addr,0);
  if(status != 0) {
     printf("VME mapping failed, status = %d\n",status);
     return status;
  }
  scthv_info[*card].address = (unsigned long) addr;
#endif

#ifdef QB_linux
  if(scthv_total == 1) {         /*only call once*/
    InitVME();
  }
  Initqb(&address ,WND_SIZE,VME_A24);
  scthv_info[*card].address = address-1;
#endif

#ifdef RCC
  if(scthv_total == 1) status = VME_Open();         /*only call once*/
/* Set up the map structure */
  master_map.vmebus_address = address;
  master_map.window_size = 0x1000;
  master_map.address_modifier = VME_A24;
  master_map.options = 0;
  status = VME_MasterMap(&master_map, &handle);
  if(status != VME_SUCCESS) {
    VME_ErrorPrint(status);
    return (int)status;
  }
  status = VME_ReadSafeUChar(handle,0x000D,&data);
  if(status != VME_SUCCESS) {
    printf (" - rcode %d => MISSING\n",status);
    scthv_info[*card].missing = 1;
    return (int)status;
  }
  status = VME_MasterMapVirtualAddress(handle,&virtual_address);
  printf("scthv_vmemap: VME address %x mapped to %x, handle %d\n",
    address, virtual_address,handle);
  scthv_info[*card].address = (unsigned long) virtual_address;
  scthv_info[*card].window = handle;
#endif

#ifdef NIVXI  /* A24 */
  if(scthv_total == 1) {
    printf("scthv_lib: initialising VXI library\n");
    InitVXIlibrary();		 /*only call once*/
    MapVXIAddressSize(0xf0000);  /* put SCTHV and SCTLV cards inside one big window 
                                  PWP 07.08.00 */
  }

  ret = VXIin(2,address+0x000D,1,&dta);

  if(ret<0){
    printf (" - rcode %d => MISSING\n",ret);
    scthv_info[*card].missing = 1;
    return (int)ret;
  }

  addr = (char *)MapVXIAddress(2,address,5000,&w,&ret);
  if(ret != 0) {
    printf("VME mapping failed, status = %d\n",ret);
    return (int)ret;
  }
  printf("scthv_vmemap: MapVXIAddress ret %d win %d addr %8x for vme %8x\n", 
		ret, w, addr, address);
  ret = GetWindowRange(w, &wstart, &wend);
  printf("scthv_vmemap: WindowRange start %lx end %lx\n", wstart, wend);
  scthv_info[*card].address = (unsigned long) addr;
  scthv_info[*card].window = w;
#endif

  return 0;
}

