
/* abc.c Simple Control Block Generator for ABC(D)     GFM Nov98 */

/* GFM 17Mar2000 Remove soft resets from abc_configure() */
/* PWP 15.08.00 replace cal_enable and test_mode bits with trim_range */
/* PWP 26.09.00 revise code to allow oddity to follow the select line */
/* PWP 29.11.00 add forward_sc_command */
/* PWP 29.05.02 merge two new options into abc_command (from Carlos) */

#include <stdio.h>

#include "abc.h"
#include "abc_config.h"

/* In this version of a cb generator the bit patterns are manipulated 
   as 16-bit unsigned shorts, and to provide functions to aid conversion to,
   and merging of, individual 1-bit-per-module FIFO patterns.  I also prefer
   to align commands on 32 bit boundaries such that the arguments are aligned
   with the LSB of the argument matching the LSB of the word:  this results
   in lots of zero-padding (and extra length) but makes the arguments easier
   to read from simple hex dumps.  */

/* cb_cat():  concatenate two control blocks */

int cb_cat(
  unsigned short *a, unsigned int *al, unsigned short *b, unsigned int bl){

  unsigned int i,imax,j;

  imax = *al + bl;
  j = 0;

  for (i= *al; i<imax; i++){
    a[i] = b[j];
    j++;
  }

  *al = imax;

  return(0);
}

/* cb_rotate(): Convert array of 16-bits to 1-bit-per-word FIFO pattern.
   Set bits are duplicated in the FIFO word according to mask. */

int cb_rotate(   
  unsigned short *in, unsigned int inl, 
  unsigned short *out, unsigned int *outl, unsigned short mask){

  unsigned int i,b,w;

  for (i=0; i<inl; i++){
    w = i*16;
    for (b=0; b<16; b++){
      if ((in[i] >> (15-b)) & 0x1) out[w+b] = out[w+b] | mask;
    }
    *outl = inl * 16;
  }
  return(0);
}


/* cb_clear(): Force array to be empty.  This is important since
   it is necessary to ensure stop bits are cleared to the full cb length */

int cb_clear(unsigned short *cb, unsigned int size){
  unsigned int i;
  for (i=0; i<size; i++) cb[i] = 0x0000;
  return(0);
}



/* cb_merge(): Merge two bit-wise cb arrays.  Note, input control blocks
   are expected already to be masked correctly. One size parameter is
   used which should be set to the maximum of the input cb sizes. */

/* ### This is bad:  must pass both cb sizes in case they are different -
if different we need to stop reading the shorter one before we risk
overrunning its array size; we need to continue reading the longer one
into the output array.  Alternatively, we could just insist that both
input and the output control block array limits are greater than or
equal to the one size parameter */

int cb_merge(unsigned short *cb0, unsigned short *cb1, unsigned short *out, unsigned int size){
  unsigned int i;
  for (i=0; i<size; i++) out[i] = cb0[i] | cb1[i];
  return(0);
}



/* abc_configure(): Configure all registers in each and every non-missing chip */

int abc_configure(module_cfg *mcfg, unsigned short *cb, 
                  unsigned int *cb_size, short full_config){
  chip_cfg ccfg;
  unsigned short chip_cb_size;
  unsigned short config[4],thresh[4],bias[4],delay[4],mask[12];
  unsigned short c[12][28];
  unsigned short i,j,k,l;
  unsigned short oddity;    /* aaaa aa00 = 1daa aa00
              EVEN => oddity = 1000 0000 = 0x0800
              ODD  => oddity = 1100 0000 = 0x0c00; */

  unsigned short death[12],missing[12];

  /* enable  0000 0101 0111 0000 1100 1111 1110 1000 */
  /*            0    5    7    0    c all chips    8 */
  unsigned short enabledata[4] = {0x0000,0x0570,0xcfe8,0x0000};

  if (mcfg->evenodd==AUTO){ /* oddity follows select line */
    if(mcfg->select==EVEN) oddity = 0x0800; 
    else oddity = 0x0c00;

  } else {                  /* oddity is hard wired */
    if (mcfg->evenodd==EVEN) oddity = 0x0800; 
    else oddity = 0x0c00;
  }

  for (i=0; i<12; i++){
    ccfg = mcfg->chips[i];
    /* revised to allow PARANOID - a bypassed, but functional LONELY chip
     * PWP 15.07.02 */
    if((ccfg.role == DEAD)||(ccfg.role == PARANOID)){
      death[i] = 1;
    } else {
      death[i] = 0;
    }
    if (ccfg.role == MISSING) missing[i] = 1;
    else                      missing[i] = 0;
  }

  for (i=0; i<12; i++){
    ccfg = mcfg->chips[i];
    l=0;

    /* config  0000 0101 0111 0001 1100 aaaa aa00 0000 dddddddddddddddd */
    /*            0    5    7    1    c a>>2 a<<2    0         config   */
    config[0] = 0x0000;
    config[1] = 0x0571;
    config[2] = 0xc000 | oddity | chipid[i];

    if (ccfg.role == LONELY)      config[3] = 0x3000;
    else if (ccfg.role == SLAVE)  config[3] = 0x0800;
    else if (ccfg.role == DEAD)   config[3] = 0x0800;
    else if (ccfg.role == MASTER) config[3] = 0x2000;
    else if (ccfg.role == END)    config[3] = 0x1800;
    else if (ccfg.role == PARANOID) config[3] = 0x3000; /* LONELY + DEATH */
    else                          config[3] = 0x0800;

    // PWP bugfix 27.06.01
    // for case where, in forward hybrid, 
    // wish to route S0 to S12 and hence to datalink 1
    if (i==0)  if (death[11]) config[3] = config[3] | 0x0400;  

    if (i>0)  if (death[i-1]) config[3] = config[3] | 0x0400;
    if (i<11) if (death[i+1]) config[3] = config[3] | 0x0200;
    if (i==11) if (death[0])  config[3] = config[3] | 0x0200;

    if (ccfg.accumulate   !=0) config[3] = config[3] | 0x0100;
    if (ccfg.mask_readout !=0) config[3] = config[3] | 0x0080;
    if (ccfg.edge_detect  !=0) config[3] = config[3] | 0x0040;
 
    config[3] = config[3] | ((ccfg.trim_range << 4) & 0x0030);
    config[3] = config[3] | ((ccfg.cal_mode << 2) & 0x000c);
    config[3] = config[3] | (ccfg.compression & 0x0003);

    for (j=0; j<4; j++) c[i][l++] = config[j];

    if(full_config){

      /* For SEU studies, we do not want to rewrite the mask until the
       * start of the next measurement cycle - hence the new argument
       * "full_config" can be used to select this option
       * PWP 09.04.00 
       */

      /* mask    0000 0101 0111 1000 1100 aaaa aa00 1000 dddd*16 */
      /*            0    5    7    8    c a>>2 a<<2    8 mask    */
      mask[0] = 0x0000;
      mask[1] = 0x0578;
      mask[2] = 0xc008 | oddity | chipid[i];
      ccfg = mcfg->chips[i];
      for (j=0; j<8; j++) mask[3+j] = ccfg.mask[j];
      mask[11] = 0x0000;
      for (j=0; j<12; j++) c[i][l++] = mask[j];

      /* delay   0000 0101 0111 0001 1100 aaaa aa01 0000 0000 0000 00dddddd */
      /*            0    5    7    1    c a>>2 a<<2    0    0    0    delay */
      delay[0] = 0x0000;
      delay[1] = 0x0571;
      delay[2] = 0xc010 | oddity | chipid[i];
      delay[3] = ((unsigned short) ((ccfg.delay+0.5)/1.0)) & 0x00ff;
      for (j=0; j<4; j++) c[i][l++] = delay[j];

      /* bias    0000 0101 0111 0001 1100 aaaa aa11 1000 00bias 00shaper */
      /*            0    5    7    1    c a>>2 a<<2    8   bias   shaper */
      bias[0] = 0x0000;
      bias[1] = 0x0571;
      bias[2] = 0xc038 | oddity | chipid[i];
      bias[3] = (((unsigned short) ((ccfg.febias+4.6)/9.2)) <<8) & 0x1f00;
      bias[3] = bias[3] | (((unsigned short) ((ccfg.feshaper+0.6)/1.2)) & 0x001f);
      for (j=0; j<4; j++) c[i][l++] = bias[j];

      /* thresh  0000 0101 0111 0001 1100 aaaa aa01 1000 vthr vcal */
      /*            0    5    7    1    c a>>2 a<<2    8 vthr vcal */
      thresh[0] = 0x0000;
      thresh[1] = 0x0571;
      thresh[2] = 0xc018 | oddity | chipid[i]; 
      thresh[3] = (((unsigned short) ((ccfg.vthr+1.25)/2.5)) <<8 ) & 0xff00;
      thresh[3] = thresh[3] | (((unsigned short) ((ccfg.vcal+0.312)/0.625)) & 0x00ff);
      for (j=0; j<4; j++) c[i][l++] = thresh[j];
    }

    chip_cb_size = l;
  }

  k=0;
  /* Append full chip-by-chip configuration */
  for (i=0; i<12; i++)
    if (!missing[i])
      for (j=0; j<chip_cb_size; j++) cb[k++] = c[i][j];

  /* Finish with all-chips re-enable data-taking if appropriate */
  if(mcfg->dtm) for (j=0; j<4; j++) cb[k++] = enabledata[j];
  for (j=0; j<20; j++) cb[k++] = 0x0000;

  *cb_size = (unsigned int)k;

  return 0;
}


/* abc_trims(): Configure ABCDT trimdacs */

int abc_trims(module_cfg *mcfg, int chip, unsigned short *cb, unsigned int *cb_size){
  chip_cfg ccfg;
  unsigned short trim[4], enabledata[4];
/*  unsigned short sreset[4]; */
  unsigned short c[512];
  unsigned short j,k,l;
  unsigned short oddity;    /* aaaa aa00 = 1daa aa00
                               EVEN => oddity = 1000 0000 = 0x0800
                               ODD  => oddity = 1100 0000 = 0x0c00; */


  if(mcfg->chipset!=ABCDT) return -1;

  if (mcfg->evenodd==AUTO){ /* oddity follows select line */
    if(mcfg->select==EVEN) oddity = 0x0800; 
    else oddity = 0x0c00;

  } else {                  /* oddity is hard wired */
    if (mcfg->evenodd==EVEN) oddity = 0x0800; 
    else oddity = 0x0c00;
  }

  ccfg = mcfg->chips[chip];

  if (ccfg.role == MISSING) return -2;
 
  l=0;

  /* trim  0000 0101 0111 0001 1100 aaaa aa00 0100 0000 0ddd dddd tttt */
  /*          0    5    7    1    c a>>2 a<<2    4    0 trim_addr trim */
  trim[0] = 0x0000;
  trim[1] = 0x0571;
  trim[2] = 0xc004 | oddity | chipid[chip];

  for(k=0;k<128;k++){
    trim[3] = ccfg.trim[k] & 0x000f;
    trim[3] = trim[3] | ((((unsigned short)k) << 4) & 0x07f0);
    for (j=0; j<4; j++) c[l++] = trim[j];
  } 

  /* enable  0000 0101 0111 0000 1100 1111 1110 1000 */
  /*            0    5    7    0    c all chips    8 */
  enabledata[0] = 0x0000;
  enabledata[1] = 0x0570;
  enabledata[2] = 0xcfe8;
  enabledata[3] = 0x0000;

  /* sreset  0000 0101 0100 0000 */
  /*            0    5    4    0 */
/*  sreset[0] = 0x0000;
  sreset[1] = 0x0540;
  sreset[2] = 0x0000;
  sreset[3] = 0x0000; */

  k=0;

  /* Append trim dac control block for this chip */
  for (j=0; j<512; j++) cb[k++] = c[j];

  /* Finish with all-chips re-enable data-taking and soft-reset */
  if(mcfg->dtm) for (j=0; j<4; j++) cb[k++] = enabledata[j];
/*  for (j=0; j<4; j++) cb[k++] = sreset[j]; */
  for (j=0; j<20; j++) cb[k++] = 0x0000;

  *cb_size = (unsigned int)k;

  return 0;
}


/* abc_allchips(): Configure each configuration register of each chip separately
                   (so cal_mode can change but other bits preserved), but use
                   all-chips addressing for DAC and delay registers.  Do not
                   change mask register as this will not normally change during
                   a scan. */

int abc_allchips(module_cfg *mcfg, unsigned short *cb, unsigned int *cb_size){
  chip_cfg ccfg;
  
  unsigned short config[12][4],thresh[4],bias[4],delay[4]; 
  unsigned short enabledata[4];
/*  unsigned short sreset[4]; */
  unsigned short i,j,l;
  unsigned short oddity;    /* aaaa aa00 = 1daa aa00
                               EVEN => oddity = 1000 0000 = 0x0800;
      	                       ODD  => oddity = 1100 0000 = 0x0c00; */

  unsigned short death[12],missing[12];

  if (mcfg->evenodd==AUTO){ /* oddity follows select line */
    if(mcfg->select==EVEN) oddity = 0x0800; 
    else oddity = 0x0c00;

  } else {                  /* oddity is hard wired */
    if (mcfg->evenodd==EVEN) oddity = 0x0800; 
    else oddity = 0x0c00;
  }

  for (i=0; i<12; i++){
    ccfg = mcfg->chips[i];
    if (ccfg.role == DEAD) death[i] = 1;
    else                   death[i] = 0;
    if (ccfg.role == MISSING) missing[i] = 1;
    else                      missing[i] = 0;
  }

  for (i=0; i<12; i++){
    ccfg = mcfg->chips[i];
    l=0;

    /* config  0000 0101 0111 0001 1100 aaaa aa00 0000 dddddddddddddddd */
    /*            0    5    7    1    c a>>2 a<<2    0         config   */
    config[i][0] = 0x0000;
    config[i][1] = 0x0571;
    config[i][2] = 0xc000 | oddity | chipid[i];

    if (ccfg.role == LONELY)      config[i][3] = 0x3000;
    else if (ccfg.role == SLAVE)  config[i][3] = 0x0800;
    else if (ccfg.role == DEAD)   config[i][3] = 0x0800;
    else if (ccfg.role == MASTER) config[i][3] = 0x2000;
    else if (ccfg.role == END)    config[i][3] = 0x1800;
    else                          config[i][3] = 0x0800;

    if (i>0)  if (death[i-1]) config[i][3] = config[i][3] | 0x0400;
    if (i<11) if (death[i+1]) config[i][3] = config[i][3] | 0x0200;

    ccfg = mcfg->chips[0];   /* Force next bits to same as chip[0] */

    if (ccfg.accumulate   !=0) config[i][3] = config[i][3] | 0x0100;
    if (ccfg.mask_readout !=0) config[i][3] = config[i][3] | 0x0080;
    if (ccfg.edge_detect  !=0) config[i][3] = config[i][3] | 0x0040;
 
    printf("i ccfg.cal_mode %x\n",i,ccfg.cal_mode);

    config[i][3] = config[i][3] | ((ccfg.trim_range << 4) & 0x0030);
    config[i][3] = config[i][3] | ((ccfg.cal_mode << 2) & 0x000c);
    config[i][3] = config[i][3] | (ccfg.compression & 0x0003);

    for (j=0; j<4; j++) cb[l++] = config[i][j];
  }

  ccfg = mcfg->chips[0];
 
  /* mask    No changes required during scans, so skip */

  /* fixes for rounding errors: add half step before conversion
     * of float to int for cases of delay, bias, shaper, thresh, cal
     * registers. PWP 07.07.99 */

  /* delay   0000 0101 0111 0001 1100 aaaa aa01 0000 0000 0000 00dddddd */
  /*            0    5    7    1    c a>>2 a<<2    0    0    0    delay */
  delay[0] = 0x0000;
  delay[1] = 0x0571;
  delay[2] = 0xc010 | 0x0fc0;  /* all-chips */
  delay[3] = ((short) ((ccfg.delay+0.5)/1.0)) & 0x00ff;
  for (j=0; j<4; j++) cb[l++] = delay[j];

  /* bias    0000 0101 0111 0001 1100 aaaa aa11 1000 00bias 00shaper */
  /*            0    5    7    1    c a>>2 a<<2    8   bias   shaper */
  bias[0] = 0x0000;
  bias[1] = 0x0571;
  bias[2] = 0xc038 | 0x0fc0;  /* all-chips */
  bias[3] = (((short) ((ccfg.febias+4.6)/9.2)) <<8) & 0x1f00;
  bias[3] = bias[3] | (((short) ((ccfg.feshaper+0.6)/1.2)) & 0x001f);
  for (j=0; j<4; j++) cb[l++] = bias[j];

  /* thresh  0000 0101 0111 0001 1100 aaaa aa01 1000 00vthr 00vcal */
  /*            0    5    7    1    c a>>2 a<<2    8   vthr   vcal */
  thresh[0] = 0x0000;
  thresh[1] = 0x0571;
  thresh[2] = 0xc018 | 0x0fc0;  /* all-chips */
  thresh[3] = (((short) ((ccfg.vthr+1.25)/2.5)) <<8) & 0xff00;
  thresh[3] = thresh[3] | (((short) ((ccfg.vcal+0.312)/0.625)) & 0x00ff);
  for (j=0; j<4; j++) cb[l++] = thresh[j];

  printf("abc_allchips: config[0] %x  delay %x  bias_reg %x  thresh_reg %x\n",
          config[0][3], delay[3], bias[3], thresh[3]);

  /* enable  0000 0101 0111 0000 1100 1111 1110 1000 */
  /*            0    5    7    0    c all chips    8 */
  enabledata[0] = 0x0000;
  enabledata[1] = 0x0570;
  enabledata[2] = 0xcfe8;
  enabledata[3] = 0x0000;

  /* sreset  0000 0101 0100 0000 */
  /*            0    5    4    0 */
/*  sreset[0] = 0x0000;
  sreset[1] = 0x0540;
  sreset[2] = 0x0000;
  sreset[3] = 0x0000; */

  /* Finish with all-chips re-enable data-taking and soft-reset */
  if (mcfg->dtm)for (j=0; j<4; j++) cb[l++] = enabledata[j];
/*  for (j=0; j<4; j++) cb[l++] = sreset[j]; */
  for (j=0; j<20; j++) cb[l++] = 0x0000;

  *cb_size = (unsigned int)l;

  return 0;
}

int abc_default(short chipset, short evenodd, short ntop, short nbot, module_cfg *md){
  short n,m;

  md->chipset = chipset;
  md->evenodd = evenodd;
  
  for (n=0; n<12; n++){
    md->chips[n].compression = ANYHIT;
    md->chips[n].cal_mode = 0;
    md->chips[n].trim_range = 0;
    md->chips[n].edge_detect = 1;
    md->chips[n].mask_readout = 0;
    md->chips[n].accumulate = 0;
    md->chips[n].role = MISSING;
    md->chips[n].active = 0;
    md->chips[n].vcal = 120.0;
    md->chips[n].vthr = 50.0;
    md->chips[n].delay = 11.0;
    md->chips[n].feshaper = (float)19.2;
    md->chips[n].febias = 74.0;
    for (m=0; m<8; m++) md->chips[n].mask[m] = 0;
  }

 if (ntop == 1) {
    md->chips[0].active = 1;
    md->chips[0].role = LONELY;
  }
  if (ntop > 1) {
    for (n=1; n<ntop; n++) md->chips[n].active = 1;
    md->chips[0].role = MASTER;
    md->chips[0].active = 1;
    for (n=1; n<ntop; n++) md->chips[n].role = SLAVE;
    md->chips[ntop-1].role = END;
  }

   
  if (nbot == 1){
    md->chips[6].active = 1;
    md->chips[6].role = LONELY;
  }
  if (nbot > 1) {
    for (n=6; n<(6+nbot); n++) md->chips[n].active = 1;
    md->chips[6].role = MASTER;
    md->chips[6].active = 1;
    for (n=7; n<(6+nbot); n++) md->chips[n].role = SLAVE;
    md->chips[6+ntop-1].role = END;
  }

  md->select = 0;

  md->vdet_set = 0.0;
  md->idet_lim = 0.0;

  md->vdd_set = 4.0;
  md->idd_lim = 500.0;

  md->vcc_set = 3.5;
  md->icc_lim = 400.0;

  return 0;
}

int abc_show_config(module_cfg *md){
  short n;

  printf("\nabc_show_config:   chipset %d   ", md->chipset);
  if (md->evenodd == 0) printf("EVEN"); else printf("ODD");
  printf("\n");

  printf("abc_show_config:   chips present:\n");
  for (n=0; n<12; n++){
    if (md->chips[n].role != MISSING){
      printf("id %2d  role %2d  \n", n, md->chips[n].role);
    }
  }
  printf("\n");
  return 0;
}

int abc_global_dacs(float vcal, float vthr, unsigned short *cbword){

  cbword[0] = 0x0000;
  cbword[1] = 0x0571;
  cbword[2] = 0xc018 | 0x0fc0;  /* all-chips */
  cbword[3] = (((short) (vthr/2.5)) <<8) & 0xff00;
  cbword[3] = cbword[3] | (((short) (vcal/0.625)) & 0x00ff);

  return 0;
}

int abc_global_delay(float delay, unsigned short *cbword){
  /* delay   0000 0101 0111 0001 1100 aaaa aa01 0000 0000 0000 00dddddd */
  /*            0    5    7    1    c a>>2 a<<2    0    0    0    delay */
  cbword[0] = 0x0000;
  cbword[1] = 0x0571;
  cbword[2] = 0xc010 | 0x0fc0;  /* all-chips */
  cbword[3] = ((short) delay) & 0x003f;

  return 0;
}

int abc_global_bias(float febias, float feshaper, unsigned short *cbword){
  /* bias    0000 0101 0111 0001 1100 aaaa aa11 1000 00bias 00shaper */
  /*            0    5    7    1    c a>>2 a<<2    8   bias   shaper */
  cbword[0] = 0x0000;
  cbword[1] = 0x0571;
  cbword[2] = 0xc038 | 0x0fc0;  /* all-chips */
  cbword[3] = (((short) (febias/9.2)) <<8) & 0x1f00;
  cbword[3] = cbword[3] | (((short) (feshaper/1.2)) & 0x001f);

  return 0;
}

int abc_global_enabledata(unsigned short *cbword){

  cbword[0] = 0x0000;
  cbword[1] = 0x0570;
  cbword[2] = 0xcfe8;
  cbword[3] = 0x0000;

  return 0;
}

int abc_pulse_input(unsigned short *cbword, unsigned int *cb_size){

  cbword[0] = 0x0000;
  cbword[1] = 0x0570;
  cbword[2] = 0xcfe0;

  /* now idle for 128 bits, = pipeline latency */
  cbword[3] = 0x0000;
  cbword[4] = 0x0000;
  cbword[5] = 0x0000;
  cbword[6] = 0x0000;
  cbword[7] = 0x0000;
  cbword[8] = 0x0000;
  cbword[9] = 0x0000;
  cbword[10] = 0x0000;

  /* send L1 trigger */

  cbword[11] = 0xc000;
  cbword[12] = 0x0000;
  cbword[13] = 0x0000;
  cbword[14] = 0x0000;

  *cb_size = 14;

  return 0; 
}

int abc_command(unsigned short *cbword, unsigned int *cb_size,
                int com1, int com2, int ntrigs,
                unsigned short delay1, unsigned short delay2, unsigned short delay3){

  /* Routine to make a control block as follows:
   *
   *    (idle)
   * Command 1  - e.g. soft reset
   *    (delay1)
   * Command 2  - e.g. calstrobe
   *    (delay2)
   * for(j=0;j<ntrigs;j++){
   *   L1 trigger
   *      (delay3)
   * }
   *    (idle)
   *
   * The routine assumes that cb has been cleared externally.
   *
   * Such a control block may be executed by SLOG to help
   * investigate pipeline effects - PWP 01.10.99 */

  /* Commands:
   * 0 - none
   * 1 - L1A
   * 2 - soft reset
   * 3 - BC reset 
   * 4 - Pulse Input Register
   * 5 - Calibration
   * 6 - Enable Data taking (all)
   * 7 - Enable Data taking (1)
   */

  int i,j;
  int bit  = 0;
  unsigned int word = 2;     /* idle for two words at start */

  unsigned int com[8] = {0x00000000, 0xc0000000, 0xa8000000, 0xa4000000, 0xae19fc00, 0xae19fe00,0xAE19FD00,0xAE198500};
  short        len[8] = {0,3,7,7,27,27,27,27};
 

  /* printf("com1:%d com2:%d delay1:%d delay2:%d\n",
   * com1, com2, delay1, delay2);
   */

  for(i=0;i<len[com1];i++){
    if(com[com1]&(0x80000000>>i)){
      cbword[word] = cbword[word]|(0x8000>>(bit%16));
    }
    bit++;
    if((bit%16)==0) word++;
  }

  for(i=0;i<delay1;i++){
    bit++;
    if((bit%16)==0) word++;
  }

  for(i=0;i<len[com2];i++){
    if(com[com2]&(0x80000000>>i)){
      cbword[word] = cbword[word]|(0x8000>>(bit%16));
    }
    bit++;
    if((bit%16)==0) word++;
  }

  for(i=0;i<delay2;i++){
    bit++;
    if((bit%16)==0) word++;
  }

  for(j=0;j<ntrigs;j++){

    for(i=0;i<len[1];i++){
      if(com[1]&(0x80000000>>i)){
        cbword[word] = cbword[word]|(0x8000>>(bit%16));
      }
      bit++;
      if((bit%16)==0) word++;
    }

    for(i=0;i<delay3;i++){
      bit++;
      if((bit%16)==0) word++;
    }

  }

  word +=2;        /* idle for two words */
  *cb_size = word;

  return 0; 
}

int abc_echo_event(short l1, short bc, short chip, short strip, short bins, unsigned short *cb){   
  /* Creates a simulated event suitable for loop-back echo to Mustard
     staying within the 64-bit limit of CLOAC */

  if (chip < 0) {
    /* This is a minimal no-hit data packet: */   
    cb[0] = 0xe800;
    cb[1] = 0x2600;
    cb[2] = 0x0000;
    cb[3] = 0x0000;
  } else {
    cb[0] = 0xe800 | ((l1 <<6)&0x0300) | ((bc>>2)&0x003f);
    cb[1] = 0x2800 | ((bc>>14)&0xc000) | ((chip<<7)&0x0780) | (strip&0x007f);
    cb[2] = 0x8800 | ((bins<<12)&0x7000);
    cb[3] = 0x0000;
  }

  return 0;
}

/* abc_sendid()
 * constructs the response expected from a module of
 * configuration mcfg in send_ID mode.  Data is returned
 * as a block of unsigned short.
 */

int abc_sendid(module_cfg *mcfg, short link, unsigned short *db, int *db_size){
  chip_cfg ccfg;
  int lo, hi, lo_r, hi_r, r_master;
  int end0 =  5;
  int end1 = 11;
  int n_bits = 0;
  unsigned short config[12];
  unsigned int datablock[12];
  unsigned short i;
  unsigned short oddity;    /* aaaa aa00 = 1daa aa00
              EVEN => oddity = 1000 0000 = 0x0800;
              ODD  => oddity = 1100 0000 = 0x0c00; */

  unsigned short death[12],missing[12];

  if (mcfg->evenodd==AUTO){ /* oddity follows select line */
    if(mcfg->select==EVEN) oddity = 0x0800; 
    else oddity = 0x0c00;

  } else {                  /* oddity is hard wired */
    if (mcfg->evenodd==EVEN) oddity = 0x0800; 
    else oddity = 0x0c00;
  }

  for (i=0; i<12; i++){
    ccfg = mcfg->chips[i];
    if (ccfg.role == DEAD) death[i] = 1;
    else                   death[i] = 0;
    if (ccfg.role == MISSING) missing[i] = 1;
    else                      missing[i] = 0;
  }

  for (i=0; i<12; i++){
    ccfg = mcfg->chips[i];

    /* Generate the configuration register contents */

    if (ccfg.role == LONELY)      config[i] = 0x3000;
    else if (ccfg.role == SLAVE)  config[i] = 0x0800;
    else if (ccfg.role == DEAD)   config[i] = 0x0800;
    else if (ccfg.role == MASTER) config[i] = 0x2000;
    else if (ccfg.role == END){
                                  config[i] = 0x1800;
      if ((i<6) && (end0 ==  5)) end0 = i;
	  if ((i>5) && (end1 == 11)) end1 = i;
    }
    else                          config[i] = 0x0800;

    if (i>0)  if (death[i-1]) config[i] = config[i] | 0x0400;
    if (i<11) if (death[i+1]) config[i] = config[i] | 0x0200;

    if (ccfg.accumulate   !=0) config[i] = config[i] | 0x0100;
    if (ccfg.mask_readout !=0) config[i] = config[i] | 0x0080;
    if (ccfg.edge_detect  !=0) config[i] = config[i] | 0x0040;
 
    config[i] = config[i] | ((ccfg.trim_range << 4) & 0x0030);
    config[i] = config[i] | ((ccfg.cal_mode << 2) & 0x000c);
    config[i] = config[i] | (ccfg.compression & 0x0003);


  /* Generate the send_id datablock for each chip */
  /* datablock = 000aaaa111cccccccc1cccccccc1xxxx */
  /* NB. contents of chip_id array are already shifted << 6 */
 
    datablock[i] = 0x01c02010;
    datablock[i] += ( chipid[i] <<  19);
    datablock[i] += ((config[i] & 0xff00) << 6);
    datablock[i] += ((config[i] & 0x00ff) << 5);
  }

  /* Assemble expected result, stored as a bit pattern in unsigned short 
     array.  Handles redundancy if second master is marked as DEAD */

  if(link==0){        /* expect chips M0 - E5  */
		lo   = 0;
		hi   = end0;
		lo_r = 7;
		hi_r = 12;
		r_master = 1;
	}
  else{               /* expect chips M8 - E13 */
		lo   = 6;
		hi   = end1;
		lo_r = 1;
		hi_r = 5;
		r_master = 0;
	}

  db_fill(db, &n_bits, HEADER, HEADER_LENGTH);

  for (i=lo; i<=hi; i++){
    if((!death[i])&&(!missing[i])){
      db_fill(db, &n_bits, datablock[i], DATABLOCK_LENGTH);
	}
  }

  if(death[r_master]){
    for (i=lo_r; i<=hi_r; i++){
	  if((!death[i])&&(!missing[i])){ db_fill(db, &n_bits, datablock[i], DATABLOCK_LENGTH); }
	}
  }

  db_fill(db, &n_bits, TRAILER, TRAILER_LENGTH);

  *db_size = n_bits;

  return 0;
}

void db_fill(unsigned short *db, int *db_size, unsigned int dword, short nbits){
  /* fill bits of each unsigned short word in the data block,
     according to content of dword */

  int i, word, the_bit;
  int bit = *db_size;
  
  for(i = 0; i <nbits; i++){
    word = bit / 0x10;
    the_bit = 0x0f - (bit%0x10);

    if(the_bit == 0x0f)  db[word] = 0;
    if(dword&0x80000000) db[word] += (1 << the_bit);

	dword = dword << 1;
	bit ++;
  }
  *db_size = bit;
}

/* abc_seu_setup(): fast all chips setup for SEU test */

int abc_seu_setup(unsigned short *cb, unsigned int *cb_size, short seu_invert){
  unsigned short j,k;

  /* mask    0000 0101 0111 1000 1100 1111 1100 1000 dddd*8 */
  /*            0    5    7    8    c all_chips    8 mask   */
  unsigned short mask[12] = {0x0000,0x0578,0xcfc8,0x0000, 
                             0x0000,0x0000,0x0000,0x0000, 
                             0x0000,0x0000,0x0000,0x0000};

  unsigned short mask2[12] = {0x0000,0x0578,0xcfc8,0xffff, 
                             0xffff,0xffff,0xffff,0xffff, 
                             0xffff,0xffff,0xffff,0x0000};

  /* sreset  0000 0101 0100 0000 */
  /*            0    5    4    0 */
  unsigned short sreset[3] = {0x0000,0x0540,0x0000};

  /* enable  0000 0101 0111 0000 1100 1111 1110 1000 */
  /*            0    5    7    0    c all chips    8 */
  unsigned short enabledata[4] = {0x0000,0x0570,0xcfe8,0x0000};

  /* L1A  0000 0011 0000 0000 */
  /*         0    3    0    0 */
  unsigned short l1a = 0x0300;

  k=0;

  if(seu_invert){
    for (j=0;j<12;j++) cb[k++] = mask2[j];
  }else{
    for (j=0;j<12;j++) cb[k++] = mask[j];
  }
  for (j=0;j<3; j++) cb[k++] = sreset[j];
  for (j=0;j<4; j++) cb[k++] = enabledata[j];

  for (j=0;j<3; j++) cb[k++] = 0x0000;
  for (j=0;j<3; j++) cb[k++] = sreset[j];
                     cb[k++] = l1a;

  for (j=0;j<480; j++) cb[k++] = 0x0000;

  *cb_size = (unsigned int)k;

  return 0;
}

int sc_configure(module_cfg *mcfg, unsigned short *cb, 
                 unsigned int *cb_size){
  /*
   * Having realised how inefficient it was to send one
   * command at a time to each forward support card, here's
   * a routine which sticks all the commands together
   * into one control block, using the contents of the
   * module configuration structure.
   * PWP 15.05.01
   */

  /* VDC_DAC 0 */
  cb[0] = 0x0000;
  cb[1] = 0x0000;
  cb[2] = 0x0575;
  cb[3] = 0x0000 + ((mcfg->vdc_dac0)&0x03ff);
  cb[3]<<=2;
  cb[4] = 0x0000;

  /* VDC_DAC 1 */
  cb[5] = 0x0000;
  cb[6] = 0x0000;
  cb[7] = 0x0575;
  cb[8] = 0x2000 + ((mcfg->vdc_dac1)&0x03ff);
  cb[8]<<=2;
  cb[9] = 0x0000;

  /* BPM_DRIVE */
  cb[10] = 0x0000;
  cb[11] = 0x0000;
  cb[12] = 0x0570 + ((7+mcfg->bpm_drive)&0x000f);
  cb[13] = 0x0000;
  cb[14] = 0x0000;

  /* SC_MODE */
  cb[15] = 0x0000;
  cb[16] = 0x0000;
  cb[17] = 0x0570 + (mcfg->sc_mode&0x000f);  /* 0000 0101 0111 wxyz */
  cb[18] = 0x0000;
  cb[19] = 0x0000;

  *cb_size = (unsigned int)20;

  return 0;
}
