
/* st_config.c: SCT Systemtest Configuration
 * originally sct_config.c
 *
 * GFM 1998
 * read module and sysmap JLB/PWP March - June 99
 * RAID compatibility GFM July 1999
 *
 * Version 0.2 PWP 04.08.99 
 *  - change function names from sct* to st*
 *  - add st_read_trims
 * PWP 12.11.99 add st_read_rc and st_show rc
 * PWP 15.12.99 additions to st_show_module for Vpin
 * PWP 14.11.00 add st_read_mask, st_mask_chan
 * PWP 29.02.00 add st_reverse_sysmap, st_find_link and st_find_module
 * PWP 16.03.00 add st_find_mustard, st_find_stream 
 * PWP 17.03.00 add GFM corrections and some more besides
 * PWP 27.06.00 add st_find_maxchips
 * PWP 28.07.00 partial additions for SCTHV and setting of Vpin
 * PWP 23.08.00 add hv_id and hv_ch to read_sysmap 
 * PWP 10.10.00 revised and expanded sysmap wrt opto settings
 * PWP 23.11.00 add st_set_rc
 * PWP 11.04.01 add support for module suspense flags (!!)
 * PWP 10.05.01 add st_get_rc
 * PWP 29.08.01 add st_write_module and revise st_read_module to read trim and mask
 *              info from the configuration file.  (Backward compatible)
 * PWP 19.02.02 revise st_write_module and st_read_module to write/read
 *              RC and c_factor info from the detector configuration file
 * JGK 06.03.02 bugfix
 * PWP 14.11.02 add st_write_module_xml
 * PWP 28.11.02 trivial bugfix re years to st_write_module_xml
 */

/* A revised version of Gareth's original notes:

   How to map detector module id to the various readout module id's when the readout 
   modules have differing numbers of channels which don't divide up simply ?

   I suggest the following scheme in the era of up to 16 detector modules
   using CLOAC/SLOG/BILED/MUSTARD readout:

   Master CNC:     1 CLOAC module with one channel connected to the SLOG cnc input and
                     thence broadcast to all detectors and readout modules, using where
                     necessary UCL ECL fanout modules (after SLOG).

   Slow Commands:  1 SLOG using only the four of its 14 channels which can be conveniently 
                     mapped to four BiLEDS without level translation.  Recall that the
                     SLOG has 12 LVDS channels (0:11) (intended for electrical testing 
                     and not directly useable with optical links) of which two can be mapped 
                     to ECL outputs (and thence to BILEDs) and 2 NIM channels (14:15)
                     which can also be connected directly to BiLEDs via their NIM inputs.

                     In order to address slow commands to 16 channels, the SLOG 32k output
                     buffer can be divided arbitrarily into four 8k pages, each sent
                     in turn when the appropriate BiLED channel is enabled.  In the first
                     time slice, channels 0,1,10,11 page 0 are sent to BiLEDs 0,1,2,3 VCSEL 1,
                     then page 1 to VCSEL 2, page 2 to VCSEL 3 and page 3 to VCSEL 4.

   Optical links:  4 BiLED modules of 4 channels each.  Recall that the BiLED modules
                     have only one common CNC input.  Triggers and fast commands are
                     sent to all four channels simultaneously from this input if they
                     are enabled.  To send slow commands to specific detector modules,
                     only the relevant BiLED channel should be enabled.  We therefore map
                     the relevant SLOG pages as described above for sequential downloading.

   Redundancy:       Redundant slog id and ch are stored for each module.

   Readout:        2 or 3 MUSTARDS each with 12 streams, two streams per detector module,
                     hence 6 detectors per MUSTARD.

   LV:               The low-voltage modules are addressed by crate, two channels per
                     card (0:1, 2:3, etc.)

   HV:               Reference by id (card number) and ch (channel) - PWP 28.07.00

   DCS:              Assume need only one DCS identifier to establish
                     communications between DAQ and DCS for any module. 


   To implement these arbitrary mappings in a versatile way, I use an array of structures
   of type sysmap, one per detector module.   Any mappings between the various
   readout module channels and the detector id can thus be accommodated, although
   the simple one proposed can be hard-coded easily as follows. */

#include <stdio.h>
#include <string.h>
#include <time.h>

#include "abc.h"
#include "abc_config.h"
#include "sct.h"
#include "sct_hardware.h"
#include "st_config.h"
#include "stlib.h"

extern sysmap sct[SCT_MAX_MODULES];
extern short  stream_2_mod[MUS_NUMBER][16];
extern short  stream_2_link[MUS_NUMBER][16];


int st_default_sysmap(){

  /* Set sysmap to defaults for Optical readout in systemtest using only
     Slog channels 0,1,12,13 each connected to one Biled wih four channels */
  
  int d;

  for (d=0; d<SCT_MAX_MODULES; d++){
    sct[d].present = 0; 
    sct[d].active = 1;    /* Default is to be active, i.e., to participate in scans when present */

    sct[d].link0 = 1;     /* Default is both uplinks reading out */
    sct[d].link1 = 1;
    sct[d].suspend0 = 0;  /* PWP 11.04.01 */
    sct[d].suspend1 = 0;

    /* DCS */

    sct[d].lv_id = d/16;  /* Crate number !!! */
    sct[d].lv_ch = d%16;  /* Channel within crate (2 per SCTLV module) */

    sct[d].hv_id = -1;    /* default - for reasons of safety */
    sct[d].hv_ch = -1;

    sct[d].dcs_id = 0;

    /* SLOG */

    sct[d].slog_id = 0;
    if ((d/4) == 0)      sct[d].slog_ch = 0;
    else if ((d/4) == 1) sct[d].slog_ch = 1;
    else                 sct[d].slog_ch = (d/4) + 10;
    sct[d].slog_pg = d%4;    

    sct[d].r_slog_id = sct[d].slog_id;
    sct[d].r_slog_ch = sct[d].slog_ch;
    sct[d].r_slog_pg = sct[d].slog_pg;

    /* OPTO */
   
    sct[d].opto_tx_type = -1;
    sct[d].opto_tx_id = -1;
	  sct[d].opto_tx_ch = -1;
    sct[d].opto_tx_iset = 0;  /* set VCSEL drive current OFF by default */
    sct[d].r_opto_tx_type = -1;
    sct[d].r_opto_tx_id = -1;
	  sct[d].r_opto_tx_ch = -1;
	  
    sct[d].opto_rx_id = -1;
    sct[d].opto_rx_ch0 = -1;
	  sct[d].opto_rx_ch1 = -1;
    sct[d].opto_rx_thr0 = -1;
	  sct[d].opto_rx_thr1 = -1;

    /* MuSTARD */    

    sct[d].mus_id = d/6;
    sct[d].mus_enable0 = 0;    
    sct[d].mus_enable1 = 0;    
    sct[d].mus_ch0 = (d*2)%12;
    sct[d].mus_ch1 = sct[d].mus_ch0 + 1;
    sct[d].mus_delay0 = 12;
    sct[d].mus_delay1 = 12;
  
    /* TODO: clear detname field
     *  PWP 15.03.00
     */
  }

  return 0;
}

int st_electro_sysmap(){

  /* Set sysmap to defaults for multi-module electrical readout system using two 
     Slog channels (primary clk0/com0 and redundant clk1/com1) for each module */

  int d;

  for (d=0; d<SCT_MAX_MODULES; d++){
    sct[d].present = 0; /* Module is not assumed to be present unless explicitly enabled */
    sct[d].active = 1;  /* Default is to be active, i.e., to participate in scans when present */

    sct[d].link0 = 1;   /* Default is both uplinks reading out */
    sct[d].link1 = 1;
    sct[d].suspend0 = 0;  /* PWP 11.04.01 */
    sct[d].suspend1 = 0;

    /* DCS */

    sct[d].lv_id = d/16;   /* Crate */
    sct[d].lv_ch = d%16;   /* Channel, two per SCTLV module */

    sct[d].hv_id = -1;   /* for reasons of safety */
    sct[d].hv_ch = -1;

    sct[d].dcs_id = 0;

    /* SLOG */

    sct[d].slog_id = d/6;
    sct[d].slog_ch = (d*2)%12; /* even channel -> primary clk/com downlinks */
    sct[d].slog_pg = 0;    /* Only one page per module as there is no fan-out */ 

    sct[d].r_slog_id = sct[d].slog_id;
    sct[d].r_slog_ch = sct[d].slog_ch + 1; /* odd channel -> redundant clk/com */
    sct[d].r_slog_pg = 0;

    /* OPTO */

    sct[d].opto_tx_type = -1;
    sct[d].opto_tx_id = -1;
	  sct[d].opto_tx_ch = -1;
    sct[d].opto_tx_iset = 0;  /* set VCSEL drive current OFF by default */
    sct[d].r_opto_tx_type = -1;
    sct[d].r_opto_tx_id = -1;
	  sct[d].r_opto_tx_ch = -1;

    sct[d].opto_rx_id = -1;
    sct[d].opto_rx_ch0 = -1;
	  sct[d].opto_rx_ch1 = -1;
    sct[d].opto_rx_thr0 = -1;
	  sct[d].opto_rx_thr1 = -1;

    /* MuSTARD */

    sct[d].mus_id = d/6;
    sct[d].mus_enable0 = 0;    
    sct[d].mus_enable1 = 0;    
    sct[d].mus_ch0 = (d*2)%12;
    sct[d].mus_ch1 = sct[d].mus_ch0 + 1;
    sct[d].mus_delay0 = 12;
    sct[d].mus_delay1 = 12;
  }

  return 0;
}


void st_menu_sysmap(){

  short det = 0;
  short var = 0;

  char c = ' ';
  short run = 1;
  short narg;
  short type, r_type, id, ch, ch0, ch1;

  printf("SCT System Mapping:  Change Settings for module # ?  \n");
  scanf("%d", &det);
  if ((det < 0) || (det >= SCT_MAX_MODULES)) {
    printf(" ... ERROR: det %d out of allowed range 0 to %d\n", det, SCT_MAX_MODULES);
    return;
  }
  
  while(run){
    printf("  Change module PRESENCE (p) ACTIVITY (a) or LINK ENABLES (l)\n");
    printf("  Change channel mappings for SLOG(s) MUSTARD(m) SCTLV(v) HV(h) OPTO_RX(r) OPTO_TX(t) EXIT(x)\n");
    scanf("%s", &c);

    if (c == 'x') run = 0;
    else if (c == 's') {
      printf("  ...Changing SLOG mapping for module %d:  enter new [id] [ch] (ints)\n",id);
      narg = scanf("%d %d", &id, &ch);
      sct[det].slog_id = id;
      sct[det].slog_ch = ch;
      printf("     ... SLOG mappings for module %d changed:  id %d  ch %d\n",
         id, sct[det].slog_id, sct[det].slog_ch);
    }
    else if (c == 'm') {
      printf("  ...Changing MUSTARD mapping for module %d:  enter new [id] [ch0] [ch1] (ints)\n", id);
      narg = scanf("%d %d %d", &id, &ch0, &ch1);
      sct[det].mus_id = id;
      sct[det].mus_ch0 = ch0;
      sct[det].mus_ch1 = ch1;
      printf("     ... MUSTARD mappings for module %d changed:  id %d  ch0 %d ch1 %d\n",
         id, sct[det].mus_id, sct[det].mus_ch0, sct[det].mus_ch1);
    }
    else if (c == 'l') {
      printf("  ...Changing LINK ENABLES for module %d:  enter new enables (1/0) for ch0 and ch1 [en0] [en1] (ints)\n",det);
      narg = scanf("%d %d", &ch0, &ch1);
      sct[det].link0 = ch0;
      sct[det].link1 = ch1;
   
      printf("     ... Module %d LINK ENABLES are now %d %d\n", 
        det, sct[det].link0, sct[det].link1);
    }

    else if (c == 'a') {
      printf("  ...Changing ACTIVITY for module %d:  flag module as  ACTIVE[1] or NOT Active[0] (ints)\n",det);
      narg = scanf("%d", &ch);
      sct[det].active = ch;
      if (sct[det].active)
        printf("     ... Module %d is now flagged as ACTIVE in scans\n");
      else
        printf("     ... Module %d is now flagged as NOT ACTIVE in scans\n");
    }
    else if (c == 'p') {
      printf("  ...Changing PRESENT for module %d:  flag module as PRESENT[1] or ABSENT[0] (ints)\n",det);
      narg = scanf("%d", &ch);
      sct[det].present = ch;
      if (sct[det].present)
        printf("     ... Module %d is now flagged as PRESENT in scans\n");
      else
        printf("     ... Module %d is now flagged as ABSENT in scans\n");
    }
    else if (c == 'v') {
      printf("  ...Changing SCTLV mapping for module %d:  enter new [id] and [ch] (ints)\n",det);
      narg = scanf("%d %d", &id, &ch);
      sct[det].lv_id = id;
      sct[det].lv_ch = ch;
      printf("     ... SCTLV mappings for module %d changed:  id %d  ch %d\n",
         det, sct[det].lv_id, sct[det].lv_ch);
    }
    else if (c == 'h') {  /* PWP 27.07.00*/
      printf("  ...Changing SCTHV mapping for module %d:  enter new [id] and [ch] (ints)\n",det);
      narg = scanf("%d %d", &id, &ch);
      sct[det].hv_id = id;
      sct[det].hv_ch = ch;
      printf("     ... SCTHV mappings for module %d changed:  id %d  ch %d\n",
         det, sct[det].hv_id, sct[det].hv_ch);
    }
    else if (c == 'r') {  /* PWP 09.10.00*/
      printf("  ...Changing OPTO RX mapping for module %d:  enter new [id], [rx_ch0] and [rx_ch1] (ints)\n",det);
      narg = scanf("%d %d %d %d %d", &id, &ch0, &ch1);
      sct[det].opto_rx_id = id;
	    sct[det].opto_rx_ch0 = ch0;
      sct[det].opto_rx_ch1 = ch1;

      printf("     ... OPTO RX mappings for module %d changed:  id %d rx_ch0 %d  rx_ch1 %d\n",
         det, sct[det].opto_rx_id,
	       sct[det].opto_rx_ch0, sct[det].opto_rx_ch1);
    }
    else if (c == 't') {  /* PWP 09.10.00*/
      printf("  ...Changing OPTO TX mapping for module %d:  enter new [type], [id], [tx_ch], [r_type], [r_id] and [r_tx_ch] (ints)\n",det);
      narg = scanf("%d %d %d %d %d %d", &type, &id, &ch, &r_type, &ch0, &ch1);
      sct[det].opto_tx_type = type;
      sct[det].opto_tx_id = id;
	    sct[det].opto_tx_ch = ch;
      sct[det].r_opto_tx_type = r_type;
	    sct[det].r_opto_tx_id = ch0;
	    sct[det].r_opto_tx_ch = ch1;
 
      printf("     ... OPTO TX mappings for module %d changed:  type %d id %d tx_ch %d  r_type %d r_id %d r_tx_ch %d\n",
         det, sct[det].opto_tx_type, sct[det].opto_tx_id, sct[det].opto_tx_ch,
	       sct[det].r_opto_tx_type, sct[det].r_opto_tx_id, sct[det].r_opto_tx_ch);
    } else {
      printf("   ... Command not valid\n");
    }
  }
}


int st_show_sysmap(){
  /*
   * Revised PWP 09.10.00
   * to suit new opto variables 
   */

  int d;

  printf("\n");
  printf("SCT System Mapping:\n");
  printf("\n");
  printf("DETECTOR        LV    HV     SLOG  RSLOG  OPTOTX          ROPTOTX     OPTORX                MUSTARD              MODULE\n");
  printf("id pr ac l0 l1  cr ch id ch  id ch id ch  type id ch iset type id ch  id ch0 ch1 thr0 thr1  id s0 s1 e0 e1 d0 d1 Name\n");
  printf("\n");

  for (d=0; d<SCT_MAX_MODULES; d++){
    printf("%2d %2d %2d %2d %2d  %2d %2d %2d %2d  %2d %2d %2d %2d  %2d %2d %2d %3d  %2d %2d %2d  %2d %2d %2d %3d %3d  %2d %2d %2d %2d %2d %2d %2d %s (%s)\n",
      d, sct[d].present, sct[d].active, sct[d].link0, sct[d].link1,
      sct[d].lv_id, sct[d].lv_ch,
      sct[d].hv_id, sct[d].hv_ch,
      sct[d].slog_id, sct[d].slog_ch,
      sct[d].r_slog_id, sct[d].r_slog_ch,
      sct[d].opto_tx_type, sct[d].opto_tx_id, sct[d].opto_tx_ch, sct[d].opto_tx_iset,
      sct[d].r_opto_tx_type, sct[d].r_opto_tx_id, sct[d].r_opto_tx_ch,
	    sct[d].opto_rx_id, sct[d].opto_rx_ch0, sct[d].opto_rx_ch1,
      sct[d].opto_rx_thr0, sct[d].opto_rx_thr1,
      sct[d].mus_id, sct[d].mus_ch0, sct[d].mus_ch1,
      sct[d].mus_enable0, sct[d].mus_enable1, sct[d].mus_delay0, sct[d].mus_delay1,
      sct[d].detname,sct[d].devtype);
  }
  printf("\n");
  printf("nBiLEDs = %d, nCLOACs = %d, nMuSTARDs = %d, nOptIFs = %d, nSCTHVs = %d, nSCTLVs = %d, nSLOGs = %d\n",
    BILED_NUMBER, CLOAC_NUMBER, MUS_NUMBER, OPTIF_NUMBER, SCTHV_NUMBER, SCTLV_NUMBER, SLOG_NUMBER);

  return 0;
} 


int st_read_sysmap() {

  /*
   * Revised PWP 09.10.00
   * to suit new opto variables
   * Note that the sysmap is now getting to be rather large
   * - close to a practical limit?
   */

  FILE *file;
  int  narg,i;
  char line[160],token[32],dname[32],dtype[32];

  short arg[30];

  file = fopen(ST_SYSMAP_FILE,"r");
  if (file == NULL){
    printf("st_read_sysmap: failed to open sysmap file %s\n",ST_SYSMAP_FILE);
    return (-1);
  }

  while(fgets(line,160,file) != (char*) NULL) {

    /* printf("st_read_sysmap: line: %s\n", line); */

    strncpy(token,line,6);
    token[6]=(char)NULL;

    if( strncmp(token,"Module",6) == 0 ){

      narg = sscanf(line,            "%s %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %s %s",
        token,
        &arg[0], &arg[1], &arg[2], &arg[3],        
        &arg[4], &arg[5], &arg[6], &arg[7],        
        &arg[8], &arg[9], &arg[10],&arg[11],        
        &arg[12],&arg[13],&arg[14],&arg[15],        
        &arg[16],&arg[17],&arg[18],&arg[19],
        &arg[20],&arg[21],&arg[22],&arg[23],
        &arg[24],&arg[25],&arg[26],&arg[27],
        &arg[28],&arg[29],dname,   dtype);
 
      printf("st_read_sysmap: narg: %ld  %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %s %s\n",
        narg,
        arg[0], arg[1], arg[2], arg[3],        
        arg[4], arg[5], arg[6], arg[7],        
        arg[8], arg[9], arg[10],arg[11],        
        arg[12],arg[13],arg[14],arg[15],        
        arg[16],arg[17],arg[18],arg[19],
        arg[20],arg[21],arg[22],arg[23],
        arg[24],arg[25],arg[26],arg[27],
        arg[28],arg[29],dname,  dtype); 

      if(arg[0]< SCT_MAX_MODULES){ /* cannot overrun sysmap size */
        sct[arg[0]].present    = arg[1];
        sct[arg[0]].active     = arg[2];
        sct[arg[0]].lv_id      = arg[3];
        sct[arg[0]].lv_ch      = arg[4];
        sct[arg[0]].hv_id      = arg[5];
        sct[arg[0]].hv_ch      = arg[6];
        sct[arg[0]].slog_id    = arg[7];
        sct[arg[0]].slog_ch    = arg[8];
        sct[arg[0]].slog_pg    = arg[9];
        sct[arg[0]].r_slog_id  = arg[10];
        sct[arg[0]].r_slog_ch  = arg[11];
        sct[arg[0]].r_slog_pg  = arg[12];
        sct[arg[0]].opto_tx_type     = arg[13];
        sct[arg[0]].opto_tx_id    = arg[14];
        sct[arg[0]].opto_tx_ch    = arg[15];
        sct[arg[0]].opto_tx_iset  = arg[16];
        sct[arg[0]].r_opto_tx_type   = arg[17];
        sct[arg[0]].r_opto_tx_id  = arg[18];
        sct[arg[0]].r_opto_tx_ch  = arg[19];
        sct[arg[0]].opto_rx_id    = arg[20];
        sct[arg[0]].opto_rx_ch0   = arg[21];
        sct[arg[0]].opto_rx_ch1   = arg[22];
        sct[arg[0]].opto_rx_thr0  = arg[23];
        sct[arg[0]].opto_rx_thr1  = arg[24];
        sct[arg[0]].mus_id     = arg[25];
        sct[arg[0]].mus_ch0    = arg[26];
        sct[arg[0]].mus_ch1    = arg[27];
        sct[arg[0]].mus_delay0 = arg[28];
        sct[arg[0]].mus_delay1 = arg[29];

        /* clear detname string before read */
        for(i=0;i<32;i++) sct[arg[0]].detname[i] = 0;
        strncpy(sct[arg[0]].detname, dname, strlen(dname));

        /* clear devtype string before read */
        for(i=0;i<32;i++) sct[arg[0]].devtype[i] = 0;
        if(narg==33){
          strncpy(sct[arg[0]].devtype, dtype, strlen(dtype));
        }else{
          strncpy(sct[arg[0]].devtype, ".", strlen("."));
        }
      }
    }
  }  /* end while */

  if(file){
    fclose(file);
  }
  printf("st_read_sysmap: finished\n");
  return(0);
}


int st_read_module(int mid) {
  /* 
   * New version PWP 28.07.00
   * extensions to read in Vpin setting (LV) and ramp rate (HV)
   * should be backwards compatible excepting "mask" which is now ignored.
   *
   * In the fullness of time, this should be replaced by something better
   * 
   * Not exactly something better, but with extensions to read in 
   * trim and mask info when appended to the configuration file
   */

  FILE *file;
  int  narg;
  int  chips_found=0;
  int  chip=0;
  int  chan,trim;
  int  start=0;
  char line[160],token[32];
  char fname[64],fname2[64];
  short trim_range;

  float arg[16];

  /* If the specified detector configuration file does not exist, an
   * attempt is made to load the default detector configuration file.
   * This is done to allow production testing to be done without the
   * creation of 2^n essentially identical files....
   * PWP 16.01.02 */

  sprintf(fname,"%s%s.det",ST_CONFIG_PATH,sct[mid].detname);
  sprintf(fname2,"%sdefault.det",ST_CONFIG_PATH);

  file = fopen(fname,"r");
  if (file == NULL){
    printf("st_read_module: file %s does not exist\n",fname);
    printf("  attempting to open the default file %s\n",fname2);

    file = fopen(fname2,"r");
    if (file == NULL){
      printf("st_read_module: failed to open file %s\n",fname2);
      return (-1);
    }
  }


  while(fgets(line,160,file) != (char*) NULL) {

    strncpy(token,line,4);
    token[4]=(char)NULL;
    narg = 1;

    if( strncmp(token,"#",1) == 0 ){    /* Ignore comments */
      narg = 0;
    }

    if( strncmp(token,"Modu",4) == 0 ){ /* First useful line has been found */
      start = 1;
      narg = 0;
    }

    if( strncmp(token,"Chip",4) == 0 ){ /* new chip configuration */
      chips_found++;
      narg = sscanf(line,"%s %d", token, &chip);
      narg = 0;
    }
      
    if( strncmp(token,"RC_D",4) == 0 ){ /* Response curve and calibration 
                                           capacitor correction factor */
      start = 2;
      narg = 0;
    }

    if( strncmp(token,"Trim",4) == 0 ){ /* trim information */
      start = 3;
      narg = 0;
      chan = 0;
    }

    if( strncmp(token,"Mask",4) == 0 ){ /* trim information */
      start = 4;
      narg = 0;
    }
      
    if((start==1)&&(narg>0)){
      narg = sscanf(line,"%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f",
        &arg[0], &arg[1], &arg[2], &arg[3],        
        &arg[4], &arg[5], &arg[6], &arg[7],        
        &arg[8], &arg[9], &arg[10],&arg[11],        
        &arg[12],&arg[13],&arg[14],&arg[15]);

      if(chips_found==0){
        if(narg==5){ 
          printf("st_read_module: stream enables, oddity, chipset and dtm\n");

          sct[mid].link0       = (short)arg[0];
          sct[mid].link1       = (short)arg[1];
          sct[mid].det.evenodd = (short)arg[2]; 
          sct[mid].det.chipset = (short)arg[3];
          sct[mid].det.dtm     = (short)arg[4];

          sct[mid].det.sc_mode   = 2;  /* default conditions for */
          sct[mid].det.bpm_drive = 3;  /*  Liverpool forward SC  */
          sct[mid].det.vdc_dac0  = 512;
          sct[mid].det.vdc_dac1  = 512;

          sct[mid].suspend0      = 0;  /* clear suspense flags */
          sct[mid].suspend1      = 0;  /* PWP 11.04.01         */

        } else if(narg==9){ 
          printf("st_read_module: stream enables, oddity, chipset, dtm and FWD SC\n");

          sct[mid].link0       = (short)arg[0];
          sct[mid].link1       = (short)arg[1];
          sct[mid].det.evenodd = (short)arg[2]; 
          sct[mid].det.chipset = (short)arg[3];
          sct[mid].det.dtm     = (short)arg[4];

          sct[mid].det.sc_mode   = (short)arg[5];
          sct[mid].det.bpm_drive = (short)arg[6];
          sct[mid].det.vdc_dac0  = (short)arg[7];
          sct[mid].det.vdc_dac1  = (short)arg[8];

          sct[mid].suspend0      = 0;  /* clear suspense flags */
          sct[mid].suspend1      = 0;  /* PWP 11.04.01         */

        } else if(narg==13){
          printf("st_read_module: LV and HV settings - old style\n");

          sct[mid].det.select = (short)arg[0];
          sct[mid].det.vdet_set   = arg[1];
          sct[mid].det.idet_lim   = arg[2];
          sct[mid].det.vcc_set    = arg[3];
          sct[mid].det.icc_lim    = arg[4];
          sct[mid].det.vdd_set    = arg[5];
          sct[mid].det.idd_lim    = arg[6];
          sct[mid].det.vi1_set    = arg[7];
          sct[mid].det.ivi1_lim   = arg[8];
          sct[mid].det.vled0_set  = arg[9];
          sct[mid].det.iled0_lim  = arg[10];
          sct[mid].det.vled1_set  = arg[11];
          sct[mid].det.iled1_lim  = arg[12];
          sct[mid].det.vpin_set   = (float)6.0; /* default */
          sct[mid].det.ramp       = (short)4; /* slowest */

        } else if(narg==14){
          printf("st_read_module: LV and HV settings - with Vpin extension\n");

          sct[mid].det.select = (short)arg[0];
          sct[mid].det.vdet_set   = arg[1];
          sct[mid].det.idet_lim   = arg[2];
          sct[mid].det.vcc_set    = arg[3];
          sct[mid].det.icc_lim    = arg[4];
          sct[mid].det.vdd_set    = arg[5];
          sct[mid].det.idd_lim    = arg[6];
          sct[mid].det.vi1_set    = arg[7];
          sct[mid].det.ivi1_lim   = arg[8];
          sct[mid].det.vled0_set  = arg[9];
          sct[mid].det.iled0_lim  = arg[10];
          sct[mid].det.vled1_set  = arg[11];
          sct[mid].det.iled1_lim  = arg[12];
          sct[mid].det.vpin_set   = arg[13];
          sct[mid].det.ramp   = (short)4; /* slowest */

        } else if(narg==15){
          printf("st_read_module: LV and HV settings - with Vpin and ramp extensions\n");

          sct[mid].det.select = (short)arg[0];
          sct[mid].det.vdet_set   = arg[1];
          sct[mid].det.idet_lim   = arg[2];
          sct[mid].det.vcc_set    = arg[3];
          sct[mid].det.icc_lim    = arg[4];
          sct[mid].det.vdd_set    = arg[5];
          sct[mid].det.idd_lim    = arg[6];
          sct[mid].det.vi1_set    = arg[7];
          sct[mid].det.ivi1_lim   = arg[8];
          sct[mid].det.vled0_set  = arg[9];
          sct[mid].det.iled0_lim  = arg[10];
          sct[mid].det.vled1_set  = arg[11];
          sct[mid].det.iled1_lim  = arg[12];
          sct[mid].det.vpin_set   = arg[13];
          sct[mid].det.ramp   = (short)arg[14];
        }

      } else if((narg==14)&&(chip<CHIPS_IN_MODULE)){
        /* old format: cal_enable and test_mode are still read
           as seperate columns but the result is stored in the 
           trim_range variable PWP 15.08.00 */

        /* printf("st_read_module: chip configuration, chip %d\n",chip); */

        sct[mid].det.chips[chip].compression  = (short)arg[0];
        sct[mid].det.chips[chip].active       = (short)arg[1];
        sct[mid].det.chips[chip].cal_mode     = (short)arg[2];

        trim_range = 0;
        if((short)arg[3]>0) trim_range += 0x0001;  /* cal_enable */
        if((short)arg[4]>0) trim_range += 0x0002;  /* test_mode  */
        sct[mid].det.chips[chip].trim_range   = trim_range;

        sct[mid].det.chips[chip].mask_readout = (short)arg[5];
        sct[mid].det.chips[chip].edge_detect  = (short)arg[6];
        sct[mid].det.chips[chip].accumulate   = (short)arg[7];
        sct[mid].det.chips[chip].delay        = arg[8];
        sct[mid].det.chips[chip].vthr         = arg[9];
        sct[mid].det.chips[chip].vcal         = arg[10];
        sct[mid].det.chips[chip].feshaper     = arg[11];
        sct[mid].det.chips[chip].febias       = arg[12];
        sct[mid].det.chips[chip].role         = (short)arg[13];

        /* no longer read the mask - enable all channels by default */
        sct[mid].det.chips[chip].mask[0] = (short)0xffff;
        sct[mid].det.chips[chip].mask[1] = (short)0xffff;
        sct[mid].det.chips[chip].mask[2] = (short)0xffff;
        sct[mid].det.chips[chip].mask[3] = (short)0xffff;
        sct[mid].det.chips[chip].mask[4] = (short)0xffff;
        sct[mid].det.chips[chip].mask[5] = (short)0xffff;
        sct[mid].det.chips[chip].mask[6] = (short)0xffff;
        sct[mid].det.chips[chip].mask[7] = (short)0xffff;

        /* set default value of c_factor - PWP 19.02.02 */
        sct[mid].det.chips[chip].c_factor     = 1.0;

      } else if((narg==13)&&(chip<CHIPS_IN_MODULE)){
        /* new format: cal_enable and test_mode merged to form 
           trim_range - one column less to read. PWP 15.08.00 */
 
        /* printf("st_read_module: chip configuration, chip %d\n",chip); */

        sct[mid].det.chips[chip].compression  = (short)arg[0];
        sct[mid].det.chips[chip].active       = (short)arg[1];
        sct[mid].det.chips[chip].cal_mode     = (short)arg[2];
        sct[mid].det.chips[chip].trim_range   = (short)arg[3];
        sct[mid].det.chips[chip].mask_readout = (short)arg[4];
        sct[mid].det.chips[chip].edge_detect  = (short)arg[5];
        sct[mid].det.chips[chip].accumulate   = (short)arg[6];
        sct[mid].det.chips[chip].delay        = arg[7];
        sct[mid].det.chips[chip].vthr         = arg[8];
        sct[mid].det.chips[chip].vcal         = arg[9];
        sct[mid].det.chips[chip].feshaper     = arg[10];
        sct[mid].det.chips[chip].febias       = arg[11];
        sct[mid].det.chips[chip].role         = (short)arg[12];

        /* no longer read the mask - enable all channels by default */
        sct[mid].det.chips[chip].mask[0] = (short)0xffff;
        sct[mid].det.chips[chip].mask[1] = (short)0xffff;
        sct[mid].det.chips[chip].mask[2] = (short)0xffff;
        sct[mid].det.chips[chip].mask[3] = (short)0xffff;
        sct[mid].det.chips[chip].mask[4] = (short)0xffff;
        sct[mid].det.chips[chip].mask[5] = (short)0xffff;
        sct[mid].det.chips[chip].mask[6] = (short)0xffff;
        sct[mid].det.chips[chip].mask[7] = (short)0xffff;

        /* set default value of c_factor - PWP 19.02.02 */
        sct[mid].det.chips[chip].c_factor     = 1.0;
      }

    } else if((start==2)&&(narg>0)){ /* Response curve and calibration 
                                        capacitor correction factor */
      narg = sscanf(line,"%d %f %f %f %f %f %f",
        &chip, &arg[0], &arg[1], &arg[2], &arg[3], &arg[4], &arg[5]);
      if((narg==2)&&(chip<CHIPS_IN_MODULE)){
        sct[mid].det.chips[chip].c_factor     = arg[0];
        sct[mid].det.chips[chip].rc_method    = 0;
        sct[mid].det.chips[chip].rc_function  = 0;
        sct[mid].det.chips[chip].rc_par[0]    = (float) 0;
        sct[mid].det.chips[chip].rc_par[1]    = (float) 0;
        sct[mid].det.chips[chip].rc_par[2]    = (float) 0;
      }
      else if((narg==6)&&(chip<CHIPS_IN_MODULE)){
        sct[mid].det.chips[chip].rc_method    = (int) arg[0];
        sct[mid].det.chips[chip].rc_function  = (int) arg[1];
        sct[mid].det.chips[chip].rc_par[0]    = arg[2];
        sct[mid].det.chips[chip].rc_par[1]    = arg[3];
        sct[mid].det.chips[chip].rc_par[2]    = arg[4];
        sct[mid].det.chips[chip].c_factor     = 1.0;
      }
      else if((narg==7)&&(chip<CHIPS_IN_MODULE)){
        sct[mid].det.chips[chip].rc_method    = (int) arg[0];
        sct[mid].det.chips[chip].rc_function  = (int) arg[1];
        sct[mid].det.chips[chip].rc_par[0]    = arg[2];
        sct[mid].det.chips[chip].rc_par[1]    = arg[3];
        sct[mid].det.chips[chip].rc_par[2]    = arg[4];
        sct[mid].det.chips[chip].c_factor     = arg[5];
      }
    } else if((start==3)&&(narg>0)){ /* trim information */
      narg = sscanf(line,"%d",&trim);
      if((narg==1)&&(chan>-1)&&(chan<1536)){
        if((trim>-1)&&(trim<16)){
          sct[mid].det.chips[(chan/128)].trim[(chan%128)]=trim;
        }
        chan++;
      }
    } else if((start==4)&&(narg>0)){ /* mask information */
      narg = sscanf(line,"%d",&chan);
      if((narg==1)&&(chan>-1)&&(chan<1536)){
        st_mask_chan(mid,(short)chan);
      }
    }
  }  /* end while */

  if(file){
    fclose(file);
  }

  /* If the module name is "default", we prompt the user 
   * to enter the serial number and device type
   */

  if(strcmp(sct[mid].detname,"default")==0){
    printf("The serial number of module %d is not known.\n",mid);

    narg=0;
    while(narg==0){
      printf("  Please enter serial number:");
      scanf("%s",&token);
      printf("  You entered %s.  Is this correct (0/1):",token);
      scanf("%d",&narg);
    }
    if(narg==1){
      sprintf(sct[mid].detname,"%s",token);
    }

    printf("The device type of module %d is not known.\n",mid);

    narg=0;
    while(narg==0){
      printf("  Which of the following device types describes this device:\n");
      printf("  0 Unknown\n");
      printf("  1 Barrel_Hybrid\n");
      printf("  2 Barrel_Module\n");
      printf("  3 Forward_Hybrid\n");
      printf("  4 Forward_Inner\n");
      printf("  5 Forward_Middle\n");
      printf("  6 Forward_Outer\n");
      printf("  Your Choice:");
      scanf("%d",&chip);
      printf("  You entered %d.  Is this correct (0/1):",chip);
      scanf("%d",&narg);
    }
    if(narg==1){
      switch(chip){
        case 1:
          sprintf(sct[mid].devtype,"Barrel_Hybrid");
          break;
        case 2:
          sprintf(sct[mid].devtype,"Barrel_Module");
          break;
        case 3:
          sprintf(sct[mid].devtype,"Forward_Hybrid");
          break;
        case 4:
          sprintf(sct[mid].devtype,"Forward_Inner");
          break;
        case 5:
          sprintf(sct[mid].devtype,"Forward_Middle");
          break;
        case 6:
          sprintf(sct[mid].devtype,"Forward_Outer");
          break;
        default:
          sprintf(sct[mid].devtype,".");
          break;
      }
    }else{
      sprintf(sct[mid].devtype,".");
    }
  }

  return(0);
}


void st_show_module(int mid)
{

  int i;

  printf("Module : %d %s (%s)\n",mid, sct[mid].detname,sct[mid].devtype);
  printf(" Present %hd,  Active  %hd\n",
       sct[mid].present, sct[mid].active);
  printf(" Use links: link0 %hd  link1  %hd\n",
       sct[mid].link0, sct[mid].link1);
  printf(" Suspend links: suspend0 %hd  suspend1  %hd\n",
       sct[mid].suspend0, sct[mid].suspend1);  /* PWP 11.04.01 */
  printf(" Chipset %hd, Oddity %hd, Select %hd, DTM %hd\n",
       sct[mid].det.chipset,sct[mid].det.evenodd,
       sct[mid].det.select,sct[mid].det.dtm);
  printf(" FWD SC mode %d, BPM drive %d, VDC DAC0 %d VDC DAC1 %d\n",
       sct[mid].det.sc_mode, sct[mid].det.bpm_drive,
       sct[mid].det.vdc_dac0, sct[mid].det.vdc_dac1);

  printf("         V      Ilim\n");
  printf(" Vdet  %6.2f  %6.2f  ramp %d\n Vcc   %6.2f  %6.2f\n Vdd   %6.2f  %6.2f\n",
       sct[mid].det.vdet_set,sct[mid].det.idet_lim,sct[mid].det.ramp,
       sct[mid].det.vcc_set, sct[mid].det.icc_lim,
       sct[mid].det.vdd_set,sct[mid].det.idd_lim);
  printf(" Vi1   %6.2f  %6.2f\n Vled0 %6.2f  %6.2f\n Vled1 %6.2f  %6.2f\n Vpin  %6.2f\n", 
       sct[mid].det.vi1_set,sct[mid].det.ivi1_lim,
       sct[mid].det.vled0_set,sct[mid].det.iled0_lim,
       sct[mid].det.vled1_set,sct[mid].det.iled1_lim,sct[mid].det.vpin_set);
  printf("         V       mA\n\n");

  printf(" Chip Role    Comp Act Cal_md T_range mask_rd edge_dt acc_md\n");
  for (i=0; i<12; i++) {
    printf(" %3d  %s  %3d %3d  %3d    %3d    %3d     %3d    %3d\n",
      i, roles[sct[mid].det.chips[i].role],
      sct[mid].det.chips[i].compression,
      sct[mid].det.chips[i].active,
      sct[mid].det.chips[i].cal_mode,
      sct[mid].det.chips[i].trim_range,
      sct[mid].det.chips[i].mask_readout,
      sct[mid].det.chips[i].edge_detect,
      sct[mid].det.chips[i].accumulate);
  }
  printf("\n Chip Delay  Vthr    Vcal  FE_S  FE_B   mk0  mk1  mk2  mk3  mk4  mk5  mk6  mk7\n");
  for (i=0; i<12; i++) {
    printf(" %3d   %2.0f  %6.2f  %6.2f  %4.1f  %5.1f  %4hx %4hx %4hx %4hx %4hx %4hx %4hx %4hx \n",
      i, sct[mid].det.chips[i].delay,
      sct[mid].det.chips[i].vthr,
      sct[mid].det.chips[i].vcal,
      sct[mid].det.chips[i].feshaper,
      sct[mid].det.chips[i].febias, 
      sct[mid].det.chips[i].mask[0],
      sct[mid].det.chips[i].mask[1],
      sct[mid].det.chips[i].mask[2],
      sct[mid].det.chips[i].mask[3],
      sct[mid].det.chips[i].mask[4],
      sct[mid].det.chips[i].mask[5],
      sct[mid].det.chips[i].mask[6],
      sct[mid].det.chips[i].mask[7]);
  }
  printf("\n");
}



int st_scanpoint_file(int index){

  FILE *f;
  char fname[128];
  int narg;
  char input_string[128];
  char c;

  short module_id;
  short typ;
  float var, newvar;

#ifdef WINNT
  sprintf(fname, "%sscans\\scan.%d", ST_PATH, index);
#else
  sprintf(fname, "%sscans/scan.%d", ST_PATH, index);
#endif

  printf("st_scanpoint_file: attempting to open file %s\n", fname);  

  if ((f = fopen(fname,"r")) == NULL){
    printf(" ... ERROR: cannot open %s for reading\n", fname);
    return -1;
  } else {
    printf(" ... opened successfully\n");
  }

  while (fgets(input_string, 128, f) != (char *) NULL){

    printf("%s\n", input_string);

    c = input_string[0];
    if (c != '#') {

      narg = sscanf(input_string,"%hd %hd %f", &module_id, &typ, &var);

      if (narg != 3) printf("problem ... narg != 3 but %d\n", narg);

      st_coerce_variable(typ, var, &newvar,0); /* PWP 15.03.00 */
      printf("st_scanpoint_file: mod %d  vartyp %d request %3.2f coerced %3.2f\n",
         module_id, typ, var, newvar);

      st_configure_variable(module_id, typ, newvar);
    }
  }

  if(f){
    fclose(f);
  }
  return 0;
}


int st_read_trims(int mid){

  char fname[128];
  char line[32];
  FILE *file;
  int i = 0;
  float trimf;
  short chip, range, trim;

  if(sct[mid].det.chipset!=ABCDT){
    printf("st_read_trims: module %d is not ABCDT - returning\n",mid);
    return -2;
  }

  sprintf(fname,"%s%s.trim",ST_CONFIG_PATH,sct[mid].detname);

  file = fopen(fname,"r");
  if (file == NULL){
    printf("st_read_trims: failed to open file %s\n",fname);
    return -1;
  }

  while (fgets(line, 32, file) != (char *) NULL){

    /* Revised PWP 10.05.01 to read trimrange settings from trimfile */

    if(strncmp(line,"Chip",4) == 0){ /* TrimRange information */

      sscanf(line,"%s %d %d", fname, &chip, &range);
      if((chip>=0)&&(chip<12)&&(range>=0)&&(range<4)){
        sct[mid].det.chips[chip].trim_range = range;
      }

    }else{                           /* TrimDAC   information */

      sscanf(line,"%f", &trimf);

      /*    trim =(short)((trimf+(float)2.5) / (float)5.0);  */
      trim = (short)((trimf + 0.01) * 15.0);      /* Carlos' format */
      if(i<1536){
        sct[mid].det.chips[(i/128)].trim[(i%128)] = trim;
      }
      /* printf("  ch %d trimf %f trim %hd\n",i,trimf,trim); */ 
      i++;
    }
  }

  printf("st_read_trims: module %d - data for %d channels\n",mid, i);

  if(file){
    fclose(file);
  }
  sct[mid].det.updatetrims = 1;
  return 0;
}

int st_show_trims(int mid){

  int chp,indx,s;
  unsigned int sum;
  float mean[12];

  if(sct[mid].det.chipset!=ABCDT){
    printf("st_show_trims: module %d is not ABCDT - returning\n",mid);
    return -2;
  }

  printf("st_show_trims: data for module %d\n",mid);
  printf("chp indx s0  - s31\n");  
  for (chp=0; chp<12; chp++){
    sum = 0;
    if (sct[mid].det.chips[chp].role != MISSING){
      for(indx=0;indx<4;indx++){
        printf("%2d %2d ",chp,indx);
          for(s=indx;s<128;s+=4){
          printf("%hd ",sct[mid].det.chips[chp].trim[s]);
          sum += sct[mid].det.chips[chp].trim[s];
        }
        printf ("\n");
      }
    }
    mean[chp] = ((float)sum)/(float)128;
  }
  printf("\nmean trim settings for chips 0 - 12\n");  

  for (chp=0; chp<12; chp++){
    printf("%3.2f ",mean[chp]);
  }
  printf ("\n");

  return 0;
}

int st_read_rc(int mid){

  char fname[128];
  char line[128];
  FILE *file;
  int a = 0;
  int i,j,k;
  float p0,p1,p2;

  sprintf(fname,"%s%s.rcdat",ST_CONFIG_PATH,sct[mid].detname);

  file = fopen(fname,"r");
  if (file == NULL){
    printf("st_read_rc: failed to open file %s\n",fname);
    return -1;
  }

  fgets(line, 128, file);  /* skip first line */

  while (fgets(line, 128, file) != (char *) NULL){

    sscanf(line,"%d %d %d %f %f %f", &i, &j, &k, &p0, &p1, &p2);

    sct[mid].det.chips[i].rc_method = j;
    sct[mid].det.chips[i].rc_function = k;
    sct[mid].det.chips[i].rc_par[0] = p0;
    sct[mid].det.chips[i].rc_par[1] = p1;
    sct[mid].det.chips[i].rc_par[2] = p2;

    printf("st_read_rc: chip %d mthd %d func %d  p0=%3.2f p1=%3.2f p2=%3.2f\n" ,i,j,k, p0,p1,p2);  
    a++;
  }

  printf("st_read_rc: module %d - data for %d chips\n",mid, a);

  if(file){
    fclose(file);
  }
  return 0;
}

int st_show_rc(int mid){

  int i;

  printf("st_show_rc: data for module %d\n",mid);
  printf("chip\tmthd\tfunc\tp0\tp1\tp2\tc_factor\n");

  for(i=0;i<12;i++){
    printf("%d\t%d\t%d\t%3.2f\t%3.2f\t%3.2f\t%3.2f\n",
      i, sct[mid].det.chips[i].rc_method,
      sct[mid].det.chips[i].rc_function, sct[mid].det.chips[i].rc_par[0],
      sct[mid].det.chips[i].rc_par[1], sct[mid].det.chips[i].rc_par[2],
      sct[mid].det.chips[i].c_factor);
  }

  return 0;
}

int st_get_rc(int mid, int cid, short* method, short* function, 
              float* p0, float* p1, float* p2){

  /* PWP 10.05.01 */

  if((mid<0) ||(mid>SCT_MAX_MODULES)) return -1;
  if((cid<0) ||(cid>11)) return -2;

  *method = sct[mid].det.chips[cid].rc_method;
  *function = sct[mid].det.chips[cid].rc_function;
  *p0 = sct[mid].det.chips[cid].rc_par[0];
  *p1 = sct[mid].det.chips[cid].rc_par[1];
  *p2 = sct[mid].det.chips[cid].rc_par[2];

  return 0;
}

int st_set_rc(int mid, int cid, short method, short function, 
              float p0, float p1, float p2){

  /* PWP 23.11.00 */

  if((mid<0) ||(mid>SCT_MAX_MODULES)) return -1;
  if((cid<0) ||(cid>11)) return -2;

  sct[mid].det.chips[cid].rc_method = method;
  sct[mid].det.chips[cid].rc_function = function;
  sct[mid].det.chips[cid].rc_par[0] = p0;
  sct[mid].det.chips[cid].rc_par[1] = p1;
  sct[mid].det.chips[cid].rc_par[2] = p2;

  return 0;
}

int st_read_mask(int mid){

  /* PWP 14.01.00 */

  char fname[128];
  char line[128];
  FILE *file;
  short i,j,chan;


  sprintf(fname,"%s%s.mask",ST_CONFIG_PATH,sct[mid].detname);

  file = fopen(fname,"r");
  if (file == NULL){
    printf("st_read_mask: failed to open file %s\n",fname);
    return -1;
  }

  /* Mask file exists: 
   * start by clearing all presently masked channels
   */

  for(i=0;i<12;i++){
    for(j=0;j<8;j++) sct[mid].det.chips[i].mask[j] = (short) 0xffff;
  }

  /* Read mask file: 
   * all channels whose numbers appear are masked
   */

  i=0;

  while (fgets(line, 128, file) != (char *) NULL){

    sscanf(line,"%hd", &chan);

    st_mask_chan(mid,chan);
    i++;
  }

  printf("st_read_mask: module %hd - %hd channels masked\n",mid, i);

  if(file){
    fclose(file);
  }

  /* return number of channels masked */
  return (int)i;
}


int st_mask_chan(int mid, short chan){

  short chip, word;
  short mask;

  if (mid >= SCT_MAX_MODULES){
    printf("st_mask_chan: module %hd outside limit of %d modules\n", mid, SCT_MAX_MODULES);
    return -1;
  }

  if ((chan<0)||(chan>1535)){
    printf("st_mask_chan: module %hd channel %hd outside range of 0 to 1535\n", mid, chan);
    return -2;
  }

  chip = chan / 128;
  word = 7-((chan % 128) / 16);
  mask = 0x1<<((chan%16));
  if (mask>0xffff) mask = 0x1;

  /* printf("  channel %3hd  chip %2hd  word %hd => 0x%4x  mask 0x%4x 0x%4x\n",
    chan, chip, word, sct[mid].det.chips[(chip)].mask[word],
	mask, (0xffff-mask)); */

  if(sct[mid].det.chips[chip].mask[word] & mask){
 
   sct[mid].det.chips[(chip)].mask[word] = 
      sct[mid].det.chips[(chip)].mask[word] &(0xffff-mask);

  }else{
    printf("st_mask_chan: module %hd channel %hd is already masked\n",mid,chan);
    return 0;
  }

  return 1;
}

void st_reverse_sysmap(){

  int m, mid, sid;

  /* RAW data comes in stream order and we need to 
   * know how to map the data to module histograms */

  /* First: initialise arrays to -1 */

  printf("st_reverse_sysmap:\n");

  for (mid=0; mid<MUS_NUMBER; mid++){
    for (sid=0; sid<12; sid++){
     stream_2_mod[mid][sid] = -1;
    }
  }

  for (m=0; m<SCT_MAX_MODULES; m++){
    if(sct[m].mus_id<MUS_NUMBER){

      /*            MuSTARD            STREAM          */
      if (sct[m].mus_enable0){
        printf("  MuSTARD %d Stream %d -> Module %d Link %d\n",
          sct[m].mus_id,sct[m].mus_ch0,m,0);
        stream_2_mod[sct[m].mus_id][sct[m].mus_ch0] = m;
        stream_2_link[sct[m].mus_id][sct[m].mus_ch0] = 0;
      }
      if (sct[m].mus_enable1){
        printf("  MuSTARD %d Stream %d -> Module %d Link %d\n",
          sct[m].mus_id,sct[m].mus_ch1,m,1);
        stream_2_mod[sct[m].mus_id][sct[m].mus_ch1] = m;
        stream_2_link[sct[m].mus_id][sct[m].mus_ch1] = 1;
      }
    }
  }
}

short st_find_link(short mid, short sid){

  if(mid<MUS_NUMBER)
    return(stream_2_link[mid][sid]);
  else
    return -1;
}

short st_find_module(short mid, short sid){

  if(mid<MUS_NUMBER)
    return(stream_2_mod[mid][sid]);
  else
    return -1;
}
  
short st_find_mustard(short m){

  if(m<0) return -1;
  if(m>=SCT_MAX_MODULES) return -2;

  return(sct[m].mus_id);
}
  
short st_find_stream(short m, short soft_link){

  if(soft_link==0){
    if (sct[m].link1==2)      /* Redundancy in use */
      return sct[m].mus_ch1;
    else                      /* Normal Readout */
      return sct[m].mus_ch0;

  } else {

    if (sct[m].link0==2)      /* Redundancy in use */
      return sct[m].mus_ch0;
    else                      /* Normal Readout */
      return sct[m].mus_ch1;
  }

}

short st_find_maxchips(short mid){

  short nchips,mchips,m;

  /*
   * How many chips worth of data might be expected from MuSTARD m?
   * We need to know this to assist timeout calculations using VBURST.
   * Allow 6 or 12 chips only - for the time being....  PWP 27.06.00
   */

  if(mid>=MUS_NUMBER)
    return 0;

  nchips=0;

  for(m=0;m<SCT_MAX_MODULES;m++){
    if ((sct[m].present)&&(sct[m].mus_id==mid)){

      mchips=0;

      if (sct[m].mus_enable0){
        if (sct[m].link0==2) mchips+=12;
        else mchips+=6;
      }

      if (sct[m].mus_enable1){
        if (sct[m].link1==2) mchips+=12;
        else mchips+=6;
      }

      if (mchips>6) nchips+=12;
      else nchips +=6;
    }
  }

  return nchips;
}

short st_find_optotx_id(short m){

  if(m<0) return -1;
  if(m>=SCT_MAX_MODULES) return -2;

  if(sct[m].det.select) return(sct[m].r_opto_tx_id);
  else                  return(sct[m].opto_tx_id);
}

short st_find_optotx_ch(short m){

  if(m<0) return -1;
  if(m>=SCT_MAX_MODULES) return -2;

  if(sct[m].det.select) return(sct[m].r_opto_tx_ch);
  else                  return(sct[m].opto_tx_ch);
}

short st_find_optif_rx_ch0(short m){

  if(m<0) return -1;
  if(m>=SCT_MAX_MODULES) return -2;

  return(sct[m].opto_rx_ch0);
}

short st_find_optif_rx_ch1(short m){

  if(m<0) return -1;
  if(m>=SCT_MAX_MODULES) return -2;

  return(sct[m].opto_rx_ch1);
}

short st_n_active_channels(short mid){

  short n_ch = 0;
  short cid, word, bit;

  if((mid<0)||(mid>=SCT_MAX_MODULES)) return 0;

  if((sct[mid].present)&&(sct[mid].active)){
    for (cid=0; cid<12; cid++){
      if((sct[mid].det.chips[cid].active) &&
         (sct[mid].det.chips[cid].role!=DEAD) &&
         (sct[mid].det.chips[cid].role!=MISSING)){

        for(word=0;word<8;word++){
          for(bit=0;bit<16;bit++){
            if(sct[mid].det.chips[cid].mask[word]&(0x0001<<bit)){
              n_ch++;
            }
          }
        }
      }
    }
  }

  return n_ch;
}

int st_write_module(int mid, int run, int scan) {
  /* 
   * PWP 29.08.01
   */

  FILE *file;
  short chip, chan;
  short word, mask;
  char fname[64];

  sprintf(fname,"%s%s_%d_%d.det",ST_CONFIG_PATH,sct[mid].detname,run,scan);

  file = fopen(fname,"w");
  if (file == NULL){
    printf("st_write_module: failed to open file %s\n",fname);
    return (-1);
  } else {

    fprintf(file, "#\n");
    fprintf(file, "# Module configuration file automatically generated by SCTDAQ\n");
    fprintf(file, "#\n");

    fprintf(file, "Module :  Lk0  Lk1 Odd Chipset DTM SCm bpm vdac0 vdac1\n");

    fprintf(file, "            %d   %d   %d    %d     %d   %d   %d   %d   %d\n",
      sct[mid].link0,
      sct[mid].link1,
      sct[mid].det.evenodd,
      sct[mid].det.chipset,
      sct[mid].det.dtm,
      sct[mid].det.sc_mode,
      sct[mid].det.bpm_drive,
      sct[mid].det.vdc_dac0,
      sct[mid].det.vdc_dac1);

    fprintf(file, "         Select  Vdet  Idet   Vcc   Icc   Vdd   Idd   Vi1  iVi1 Vld0 Ild0 Vld1 Ild1 Vpin Ramp\n");

    fprintf(file, "            %d   %3.1f  %3.1f  %3.2f  %4.0f  %3.2f  %4.0f  ",
      sct[mid].det.select,
      sct[mid].det.vdet_set,
      sct[mid].det.idet_lim,
      sct[mid].det.vcc_set,
      sct[mid].det.icc_lim,
      sct[mid].det.vdd_set,
      sct[mid].det.idd_lim);

    fprintf(file, "%3.2f  %2.0f  %3.2f  %2.0f  %3.2f  %2.0f  %3.2f  %d\n",
      sct[mid].det.vi1_set,
      sct[mid].det.ivi1_lim,
      sct[mid].det.vled0_set,
      sct[mid].det.iled0_lim,
      sct[mid].det.vled1_set,
      sct[mid].det.iled1_lim,
      sct[mid].det.vpin_set,
      sct[mid].det.ramp);

    fprintf(file, "          Cmp Act Cal Tr_r Mk_r Edge Acc Del Vth     Vcal    FEShp  FEBias Role\n");

    for(chip=0; chip<12;chip++){
      fprintf(file,"Chip %2d :\n",chip);

      fprintf(file, "            %d   %d   %d   %d   %d   %d   %d   ",
        sct[mid].det.chips[chip].compression,
        sct[mid].det.chips[chip].active,
        sct[mid].det.chips[chip].cal_mode,
        sct[mid].det.chips[chip].trim_range,
        sct[mid].det.chips[chip].mask_readout,
        sct[mid].det.chips[chip].edge_detect,
        sct[mid].det.chips[chip].accumulate);


      fprintf(file, "%2.0f  %3.2f   %2.2f   %2.1f   %3.1f   %d\n",
        sct[mid].det.chips[chip].delay,
        sct[mid].det.chips[chip].vthr,
        sct[mid].det.chips[chip].vcal,
        sct[mid].det.chips[chip].feshaper,
        sct[mid].det.chips[chip].febias,
        sct[mid].det.chips[chip].role);
    }

    fprintf(file, "RC_Data\n");
    for(chip=0; chip<12;chip++){
      fprintf(file, "%d  %d  %d  %3.2f  %3.2f  %3.2f  %3.2f\n",
      chip,
      sct[mid].det.chips[chip].rc_method,
      sct[mid].det.chips[chip].rc_function,
      sct[mid].det.chips[chip].rc_par[0],
      sct[mid].det.chips[chip].rc_par[1],
      sct[mid].det.chips[chip].rc_par[2],
      sct[mid].det.chips[chip].c_factor);
    }

    fprintf(file, "Trim Information\n");
    for(chan=0;chan<1536;chan++){
      fprintf(file,"%d\n",sct[mid].det.chips[(chan/128)].trim[(chan%128)]);
    }

    fprintf(file, "Masked Channels\n");
    for(chan=0;chan<1536;chan++){
      chip = chan / 128;
      word = 7-((chan % 128) / 16);
      mask = 0x1<<((chan%16));
      if (mask>0xffff) mask = 0x1;

      if((sct[mid].det.chips[chip].mask[word] & mask)==0){
        fprintf(file,"%d\n",chan);
      }
    }

    fclose(file);
  }

  return(0);
}

int st_write_module2(int mid, int run, int scan) {
  /* 
   * Generate configuration file in format suited to RODDAQ
   * PWP 15.07.02
   * fixes JCH 26.07.02
   */

  FILE *file;
  short chip, lastChip, nextChip;
  short start, word;
  short channel, nibble;
  char fname[64];
  unsigned int data;

  short inputBypass[12], outputBypass[12];
  short master[12], end[12], feedThrough[12];

  unsigned short oddity, address;

  sprintf(fname,"%s%s_%d_%d.det2",ST_RESULT_PATH,sct[mid].detname,run,scan);

  file = fopen(fname,"w");
  if (file == NULL){
    printf("st_write_module: failed to open file %s\n",fname);
    return (-1);
  } else {

    fprintf(file, "# Module configuration file automatically generated by SCTDAQ\n");
    fprintf(file, "# Run %d Scan %d\n",run,scan);
    fprintf(file, "#\n");


    fprintf(file,"Module active %d select %d\n",
     1, sct[mid].det.select);

    if (sct[mid].det.evenodd==AUTO){ /* oddity follows select line */
      if(sct[mid].det.select==EVEN) oddity = 0x0800; 
      else oddity = 0x0c00;
    } else {                         /* oddity is hard wired */
      if (sct[mid].det.evenodd==EVEN) oddity = 0x0800; 
      else oddity = 0x0c00;
    }

    /* Determine master, end, bypass settings */
    for(chip=0;chip<12;chip++){
      outputBypass[chip] = 0;
      inputBypass[chip]  = 0;
      master[chip]       = 0;  /* active high */
      end[chip]          = 0;  /* active high */
      feedThrough[chip]  = 1;  /* active low  */
    }


    for(chip=0;chip<12;chip++){
      lastChip = chip-1;
      nextChip = chip+1;

      if(chip==0)  lastChip = 11;
      if(chip==11) nextChip =  0;

      switch(sct[mid].det.chips[chip].role){
        case MISSING:
          /* who cares! */
          break;
        case DEAD:
          inputBypass[lastChip]  = 1;
          outputBypass[nextChip] = 1;
          break;
        case END:
          master[chip] = 0;
          end[chip]    = 1;
          feedThrough[chip] = 1;
          break;
        case MASTER:
          master[chip] = 1;
          end[chip]    = 0;
          feedThrough[chip] = 0;  /* active low */
          break;
        case SLAVE:
          master[chip] = 0;
          end[chip]    = 0;
          feedThrough[chip] = 1;
          break;
        case LONELY:
          master[chip] = 1;
          end[chip]    = 1;
          feedThrough[chip] = 1;
        case PARANOID:
          inputBypass[lastChip]  = 1;
          outputBypass[nextChip] = 1;
          master[chip] = 1;
          end[chip]    = 1;
          feedThrough[chip] = 1;
          break;
        default:
          break;
      }
    }



    for(chip=0;chip<12;chip++){

      address = (oddity | chipid[chip])>>6;

      fprintf(file,"Chip %02d address 0x%02x active %d\n",
        chip, address,
        sct[mid].det.chips[chip].active);

      fprintf(file,"  config %d %d %d %d %d %d %d %d %d %d %d\n",
        sct[mid].det.chips[chip].compression,
        sct[mid].det.chips[chip].cal_mode,
        sct[mid].det.chips[chip].trim_range,
        sct[mid].det.chips[chip].edge_detect,
        sct[mid].det.chips[chip].mask_readout,
        sct[mid].det.chips[chip].accumulate,
        inputBypass[chip],
        outputBypass[chip],
        master[chip],
        end[chip],
        feedThrough[chip]);

      fprintf(file,"  mask 0x%04hx%04hx 0x%04hx%04hx 0x%04hx%04hx 0x%04hx%04hx\n",
        sct[mid].det.chips[chip].mask[0],
        sct[mid].det.chips[chip].mask[1],
        sct[mid].det.chips[chip].mask[2],
        sct[mid].det.chips[chip].mask[3],
        sct[mid].det.chips[chip].mask[4],
        sct[mid].det.chips[chip].mask[5],
        sct[mid].det.chips[chip].mask[6],
        sct[mid].det.chips[chip].mask[7]);

      fprintf(file,"  vthr 0x%02x vcal 0x%02x delay 0x%02x preamp 0x%02x shaper 0x%02x\n",
        (int)((sct[mid].det.chips[chip].vthr+1.25)/2.5),
        (int)((sct[mid].det.chips[chip].vcal+0.312)/0.625),
        (int)((sct[mid].det.chips[chip].delay+0.5)/1.0),
        (int)((sct[mid].det.chips[chip].febias+4.6)/9.2),
        (int)((sct[mid].det.chips[chip].feshaper+0.6)/1.2) );

      fprintf(file,"  rc_function %d p0 %3.2e p1 %3.2e p2 %3.2e c_factor %3.2f\n",
        sct[mid].det.chips[chip].rc_function,
        sct[mid].det.chips[chip].rc_par[0],
        sct[mid].det.chips[chip].rc_par[1],
        sct[mid].det.chips[chip].rc_par[2],
        sct[mid].det.chips[chip].c_factor);

      for(start=0;start<128;start+=32){
        fprintf(file,"  trim %02x-%02x ",start,start+31);

        channel = start;
        for(word=0;word<4;word++){
          data = 0;
          for(nibble=0;nibble<7;nibble++){
            data|=(sct[mid].det.chips[chip].trim[channel++]&0x0f);
            data<<=4;
          }
          data|=(sct[mid].det.chips[chip].trim[channel++]&0x0f);
          fprintf(file," 0x%08x",data);
        }
        fprintf(file,"\n");
      }
    }

    fclose(file);
  }

  return (0);
}


int st_write_module_xml(int mid, int run, int scan, float version, char* location) {
  /* 
   * Generate module configuration file in xml format
   * PWP 14.12.02
   */

  FILE *file;
  short chip, lastChip, nextChip;
  short start, word;
  short channel, nibble;
  char fname[64];
  unsigned int data;

  short inputBypass[12], outputBypass[12];
  short master[12], end[12], feedThrough[12];
  unsigned short oddity, address;

  struct tm *cookedtime;
  time_t rawtime;

  sprintf(fname,"%s%s_%d_%d.xml",ST_RESULT_PATH,sct[mid].detname,run,scan);

  file = fopen(fname,"w");
  if (file == NULL){
    printf("st_write_module_xml: failed to open file %s\n",fname);
    return (-1);
  } else {

    /* Generate time stamp */

    time( &rawtime);
    cookedtime = localtime( &rawtime);


    if (sct[mid].det.evenodd==AUTO){ /* oddity follows select line */
      if(sct[mid].det.select==EVEN) oddity = 0x0800; 
      else oddity = 0x0c00;
    } else {                         /* oddity is hard wired */
      if (sct[mid].det.evenodd==EVEN) oddity = 0x0800; 
      else oddity = 0x0c00;
    }

    /* Determine master, end, bypass settings */
    for(chip=0;chip<12;chip++){
      outputBypass[chip] = 0;
      inputBypass[chip]  = 0;
      master[chip]       = 0;  /* active high */
      end[chip]          = 0;  /* active high */
      feedThrough[chip]  = 1;  /* active low  */
    }

    for(chip=0;chip<12;chip++){
      lastChip = chip-1;
      nextChip = chip+1;

      if(chip==0)  lastChip = 11;
      if(chip==11) nextChip =  0;

      switch(sct[mid].det.chips[chip].role){
        case MISSING:
          /* who cares! */
          break;
        case DEAD:
          inputBypass[lastChip]  = 1;
          outputBypass[nextChip] = 1;
          break;
        case END:
          master[chip] = 0;
          end[chip]    = 1;
          feedThrough[chip] = 1;
          break;
        case MASTER:
          master[chip] = 1;
          end[chip]    = 0;
          feedThrough[chip] = 0;  /* active low */
          break;
        case SLAVE:
          master[chip] = 0;
          end[chip]    = 0;
          feedThrough[chip] = 1;
          break;
        case LONELY:
          master[chip] = 1;
          end[chip]    = 1;
          feedThrough[chip] = 1;
        case PARANOID:
          inputBypass[lastChip]  = 1;
          outputBypass[nextChip] = 1;
          master[chip] = 1;
          end[chip]    = 1;
          feedThrough[chip] = 1;
          break;
        default:
          break;
      }
    }


    /* Write header */
    fprintf(file,"<module>\n");
    fprintf(file,"  <sn>%s</sn>\n",sct[mid].detname);
    fprintf(file,"  <source>generated by SCTDAQ</source>\n");

    fprintf(file,"  <location>%s</location>\n",location);

    fprintf(file,"  <date>%02d/%02d/%02d %02d:%02d:%02d</date>\n",
      cookedtime->tm_mday,
      cookedtime->tm_mon,
      (cookedtime->tm_year%100), /*tm_year is 102 for 2002*/
      cookedtime->tm_hour,
      cookedtime->tm_min,
      cookedtime->tm_sec);


    fprintf(file,"  <meta><version>%1.2f</version><run>%d</run><scan>%d</scan></meta>\n",
      version,run,scan);
    fprintf(file,"  <active>%d</active>\n",1);
    fprintf(file,"  <select>%d</select>\n",sct[mid].det.select);

    /* Write DCS info */
    fprintf(file,"  <dcs name=\"vdet\">%3.0f %4.2f</dcs>\n",
      sct[mid].det.vdet_set, sct[mid].det.idet_lim);
    fprintf(file,"  <dcs name=\"vcc\">%1.2f %4.0f</dcs>\n",
      sct[mid].det.vcc_set, sct[mid].det.icc_lim);
    fprintf(file,"  <dcs name=\"vdd\">%1.2f %4.0f</dcs>\n",
      sct[mid].det.vdd_set, sct[mid].det.idd_lim);

    /* Write chip info */
    for(chip=0;chip<12;chip++){

      address = (oddity | chipid[chip])>>6;

      fprintf(file,"  <chip id=\"%02d\" address=\"0x%02x\" active=\"%d\">\n",
        chip, address, sct[mid].det.chips[chip].active);

      fprintf(file,"    <config> %d %d %d %d %d %d %d %d %d %d %d</config>\n",
        sct[mid].det.chips[chip].compression,
        sct[mid].det.chips[chip].cal_mode,
        sct[mid].det.chips[chip].trim_range,
        sct[mid].det.chips[chip].edge_detect,
        sct[mid].det.chips[chip].mask_readout,
        sct[mid].det.chips[chip].accumulate,
        inputBypass[chip],
        outputBypass[chip],
        master[chip],
        end[chip],
        feedThrough[chip]);

      fprintf(file,"    <mask> 0x%04hx%04hx 0x%04hx%04hx 0x%04hx%04hx 0x%04hx%04hx</mask>\n",
        sct[mid].det.chips[chip].mask[0],
        sct[mid].det.chips[chip].mask[1],
        sct[mid].det.chips[chip].mask[2],
        sct[mid].det.chips[chip].mask[3],
        sct[mid].det.chips[chip].mask[4],
        sct[mid].det.chips[chip].mask[5],
        sct[mid].det.chips[chip].mask[6],
        sct[mid].det.chips[chip].mask[7]);

      fprintf(file,"    <reg name=\"vthr\">0x%02x</reg>\n",
        (int)((sct[mid].det.chips[chip].vthr+1.25)/2.5));
      fprintf(file,"    <reg name=\"vcal\">0x%02x</reg>\n",
        (int)((sct[mid].det.chips[chip].vcal+0.312)/0.625));
      fprintf(file,"    <reg name=\"delay\">0x%02x</reg>\n",
        (int)((sct[mid].det.chips[chip].delay+0.5)/1.0));
      fprintf(file,"    <reg name=\"preamp\">0x%02x</reg>\n",
        (int)((sct[mid].det.chips[chip].febias+4.6)/9.2));
      fprintf(file,"    <reg name=\"shaper\">0x%02x</reg>\n",
        (int)((sct[mid].det.chips[chip].feshaper+0.6)/1.2));

      fprintf(file,"    <c_factor>%3.2f</c_factor>\n",
        sct[mid].det.chips[chip].c_factor);

      fprintf(file,"    <rc_function type=\"%d\">p0 %3.2e p1 %3.2e p2 %3.2e</rc_function>\n",
        sct[mid].det.chips[chip].rc_function,
        sct[mid].det.chips[chip].rc_par[0],
        sct[mid].det.chips[chip].rc_par[1],
        sct[mid].det.chips[chip].rc_par[2],
        sct[mid].det.chips[chip].c_factor);

      for(start=0;start<128;start+=32){
        fprintf(file,"    <trim start=\"%02x\" end=\"%02x\">",start,start+31);

        channel = start;
        for(word=0;word<4;word++){
          data = 0;
          for(nibble=0;nibble<7;nibble++){
            data|=(sct[mid].det.chips[chip].trim[channel++]&0x0f);
            data<<=4;
          }
          data|=(sct[mid].det.chips[chip].trim[channel++]&0x0f);
          fprintf(file," 0x%08x",data);
        }
        fprintf(file,"</trim>\n");
      }
      fprintf(file,"  </chip>\n");
    }
    fprintf(file,"  </module>\n");

    fclose(file);
  }

  return (0);
}



