/*
 * stout.c - file output functions
 *
 * PWP 04.07.99 revised 16.07.99 
 * PWP 28.07.99 - revisions to match new st data format
 * PWP 13.08.99 - Version 0.2 - revised data format again...
 * PWP 25.08.99 - Version 0.21 - add protection against case
 *                where stout routines called when no file open;
 *                several variables moved to be arguments of
 *                stout_sob - no need for them to be global.
 * PWP 12.10.99 - Version 0.22 - change tmax to type int.
 *
 * Intended usage
 * ==============
 * At the start of a session, call: 
 *    stout_open()      to open the data file
 *    stout_sor()       to record the initial configuration
 *                      of the system.  (In turn this calls
 *                      stout_module_sysmap, stout_module_dcs
 *                      and stout_chip)
 *    stout_close()     to close the data file.
 *
 * At the start of each scan, call:
 *    stout_open()      to open the data file
 *
 * Prior to each burst / scanpoint, cal:
 *    stout_sob()       to record the intended burst settings
 *                      PWP 25.08.99 sob now called by tburst
 *
 * After each burst/scanpoint, call:
 *    stout_eob()       to record the occupancy data
 *                      PWP 25.08.99 eob now called by tburst
 *
 * At the end of each scan, call:
 *    stout_close()     to close the data file.
 *
 * Reference:
 * http://www.cern.ch/Atlas/GROUPS/INNER_DETECTOR/SCT/systemtest/st_data_format.html 
 * GFM 990724: st_data_format revised and updated.
 * *** PWP 25.08.99 *** does not presently match this format ***
 */

#include <stdio.h>
#include <string.h>
#include <time.h>

#ifdef RAID
#include <posix/time.h>
#endif

#include "abc.h"
#include "abc_config.h"
#include "sct.h"
#include "sct_hardware.h"
#include "st_config.h"
#include "stout.h"

extern short ascii_nt;

extern sysmap sct[SCT_MAX_MODULES];
extern short sct_histogram[SCT_MAX_MODULES][2][2048];
extern short mus_error_count[MUS_NUMBER];
extern int st_trigger_count; 

int run_number;  /* can be local to this module */
extern int scan_number;
extern int burst_number;

extern short scan_pnt[4];
extern short scan_typ[4];
extern float scan_var[4];

extern float scan_start[4];
extern float scan_stop[4];
extern float scan_step[4];
extern short scan_points[4];

extern FILE* outf;
extern FILE* musf;


int stout_open(short run){

  /* Open file for writing - append if file exists */

  char fname[64];

  run_number = run;

  sprintf(fname,"%sstrun%d.dat",ST_DATA_PATH,run);

  outf=fopen(fname,"a+");
  if (outf == NULL){ 
    printf("stout_open: Failed to open st data file %s\n",fname);
    return(-1);
  }
  printf("stout_open: File %s opened for writing\n",fname);
  return(0);
}


int stout_close(){

  /* Close file */

  if (outf != NULL){
    fclose(outf);
    outf = NULL;
  }
  return(0);
}


int stout_sor(){

  /* Write the complete system configuration to the ascii ntuple file */

  short mid,chp;

  if (outf == NULL){
    printf("stout_sor: called when no file open\n");
    return(-1);
  }

  for (mid=0; mid<SCT_MAX_MODULES; mid++){  /* loop over mid  */
    if (sct[mid].present){
      stout_module_sysmap(mid);
      stout_module_dcs(mid);
      for(chp=0; chp<12; chp++){            /* loop over chp  */
        if (sct[mid].det.chips[chp].role!=MISSING){
          stout_chip(mid,chp);
        }
      }
    }
  }
  return(0);
}


int stout_module_sysmap(short mid){  /* cal_mode = -5 */

  /* Write a module sysmap configuration record to the ascii ntuple file */

  int date_int, time_int;
  struct tm *cookedtime;
  time_t rawtime;

  if (outf == NULL){
    printf("stout_module_sysmap: called when no file open\n");
    return(-1);
  }

  time( &rawtime);
  cookedtime = localtime( &rawtime);

  date_int = ((cookedtime->tm_year + 1900) * 10000)
           + ((cookedtime->tm_mon+1) * 100) 
           +   cookedtime->tm_mday;

  time_int =  (cookedtime->tm_hour *3600) 
           +  (cookedtime->tm_min  * 60 ) 
           +   cookedtime->tm_sec;

  /* columns 1 to 11, a standard data row with the following 
   * hardcoded elements:
   * chip_id = -1:    applies to all chips equally
   * cal_mode = -5:   module sysmap configuration record
   * index_base = -1: irrelevant
   */

  fprintf(outf, "%ld %hd %hd %hd %3.2f %3.2f %3.2f %3.2f %ld %hd %hd ", 
    scan_number, mid, (short) -1, (short) -5,
    scan_var[4], scan_var[3], scan_var[2], scan_var[1],
    st_trigger_count, mus_error_count[sct[mid].mus_id], (short) -1);

  /* columns s0 to s11 */

  fprintf(outf, "%ld %ld %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd ",
    date_int, time_int,
    sct[mid].present, sct[mid].active,
    sct[mid].link0,
    sct[mid].link1,
    sct[mid].mus_id,
    sct[mid].mus_enable0,
    sct[mid].mus_enable1,
    sct[mid].mus_ch0,
    sct[mid].mus_ch1,
    sct[mid].mus_delay0,
    sct[mid].mus_delay1);

  /* columns s12 to s23 */

  fprintf(outf,"%hd %hd %4hd %4hd %hd %hd %hd %hd %4hd %4hd %hd %hd ",
    (short) 0,(short) 0,   /* TODO: rx0, rx1 */
    ((sct[mid].slog_id*100) + sct[mid].slog_ch),
    ((sct[mid].r_slog_id*100) + sct[mid].r_slog_ch),
    (short) 0,(short) 0,   /* TODO: clk phase and clk duty */
    (short) 0,(short) 0,
    ((sct[mid].opto_tx_id*100) + sct[mid].opto_tx_ch),
    ((sct[mid].r_opto_tx_id*100) + sct[mid].r_opto_tx_ch),
    (short) 0,(short) 0);  /* TODO: tx, r_tx */

  /*columns s24 to s31 */

  fprintf(outf,"%hd %hd %4hd %hd %hd %4hd %hd %hd\n",
    (short) 0,(short) 0,
    ( (sct[mid].lv_id*100) + sct[mid].lv_ch),
    (short) 0,(short) 0,
    ( (sct[mid].hv_id*100) + sct[mid].hv_ch),
    (short) 0,(short) 0);

  fflush(outf);
  return (0);
}


int stout_module_dcs(short mid){  /* cal_mode = -4 */

  /* Write a module environment record to the ascii ntuple file */

  int date_int, time_int;
  struct tm *cookedtime;
  time_t rawtime;

  if (outf == NULL){
    printf("stout_module_dcs: called when no file open\n");
    return(-1);
  }

  time( &rawtime);
  cookedtime = localtime( &rawtime);

  date_int = ((cookedtime->tm_year + 1900) * 10000)
           + ((cookedtime->tm_mon+1) * 100) 
           +   cookedtime->tm_mday;

  time_int =  (cookedtime->tm_hour *3600) 
           +  (cookedtime->tm_min  * 60 ) 
           +   cookedtime->tm_sec;

  /* columns 1 to 11, a standard data row with the following 
   * hardcoded elements:
   * chip_id = -1:    applies to all chips equally
   * cal_mode = -4:   module DCS configuration record
   * index_base = -1: irrelevant
   */

  fprintf(outf, "%ld %hd %hd %hd %3.2f %3.2f %3.2f %3.2f %ld %hd %hd ", 
    scan_number, mid, (short) -1, (short) -4,
    scan_var[4], scan_var[3], scan_var[2], scan_var[1],
    st_trigger_count, mus_error_count[sct[mid].mus_id], (short) -1);

  /* columns s0 to s3 */

  fprintf(outf, "%ld %ld %hd %hd",
    date_int, time_int,
    sct[mid].present, sct[mid].active);

  /* following assumes voltages in V and currents in mA */
  /* columns s4 to s10 */

  fprintf(outf,"%hd %3.2f %2.3f %3.2f %3.2f %3.2f %3.2f ",
    sct[mid].det.select,
    sct[mid].det.vdet_set, 
    sct[mid].det.idet_lim,
    sct[mid].det.vdd_set,
    sct[mid].det.idd_lim,
    sct[mid].det.vcc_set, 
    sct[mid].det.icc_lim);

  /*columns s11 to s18 */

  fprintf(outf,"%3.2f %3.2f %3.2f %3.2f %3.2f %3.2f %3.2f %3.2f ",
    sct[mid].det.vi1_set, 
    sct[mid].det.ivi1_lim,
    sct[mid].det.vpin_set, 
    sct[mid].det.ipin_lim,
    sct[mid].det.vled0_set, 
    sct[mid].det.iled0_lim,
    sct[mid].det.vled1_set, 
    sct[mid].det.iled1_lim);

  /* columns s19 to s24 */

  fprintf(outf,"%3.2f %3.2f %hd %hd %hd %hd",
    sct[mid].det.t1_lim, /* N.B. this gives the temperature limit not */
    sct[mid].det.t2_lim, /* the actual monitored temperature PWP 11.12.00 */
    (short) 0, (short) 0, (short) 0, (short) 0);

  /* columns s25 to s31 */

  fprintf(outf,"%hd %hd %hd %hd %hd %hd %hd\n",
    (short) 0, (short) 0, (short) 0, (short) 0,
    (short) 0, (short) 0, (short) 0);

  fflush(outf);
  return (0);
}


int stout_trims(short mid,short chp){  /* cal_mode = -3 */

  /* Write ABCD2NT trimdac record to the ascii ntuple file */

  short chan, indx;
  
  if (outf == NULL){
    printf("stout_trims: called when no file open\n");
    return(-1);
  }

  if(sct[mid].det.chipset!=ABCDT) return -1;

  for(indx=0;indx<4;indx++){  /* loop over indx */  

    /* columns 1 to 11, a standard data row with the following 
     * hardcoded element:
     * cal_mode = -3:   ABCDNT trimdac record
     */

    fprintf(outf, "%ld %hd %hd %hd %3.2f %3.2f %3.2f %3.2f %ld %hd %hd ", 
      scan_number, mid, chp, (short) -3,
      scan_var[4], scan_var[3], scan_var[2], scan_var[1],
      st_trigger_count, mus_error_count[sct[mid].mus_id], indx);

    /* columns s0 to s31 */
  
    for(chan=indx; chan<128; chan+=4){
      fprintf(outf," %hd", sct[mid].det.chips[chp].trim[chan]);
    }
    fprintf(outf,"\n");
  }

  fflush(outf);
  return(0);
}

int stout_chip(short mid,short chp){  /* cal_mode = -2 */

  /* Write a chip configuration record to the ascii ntuple file */

  int date_int, time_int; 
  short present;
  int bit, nword, word, num_masks;
  unsigned int mask[4];
  double mean_trim;
  struct tm *cookedtime;
  time_t rawtime;

  if (outf == NULL){
    printf("stout_chip: called when no file open\n");
    return(-1);
  }

  time( &rawtime);
  cookedtime = localtime( &rawtime);

  date_int = ((cookedtime->tm_year + 1900) * 10000)
           + ((cookedtime->tm_mon+1) * 100) 
           +   cookedtime->tm_mday;

  time_int =  (cookedtime->tm_hour *3600) 
           +  (cookedtime->tm_min  * 60 ) 
           +   cookedtime->tm_sec;

  if (sct[mid].det.chips[chp].role!=MISSING) present = 1;
  else present = 0;

  /* Mask section.  Note that the mask is presently stored in an
   * array of 8 ints, only the lowest 16 bits of each word are
   * presently used.  It would make sense to either use unsigned ints
   * or to compress the masks into 4 unsigned ints, as done here
   */

  num_masks = 0;

  for (nword=0; nword<4; nword+=2){
    word  =  (sct[mid].det.chips[chp].mask[(nword*2)]&&0xffff);
    word += ((sct[mid].det.chips[chp].mask[(nword*2+1)]&&0xffff)<<16);
    mask[nword] = !word;
    for(bit=0; bit<32; bit++){
      num_masks += (word & 0x0001);
      word = word >> 1;
    }
  }

  /* Mean trim DAC value - applies only in the case of ABCBT
   * chipset 3 - ABCD  ( and ABCDNT???? - should really be different)
   *         4 - ABCDT
   *         5 - ABC/CAFE-P
   */

  mean_trim =0;

  if(sct[mid].det.chipset==ABCDT){
    for(nword=0; nword<128; nword++){
      mean_trim += sct[mid].det.chips[chp].trim[nword];
    }
    mean_trim = mean_trim / 128;
  }

  /* columns 1 to 11, a standard data row with the following 
   * hardcoded elements:
   * cal_mode = -2:   chip configuration record
   * index_base = -1: irrelevant
   */

  fprintf(outf, "%ld %hd %hd %hd %3.2f %3.2f %3.2f %3.2f %ld %hd %hd ", 
    scan_number, mid, chp, (short)-2,
    scan_var[4], scan_var[3], scan_var[2], scan_var[1],
    st_trigger_count, mus_error_count[sct[mid].mus_id], (short)-1);

  /* columns s0 to s8 */

  fprintf(outf, "%ld %ld %hd %hd %3.2f %3.2f %3.2f %3.2f %3.2f ",
    date_int, time_int, present,
    sct[mid].det.chips[chp].active,
    sct[mid].det.chips[chp].vthr,
    sct[mid].det.chips[chp].vcal,
    sct[mid].det.chips[chp].delay,
    sct[mid].det.chips[chp].febias,
    sct[mid].det.chips[chp].feshaper);

  /* columns s9 to s19 */

  fprintf(outf, "%hd %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd ",
    (short) 0, (short) 0, (short) 0,
    sct[mid].det.chips[chp].role,
    sct[mid].det.chips[chp].compression,
    sct[mid].det.chips[chp].cal_mode,
    sct[mid].det.chips[chp].trim_range,
    0,   /* test_mode was dumped here, now merged with cal_enable 
            to form trim_range PWP 15.08.00 */
    sct[mid].det.chips[chp].edge_detect,
    sct[mid].det.chips[chp].mask_readout,
    sct[mid].det.chips[chp].accumulate);

  /*columns s20 to s27 */

  fprintf(outf,"%hd %ld %ld %ld %ld %3.2f %hd %hd ",
    num_masks,
    mask[0], mask[1], mask[2], mask[3],
    mean_trim, (short) 0, (short) 0);
  /* columns s28 to s31 */

  fprintf(outf,"%hd %hd %hd %hd\n",
    (short) 0, (short) 0, (short) 0, (short) 0);

  fflush(outf);
  return(0);
}


int stout_sob(short burst_type, short loop_cals, 
              int tmax, short ttype,
              short cfgfreq, short srfreq, short bcfreq){  /* cal_mode = -1 */

  /* Write a "burst intention" record to the ascii ntuple file */

  if (outf == NULL){
    printf("stout_sob: called when no file open\n");
    return(-1);
  }

  /* columns 1 to 11, a standard data row with the following 
   * hardcoded elements:
   * module_id = -1:  applies to all modules equally
   * chip_id = -1:    applies to all chips equally
   * cal_mode = -1:   sob record
   * index_base = 0:  irrelevant
   */

  fprintf(outf, "%ld %hd %hd %hd %3.2f %3.2f %3.2f %3.2f %ld %hd %hd ", 
    scan_number, (short) -1, (short) -1, (short) -1,
    scan_var[3], scan_var[2], scan_var[1], scan_var[0],
   /* TODO: complete implementation!! */
    (int) 0,(short) 0,(short) 0);  

  /* columns s0 to s7 */
  fprintf(outf, "%hd %hd %hd %hd %hd %hd %hd %hd ",
    scan_pnt[3], scan_pnt[2], scan_pnt[1], scan_pnt[0],
    scan_typ[3], scan_typ[2], scan_typ[1], scan_typ[0]);
 
  /* columns s8 to s15 */
  fprintf(outf, "%ld %hd %hd %ld %hd %hd %hd %hd ",
    burst_number, burst_type,
    loop_cals, tmax,
    ttype, cfgfreq,
    srfreq, bcfreq);
 
  /* columns s16 to s23 */
  fprintf(outf, "%3.2f %3.2f %3.2f %hd %3.2f %3.2f %3.2f %hd ",
    scan_start[0], scan_stop[0],
    scan_step[0],  scan_points[0],
    scan_start[1], scan_stop[1],
    scan_step[1],  scan_points[1]);
 
  /* columns s24 to s31 */
  fprintf(outf, "%3.2f %3.2f %3.2f %hd %3.2f %3.2f %3.2f %hd\n",
    scan_start[2], scan_stop[2],
    scan_step[2],  scan_points[2],
    scan_start[3], scan_stop[3],
    scan_step[3],  scan_points[3]);
 
  fflush(outf);
  return (0);
}


int stout_eob(){

  /* Write burst data to the ascii ntuple file */

  short mid,chp,indx,s,link;

  if (outf == NULL){
    printf("stout_eob: called when no file open\n");
    return(-1);
  }

  for (mid=0; mid<SCT_MAX_MODULES; mid++){  /* loop over mid  */
    /*
     * Output data only if module if present
     */
      for(chp=0; chp<12; chp++){            /* loop over chp  */
      /*
       * Output data only if chip is not missing
       */
        if (sct[mid].det.chips[chp].role!=MISSING){
          if(chp>5) link = 1; 
          else      link = 0;
          for(indx=0; indx<4; indx++){      /* loop over indx */

            fprintf(outf, "%ld %hd %hd %hd %3.2f %3.2f %3.2f %3.2f %ld %hd %hd ", 
              scan_number, mid, chp, sct[mid].det.chips[chp].cal_mode,
             /* TODO: write actual var settings for each chip!! */
              scan_var[3], scan_var[2], scan_var[1], scan_var[0], /* PWP bugfix 01.09.99 */
              st_trigger_count, mus_error_count[sct[mid].mus_id], indx);

            for (s=0; s<32; s++) {          /* loop over data */
              fprintf(outf," %hd", sct_histogram[mid][link][((chp%6)*128)+(s*4)+indx]);
            }
            fprintf(outf,"\n");
          }
        }
      }
    }
  fflush(outf);
  return (0);
}

int mout_open(short run){

  /* Open file for writing - replace if file exists */

  char fname[64];

  run_number = run;

  sprintf(fname,"%sstrun%dburst%d.evdump",ST_DATA_PATH,run,burst_number);

  musf=fopen(fname,"w");
  if (musf == NULL){ 
    printf("mout_open: Failed to open event dump file %s\n",fname);
    return(-1);
  }
  printf("mout_open: File %s opened for writing\n",fname);
  return(0);
}


int mout_close(){

  /* Close file */

  if (musf != NULL){
    fclose(musf);
    musf = NULL;
  }
  return(0);
}

