/* 
   TST.cpp  :  Atlas SCT Systemtest / Multimodule Root-enabled Acquisition

   Prototype class  GFM 990421

   This class provides a DLL with root-enabled acquisition functions which
   can be called from Root C++ macros for run control, analysis and display.

   All physical i/o is handled in device or task specific standard C static
   libraries which can also be compiled into non-root applications, eg, on 
   the SCT RAID.
*/


#include <iostream.h>

extern "C" {

#include "../CLOAC/cloac_lib.h"
#include "../CLOAC/cloac_menu.h"

#include "../optif/optif_lib.h"

#include "../MUSTARD/mus_lib.h"
#include "../MUSTARD/mus_user.h"

#include "../SLOG/slog_lib.h"

#include "../stlib/st_config.h"
#include "../stlib/stlib.h"
#include "../stlib/stout.h"

#ifdef NIVXI
#include "../include/nivxi.h"
#endif

#ifdef QB_linux
#include "../QB/InitVME.h"
#endif
}

#include "../stlib/sct_hardware.h"

#include "TDatime.h"

#include "TCloac.h"
#include "TModule.h"
#include "TST.h"
#include "dcs_message.h"

ClassImp(TST)

TST::TST(){
  char h_title[32],h_name[32];
  short n;

  strcat(sct_path, ST_PATH);
  strcat(sct_var,  ST_VAR);
  strcat(sct_etc,  ST_ETC);
  strcat(sct_data, ST_DATA_PATH);
  strcat(dedication, "Dedicated to the dough balls at Pizza d'Oro!");
  version = (float)3.31;

  nCLOAC    = CLOAC_NUMBER;  // to be used later as part of runtime
  nMuSTARDs = MUS_NUMBER;    // configuration system
  nSLOGs    = SLOG_NUMBER;   // PWP 21.08.00
  nSCTLV    = SCTLV_NUMBER;
  nSCTHV    = SCTHV_NUMBER;
  nOPTIFs   = OPTIF_NUMBER;

  cc = 0;                    // Set pointer to instance of TCloac to be zero

  // Universal counters

  runnum = 0;
  burstnum = 0;
  scannum = 0;

  burst_count = 0;
  scan_count = 0;

  run_tsent = 0;
  run_evcnt = 0;
  run_ercnt = 0;
  run_tocnt = 0;
  run_xecnt = 0;

  //
  // Set default values of burst/scan control parameters
  //

  abort = 0;                 // clear abort flag
  stopper = 1;               // ENABLE use of the STOP button
  burst_max_timeouts = 6;    // maximum number of timeouts/burst

  ascii_nt = 0;              // default - ascii ntuples OFF

  config_frequency = 0;        // Auto chip reconfiguration
  soft_reset_frequency = 1000; // Auto soft reset after this many triggers sent
  bc_reset_frequency = 0;      // Auto bc resets

  do_action_DCS = 1;         // ENABLE pseudo DCS system
  do_autostop = 1;           // ENABLE Automatic Stop once a scan has covered the interesting bit
  do_cal_loop = 1;           // ENABLE loop over the four cal lines (where cal strobed)
  do_fits = 1;               // ENABLE automatic fitting of scans
  do_mask_test = 0;          // DISABLE alternating mask patterns on adjacent events
  do_preburst = 0;           // DISABLE automatic suppression of dead links
  do_vary_ntrigs = 0;        // DISABLE variation in trigger count in response to occupancy
  do_fill_evtree = 0;        // DISABLE filling of event tree

  burst_type = 1;
  burst_trtype = 3;
  burst_ntrigs = 1000;
  burst_maxtrigs = 1000000;
  burst_target = 10;

  paf = 3;   // programmable almost full flag 8k
  trlim = 0; // L1A suppressed when more than 4 L1A outstanding

  L1A_per_loop = 1;
  throw_away   = 0;

  rate_limit_delay = 0;       // Delay in ms using TSystem::Sleep() in some bursts to limit
                              // repetition rate for some applications, e.g., laser scanner
  seu_duration=1000;          // duration of SEU test with software timer

  debug_level = 6;

  //
  // Set the default scan to be a threshold scan
  //

  ConfigureScanDefaults(1);

  //
  // clear buffers and counters used by AppendEvent
  // 

  buf[0] = 0;
  bufpnt = 0;
  bufevt = 0;

  last_BPT=-1;
  last_SPT=-1;

  BurstPaveText=0;
  ScanPaveText=0;

  // Initialise pointers to TModule objects to zero
  for (n=0; n<SCT_MAX_MODULES; n++){
    m[n]=0;
  }

  // Book histograms for global event and error counters
  // TODO: replace m_err by one histogram per MuSTARD??

  h_merr = new TH1F("h_merr","Mustard Counters",5,0.0,5.0);
  h_merr->SetFillColor(46);

  h_scan_tsent = new TH1F("h_scan_tsent", "Triggers Sent", ST_MAX_SCANPOINTS, 0.0, (float)ST_MAX_SCANPOINTS);
  h_scan_tsent->SetFillColor(46);

  for(n=0;n<MUS_NUMBER;n++){
    sprintf(h_title,"Mustard %d Total event length",n);
    sprintf(h_name,"h_evlen%d",n);
    h_evlen[n] = new TH1F(h_name,h_title,100,0.0,6000.0);  // range was 0 to 1000
    h_evlen[n]->SetFillColor(46);

    sprintf(h_title,"Mustard %d Events Decoded",n);
    sprintf(h_name,"h_scan_evcnt%d",n);
    h_scan_evcnt[n] = new TH1F(h_name, h_title, ST_MAX_SCANPOINTS, 0.0, (float)ST_MAX_SCANPOINTS);
    h_scan_evcnt[n]->SetFillColor(46);

    sprintf(h_title,"Mustard %d Decoding Errors",n);
    sprintf(h_name,"h_scan_ercnt%d",n);
    h_scan_ercnt[n] = new TH1F(h_name, h_title, ST_MAX_SCANPOINTS, 0.0, (float)ST_MAX_SCANPOINTS);
    h_scan_ercnt[n]->SetFillColor(46);

    sprintf(h_title,"Mustard %d Timeouts",n);
    sprintf(h_name,"h_scan_tocnt%d",n);
    h_scan_tocnt[n] = new TH1F(h_name, h_title, ST_MAX_SCANPOINTS, 0.0, (float)ST_MAX_SCANPOINTS);
    h_scan_tocnt[n]->SetFillColor(46);

    sprintf(h_title,"Mustard %d Cntrl Errors",n);
    sprintf(h_name,"h_scan_xecnt%d",n);
    h_scan_xecnt[n] = new TH1F(h_name, h_title, ST_MAX_SCANPOINTS, 0.0, (float)ST_MAX_SCANPOINTS);
    h_scan_xecnt[n]->SetFillColor(46);
  }
}

  
TST::~TST(){

}


void TST::PrintStatus(){
  // Print information about the present system status to 
  // the CINT window
  TDatime *dt;
  dt = new TDatime;

  cout << endl;
  cout << "TST::PrintStatus()" << endl;
  cout << endl;
  cout << " Version      " << version << ": " << dedication << endl;
  cout << " Date         " << dt->GetDate() << "  Time " << dt->GetTime() << endl;
  cout << " Run number   " << runnum << endl;
  cout << " Scan number  " << scannum << endl;
  cout << " Burst number " << burstnum << endl;

  cout << " ascii_nt     " << ascii_nt << endl;

  cout << " config_frequency     " << config_frequency << endl;
  cout << " soft_reset_frequency " << soft_reset_frequency << endl;
  cout << " bc_reset_frequency   " << bc_reset_frequency << endl;

  cout << " do_action_DCS  " << do_action_DCS << endl;
  cout << " do_autostop    " << do_autostop << endl;
  cout << " do_cal_loop    " << do_cal_loop << endl;
  cout << " do_fits        " << do_fits << endl;
  cout << " do_mask_test   " << do_mask_test << endl;
  cout << " do_preburst    " << do_preburst << endl;
  cout << " do_vary_ntrigs " << do_vary_ntrigs << endl;
  cout << " do_fill_evtree " << do_fill_evtree << endl;

  cout << " burst_type     " << burst_type << endl;
  cout << " burst_trtype   " << burst_trtype << endl;
  cout << " burst_ntrigs   " << burst_ntrigs << endl;
  cout << " burst_maxtrigs " << burst_maxtrigs << endl;
  cout << " burst_target   " << burst_target << endl;
  cout << " paf " << paf << " trlim " << trlim << endl;

  cout << " L1A_per_loop   " << L1A_per_loop << endl;
  cout << " throw_away     " << throw_away << endl;

  cout << " rate_limit_delay  " << rate_limit_delay << endl;
  cout << " seu_duration  " << seu_duration << endl;

  cout << " Debug level " << debug_level << endl;

  cout << " num_modules  " << num_modules << endl;
  cout << " num_showmods " << num_showmods << endl;
  cout << " num_panelmods " << num_panelmods << endl;
  cout << " abort " << abort << endl;
  cout << " stopper " << stopper << endl;

  cout << endl;

  delete dt; // PWP 20.12.01
}


void TST::RestoreState(){
  // Restore the previously saved settings of parameters 
  // which control burst and/or scan behaviour

  ascii_nt = old_ascii_nt;

  config_frequency = old_config_frequency;
  soft_reset_frequency = old_soft_reset_frequency;
  bc_reset_frequency = old_bc_reset_frequency;

  do_action_DCS = old_do_action_DCS;
  do_autostop = old_do_autostop;
  do_cal_loop = old_do_cal_loop;
  do_fits = old_do_fits;
  do_mask_test = old_do_mask_test;
  do_preburst = old_do_preburst;
  do_vary_ntrigs = old_do_vary_ntrigs;
  do_fill_evtree = old_do_fill_evtree;

  burst_type = old_burst_type;
  burst_trtype = old_burst_trtype;
  burst_ntrigs = old_burst_ntrigs;
  burst_maxtrigs = old_burst_maxtrigs;
  burst_target = old_burst_target;

  paf = old_paf;
  trlim = old_trlim;

  L1A_per_loop = old_L1A_per_loop;
  throw_away = old_throw_away;

  rate_limit_delay = old_rate_limit_delay;
  seu_duration = old_seu_duration;

  debug_level = old_debug_level;
}


void TST::SaveState(){
  // Record the present settings of parameters 
  // which control burst and/or scan behaviour

  old_ascii_nt = ascii_nt;

  old_config_frequency = config_frequency;
  old_soft_reset_frequency = soft_reset_frequency;
  old_bc_reset_frequency = bc_reset_frequency;

  old_do_action_DCS = do_action_DCS;
  old_do_autostop = do_autostop;
  old_do_cal_loop = do_cal_loop;
  old_do_fits = do_fits;
  old_do_mask_test = do_mask_test;
  old_do_preburst = do_preburst;
  old_do_vary_ntrigs = do_vary_ntrigs;
  old_do_fill_evtree = do_fill_evtree;

  old_burst_type = burst_type;
  old_burst_trtype = burst_trtype;
  old_burst_ntrigs = burst_ntrigs;
  old_burst_maxtrigs = burst_maxtrigs;
  old_burst_target = burst_target;

  old_paf = paf;
  old_trlim = trlim;

  old_L1A_per_loop = L1A_per_loop;
  old_throw_away = throw_away;

  old_rate_limit_delay = rate_limit_delay;
  old_seu_duration = seu_duration;

  old_debug_level = debug_level;
}


int TST::VMETest(){

  // A short routine to test the VME interface
  // if this fails, we probably need to run resman
  // PWP 28.11.00

#ifdef NIVXI
  short val;
  NIVXI_STATUS ret;

  ret = InitVXIlibrary();

  cout << "Init VXI Library returns " << ret;
  if(ret>0){
    cout << " - it was already initialised!!";
  }
  cout << endl;

  ret = VXIin(3,MUS_ADDRESS,2,&val);

  cout << "Read from MuSTARD base address returns " << ret;
  if(ret<0){
    cout << " - ERROR!!!";
  }
  cout << endl;

  if(ret<0) CloseVXIlibrary();

  return (int) ret;
#endif

#ifdef QB_linux

  InitVME();
  ResetVME();

#endif

  return 0;
}
 
// These functions act for all modules defined in sysmap, and are
// primarily wrappers for stlib functions with the addition of
// Root histogram booking and filling.

void TST::Startup(){
  char mname[32];
  short n;

  st_startup();

 
  // Create an object of type TCloac - 
  // useful interface to CLOAC, bringing many functions to RINT
  // (OK, this is experimental - PWP 07.08.00)

  if(CLOAC_NUMBER){
    cc = new TCloac(0);
    cc->n_trigs=burst_ntrigs;
  }

  // Create array of objects of type "TModule":
  // copy across the module name from stlib
  // and mark present modules as present

  num_modules=0;

  for (n=0; n<SCT_MAX_MODULES; n++){
    if(m[n]){
      delete m[n];
    }
    m[n] = new TModule(n);
    if(st_give_modname(n,mname)==0){
      m[n]->present = 1;
      m[n]->SetModuleName(mname);
      m[n]->n_active_channels=st_n_active_channels(n);
      num_modules++;
    }
  }

  // Assume all power supplies have started OK,
  // and hence start the stopwatches PWP 13.12.00

  for (n=0; n<SCT_MAX_MODULES; n++) {
    if(m[n]->present){
      m[n]->DCS->LVStopwatch->Continue();
      m[n]->DCS->LVStopwatchRunning=kTRUE;
    }
  }

  //num_showmods = 3; Commented out by LE 21.11.00, to avoid resetting value to 3.

  printf("\nTST::Startup - Number of modules present: %d", num_modules);
  //  printf("\n               num_showmods will be %d\n\n", num_showmods);


}


void TST::Restart(){
  short n;

  burst_type = 1;
  burst_ntrigs = 1000;
  burst_trtype = 3;          // Reselect CAL+ L1A as trigger type
  burst_maxtrigs = 1000000;
  burst_target = 10;
  do_cal_loop = 1;

  paf = 3;
  trlim = 0;

  st_configure_modules();    // Reload module configurations into st according to current sysmap
  st_init();                 // Re-init all VME modules, re-power system and execute all configs

  for (n=0; n<SCT_MAX_MODULES; n++){
    if(m[n]->present){
      m[n]->n_active_channels=st_n_active_channels(n);
    }
  }
}


void TST::Shutdown(){
  st_shutdown();
}


void TST::ModifyEnables(int imod, short present, short active, short link0, short link1){
  // If a module is present, active can be altered; if links
  // are active in the sysmap, stream channels can be en/disabled

  st_configure_enable(imod, present, active, link0, link1);  
}


void TST::ConfigureMask(int imod, short chan){
  st_mask_chan(imod, chan);
}


void TST::ConfigureRC(int imod, int chip, short method, short function,
                      float p0, float p1, float p2){
  st_set_rc(imod, chip, method, function, p0, p1, p2);
}


void TST::ConfigureTrim(int imod, short chan, short var){
  st_configure_trim((short)imod,(chan/128),(chan%128),var);
}


void TST::ConfigureVariable(int imod, int chip, int typ, float var){
  float newvar;

  st_coerce_variable(typ,var,&newvar,0);
  st_configure_chip_variable(imod,chip,typ,newvar);

  if((typ==ST_MASK)||(typ==ST_NMASK)){
    if(m[imod]->present){
      m[imod]->n_active_channels=st_n_active_channels(imod);
    }
  }
}

void TST::ConfigureVariable(int imod, int typ, float var){
  float newvar;

  st_coerce_variable(typ,var,&newvar,0);
  st_configure_variable(imod,typ,newvar);

  if((typ==ST_MASK)||(typ==ST_NMASK)){
    if(m[imod]->present){
      m[imod]->n_active_channels=st_n_active_channels(imod);
    }
  }
}

void TST::ConfigureVariable(int typ, float var){
  float newvar;
  short n;

  st_coerce_variable(typ,var,&newvar,0);
  st_configure_variable(-1,typ,newvar); 

  if((typ==ST_MASK)||(typ==ST_NMASK)){
    for (n=0; n<SCT_MAX_MODULES; n++){
      if(m[n]->present){
        m[n]->n_active_channels=st_n_active_channels(n);
      }
    }
  }
}

void TST::RetrieveRC(int imod, int chip, short method, short function,
                     float p0, float p1, float p2){
  st_get_rc(imod, chip, &method, &function, &p0, &p1, &p2);
}


float TST::RetrieveVariable(int imod, int chip, int chan, int typ){
  float var;

  st_retrieve_variable(imod,chip,chan,typ,&var);
  return var;
}

float TST::RetrieveVariable(int imod, int chip, int typ){
  float var;

  st_retrieve_variable(imod,chip,0,typ,&var);
  return var;
}

float TST::RetrieveVariable(int imod, int typ){
  float var;

  st_retrieve_variable(imod,0,0,typ,&var);
  return var;
}

void TST::ExecuteConfigs(){
  st_execute_configs(0,1);
}

void TST::ExecuteConfigs(short dump_level){
  st_execute_configs(dump_level,1);
}

void TST::ExecuteSCConfigs(){ // may change
  st_execute_sc_configs(1,1);
}

void TST::ExecuteSCConfigs(short hard_reset){
  st_execute_sc_configs(1,hard_reset);
}

void TST::ExecuteBurst(){
  ExecuteBurst(-1);  // cal_mode = -1 => read data for all cal lines
}

void TST::ExecuteBurst(short cal_mode){
  short i,j;
  short stream;
  short mid;
  short FillEvcnt, FillErcnt, FillTocnt;

  if (burst_count >= ST_MAX_SCANPOINTS) burst_count = 0;  // Not normally reached but needed for security

  // Reset burst data and histograms 

  tsent = 0;
  ercnt_sum = 0;
  tocnt_sum = 0;
  xecnt_sum = 0;

  if (do_action_DCS) DCSAction(); // PWP 13.12.00

  if (ResetMustards() != 0) return;

  if(do_preburst){  // Preburst to facilitiate automatic masking
                    // of streams causing timeout

    st_clear_suspense();
    PreBurst();
  }


  for(mid=0;mid<MUS_NUMBER;mid++){
    evcnt[mid] = 0;
    ercnt[mid] = 0;
    tocnt[mid] = 0;
    xecnt[mid] = 0;
  }

  if(cal_mode<=0) {
    h_merr->Reset();

    for(i=0;i<MUS_NUMBER;i++){
      h_evlen[i]->Reset();
    }

    for (i=0; i<SCT_MAX_MODULES; i++) {
      if(m[i]->present){ // fill hists only for present modules 
        m[i]->SetBurstCount(burst_count);
        m[i]->ResetBurstHist();
        m[i]->SetPoint(scan_current);

        if(burst_type==ST_TIMEEVBURST){ // must fill scanpoint array for bins 1-23
          for(j=1;j<24;j++){
            m[i]->SetBurstCount(burst_count+j);
            m[i]->SetPoint(scan_current+(float)j);
          }
          m[i]->SetBurstCount(burst_count);
        }
      }
    }
  }


  // Actually execute the burst and store results in TModule objects

  switch (burst_type) {
    case ST_TBURST:      // normal running; single CLOAC triggers; use MuSTARD Histogram
      FillEvcnt = 1;
      FillErcnt = 1;
      FillTocnt = 1;
      TriggerBurst(cal_mode);
      break;
    case ST_RAWBURST:    // single CLOAC triggers; histogram RAW DATA in software
      FillEvcnt = 1;
      FillErcnt = 1;
      FillTocnt = 1;
      RawBurst();
      break;
    case ST_SENDIDBURST: // raw burst in SENDID mode, config frequency = 1 to test downlink
      FillEvcnt = 1;
      FillErcnt = 1;
      FillTocnt = 1;
      SendIDBurst();
      break;
    case ST_NSEBURST:    // Noise occupancy scan with single triggers and MuSTARD Histogram
      FillEvcnt = 1;
      FillErcnt = 1;
      FillTocnt = 1;
      NoiseBurst();
      break;
    case ST_TBURST2:     // running with error decoding; use MuSTARD Histogram
      FillEvcnt = 1;
      FillErcnt = 0;
      FillTocnt = 0;
      TriggerBurst2(cal_mode);
      break;
    case ST_VBURST2:     // running with CLOAC periodic or pseudo random trigger
      FillEvcnt = 1;
      FillErcnt = 1;
      FillTocnt = 1;
      VBurst2();
      break;
    case ST_DBURST:      // Decoded Trigger Burst: decode and histogram in stdll, not Mustard
      FillEvcnt = 0;
      FillErcnt = 0;
      FillTocnt = 0;
      DecodedBurst(cal_mode);
      break;
    case ST_DVBURST:     // running with CLOAC periodic or pseudo random trigger
      FillEvcnt = 0;     // but with histogramming in software
      FillErcnt = 0;
      FillTocnt = 0;
      DecodedVBurst();
      break;
    case ST_EBURST:      // running with CLOAC external trigger
      FillEvcnt = 0;
      FillErcnt = 0;
      FillTocnt = 0;
      EBurst();
      break;
    case ST_DEBURST:     // running with CLOAC external trigger
      FillEvcnt = 0;     // but with histogramming in software
      FillErcnt = 0;
      FillTocnt = 0;
      DecodedEBurst();
      break;
    case ST_DSBURST:     // running with SLOG periodic trigger from slog_ctest.txt
      FillEvcnt = 0;     // but with histogramming in software
      FillErcnt = 0;
      FillTocnt = 0;
      DecodedSBurst();
      break;
    // DumpBurst deleted PWP 26.04.01
    case ST_TESTBURST:   // Decoded Trigger Burst: links 0,1 only; Mustard and external hists
      FillEvcnt = 0;
      FillErcnt = 0;
      FillTocnt = 0;
      TestBurst();
      break;
    case ST_TIMEEVBURST: // up to 8 consecutive L1a decoded to up to 24 timebins
      FillEvcnt = 0;
      FillErcnt = 0;
      FillTocnt = 0;
      TimeEvBurst(cal_mode);
      break;
    case ST_SEUBURST:    // Burst of SEUTest() calls: note
      FillEvcnt = 0;
      FillErcnt = 0;
      FillTocnt = 0;
      SEUBurst();
      break;    
    default:
      FillEvcnt = 1;
      FillErcnt = 1;
      FillTocnt = 1;
      printf("Unknown burst type %d requested: no action taken\n", burst_type);
      break;
  }
  

  // Update arrays and histograms for whole burst
  // Note: some histograms are filled at each event

  printf("ExecuteBurst type %d: tsent %d\n", burst_type, tsent);

  for(mid=0;mid<MUS_NUMBER;mid++){
    printf("  MuSTARD %d:  evcnt %d  ercnt %d  tocnt %d  xecnt %d\n",
        mid, evcnt[mid], ercnt[mid], tocnt[mid], xecnt[mid]);
  }

  printf("  Totals:  evcnt %d  ercnt %d  tocnt %d  xecnt %d\n",
      evcnt[0], ercnt_sum, tocnt_sum, xecnt_sum);

  run_tsent += tsent;
  run_evcnt += evcnt[0];
  run_ercnt += ercnt_sum;
  run_tocnt += tocnt_sum;
  run_xecnt += xecnt_sum;

  h_merr->Fill(0.5, tsent);
  h_merr->Fill(1.5, evcnt[0]);
  h_merr->Fill(2.5, ercnt_sum);
  h_merr->Fill(3.5, tocnt_sum);
  h_merr->Fill(4.5, xecnt_sum);

  //h_scan_tsent->Fill(burst_count, tsent);
  h_scan_tsent->Fill(scan_point, tsent); //PWP 31.07.01

  for(i=0;i<MUS_NUMBER;i++){
    //h_scan_evcnt[i]->Fill(burst_count, evcnt[i]);
    //h_scan_ercnt[i]->Fill(burst_count, ercnt[i]);
    //h_scan_tocnt[i]->Fill(burst_count, tocnt[i]);
    //h_scan_xecnt[i]->Fill(burst_count, xecnt[i]);

    h_scan_evcnt[i]->Fill(scan_point, evcnt[i]); //PWP 25.05.01
    h_scan_ercnt[i]->Fill(scan_point, ercnt[i]);
    h_scan_tocnt[i]->Fill(scan_point, tocnt[i]);
    h_scan_xecnt[i]->Fill(scan_point, xecnt[i]);
  }

  for (i=0; i<SCT_MAX_MODULES; i++) {
    if(m[i]->present){ // fill hists only for present modules
      m[i]->SetTrigs(tsent);

      // Fill counters from system wide records in cases
      // where they have not already been filled during the burst:

      for(stream=0;stream<2;stream++){
        if(FillEvcnt) m[i]->SetEvcnt(stream,evcnt[st_find_mustard(i)]);
        if(FillErcnt) m[i]->SetErcnt(stream,ercnt[st_find_mustard(i)]);
        if(FillTocnt) m[i]->SetTocnt(stream,tocnt[st_find_mustard(i)]);
      }

      m[i]->FillBurstHist(cal_mode);
    }
  }

  if((cal_mode<0)||(cal_mode>=3)){

    // Read DCS values BEFORE incrementing burst counters
    // moved here by PWP 28.11.00

    if(do_action_DCS){  // Read LV, HV values
      DCSAction();
      for(mid=0;mid<SCT_MAX_MODULES;mid++){
        if(m[mid]->present){;
          m[mid]->ReadDCS(burst_count);
        }
      }
    }

    burstnum++;       // Bursts since start of session
    burst_count++;    // Bursts since last InitScan
  }
}

void TST::DCSAction(){
  int mid;
  for(mid=0;mid<SCT_MAX_MODULES;mid++){
    if(m[mid]->present){
      m[mid]->DCS->Action();
    }
  }
}

void TST::DCSQuery(){
  int mid;
  for(mid=0;mid<SCT_MAX_MODULES;mid++){
    if(m[mid]->present){
      m[mid]->DCS->Query();
    }
  }
}

void TST::DCSPrint(){
  int mid;
  for(mid=0;mid<SCT_MAX_MODULES;mid++){
    if(m[mid]->present){
      m[mid]->DCS->Print();
    }
  }
}

float TST::DCSMessg(char* messg){
  return dcs_message(messg);
}

void TST::TriggerBurst(short cal_mode){
  short mid;
  st_tburst(burst_ntrigs, burst_trtype, config_frequency, soft_reset_frequency, bc_reset_frequency, debug_level);
  for (mid=0; mid<MUS_NUMBER; mid++) {
    st_give_counters(mid, &tsent, &evcnt[mid], &ercnt[mid], &tocnt[mid], &xecnt[mid]);
    ercnt_sum += ercnt[mid];
    tocnt_sum += tocnt[mid];
    xecnt_sum += xecnt[mid];
  }

  for (mid=0; mid<SCT_MAX_MODULES; mid++) {
    if(m[mid]->present){ 
      st_give_data(mid, cal_mode, m[mid]->scan[burst_count][0], m[mid]->scan[burst_count][1]);
    }
  }
}

void TST::RawBurst(){
  short mid;
  st_rawburst(burst_ntrigs, burst_trtype, config_frequency, soft_reset_frequency, bc_reset_frequency, debug_level);
  for (mid=0; mid<MUS_NUMBER; mid++) {
    st_give_counters(mid, &tsent, &evcnt[mid], &ercnt[mid], &tocnt[mid], &xecnt[mid]);
    ercnt_sum += ercnt[mid];
    tocnt_sum += tocnt[mid];
    xecnt_sum += xecnt[mid];
  }

  for (mid=0; mid<SCT_MAX_MODULES; mid++) {
    if(m[mid]->present){ 
      st_give_data(mid, -1, m[mid]->scan[burst_count][0], m[mid]->scan[burst_count][1]);
    }
  }
}

void TST::SendIDBurst(){
  short mid;

  // st_compare_sendid will return the error tally in the MuSTARD variable ercnt,
  // hence it must be called before st_give_counters...
  // (Sorry, will do better next time, full software decoding and tally of errors per chip..)
  st_rawburst(burst_ntrigs, burst_trtype, 1, soft_reset_frequency, bc_reset_frequency, debug_level);
  st_compare_sendid(burst_ntrigs, debug_level);
  for (mid=0; mid<MUS_NUMBER; mid++) {
    st_give_counters(mid, &tsent, &evcnt[mid], &ercnt[mid], &tocnt[mid], &xecnt[mid]);
    ercnt_sum += ercnt[mid];
    tocnt_sum += tocnt[mid];
    xecnt_sum += xecnt[mid];
  }

  for (mid=0; mid<SCT_MAX_MODULES; mid++) {
    if(m[mid]->present){ 
      st_give_data(mid, -1, m[mid]->scan[burst_count][0], m[mid]->scan[burst_count][1]);
    }
  }
}

void TST::NoiseBurst(){
  short mid;

  st_nseburst(burst_ntrigs, burst_maxtrigs, burst_target, burst_trtype, debug_level);
  for (mid=0; mid<MUS_NUMBER; mid++) {
    st_give_counters(mid, &tsent, &evcnt[mid], &ercnt[mid], &tocnt[mid], &xecnt[mid]);
    ercnt_sum += ercnt[mid];
    tocnt_sum += tocnt[mid];
    xecnt_sum += xecnt[mid];
  }

  for (mid=0; mid<SCT_MAX_MODULES; mid++) {
    if(m[mid]->present){ 
      st_give_data(mid, -1, m[mid]->scan[burst_count][0], m[mid]->scan[burst_count][1]);
    }
  }
}

int TST::DecodedBurst(short cal_mode){   // Burst of single triggers, decoded in software
    
  short mid;
  short rcode;
  short running = 1;
  short next_config;
  short next_soft_reset;
  short next_bc_reset;

  short inner_loop_running=0;
  int   inner_loop_tsent=0;

  int   events_expected = 0;
  int   events_read = 0;
  int   events_decoded = 0;
  int   events_thrown = 0;

  unsigned short n_trigs;

  short n,n_under;

  char fname[64];
  FILE* dumpfile = NULL;

  // Option to dump events to a file, one per burst, for diagnostic purposes
  if((debug_level==998)||(debug_level==999)){
    sprintf(fname,"%sstrun%dburst%d.evdump",sct_data,runnum,burstnum);
    dumpfile=fopen(fname,"w");
    if (dumpfile == NULL){ 
      printf("DecodedBurst: Failed to open event dump file %s\n",fname);
      return(-1);
    }
  }

  SingleSoftReset();       /* Always do a global synchronous soft reset at SOB */
 
  for(mid=0; mid<MUS_NUMBER; mid++) {
    if ((rcode = mus_set_readout_modes(mid,0,1)) != 0) { /* DECODED, SAMPLE ALL */
      printf("DecodedBurst: ERROR for Mustard %d: mus_set_readout_modes returns %d\n", mid, rcode);
      return -5;
    }
   
    if ((rcode = mus_set_readout_cntl(mid,1,1)) != 0) {
      printf("DecodedBurst: ERROR for Mustard %d: mus_set_readout_cntl returns %d\n", mid, rcode);
      return -6;
    }
  }

  // The data histogram has only 15 useful bits.  To avoid 
  // overflow on the first sub-burst we must limit the
  // number of triggers to less than 2^15.  32500 should
  // be safe.  PWP 18.03.02
  if(burst_ntrigs>32500) n_trigs = 32500;
  else n_trigs = burst_ntrigs;

  Sleep(10);   // Wait 10ms to ensure Mustards ready to receive first event

  while (running) {     // Enter event loop 
    // Repeatedly request bursts of "ntrigs" L1A 
    // until "burst_ntrigs" has been reached

    if(do_vary_ntrigs){
      if(burst_maxtrigs>n_trigs){
        if((burst_maxtrigs-tsent)<n_trigs) {
        n_trigs = (burst_maxtrigs-tsent); 
        }
      }
    } else {
      if(burst_ntrigs>n_trigs){
        if((burst_ntrigs-tsent)<n_trigs) {
        n_trigs = (burst_ntrigs-tsent); 
        }
      }
    }

    if (config_frequency > 0) next_config = config_frequency;
    else                      next_config = -1;

    if (soft_reset_frequency > 0) next_soft_reset = soft_reset_frequency;
    else                          next_soft_reset = -1;

    if (bc_reset_frequency > 0) next_bc_reset = bc_reset_frequency;
    else                        next_bc_reset = -1;

    inner_loop_tsent=0;
    inner_loop_running=1;

    while (inner_loop_running){  // sub-burst
      next_config--;
      next_soft_reset--;
      next_bc_reset--;

      if (!next_config) {
        if(do_mask_test){
          ConfigureVariable(7,(float)((tsent%2)+1)); // toggle mask between 1010 and 0101
        }
        ExecuteConfigs(0);
        next_config = config_frequency;
        Sleep(1);
      }

      if (!next_soft_reset) {
        SingleSoftReset();
        next_soft_reset = soft_reset_frequency;
        Sleep(1);
      }

      if (!next_bc_reset) {
        SingleBCReset();
        next_bc_reset = bc_reset_frequency;
        Sleep(1);
      }
	
      Single(burst_trtype);   // Send the single trigger
      tsent+=L1A_per_loop;
      events_expected+=L1A_per_loop;
      inner_loop_tsent+=L1A_per_loop;
      if(debug_level>7) {
        printf("Sending %d triggers - now up to %d\n",L1A_per_loop,tsent);
      }
      if (debug_level>8) {
        printf("Events exp %d, events read %d, events dec. %d, events thr. %d\n",
               events_expected,events_read,events_decoded,events_thrown);
      } 
 
      while(events_read<events_expected){

        if(events_decoded<events_read){  // event stored in host memory

          if((L1A_per_loop>1)&&(((events_read+1)%L1A_per_loop)<throw_away)){
            events_decoded++;            // event to be thrown away
            events_thrown++;
          } else {
            if(dumpfile){
              DumpEvent(dumpfile);
            }
            DecodeEvent(cal_mode);       // event to be decoded
            events_decoded++;
          }
        }

        ReadEvent();
        events_read++;
      }

      if(inner_loop_tsent>=n_trigs){
        inner_loop_running = 0;
      }

      if (tocnt_sum >= burst_max_timeouts) {
        running = 0;
        inner_loop_running = 0;
        printf("DecodedBurst: exceeded max number timeouts %d\n", tocnt_sum);
      }
    } // end of inner loop


    // Decode Last Event
    if(dumpfile){
      DumpEvent(dumpfile);
      fclose(dumpfile);
    }

    DecodeEvent(cal_mode);
    events_decoded++;


    if((do_vary_ntrigs)&&(do_cal_loop==0)){
//      if(((tsent-events_thrown)%burst_ntrigs)<L1A_per_loop){
      if(((tsent-events_thrown)%n_trigs)<L1A_per_loop){ // PWP 20.03.02
        n_under=0;
        running=0;
        for (n=0; n<SCT_MAX_MODULES; n++){
          if(m[n]->present){
            // Here, Event counter has been filled by DecodeEvent
            m[n]->FillNOScanHist(burst_count);
            // Keep running if less than half the channels
            // have exceeded the target number of hits
            if(m[n]->ChansOverTarget(burst_target)<(st_n_active_channels(n)/2)){
              n_under++;
              running = 1;
            }
          }          
        }
        printf("DecodedBurst: %d modules under target after %d triggers\n",n_under, tsent);
      }
      if (tsent >= burst_maxtrigs) running = 0;
    } else {
      if ((tsent-events_thrown) >= burst_ntrigs) running = 0;  // PWP 21.07.00
    }

  } // end of outer loop


  for(mid=0; mid<MUS_NUMBER; mid++) {
    mus_set_readout_cntl(mid,0,0);  
  }
  return 0;
}


int TST::TriggerBurst2(short cal_mode){   // Burst of single triggers, decode error events only

  short mid;
  short rcode;
  short running = 1;
  short next_config;
  short next_soft_reset;
  short next_bc_reset;

  int   events_expected = 0;
  int   events_read = 0;
  int   events_decoded = 0;

  
  int n_errs = 0;

  char fname[64];
  FILE* dumpfile = NULL;

  // Option to dump events to a file, one per burst, for diagnostic purposes
  if((debug_level==998)||(debug_level==999)){
    sprintf(fname,"%sstrun%dburst%d.evdump",sct_data,runnum,burstnum);

    dumpfile=fopen(fname,"w");
    if (dumpfile == NULL){ 
      printf("TriggerBurst2: Failed to open event dump file %s\n",fname);
      return(-1);
    }
  }

  SingleSoftReset();       /* Always do a global synchronous soft reset at SOB */
  

  for(mid=0; mid<MUS_NUMBER; mid++) {
    if ((rcode = mus_set_readout_modes(mid,0,1)) != 0) { /* DECODED, SAMPLE ALL */
      printf("TriggerBurst2: ERROR for Mustard %d: mus_set_readout_modes returns %d\n", mid, rcode);
      return -5;
    }
   
    if ((rcode = mus_set_readout_cntl(mid,1,1)) != 0) {
      printf("TriggerBurst2: ERROR for Mustard %d: mus_set_readout_cntl returns %d\n", mid, rcode);
      return -6;
    }
  }
  
  if (config_frequency > 0) next_config = config_frequency;
  else                      next_config = -1;

  if (soft_reset_frequency > 0) next_soft_reset = soft_reset_frequency;
  else                          next_soft_reset = -1;

  if (bc_reset_frequency > 0) next_bc_reset = bc_reset_frequency;
  else                        next_bc_reset = -1;

  Sleep(10);   // Wait 10ms to ensure Mustards ready to receive first event

  while (running) {     // Enter event loop 
    next_config--;
    next_soft_reset--;
    next_bc_reset--;

    if (!next_config) {
      if(do_mask_test){
        ConfigureVariable(7,(float)((tsent%2)+1)); // toggle mask between 1010 and 0101
      }
      ExecuteConfigs(0);
      next_config = config_frequency;
      Sleep(1);
    }

    if (!next_soft_reset) {
      SingleSoftReset();
      next_soft_reset = soft_reset_frequency;
      Sleep(1);
    }

    if (!next_bc_reset) {
      SingleBCReset();
      next_bc_reset = bc_reset_frequency;
      Sleep(1);
    }

    // Delay to limit repetition rate in certain applications, 
    // e.g., for laser scanning
    if (rate_limit_delay > 0) Sleep(rate_limit_delay);

    // Enable CLOAC Internal Busy Out for use as a prompt trigger, 
    // e.g., for an oscilloscope or for a laser scanner
    if(cc) cc->EnableInternalBusy();

    Single(burst_trtype);   // Send the single trigger
    tsent+=L1A_per_loop;
    events_expected+=L1A_per_loop;
    if(debug_level>7) {
      printf("Sending %d triggers - now up to %d\n",L1A_per_loop,tsent);
    }
 
    // Disable the CLOAC Internal Busy Out so that any further command 
    // activity does not fire any external trigger
    if(cc) cc->DisableInternalBusy();

    while(events_read<events_expected){
      if(n_errs>0){  // error event to decode
        if(dumpfile){
          DumpEvent(dumpfile);
        }
        DecodeErrors();
        events_decoded++;
      }
      n_errs=TestEvent();
      events_read++;
    }

    if (tocnt_sum >= burst_max_timeouts) {
      running = 0;
      printf("TriggerBurst2: exceeded max number timeouts %d\n", tocnt_sum);
    }
    if (tsent >= burst_ntrigs) running = 0;
  }
  // Decode Last Event
  if(n_errs>0){
    if(dumpfile){
      DumpEvent(dumpfile);
    }
    DecodeErrors();  // last event is error event
    events_decoded++;
  }
  if(dumpfile){
    fclose(dumpfile);
  }

  for(mid=0; mid<MUS_NUMBER; mid++) {
    mus_set_readout_cntl(mid,0,0);
  }

  ReadHistograms(cal_mode);
  
  return 0;
}


int TST::VBurst2(){
  // use CLOAC to generate bursts of
  //   burst_ntrigs periodic or pseudo random triggers
  //
  // Use MuSTARD stream trigger limit register / CLOAC busy in
  // to throttle triggers when ABCD pipeline has received >7 L1A
  // (N.B. if burst_ntrigs =7, this condition should not occur)
  //
  // Use MuSTARD programmable almost full flag / CLOAC busy in
  // to throttle triggers when input FIFO is nearly full (<8k left)
  //
  // Detect and release each event as it becomes available at the MuSTARD output 
  // buffer -  but obtain occupancy information from the MuSTARD histogram

  short mid,n;
  short loop,status;
  short running = 1;

  short EOB, SOB;
  short empty, not_empty;
  short almost_full, overflow;
  short trigs_stopped;

  short n_under;

  unsigned short status1;

  short n_chips, max_chips;
  long event_duration, event_interval, trigger_period;
  long t_zero, t_max;
  double t_burst;

  unsigned short n_trigs;

  int events_expected = 0;
  int events_read = 0;
  int events_decoded = 0;
  int n_errs;

  int   old_evcnt[MUS_NUMBER];
  short old_ercnt[MUS_NUMBER];

  char fname[64];
  FILE* dumpfile = NULL;

  // This burst type requires CLOAC
  // If there is no CLOAC, return now!!
  if(cc==0){
    cout << "VBurst2 called with no CLOAC!!" << endl;
    cout << "  exiting now...." << endl;
    return -99;
  }

  // clear local event and error counters
  for(mid=0; mid<MUS_NUMBER; mid++) {
    old_ercnt[mid]=0;
    old_evcnt[mid]=0;
  }

  // Max time to process one event in MuSTARD:
  //   single chip, 1/2 occupancy, no adjacent hits => event length ~1100 bits
  //   clock period = 25 nS
  //   => allow 28 microS per chip per event

  max_chips=0;
  for(mid=0; mid<MUS_NUMBER; mid++) {
    n_chips = st_find_maxchips(mid);
    if (n_chips>max_chips) max_chips=n_chips;
  }
  event_duration = (max_chips*28); // in microS

  // For timeout calculations, we must also consider the requested
  // Trigger frequency:  at high trigger frequencies the time taken to
  // process events in MuSTARD will dominate, but at low frequencies
  // the decoding time is small with respect to the time between events.

  trigger_period = cc->GetTriggerPeriod();

  if (trigger_period>event_duration) event_interval = trigger_period;
  else                               event_interval = event_duration;

  printf("ev_duration %ldmS, tr_period %ldmS, =>ev_interval %ldmS\n",
    event_duration, trigger_period, event_interval);


  // Option to dump events to a file, one per burst, for diagnostic purposes
  if((debug_level==998)||(debug_level==999)){
    sprintf(fname,"%sstrun%dburst%d.evdump",sct_data,runnum,burstnum);

    dumpfile=fopen(fname,"w");
    if (dumpfile == NULL){ 
      printf("VBurst: Failed to open event dump file %s\n",fname);
      return(-1);
    }
  }

  // CLOAC's VBURST counter may only have 16 bits, but the
  // data histogram has only 15 useful bits.  To avoid 
  // overflow on the first sub-burst we must limit the
  // number of triggers to less than 2^15.  32500 should
  // be safe.  PWP 18.03.02
  if(burst_ntrigs>32500) n_trigs = 32500;
  else n_trigs = burst_ntrigs;


  // Enable CLOAC external busy input(s) 1 (and 2) to accept busy from MuSTARD trigger 
  // overflow counter(s)
  //   NB. MuSTARD0 must be connected to CLOAC output A
  //   and MuSTARD1 must be connected to CLOAC output B

  cc_menu_enable(ENEXTBUSY1); 
  if(MUS_NUMBER>1){
    cc_menu_enable(ENEXTBUSY2);
  } 

  SingleSoftReset();       /* Always do a global synchronous soft reset at SOB */

  loop = 0;
  running = 1; 
  while (running) {     // Enter event loop 
    loop++;
    ResetMustards();

    for(mid=0; mid<MUS_NUMBER; mid++) {
      if ((status = mus_set_readout_modes(mid,0,1)) != 0) { /* DECODED, SAMPLE ALL */
        printf("VBURST: loop %d ERROR for Mustard %d: mus_set_readout_modes returns %d\n", loop, mid, status);
        return -5;
      }
      if ((status = mus_set_readout_cntl(mid,1,1)) != 0) {  /* DAQ ENABLE, READOUT ENABLE */
        printf("VBURST: loop %d ERROR for Mustard %d: mus_set_readout_cntl returns %d\n", loop, mid, status);
        return -6;
      }
    }

    Sleep(10);   // Wait 10ms to ensure Mustards ready to receive first event


    // Repeatedly request bursts of "ntrigs" L1A 
    // until "burst_ntrigs" has been reached

    if(do_vary_ntrigs){
      if(burst_maxtrigs>n_trigs){
        if((burst_maxtrigs-tsent)<n_trigs) {
        n_trigs = (burst_maxtrigs-tsent); 
        }
      }
    } else {
      if(burst_ntrigs>n_trigs){
        if((burst_ntrigs-tsent)<n_trigs) {
        n_trigs = (burst_ntrigs-tsent); 
        }
      }
    }

    cc->n_trigs=n_trigs;
    events_expected+=n_trigs;
    tsent+= n_trigs;

    printf("VBURST: loop %d requesting %d L1A triggers\n", loop, n_trigs);
    trigs_stopped = 0;

    // Wait for start of CLOAC VBURST

    t_zero = (long) clock();
    t_max  = t_zero + CLOCKS_PER_SEC; 

    SOB=0;
    cc->Burst();
    while (((long)clock()<t_max)&&(SOB==0)){
      cc_read_state(&status1);
      if ((status1&0x80)==0) SOB=1;
    }

    if(SOB){
      printf("VBURST: SOB\n");
    } else {
      printf("VBURST: TIMEOUT waiting for SOB (clk %ld tmax %ld tzero %ld)\n", (long) clock(), t_max, t_zero);
      cc->PrintStatus();
      tocnt_sum++;
    }

    // Wait for end of CLOAC VBURST

    t_zero  = (long) clock();
    t_burst = (double)event_duration * (double)n_trigs / (double)1000; // burst duration in seconds
    t_max  = t_zero + (long)((t_burst+1) * CLOCKS_PER_SEC);

    EOB=0;
    while (((long)clock()<t_max)&&(EOB==0)){
      cc_read_state(&status1);
      if (status1&0x80){
         EOB=1;
         printf("VBURST: EOB\n");
       }

      // meanwhile, check for FIFO ALMOST FULL flags
      // if found, inhibit triggers until all FIFOs are empty

      almost_full = 0;
      overflow = 0;
      StreamStatus(&empty,&not_empty,&almost_full,&overflow);

      if(almost_full){ 
        printf("VBURST: %d FIFOs ALMOST FULL\n", almost_full);
        if(overflow){
          printf("VBURST: ERROR: %d FIFOs have OVERFLOWED\n", overflow);
        }

        if(!EOB) {
          cc_menu_vbusy(1);  // set VBUSY to suppress triggers
          trigs_stopped = 1;
          printf("VBURST: triggers stopped to drain FIFOs\n");
        }
      }

      // Read and decode events until all FIFOs are empty at EOB, or if any streams are almost_full

      if(EOB||almost_full){

        empty = 0;
        not_empty = 1;
        while (((long)clock()<t_max)&&(not_empty>0)){
  
          // Clear one event from output buffer
          n_errs=TestEvent();
          events_read++;
          if(n_errs>0){  // error event to be decoded
            if(dumpfile){
              DumpEvent(dumpfile);
            }
            DecodeErrors();
            events_decoded++;
          }

          // Check stream status
          StreamStatus(&empty,&not_empty,&almost_full,&overflow);

        } // end while(not_empty>0)

        if(not_empty>0) printf("VBURST: Timeout waiting for FIFOs to drain\n");
      } // end if (EOB||almost_full)


      if((!EOB)&&(not_empty==0)&&(trigs_stopped)){ // not EOB, but all FIFOs are empty
        trigs_stopped = 0;
        cc_menu_vbusy(0);  // set VBUSY to restart triggers
        printf("VBURST: triggers restarted\n");
      }
    }

    if (!EOB){
      printf("VBURST: TIMEOUT waiting for EOB (clk %ld tmax %ld tburst %f tzero %ld)\n", 
              (long)clock(), t_max, t_burst, t_zero);
      tocnt_sum++;
    }

    // Print out FIFO status at EOB
    if (empty)       printf("VBURST: %d FIFOs are empty\n",empty); 
    if (not_empty)   printf("VBURST: ERROR: %d FIFOs NOT EMPTY\n", not_empty);
    if (almost_full) printf("VBURST: ERROR: %d FIFOs ALMOST FULL (PAF flag)\n", almost_full);
    if (overflow)    printf("VBURST: ERROR: %d FIFOs have OVERFLOWED (fifo full flag)\n", overflow);

    // Stop CLOAC
    cc->Stop();
    cc_load_command(0);

    // Clear any events left in the output pages
    while((events_read<events_expected)&&(tocnt_sum<burst_max_timeouts)){
      n_errs=TestEvent();
      events_read++;
      if(n_errs>0){  // error event to be decoded
        if(dumpfile){
          DumpEvent(dumpfile);
        }
        DecodeErrors();
        events_decoded++;
      }
    }

    // Read Hardware Histogram
    for(mid=0; mid<MUS_NUMBER; mid++) {
      mus_set_readout_cntl(mid,1,0); // DAQ enable, READOUT disable

      old_ercnt[mid]+=ercnt[mid];
      old_evcnt[mid]+=evcnt[mid];
      ercnt[mid]=old_ercnt[mid];
      evcnt[mid]=old_evcnt[mid];
    }

    if(loop==1){
      ReadHistograms(-1);
    } else {
      AddHistograms(-1);
    }

    // Fill Software histograms
    n_under=0;
    running=0;
    for (n=0; n<SCT_MAX_MODULES; n++){
      if(m[n]->present){
        m[n]->SetEvcnt(0,evcnt[st_find_mustard(n)]);
        m[n]->SetEvcnt(1,evcnt[st_find_mustard(n)]);
        m[n]->FillNOScanHist(burst_count);
        if(do_vary_ntrigs){
          // Keep running if, for any module, less than half of
          // the channels have exceeded the target number of hits
          if(m[n]->ChansOverTarget(burst_target)<(st_n_active_channels(n)/2)){
            n_under++;
            running = 1;
          }
        }
      }
    }


    if(do_vary_ntrigs){
      printf("VBURST: %d modules under target after %d triggers\n",n_under, tsent);
      if (tsent >= burst_maxtrigs) running = 0;
    } else {
      if (tsent >= burst_ntrigs) running = 0;
    }

    if (tocnt_sum >= burst_max_timeouts) {
      running = 0;
      printf("VBURST: exceeded max number timeouts %d\n", tocnt_sum);
    }
  } // end while(running)


  // Close dumpfile (if open)
  if(dumpfile){
    fclose(dumpfile);
  }

  // Disable External Busy
  cc_menu_disable(ENEXTBUSY1);
  if(MUS_NUMBER>1){
    cc_menu_disable(ENEXTBUSY2);
  } 

  for(mid=0; mid<MUS_NUMBER; mid++) {
    mus_set_readout_cntl(mid,0,0);
  }

  return 0;
}


int TST::DecodedVBurst(){
  // use CLOAC to generate bursts of
  //   burst_ntrigs periodic or pseudo random triggers
  //
  // Use MuSTARD stream trigger limit register / CLOAC busy in
  // to throttle triggers when ABCD pipeline has received >7 L1A
  // (N.B. if burst_ntrigs =7, this condition should not occur)
  //
  // Use MuSTARD programmable almost full flag / CLOAC busy in
  // to throttle triggers when input FIFO is nearly full (<8k left)
  //
  // Read each event into host memory and process the decoded event
  // such that histograms of n(hits) vs n(events) can be filled

  short mid,n;
  short status;
  short running = 1;

  short n_chips, max_chips;

  short EOB, SOB;
  short empty, not_empty;
  short almost_full, overflow;
  short trigs_stopped;

  short n_under;

  unsigned short status1;

  long event_duration, event_interval, trigger_period;
  long t_zero, t_max;
  double t_burst;

  unsigned short n_trigs;

  int events_expected = 0;
  int events_read = 0;

  char fname[64];
  FILE* dumpfile = NULL;

  // This burst type requires CLOAC
  // If there is no CLOAC, return now!!
  if(cc==0){
    cout << "VBurst2 called with no CLOAC!!" << endl;
    cout << "  exiting now...." << endl;
    return -99;
  }

  // Option to dump events to a file, one per burst, for diagnostic purposes
  if((debug_level==998)||(debug_level==999)){
    sprintf(fname,"%sstrun%dburst%d.evdump",sct_data,runnum,burstnum);

    dumpfile=fopen(fname,"w");
    if (dumpfile == NULL){ 
      printf("DecodedVBurst: Failed to open event dump file %s\n",fname);
      return(-1);
    }
  }

  SingleSoftReset();       /* Always do a global synchronous soft reset at SOB */
 
  for(mid=0; mid<MUS_NUMBER; mid++) {

    if ((status = mus_set_readout_modes(mid,0,1)) != 0) { /* DECODED, SAMPLE ALL */
      printf("DecodedVBurst: ERROR for Mustard %d: mus_set_readout_modes returns %d\n", mid, status);
      return -5;
    }
   
    if ((status = mus_set_readout_cntl(mid,1,1)) != 0) {  /* DAQ ENABLE, READOUT ENABLE */
      printf("DecodedVBurst: ERROR for Mustard %d: mus_set_readout_cntl returns %d\n", mid, status);
      return -6;
    }

  }

  Sleep(10);   // Wait 10ms to ensure Mustards ready to receive first event

  // Enable CLOAC external busy input(s) 1 (and 2) to accept busy from MuSTARD trigger 
  // overflow counter(s)
  //   NB. MuSTARD0 must be connected to CLOAC output A
  //   and MuSTARD1 must be connected to CLOAC output B

  cc_menu_enable(ENEXTBUSY1); 
  if(MUS_NUMBER>1){
    cc_menu_enable(ENEXTBUSY2);
  } 

  // TIMEOUT CALCULATIONS

  // Max time to process one event in MuSTARD:
  //   single chip, 1/2 occupancy, no adjacent hits => event length ~1100 bits
  //   clock period = 25 nS
  //   => allow 28 microS per chip per event

  max_chips=0;
  for(mid=0; mid<MUS_NUMBER; mid++) {
    n_chips = st_find_maxchips(mid);
    if (n_chips>max_chips) max_chips=n_chips;
  }
  event_duration = (max_chips*28); // in microS

  // For timeout calculations, we must also consider the requested
  // Trigger frequency:  at high trigger frequencies the time taken to
  // process events in MuSTARD will dominate, but at low frequencies
  // the decoding time is small with respect to the time between events.

  trigger_period = cc->GetTriggerPeriod();

  if (trigger_period>event_duration) event_interval = trigger_period;
  else                               event_interval = event_duration;

  printf("ev_duration %ldmS, tr_period %ldmS, =>ev_interval %ldmS\n",
    event_duration, trigger_period, event_interval);


  // CLOAC's VBURST counter may only have 16 bits, but the
  // data histogram has only 15 useful bits.  To avoid 
  // overflow on the first sub-burst we must limit the
  // number of triggers to less than 2^15.  32500 should
  // be safe.  PWP 18.03.02
  if(burst_ntrigs>32500) n_trigs = 32500;
  else n_trigs = burst_ntrigs;


  while (running) {     // Enter event loop 

    // Repeatedly request bursts of "ntrigs" L1A 
    // until "burst_ntrigs" has been reached

    if(do_vary_ntrigs){
      if(burst_maxtrigs>n_trigs){
        if((burst_maxtrigs-tsent)<n_trigs) {
        n_trigs = (burst_maxtrigs-tsent); 
        }
      }
    } else {
      if(burst_ntrigs>n_trigs){
        if((burst_ntrigs-tsent)<n_trigs) {
        n_trigs = (burst_ntrigs-tsent); 
        }
      }
    }

    cc->n_trigs=n_trigs;
    events_expected+=n_trigs;
    tsent+= n_trigs;

    printf("DECODEDVBURST: requesting %d L1A triggers\n", n_trigs);
    trigs_stopped = 0;

    // Wait for start of CLOAC DECODEDVBURST

    t_zero = (long) clock();
    t_max  = t_zero + CLOCKS_PER_SEC; 

    SOB=0;
    cc->Burst();
    while (((long)clock()<t_max)&&(SOB==0)){
      cc_read_state(&status1);
      if ((status1&0x80)==0) SOB=1;
    }

    if(SOB){
      printf("DECODEDVBURST: SOB\n");
    } else {
      printf("DECODEDVBURST: TIMEOUT waiting for SOB (clk %ld tmax %ld tzero %ld)\n", (long) clock(), t_max, t_zero);
      cc->PrintStatus();
      tocnt_sum++;
    }

    // Wait for end of CLOAC DECODEDVBURST

    t_zero  = (long) clock();
    t_burst = (double)event_duration * (double)n_trigs / (double)1000; // burst duration in seconds
    t_max  = t_zero + (long)((t_burst+1) * CLOCKS_PER_SEC);

    EOB=0;
    while (((long)clock()<t_max)&&(EOB==0)){
      cc_read_state(&status1);
      if (status1&0x80){
         EOB=1;
         printf("DECODEDVBURST: EOB\n");
       }

      // meanwhile, check for FIFO ALMOST FULL flags
      // if found, inhibit triggers until all FIFOs are empty

      almost_full = 0;
      overflow = 0;
      StreamStatus(&empty,&not_empty,&almost_full,&overflow);

      if(almost_full){ 
        printf("DECODEDVBURST: %d FIFOs ALMOST FULL\n", almost_full);
        if(overflow){
          printf("DECODEDVBURST: ERROR: %d FIFOs have OVERFLOWED\n", overflow);
        }

        if(!EOB) {
          cc_menu_vbusy(1);  // set VBUSY to suppress triggers
          trigs_stopped = 1;
          printf("DECODEDVBURST: triggers stopped to drain FIFOs\n");
        }
      }

      // Read and decode events until all FIFOs are empty at EOB, or if any streams are almost_full

      if(EOB||almost_full){

        empty = 0;
        not_empty = 1;
        while (((long)clock()<t_max)&&(not_empty>0)){
  
          empty = 0;
          not_empty = 0;
          almost_full = 0;
          overflow = 0;

          // Read and decode one event
          ReadEvent();
          events_read++;
          if(dumpfile){
            DumpEvent(dumpfile);
          }
          DecodeEvent(-1);
          
          StreamStatus(&empty,&not_empty,&almost_full,&overflow);

        } // end while(not_empty>0)

        if(not_empty>0) printf("DECODEDVBURST: Timeout waiting for FIFOs to drain\n");
      } // end if (EOB||almost_full)


      if((!EOB)&&(not_empty==0)&&(trigs_stopped)){ // not EOB, but all FIFOs are empty
        trigs_stopped = 0;
        cc_menu_vbusy(0);  // set VBUSY to restart triggers
        printf("DECODEDVBURST: triggers restarted\n");
      }
    }

    if (!EOB){
      printf("DECODEDVBURST: TIMEOUT waiting for EOB (clk %ld tmax %ld tburst %f tzero %ld)\n", 
              (long)clock(), t_max, t_burst, t_zero);
      tocnt_sum++;
    }

    // Read and decode any events left in the output pages (should be <=4)
 
    while((events_read<events_expected)&&(tocnt_sum<burst_max_timeouts)){
      ReadEvent();
      events_read++;
      if(dumpfile){
        DumpEvent(dumpfile);
      }
      DecodeEvent(-1);
    }

    // Print out FIFO status at EOB
    if (empty)       printf("DECODEDVBURST: %d FIFOs are empty\n",empty); 
    if (not_empty)   printf("DECODEDVBURST: ERROR: %d FIFOs NOT EMPTY\n", not_empty);
    if (almost_full) printf("DECODEDVBURST: ERROR: %d FIFOs ALMOST FULL (PAF flag)\n", almost_full);
    if (overflow)    printf("DECODEDVBURST: ERROR: %d FIFOs have OVERFLOWED (fifo full flag)\n", overflow);

    // Stop CLOAC
    cc->Stop();
    cc_load_command(0);

    n_under=0;
    running=0;
    for (n=0; n<SCT_MAX_MODULES; n++){
      if(m[n]->present){
        // Here, Event counter has been filled by DecodeEvent
        m[n]->FillNOScanHist(burst_count);
        // Keep running if less than half the channels
        // have exceeded the target number of hits
        if(m[n]->ChansOverTarget(burst_target)<(st_n_active_channels(n)/2)){
          n_under++;
          running = 1;
        }
      }          
    }

    if(do_vary_ntrigs){
      printf("DECODEDVBURST: %d modules under target after %d triggers\n",n_under, tsent);
      if (tsent >= burst_maxtrigs) running = 0;
    } else {
      if (tsent >= burst_ntrigs) running = 0;
    }

    if (tocnt_sum >= burst_max_timeouts) {
      running = 0;
      printf("DECODEDVBURST: exceeded max number timeouts %d\n", tocnt_sum);
    }


  } // end while(running)

  // Close dumpfile (if open)
  if(dumpfile){
    fclose(dumpfile);
  }

  // Disable External Busy
  cc_menu_disable(ENEXTBUSY1);
  if(MUS_NUMBER>1){
    cc_menu_disable(ENEXTBUSY2);
  } 

  Sleep(10);

  for(mid=0; mid<MUS_NUMBER; mid++) {
    mus_set_readout_cntl(mid,0,0);
  }

  return 0;
}

int TST::EBurst(){

  // A trigger burst using external triggers
  // using HARDWARE histogramming
  // PWP 27.09.01
  //
  // CLOAC's external and VME busy input DO NOT suppress external
  // triggers, so the best we can do here is:
  //   o enable external triggers
  //   o poll MuSTARD until any stream is "not_empty"
  //   o disable triggers
  //   o read out events until MuSTARD is empty

  short mid,n;
  short rcode;
  short running = 1;

  short empty, not_empty;
  short almost_full, overflow;

  int events_expected = 0;
  int events_read = 0;
  int events_decoded = 0;

  int n_errs = 0;
  long t_max;

  char fname[64];
  FILE* dumpfile = NULL;

  // This burst type requires CLOAC
  // If there is no CLOAC, return now!!
  if(cc==0){
    cout << "EBurst called with no CLOAC!!" << endl;
    cout << "  exiting now...." << endl;
    return -99;
  }

  // Option to dump events to a file, one per burst, for diagnostic purposes
  if((debug_level==998)||(debug_level==999)){
    sprintf(fname,"%sstrun%dburst%d.evdump",sct_data,runnum,burstnum);

    dumpfile=fopen(fname,"w");
    if (dumpfile == NULL){ 
      printf("EBurst: Failed to open event dump file %s\n",fname);
      return(-1);
    }
  }

  // Enable CLOAC external busy input(s) 1 (and 2) to accept busy from MuSTARD trigger 
  // overflow counter(s)
  //   NB. MuSTARD0 must be connected to CLOAC output A
  //   and MuSTARD1 must be connected to CLOAC output B

  cc_menu_enable(ENEXTBUSY1); 
  if(MUS_NUMBER>1){
    cc_menu_enable(ENEXTBUSY2);
  } 

  SingleSoftReset();       /* Always do a global synchronous soft reset at SOB */
  

  for(mid=0; mid<MUS_NUMBER; mid++) {
    if ((rcode = mus_set_readout_modes(mid,0,1)) != 0) { /* DECODED, SAMPLE ALL */
      printf("EBurst: ERROR for Mustard %d: mus_set_readout_modes returns %d\n", mid, rcode);
      return -5;
    }
   
    if ((rcode = mus_set_readout_cntl(mid,1,1)) != 0) {
      printf("EBurst: ERROR for Mustard %d: mus_set_readout_cntl returns %d\n", mid, rcode);
      return -6;
    }
  }

  Sleep(10);   // Wait 10ms to ensure Mustards ready to receive first event

  events_expected = burst_ntrigs;
  n=0;

  while (running) {     // Enter event loop 

    // Enable external triggers 
    // and wait up to 5 seconds for a trigger to be seen

    not_empty=0;
    t_max = clock()+ 5*CLOCKS_PER_SEC;
    cc->EnableExternalTrigger();
    while (((long)clock()<t_max)&&(not_empty==0)){
      StreamStatus(&empty,&not_empty,&almost_full,&overflow);
    }
    cc->DisableExternalTrigger();

    if(not_empty==0){
      printf("EBurst: timeout waiting for event %d\n",events_read+1);
      tocnt_sum++;
    }

    // allow up to 5 seconds to empty MuSTARD's buffers

    if(not_empty){
      t_max = clock()+ 5*CLOCKS_PER_SEC; // 5 seconds timeout
      while (((long)clock()<t_max)&&(not_empty>0)){

        // Read and decode one event
        n_errs=TestEvent();
        events_read++;
        if(n_errs>0){  // error event to decode
          if(dumpfile){
            DumpEvent(dumpfile);
          }
          DecodeErrors();
          events_decoded++;
        }
        StreamStatus(&empty,&not_empty,&almost_full,&overflow);
      }

      if(not_empty){
        printf("EBurst: timeout waiting for FIFOs to drain\n");
        running = 0;
      }
    } // end if(not_empty)

    if (tocnt_sum >= burst_max_timeouts) {
      printf("EBurst: exceeded max number timeouts %d\n", tocnt_sum);
      running = 0;
    }

    if(events_read>=events_expected){
      printf("EBurst: %d events processed\n",events_read);
      running=0;
    }
  }

  // Print out FIFO status at EOB
  StreamStatus(&empty,&not_empty,&almost_full,&overflow);
  if (empty)       printf("EBurst: %d FIFOs are empty\n",empty); 
  if (not_empty)   printf("EBurst: ERROR: %d FIFOs NOT EMPTY\n", not_empty);
  if (almost_full) printf("EBurst: ERROR: %d FIFOs ALMOST FULL (PAF flag)\n", almost_full);
  if (overflow)    printf("EBurst: ERROR: %d FIFOs have OVERFLOWED (fifo full flag)\n", overflow);

  Sleep(10);
  tsent = events_read;

  for(mid=0; mid<MUS_NUMBER; mid++) {
    mus_set_readout_cntl(mid,0,0);
  }

  // Disable External Busy
  cc_menu_disable(ENEXTBUSY1);
  if(MUS_NUMBER>1){
    cc_menu_disable(ENEXTBUSY2);
  }
 
  ReadHistograms(-1);
  
  return 0;
}

int TST::DecodedEBurst(){

  // A trigger burst using external triggers
  // using HARDWARE histogramming
  // PWP 27.09.01
  //
  // CLOAC's external and VME busy input DO NOT suppress external
  // triggers, so the best we can do here is:
  //   o enable external triggers
  //   o poll MuSTARD until any stream is "not_empty"
  //   o disable triggers
  //   o read out events until MuSTARD is empty

  short mid;
  short rcode;
  short running = 1;

  short empty, not_empty;
  short almost_full, overflow;

  int events_expected = 0;
  int events_read = 0;
  int events_decoded = 0;

  int n_errs = 0;
  long t_max;

  char fname[64];
  FILE* dumpfile = NULL;

  // This burst type requires CLOAC
  // If there is no CLOAC, return now!!
  if(cc==0){
    cout << "DecodedEBurst called with no CLOAC!!" << endl;
    cout << "  exiting now...." << endl;
    return -99;
  }

  // Option to dump events to a file, one per burst, for diagnostic purposes
  if((debug_level==998)||(debug_level==999)){
    sprintf(fname,"%sstrun%dburst%d.evdump",sct_data,runnum,burstnum);

    dumpfile=fopen(fname,"w");
    if (dumpfile == NULL){ 
      printf("DecodedEBurst: Failed to open event dump file %s\n",fname);
      return(-1);
    }
  }

  // Enable CLOAC external busy input(s) 1 (and 2) to accept busy from MuSTARD trigger 
  // overflow counter(s)
  //   NB. MuSTARD0 must be connected to CLOAC output A
  //   and MuSTARD1 must be connected to CLOAC output B

  cc_menu_enable(ENEXTBUSY1); 
  if(MUS_NUMBER>1){
    cc_menu_enable(ENEXTBUSY2);
  } 

  SingleSoftReset();       /* Always do a global synchronous soft reset at SOB */
  

  for(mid=0; mid<MUS_NUMBER; mid++) {
    if ((rcode = mus_set_readout_modes(mid,0,1)) != 0) { /* DECODED, SAMPLE ALL */
      printf("DecodedEBurst: ERROR for Mustard %d: mus_set_readout_modes returns %d\n", mid, rcode);
      return -5;
    }
   
    if ((rcode = mus_set_readout_cntl(mid,1,1)) != 0) {
      printf("DecodedEBurst: ERROR for Mustard %d: mus_set_readout_cntl returns %d\n", mid, rcode);
      return -6;
    }
  }

  Sleep(10);   // Wait 10ms to ensure Mustards ready to receive first event

  events_expected = burst_ntrigs;

  while (running) {     // Enter event loop 

    // Enable external triggers 
    // and wait up to 5 seconds for a trigger to be seen

    not_empty=0;
    t_max = clock()+ 5*CLOCKS_PER_SEC;
    cc->EnableExternalTrigger();
    while (((long)clock()<t_max)&&(not_empty==0)){
      StreamStatus(&empty,&not_empty,&almost_full,&overflow);
    }

    cc->DisableExternalTrigger();

    if(not_empty==0){
      printf("DecodedEBurst: timeout waiting for event %d\n",events_read+1);
      tocnt_sum++;
    }else{
      StreamStatus(&empty,&not_empty,&almost_full,&overflow);
    }

    // allow up to 5 seconds to empty MuSTARD's buffers

    if(not_empty){
      t_max = clock()+ 5*CLOCKS_PER_SEC; // 5 seconds timeout
      while (((long)clock()<t_max)&&(not_empty>0)){

        // Read and decode one event
        n_errs=ReadEvent();
        events_read++;
        if(dumpfile){
          DumpEvent(dumpfile);
        }
        DecodeEvent(-1);
        events_decoded++;

        StreamStatus(&empty,&not_empty,&almost_full,&overflow);
      }

      if(not_empty){
        printf("DecodedEBurst: timeout waiting for FIFOs to drain\n");
        running = 0;
      }
    } // end if(not_empty)

    if (tocnt_sum >= burst_max_timeouts) {
      printf("DecodedEBurst: exceeded max number timeouts %d\n", tocnt_sum);
      running = 0;
    }

    if(events_read>=events_expected){
      printf("DecodedEBurst: %d events processed\n",events_read);
      running=0;
    }
  }

  // Print out FIFO status at EOB
  if (empty)       printf("DecodedEBurst: %d FIFOs are empty\n",empty); 
  if (not_empty)   printf("DecodedEBurst: ERROR: %d FIFOs NOT EMPTY\n", not_empty);
  if (almost_full) printf("DecodedEBurst: ERROR: %d FIFOs ALMOST FULL (PAF flag)\n", almost_full);
  if (overflow)    printf("DecodedEBurst: ERROR: %d FIFOs have OVERFLOWED (fifo full flag)\n", overflow);

  Sleep(10);
  tsent = events_read;

  for(mid=0; mid<MUS_NUMBER; mid++) {
    mus_set_readout_cntl(mid,0,0);
  }

//  ReadHistograms(-1);
  
  return 0;
}


int TST::DecodedSBurst(){
  // use SLOG to generate bursts of
  //   burst_ntrigs periodic triggers
  //
  // Use MuSTARD stream trigger limit register / SLOG busy in
  // to throttle triggers when ABCD pipeline has received >7 L1A
  // (N.B. if burst_ntrigs =7, this condition should not occur)
  //
  // Use MuSTARD programmable almost full flag / SLOG busy in
  // to throttle triggers when input FIFO is nearly full (<8k left)
  //
  // Read each event into host memory and process the decoded event
  // such that histograms of n(hits) vs n(events) can be filled

  short mid,n,sid;
  short status;
  short running = 1;

  short n_chips, max_chips;

  short EOB, SOB;
  short empty, not_empty;
  short almost_full, overflow;
  short trigs_stopped;

  short n_under;

  long event_duration, event_interval, trigger_period;
  long t_zero, t_max;
  double t_burst;

  unsigned short n_trigs;
  unsigned short runstate;

  int events_expected = 0;
  int events_read = 0;

  char fname[64];
  FILE* dumpfile = NULL;

  // This burst type does not require CLOAC

  // Option to dump events to a file, one per burst, for diagnostic purposes
  if((debug_level==998)||(debug_level==999)){
    sprintf(fname,"%sstrun%dburst%d.evdump",sct_data,runnum,burstnum);

    dumpfile=fopen(fname,"w");
    if (dumpfile == NULL){ 
      printf("DecodedSBurst: Failed to open event dump file %s\n",fname);
      return(-1);
    }
  }

  SingleSoftReset();       /* Always do a global synchronous soft reset at SOB */
 
  for(mid=0; mid<MUS_NUMBER; mid++) {

    if ((status = mus_set_readout_modes(mid,0,1)) != 0) { /* DECODED, SAMPLE ALL */
      printf("DecodedSBurst: ERROR for Mustard %d: mus_set_readout_modes returns %d\n", mid, status);
      return -5;
    }
   
    if ((status = mus_set_readout_cntl(mid,1,1)) != 0) {  /* DAQ ENABLE, READOUT ENABLE */
      printf("DecodedSBurst: ERROR for Mustard %d: mus_set_readout_cntl returns %d\n", mid, status);
      return -6;
    }

  }

  Sleep(10);   // Wait 10ms to ensure Mustards ready to receive first event

  // Enable SLOG external busy input 1  to accept busy from MuSTARD trigger 
  // overflow counter(s)
  //   NB. MuSTARD0 must be connected to SLOG0 ECLA
  //   and MuSTARD1 must be connected to SLGO1 ECLB
  //   (This is very crude and not robust against all possible uses,
  //   but it will do for now!  PWP 13.08.2002)
  for(sid=0;sid<SLOG_NUMBER;sid++){
    slog_set_busy(sid,1);
  } 

  // TIMEOUT CALCULATIONS

  // Max time to process one event in MuSTARD:
  //   single chip, 1/2 occupancy, no adjacent hits => event length ~1100 bits
  //   clock period = 25 nS
  //   => allow 28 microS per chip per event

  max_chips=0;
  for(mid=0; mid<MUS_NUMBER; mid++) {
    n_chips = st_find_maxchips(mid);
    if (n_chips>max_chips) max_chips=n_chips;
  }
  event_duration = (max_chips*28); // in microS

  // For timeout calculations, we must also consider the requested
  // Trigger frequency:  at high trigger frequencies the time taken to
  // process events in MuSTARD will dominate, but at low frequencies
  // the decoding time is small with respect to the time between events.

  trigger_period = 10;

  if (trigger_period>event_duration) event_interval = trigger_period;
  else                               event_interval = event_duration;

  printf("ev_duration %ldmS, tr_period %ldmS, =>ev_interval %ldmS\n",
    event_duration, trigger_period, event_interval);


  // SLOG's cycle counter may have 16 bits, but the
  // data histogram has only 15 useful bits.  To avoid 
  // overflow on the first sub-burst we must limit the
  // number of triggers to less than 2^15.  32500 should
  // be safe.  PWP 18.03.02
//  if(burst_ntrigs>32500) n_trigs = 32500;
  if(burst_ntrigs>16000) n_trigs = 16000;  // The cycle counter seems to decrement
                                           // twice each cycle, hence the maximum  
                                           // number of events is halved PWP 20.08.02
  else n_trigs = burst_ntrigs;

  // load trigger pattern

  st_set_slog_loadtrig(1);
  st_slog_load("slog_ctest.txt");

  while (running) {     // Enter event loop 

    // Repeatedly request bursts of "ntrigs" L1A 
    // until "burst_ntrigs" has been reached

    if(do_vary_ntrigs){
      if(burst_maxtrigs>n_trigs){
        if((burst_maxtrigs-tsent)<n_trigs) {
        n_trigs = (burst_maxtrigs-tsent); 
        }
      }
    } else {
      if(burst_ntrigs>n_trigs){
        if((burst_ntrigs-tsent)<n_trigs) {
        n_trigs = (burst_ntrigs-tsent); 
        }
      }
    }

    events_expected+=n_trigs;
    tsent+= n_trigs;

    printf("DECODEDSBURST: requesting %d L1A triggers\n", n_trigs);
    trigs_stopped = 0;

    // start burst of SLOG triggers

    t_zero = (long) clock();
    t_max  = t_zero + CLOCKS_PER_SEC; 

    SOB=0;

    st_slog_start(1,(n_trigs-1));
//    st_slog_start(1,2*(n_trigs-1));  // Wierdly, the cycle counter seems to 
                                     // decrement twice each cycle PWP 20.08.02
 
    while (((long)clock()<t_max)&&(SOB==0)){
      SOB=1;
      for(sid=0;sid<SLOG_NUMBER;sid++){
        slog_get_runstate(sid, &runstate); // returns TRUE if running
        if ((runstate&0x01)==0x00) SOB=0;
      }
    }

    if(SOB){
      printf("DECODEDSBURST: SOB\n");
    } else {
      printf("DECODEDSBURST: TIMEOUT waiting for SOB (clk %ld tmax %ld tzero %ld)\n", (long) clock(), t_max, t_zero);
      tocnt_sum++;
    }

    // Wait for end of SLOG DECODEDSBURST

    t_zero  = (long) clock();
    t_burst = (double)event_duration * (double)n_trigs / (double)1000; // burst duration in seconds
    t_max  = t_zero + (long)((t_burst+1) * CLOCKS_PER_SEC);

    EOB=0;
    while (((long)clock()<t_max)&&(EOB==0)){
      EOB=1;
      for(sid=0;sid<SLOG_NUMBER;sid++){
        slog_get_runstate(sid, &runstate); // returns TRUE if running
        if ((runstate&0x01)==0x01) EOB=0;
      }

      if (EOB){
         printf("DECODEDSBURST: EOB\n");
      }

      // meanwhile, check for FIFO ALMOST FULL flags
      // if found, inhibit triggers until all FIFOs are empty

      almost_full = 0;
      overflow = 0;
      StreamStatus(&empty,&not_empty,&almost_full,&overflow);

      if(almost_full){ 
        printf("DECODEDSBURST: %d FIFOs ALMOST FULL\n", almost_full);
        if(overflow){
          printf("DECODEDSBURST: ERROR: %d FIFOs have OVERFLOWED\n", overflow);
        }

        if(!EOB) {
          st_slog_stop();
          trigs_stopped = 1;
          printf("DECODEDSBURST: triggers stopped to drain FIFOs\n");
        }
      }

      // Read and decode events until all FIFOs are empty at EOB, or if any streams are almost_full

      if(EOB||almost_full){

        empty = 0;
        not_empty = 1;
        while (((long)clock()<t_max)&&(not_empty>0)){
  
          empty = 0;
          not_empty = 0;
          almost_full = 0;
          overflow = 0;

          // Read and decode one event
          ReadEvent();
          events_read++;
          if(dumpfile){
            DumpEvent(dumpfile);
          }
          DecodeEvent(-1);
          
          StreamStatus(&empty,&not_empty,&almost_full,&overflow);

        } // end while(not_empty>0)

        if(not_empty>0) printf("DECODEDSBURST: Timeout waiting for FIFOs to drain\n");
      } // end if (EOB||almost_full)


      if((!EOB)&&(not_empty==0)&&(trigs_stopped)){ // not EOB, but all FIFOs are empty
        trigs_stopped = 0;
        st_slog_start(0,0);
        printf("DECODEDSBURST: triggers restarted\n");
      }
    }

    if (!EOB){
      printf("DECODEDSBURST: TIMEOUT waiting for EOB (clk %ld tmax %ld tburst %f tzero %ld)\n", 
              (long)clock(), t_max, t_burst, t_zero);
      tocnt_sum++;
    }

    // Read and decode any events left in the output pages (should be <=4)
 
    while((events_read<events_expected)&&(tocnt_sum<burst_max_timeouts)){
      ReadEvent();
      events_read++;
      if(dumpfile){
        DumpEvent(dumpfile);
      }
      DecodeEvent(-1);
    }

    // Print out FIFO status at EOB
    if (empty)       printf("DECODEDSBURST: %d FIFOs are empty\n",empty); 
    if (not_empty)   printf("DECODEDSBURST: ERROR: %d FIFOs NOT EMPTY\n", not_empty);
    if (almost_full) printf("DECODEDSBURST: ERROR: %d FIFOs ALMOST FULL (PAF flag)\n", almost_full);
    if (overflow)    printf("DECODEDSBURST: ERROR: %d FIFOs have OVERFLOWED (fifo full flag)\n", overflow);

    // Stop SLOG 
    st_slog_stop();

    n_under=0;
    running=0;
    for (n=0; n<SCT_MAX_MODULES; n++){
      if(m[n]->present){
        // Here, Event counter has been filled by DecodeEvent
        m[n]->FillNOScanHist(burst_count);
        // Keep running if less than half the channels
        // have exceeded the target number of hits
        if(m[n]->ChansOverTarget(burst_target)<(st_n_active_channels(n)/2)){
          n_under++;
          running = 1;
        }
      }          
    }

    if(do_vary_ntrigs){
      printf("DECODEDSBURST: %d modules under target after %d triggers\n",n_under, tsent);
      if (tsent >= burst_maxtrigs) running = 0;
    } else {
      if (tsent >= burst_ntrigs) running = 0;
    }

    if (tocnt_sum >= burst_max_timeouts) {
      running = 0;
      printf("DECODEDSBURST: exceeded max number timeouts %d\n", tocnt_sum);
    }


  } // end while(running)

  // Close dumpfile (if open)
  if(dumpfile){
    fclose(dumpfile);
  }

  // Disable External Busy
  for(sid=0;sid<SLOG_NUMBER;sid++){
    slog_set_busy(sid,0);
  } 

  Sleep(10);

  for(mid=0; mid<MUS_NUMBER; mid++) {
    mus_set_readout_cntl(mid,0,0);
  }

  return 0;
}



int TST::TimeEvBurst(short cal_mode){
  // Time Evolution Burst
  // Decode a block of up to 8 arbitrarily (but normally zero) spaced triggers to 
  // up to 24 bins of the scan histogram.  The idea: to observe the time evolution
  // of a signal in as close as possible to an analogue form as you can get without 
  // a picoprobe.... (Still not very close, is it??)
  // 
  // In practice, one burst is equivalent to a full scan - makes life complicated
  // at the scan level, but not impossible....
  //
  // PWP 16.03.00

  short mid;
  short rcode;
  short running = 1;
  short next_config;
  short next_soft_reset;
  short next_bc_reset;
  short start_burst = burst_count;

  int   events_expected = 0;
  int   events_read = 0;
  int   events_decoded = 0;

  SingleSoftReset();       /* Always do a global synchronous soft reset at SOB */
  
  for(mid=0; mid<MUS_NUMBER; mid++) {
    if ((rcode = mus_set_readout_modes(mid,0,1)) != 0) { /* DECODED, SAMPLE ALL */
      printf("TimeEvBurst: ERROR for Mustard %d: mus_set_readout_modes returns %d\n", mid, rcode);
      return -5;
    }
   
    if ((rcode = mus_set_readout_cntl(mid,1,1)) != 0) {
      printf("TimeEvBurst: ERROR for Mustard %d: mus_set_readout_cntl returns %d\n", mid, rcode);
      return -6;
    }
  }

  if (config_frequency > 0) next_config = config_frequency;
  else                      next_config = -1;

  if (soft_reset_frequency > 0) next_soft_reset = soft_reset_frequency;
  else                          next_soft_reset = -1;

  if (bc_reset_frequency > 0) next_bc_reset = bc_reset_frequency;
  else                        next_bc_reset = -1;

  Sleep(10);   // Wait 10ms to ensure Mustards ready to receive first event

  while (running) {     // Enter event loop 
    next_config--;
    next_soft_reset--;
    next_bc_reset--;

    if (!next_config) {
      if(do_mask_test){
        ConfigureVariable(7,(float)((tsent%2)+1)); // toggle mask between 1010 and 0101
      }
      ExecuteConfigs(0);
      next_config = config_frequency;
      Sleep(1);
    }

    if (!next_soft_reset) {
      SingleSoftReset();
      next_soft_reset = soft_reset_frequency;
      Sleep(1);
    }

    if (!next_bc_reset) {
      SingleBCReset();
      next_bc_reset = bc_reset_frequency;
      Sleep(1);
    }

    Single(burst_trtype);   // Send a burst of triggers
    tsent+=L1A_per_loop;
    events_expected+=L1A_per_loop;
    if(debug_level>7) {
      printf("Sending %d triggers - now up to %d\n",L1A_per_loop,tsent);
    }

    while(events_read<events_expected){
      if(events_decoded<events_read){  // event in memory to be decoded
        if((events_decoded%L1A_per_loop)==0)
          burst_count = start_burst;
        DecodeEvent3bins(cal_mode);
        events_decoded++;
        burst_count+=3;
      }
      ReadEvent();
      events_read++;
    }

    if (tocnt_sum >= burst_max_timeouts) {
      running = 0;
      printf("TimeEvBurst: exceeded max number timeouts %d\n", tocnt_sum);
    }
    if (tsent >= burst_ntrigs) running = 0;
  }

  for(mid=0; mid<MUS_NUMBER; mid++) {
    mus_set_readout_cntl(mid,0,0);  
  }
  return 0;
}

//
// TST::DumpBurst has been deleted - option to dump events now added
//                                   to DecodedBurst and DecodedVBurst
//                                   PWP 26.04.01
//

int TST::TestBurst(){   // Burst of single triggers, decoded in software AND read from
                        // Mustard.  Only one module (two links, 0 and 1) supported:
                        // higher module space used for detailed histograms as follows:
                        //  0:1  Mustard histograms
                        //  2:3  Decoded histogram, all timebins merged
                        //  4:5      "       "      timebin 0
                        //  6:7      "       "         "    1
                        //  8:9      "       "         "    2
                        // For event-by-event comparison, use ntrigs = 1     
  short mid = 0;
  short rcode,status;
  int   trigger = 0;
  short running = 1;
  short next_config;
  short next_soft_reset;
  short next_bc_reset;

  short evdata[0x8000];  // This is the size of Mustard's event data buffer
                         // ### What is the most efficient method of reserving this space ?
  short *evpointer;
  short evlength;

  short hdata[1024];     // We read only 768 channels at a time
  short *hpointer;
  short hlength;

  // Variables filled from decoded event data
  short n;
  // short word_count;
  // short error_count;
  // int   event_count;

  short stream;
  short strip_address = 0;
  short L1_count;
  short BC_count;
  short module = -1;
  short soft_link = -1;

  for (n=0; n<6; n++) m[n]->ResetData();

  SingleSoftReset();       /* Always do a global synchronous soft reset at SOB */
  
  if ((rcode = mus_set_readout_modes(mid,0,1)) != 0) { /* DECODED, SAMPLE ALL */
    printf("TST::TestBurst: ERROR for Mustard %d: mus_set_readout_modes returns %d\n", mid, rcode);
    return -5;
  }
   
  if ((rcode = mus_set_readout_cntl(mid,1,1)) != 0) {
    printf("TST::TestBurst: ERROR for Mustard %d: mus_set_readout_cntl returns %d\n", mid, rcode);
    return -6;
  }

  if (config_frequency > 0) next_config = config_frequency;
  else                      next_config = -1;

  if (soft_reset_frequency > 0) next_soft_reset = soft_reset_frequency;
  else                          next_soft_reset = -1;

  if (bc_reset_frequency > 0) next_bc_reset = bc_reset_frequency;
  else                        next_bc_reset = -1;

  Sleep(10);   // Wait 10ms to ensure Mustard ready to receive first event

  running = 1;
  while (running) {     // Enter event loop 
    next_config--;
    next_soft_reset--;
    next_bc_reset--;

    if (!next_config) {
      if(do_mask_test){
        ConfigureVariable(7,(float)((tsent%2)+1)); // toggle mask between 1010 and 0101
      }
      ExecuteConfigs(0);
      next_config = config_frequency;
      Sleep(1);
    }

    if (!next_soft_reset) {
      SingleSoftReset();
      next_soft_reset = soft_reset_frequency;
      Sleep(1);
    }

    if (!next_bc_reset) {
      SingleBCReset();
      next_bc_reset = bc_reset_frequency;
      Sleep(1);
    }
	
    Single(burst_trtype);   // Send the single trigger
    tsent++;

    evpointer = evdata;
    rcode = mus_read_event(mid,&evlength,&evpointer);

    // Update MuSTARD hardware event and error counters 
    ercnt[mid] =  evdata[1];
    evcnt[mid] = (evdata[3]<<16) + (unsigned short) evdata[2]; // PWP 19.03.02
  
    if (debug_level > 8) {
      printf("TestBurst: burst %d tsent %d evcnt %d got evlength %d\n", burst_count,tsent,evcnt[mid],evlength);
    }

    if (rcode != 0) {
      if (debug_level > 5) printf("TestBurst: ERROR %d from Mustard %d\n", rcode,mid);
      if (rcode == -4) {
        tocnt_sum++;
        tocnt[0]++;
        evlength = 0;  
      } else {
        xecnt_sum++;
        xecnt[0]++;
      }
      mus_read_status(mid,&status);
      mus_print_status(status);
    }

    h_evlen[mid]->Fill((float)evlength);   // Fill even if zero

    if (evlength > 0){   // Decode the DECODED mode stream data
                         // ### ignore timebins for the moment
      // word_count = evdata[0];
      // error_count = evdata[1];
      // event_count = evdata[3];
      // event_count = (event_count<<16) + evdata[2];

      for (n=4; n<evlength; n++){

        if (evdata[n]&0x8000) {
          BC_count = evdata[n] & 0x00ff;
          L1_count = (evdata[n] & 0x0f00) >> 8;
          stream = evdata[n++]&0x000f;
          if (stream > 1) break;  // ### Special decoding for this burst type: module 0 only
          module = 0;
          // TODO:  Histogram BC and L1 counters
        }
        if ((evdata[n]&0x4000)&&(debug_level>6)){
          printf("TestBurst: stream %d error at short %d val %4x\n", stream, n, evdata[n]);
        }
        if (evdata[n]&0x2000) { 
          // For the time being, we have two scan histograms, one per "link"
          // in practice, "link 0" handles strips    0 to  768
          //              "link 1" handles strips 1024 to 1792
          // We may come to regret having taken this approach....

          strip_address =  evdata[n] & 0x03ff;
          soft_link   = (evdata[n] & 0x0400)>>10; 
        }
        if (evdata[n]&0x1000) {   
          if (strip_address<768){
            m[1]->scan[burst_count][soft_link][strip_address]++;
            if (evdata[n]&0x0001)  m[2]->scan[burst_count][soft_link][strip_address]++;
            if (evdata[n]&0x0002)  m[3]->scan[burst_count][soft_link][strip_address]++;
            if (evdata[n]&0x0004)  m[4]->scan[burst_count][soft_link][strip_address]++;
          } else if (debug_level>6) {
            printf("TestBurst: Stream %d non-existent strip %d\n",stream,strip_address);
          } 
          strip_address++;
        }
      }
    }

    if (tocnt_sum >= burst_max_timeouts) {
      running = 0;
      printf("TestBurst: exceeded max number timeouts %d\n", tocnt_sum);
    }
    if (tsent >= burst_ntrigs) running = 0;
  }

  mus_set_readout_cntl(mid,0,0);

  // ### Now read the Mustard histograms for module 0 into the space normally
  // used for module 0.  No attempts to cater to redundancy.

  hpointer = hdata;
  mus_read_histogram(mid,0,0,&hlength,&hpointer);  // hilo=0, stream=0
  printf("TestBurst: hlength %d  first entries %d %d %d %d %d %d\n",
    hlength,hdata[0],hdata[1],hdata[2],hdata[3],hdata[4],hdata[5]);
  if (hlength>768) hlength = 768;
  for (n=0; n<hlength; n++){
    m[0]->scan[burst_count][0][n] = hdata[n];
  }  

  hpointer = hdata;
  mus_read_histogram(mid,1,1,&hlength,&hpointer);  // hilo=1, stream=1
  printf("TestBurst: hlength %d  first entries %d %d %d %d %d %d\n",
    hlength,hdata[0],hdata[1],hdata[2],hdata[3],hdata[4],hdata[5]);
  if (hlength>768) hlength = 768;
  for (n=0; n<hlength; n++){
    m[0]->scan[burst_count][1][n] = hdata[n];
  }  

  mus_reset(mid); 
  for (n=0;n<12;n++) mus_clear_histogram(mid,n);

  // h_merr is filled by ExecuteBurst

  for (n=0; n<6; n++){
    // Why don't we fill scan histograms here ??
    // ExecuteBurst only fills scans histograms for "present" modules.
    // Confused...
    m[n]->ResetBurstHist();
    m[n]->FillBurstHist(-1); // was burst_count, but argument is cal_mode PWP 15.08.00
    printf("Module %d  Total occupancies link0 %d  link1 %d\n",n,
      m[n]->scan_occupancy[0][burst_count],m[n]->scan_occupancy[1][burst_count]);
  }

  for (n=0;n<128;n++) 
    printf("ch %3d  mushis %6d   decoded (addr) %6d  (hits) %6d  %6d  %6d\n", n, 
      m[0]->scan[burst_count][0][n], m[1]->scan[burst_count][0][n],
      m[2]->scan[burst_count][0][n], m[3]->scan[burst_count][0][n],
      m[4]->scan[burst_count][0][n]);

  return 0;
}


void TST::SEUClear(){

  short event,n,link,chip;

  // reset counters for SEUBurst error totals
  // should perhaps be scan wide??
  // or could use counters within TModule??

  for (link=0; link<SEU_NLINKS; link++){
    for (n=0;n<4;n++){
      seu_bad_count[n]=0;
      seu_good_count[n]=0;

      for (event=0;event<SEU_NEVENTS;event++){
        seu_tot_l1err[n][event][link]=0;
        seu_tot_wcerr[n][event][link]=0;
      }   
      for(chip=0;chip<6;chip++){
        seu_tot_conf[n][link][chip]=0;
        seu_tot_mask[n][link][chip]=0;
        seu_tot_pipe[n][link][chip]=0;
        seu_tot_mystery[n][link][chip]=0;
      }
    }
  }
}


void TST::SEUBurst(){

  int save_do_cal;

  int rcode;
  int running = 1;
  short n, event,link,chip;
  int seu_burst_header[14];

  save_do_cal = do_cal_loop;
  do_cal_loop = 0;

  tsent = 0;
  ercnt_sum = 0;
  tocnt_sum = 0;
  xecnt_sum = 0;

  burstnum++;
  seu_test_count=0; // Total number of SEU tests performed

  // reset SEU specific counters

  for (link=0; link<SEU_NLINKS; link++){
    for (n=0;n<2;n++){
      seu_bad_count[n]=0;   // SEU cycles with protocol violations
      seu_good_count[n]=0;  // SEU cycles without protocol violations

      for (event=2;event<SEU_NEVENTS;event++){
        seu_tot_l1err[n][event][link]=0;
        seu_tot_wcerr[n][event][link]=0;
      }   
      for(chip=0;chip<6;chip++){
        seu_tot_conf[n][link][chip]=0;
        seu_tot_mask[n][link][chip]=0;
        seu_tot_pipe[n][link][chip]=0;
        seu_tot_mystery[n][link][chip]=0;
      }
    }
  }


  n=0;  // start with SEU 0->1


  while (running){

    seu_invert = n;

    if (ResetMustards() != 0) return;

    // SEUTest(1)  => SEU Measurement with HARDWARE timing
    // SEUTest(0)  =>                 with SOFTWARE timing

    rcode = SEUTest(1);  // HARDWARE timing

    if(rcode==0){  // No protocol violation

      for (link=0; link<SEU_NLINKS; link++){
        for (event=2;event<SEU_NEVENTS;event++){
          seu_tot_l1err[n][event][link] += seu_l1err[event][link];
          seu_tot_wcerr[n][event][link] += seu_wcerr[event][link];
        }
        for(chip=0;chip<6;chip++){
          seu_tot_conf[n][link][chip] += seu_conf[link][chip];
          seu_tot_mask[n][link][chip] += seu_mask[link][chip];
          seu_tot_pipe[n][link][chip] += seu_pipe[link][chip];
          if(seu_mystery[link][chip]) seu_tot_mystery[n][link][chip]++; 
        }
      }
      seu_good_count[n]++;
    } else { // Protocol Violation
      seu_bad_count[n]++;
    }

// for use at PS uncomment the next two lines PWP 06.07.00

    // Swap to "other type" of SEU measurement

    if (n==1) n=0;
    else      n=1;

    if (seu_test_count >= burst_ntrigs) running = 0;
    if (tocnt_sum > burst_max_timeouts) running = 0;
  }

  // update run error counters

  run_tsent += tsent;
  run_evcnt += tsent;
  run_ercnt += ercnt_sum;
  run_tocnt += tocnt_sum;
  run_xecnt += xecnt_sum;

  // update SEU specific counters

  for (n=0;n<2;n++){
    seu_bad_count[n+2] += seu_bad_count[n];
    seu_good_count[n+2] += seu_good_count[n];

    for (link=0; link<SEU_NLINKS; link++){
      for (event=2;event<SEU_NEVENTS;event++){
        seu_tot_l1err[n+2][event][link] += seu_tot_l1err[n][event][link];
        seu_tot_wcerr[n+2][event][link] += seu_tot_wcerr[n][event][link];
      }
      for(chip=0;chip<6;chip++){
        seu_tot_conf[n+2][link][chip] += seu_tot_conf[n][link][chip];
        seu_tot_mask[n+2][link][chip] += seu_tot_mask[n][link][chip];
        seu_tot_pipe[n+2][link][chip] += seu_tot_pipe[n][link][chip];
        seu_tot_mystery[n+2][link][chip] += seu_tot_mystery[n][link][chip];
      }
    }
  }

  printf("SEUBurst Error Totals: ercnt %d tocnt %d xecnt %d\n\n",
    ercnt_sum, tocnt_sum, xecnt_sum);
  for (n=0;n<2;n++){
    for (link=0; link<SEU_NLINKS; link++){
      if (seu_wc[0][link]>0){
        printf("Link %d\n",link);
        for (event=2;event<7;event++){
          printf("  Event:%d  l1_errors:%d  wc_errors:%d\n", event,
             seu_tot_l1err[n][event][link], seu_tot_wcerr[n][event][link]);
        }
      }
    }
  }

  for (n=0;n<2;n++){
    for (link=0; link<SEU_NLINKS; link++){
      printf("Lk %d  ",link);
      for(chip=0;chip<6;chip++) printf("%d ",seu_tot_conf[n][link][chip]);
      printf("  ");
      for(chip=0;chip<6;chip++) printf("%d ",seu_tot_pipe[n][link][chip]);
      printf("  ");
      for(chip=0;chip<6;chip++) printf("%d ",seu_tot_mask[n][link][chip]);
      printf("  ");
      for(chip=0;chip<6;chip++) printf("%d ",seu_tot_mystery[n][link][chip]);
      printf("\n");
    }
  }

  int date, time;
  TDatime dt;

  date = dt.GetDate();
  time = dt.GetTime();
  
  seu_burst_header[0] = runnum;
  seu_burst_header[1] = scannum;
  seu_burst_header[2] = burstnum;
  seu_burst_header[3] = seu_test_count;
  seu_burst_header[4] = date;
  seu_burst_header[5] = time;
  seu_burst_header[6] = tsent;
  seu_burst_header[7] = ercnt_sum;
  seu_burst_header[8] = tocnt_sum;
  seu_burst_header[9] = xecnt_sum;
  seu_burst_header[10] = seu_good_count[0];
  seu_burst_header[11] = seu_good_count[1];
  seu_burst_header[12] = seu_bad_count[0];
  seu_burst_header[13] = seu_bad_count[1];


  SEUAppendSummary(14, seu_burst_header,0); // not grand total
  do_cal_loop = save_do_cal;

  return;
}


void TST::TranslateVariableName(short typ, char *nam){

  switch (typ) {
    case ST_ELAPSED: 
      sprintf(nam,"%s","ELAPSED SEQUENCE");
      break;
    case ST_VTHR:
      sprintf(nam,"%s","THRESHOLD (mV)");
      break;
    case ST_VCAL:
      sprintf(nam,"%s","VCAL (mV)");
      break;
    case ST_STROBE_DELAY:
      sprintf(nam,"%s","STROBE DELAY");
      break;
    case ST_PREAMP:
      sprintf(nam,"%s","PREAMP CURRENT(uA)");
      break;
    case ST_SHAPER:
      sprintf(nam,"%s","SHAPER CURRENT(uA)");
      break;
    case ST_TRIM:
      sprintf(nam,"%s","TRIMDAC");
      break;
    case ST_MASK:
      sprintf(nam,"%s","MASK PATTERN (0-3)");
      break;
    case ST_ROLE:
      sprintf(nam,"%s","CHIP ROLE (0-5)");
      break;
    case ST_NMASK:
      sprintf(nam,"%s","N MASKED CHANNELS (0-1535)");
      break;
    case ST_CAL_MODE:
      sprintf(nam,"%s","CAL_MODE (0-3)");
      break;
    case ST_COMPRESSION:
      sprintf(nam,"%s","COMPRESSION (0-3)");
      break;
    case ST_TRIM_RANGE:
      sprintf(nam,"%s","TRIM_RANGE (0-3)");
      break;
    case ST_EDGE_DETECT:
      sprintf(nam,"%s","EDGE_DETECT (0/1)");
      break;
    case ST_SEND_MASK:
      sprintf(nam,"%s","SEND_MASK (0/1)");
      break;
    case ST_ACCUMULATE:
      sprintf(nam,"%s","ACCUMULATE (0/1)");
      break;
    case ST_BYPASS:
      sprintf(nam,"%s","BYPASS (0-12)");
      break;
    case ST_TOKEN:
      sprintf(nam,"%s","TOKEN (0-61)");
      break;
    case ST_QTHR:
      sprintf(nam,"%s","THRESHOLD (fC)");
      break;
    case ST_QCAL:
      sprintf(nam,"%s","QCAL (fC)");
      break;
    case ST_STREAM_DELAY:
      sprintf(nam,"%s","STREAM DELAY (ns)");
      break;
    case ST_STREAM_DELAY0:
      sprintf(nam,"%s","STREAM DELAY0 (ns)");
      break;
    case ST_STREAM_DELAY1:
      sprintf(nam,"%s","STREAM DELAY1 (ns)");
      break;
    case ST_CLOCK_DUTY:
      sprintf(nam,"%s","CLOCK DUTY");
      break;
    case ST_CLOCK_PHASE:
      sprintf(nam,"%s","CLOCK PHASE");
      break;
    case ST_COM_DELAY:
      sprintf(nam,"%s","COMMAND DELAY (BCO)");
      break;
    case ST_PULSE_DELAY:
      sprintf(nam,"%s","PULSE INPUT TRIGGER DELAY (BCO)");
      break;
    case ST_TRIG_DELAY:
      sprintf(nam,"%s","TRIGGER DELAY (BCO)");
      break;
    case ST_NL1A:
      sprintf(nam,"%s","Number of L1A");
      break;
    case ST_L1A_SPACING:
      sprintf(nam,"%s","L1A spacing (BCO)");
      break;
    case ST_VDET:
      sprintf(nam,"%s","Vdet (V)");
      break;
    case ST_VCC:
      sprintf(nam,"%s","Vcc (V)");
      break;
    case ST_VI1:
      sprintf(nam,"%s","Vi1 (V)");
      break;
    case ST_VDD:
      sprintf(nam,"%s","Vdd (V)");
      break;
    case ST_VPIN:
      sprintf(nam,"%s","Vpin (V)");
      break;
    case ST_VLED:
      sprintf(nam,"%s","Vled (V)");
      break;
    case ST_VLED0:
      sprintf(nam,"%s","Vled0 (V)");
      break;
    case ST_VLED1:
      sprintf(nam,"%s","Vled1 (V)");
      break;
    case ST_SELECT:
      sprintf(nam,"%s","SELECT (0/1)");
      break;
    case ST_OPTO_RX_THR:
      sprintf(nam,"%s","Opto RX threshold (bits)");
      break;
    case ST_OPTO_RX_THR0:
      sprintf(nam,"%s","Opto RX threshold0 (bits)");
      break;
    case ST_OPTO_RX_THR1:
      sprintf(nam,"%s","Opto RX threshold1 (bits)");
      break;
    case ST_OPTO_TX_DAC:
      sprintf(nam,"%s","Opto TX DAC (bits)");
      break;
    case ST_OPTO_TX_MKSPNEG:
      sprintf(nam,"%s","Opto TX MKSPNEG (bits)");
      break;
    case ST_OPTO_TX_MKSPPOS:
      sprintf(nam,"%s","Opto TX MKSPPOS (bits)");
      break;
    case ST_OPTO_TX_CDELAY:
      sprintf(nam,"%s","Opto TX coarse Delay (bits)");
      break;
    case ST_OPTO_TX_FDELAY:
      sprintf(nam,"%s","Opto TX fine Delay (bits)");
      break;
    case ST_FWD_SC_MODE:
      sprintf(nam,"%s","FWD SC MODE (0-4)");
      break;
    case ST_BPM_DRIVE:
      sprintf(nam,"%s","FWD SC BPM DRIVE (0-3)");
      break;
    case ST_VDC_DAC:
      sprintf(nam,"%s","FWD SC VDC DAC (bits)");
      break;
    case ST_VDC_DAC0:
      sprintf(nam,"%s","FWD SC VDC DAC0 (bits)");
      break;
    case ST_VDC_DAC1:
      sprintf(nam,"%s","FWD SC VDC DAC1 (bits)");
      break;
    
    default:
      sprintf(nam,"%s","UKNOWN");
      break;
  }
}

void TST::ConfigureScan(short new_type, float new_start, float new_stop, float new_step){
  // Set scan_type to new_type, scan range settings to coerced new scan parameters

  scan_type = new_type;
  if (scan_step<0) scan_step = (float) -1.0 * scan_step;  // code assumes positive scan_step
                                                          // whether scanning up or down
  // Where scanning SCTLV parameters, it is nonsensical
  // to coerce step size to minimum permissible setting
  // hence introduction of extra argument "is_step"
  // within st_coerce_variable PWP 11.03.00

  st_coerce_variable(new_type,new_start,&scan_start,0);
  st_coerce_variable(new_type,new_stop,&scan_stop,0);
  st_coerce_variable(new_type,new_step,&scan_step,1); 
  
  // calculate number of scan points, allowing for rounding errors in float arithmetic
  // Trim to value within bounds of scan histogram, then recalculate scan_stop

  if (scan_start>scan_stop){
    scan_max_points = 1 + abs((int)(((scan_start-scan_stop )+(scan_step/2))/scan_step));
    if (scan_max_points > ST_MAX_SCANPOINTS) scan_max_points = ST_MAX_SCANPOINTS;
    scan_stop = scan_start - ((scan_max_points-1) * scan_step);
  } else {
    scan_max_points = 1 + abs((int)(((scan_stop -scan_start)+(scan_step/2))/scan_step));
    if (scan_max_points > ST_MAX_SCANPOINTS) scan_max_points = ST_MAX_SCANPOINTS;
    scan_stop = scan_start + ((scan_max_points-1) * scan_step);
  }

  TranslateVariableName(scan_type, scan_name);

  sprintf(scan_description_string,"Scan of %s from %3.1f to %3.1f steps %3.1f", scan_name, scan_start, scan_stop, scan_step);
}


void TST::ConfigureScanDefaults(short new_type){
  // Set scan_type to new_type, and set scan settings to defaults for that type

  switch (new_type) {
    case ST_ELAPSED:
      ConfigureScan(new_type, 0.0, (float)(ST_MAX_SCANPOINTS-1), 1.0);
      break;
    case ST_VTHR:
      ConfigureScan(new_type, (float)50.0, (float)400.0,   (float)5.0);
      break;
    case ST_VCAL:
      ConfigureScan(new_type, (float)10.0, (float) 51.250, (float) 1.875);
      break;
    case ST_STROBE_DELAY:
      ConfigureScan(new_type, (float) 1.0, (float) 63.0,   (float) 2.0);
      break; 
    case ST_PREAMP:
      ConfigureScan(new_type, (float)46.0, (float)285.2,   (float) 9.2);
      break;
    case ST_SHAPER:
      ConfigureScan(new_type, (float) 7.2, (float) 37.2,   (float) 1.2);
      break;
    case ST_TRIM:
      ConfigureScan(new_type, (float) 0,   (float) 15,   (float) 1.0);
      break;
    case ST_QTHR:
    case ST_QCAL:
      ConfigureScan(new_type, (float) 0.0, (float) 3.0, (float) 0.1250);
      break;
    case ST_STREAM_DELAY:
      ConfigureScan(new_type, (float) 0.0, (float) 25.0,   (float) 1.0);
      break;
    case ST_STREAM_DELAY0:
      ConfigureScan(new_type, (float) 0.0, (float) 25.0,   (float) 1.0);
      break;
    case ST_STREAM_DELAY1:
      ConfigureScan(new_type, (float) 0.0, (float) 25.0,   (float) 1.0);
      break;
    case ST_CLOCK_DUTY:     // failure more likely at low settings - START HIGH!!
      ConfigureScan(new_type, (float) 15.0, (float) 0.0,   (float) 1.0);
      break;
    case ST_CLOCK_PHASE:
      // not yet implemented;
      break;
    case ST_COM_DELAY:
      ConfigureScan(new_type, (float) 10, (float) 34, (float)1);
      break;
    case ST_PULSE_DELAY:
    case ST_TRIG_DELAY:
      ConfigureScan(new_type, (float)120, (float)140, (float)1);
      break;
    case ST_VDET:
      // not yet implemented;
      break;
    case ST_VCC:
      ConfigureScan(new_type, (float)3.2, (float)3.8, (float)0.1);
      break;
    case ST_VI1:
      ConfigureScan(new_type, (float)0.0, (float)1.2, (float)0.1);
      break;
    case ST_VDD:
      ConfigureScan(new_type, (float)4.5, (float)3.5, (float)0.1);
      break;
    case ST_VPIN:
      // not yet implemented;
      break;
    case ST_VLED:
      ConfigureScan(new_type, (float)6.0, (float)2.0, (float)0.4); // range to be confirmed
      break;
    case ST_VLED0:
      ConfigureScan(new_type, (float)6.0, (float)2.0, (float)0.4); // range to be confirmed
      break;
    case ST_VLED1:
      ConfigureScan(new_type, (float)6.0, (float)2.0, (float)0.4); // range to be confirmed
      break;
    case ST_OPTO_RX_THR:
    case ST_OPTO_RX_THR0:
    case ST_OPTO_RX_THR1:
      ConfigureScan(new_type, (float)5.0, (float)255.0, (float)5.0); // range to be confirmed
      break;
    case ST_OPTO_TX_DAC:
      ConfigureScan(new_type, (float)10.0, (float)250.0, (float)10.0); // range to be confirmed
      break;
    default:  // use existing values of scan_type, scan_start, scan_stop, scan_step
      ConfigureScan(scan_type, scan_start, scan_stop, scan_step);
      break;
  }
}

void TST::InitScan(){  
  // Prepare for new scan
  // Revised PWP 10.05.01
 
  float half_step;
  float scan_hi, scan_lo;
  short n;

  scannum++;
  st_scan_increment(); // increment scan counter in stlib to match scannum
  burst_count = 0;
  scan_count++;
  scan_running = 1;

  // Assume that coersion of scan limits and calculation of scan_max_points
  // have been done by ConfigureScan.  Here we check only that scan will
  // remain within the bounds of the scan histogram, for reasons of safety.
  
  if (scan_max_points > ST_MAX_SCANPOINTS) scan_max_points = ST_MAX_SCANPOINTS;

  scan_point = scan_start;
  scan_current = scan_start;
  sprintf(scan_status_string,"%s", scan_description_string);
  printf("%s\n", scan_status_string);
  printf("%s   %d\n", scan_name, scan_max_points);


  // Now reset and rebin the scan histograms...

  half_step = scan_step/2;
  if(half_step<0) half_step = half_step*-1;

  if (scan_stop > scan_start){ 
    scan_lo = scan_start;
    scan_hi = scan_stop;
  }else{
    scan_lo = scan_stop;
    scan_hi = scan_start;
  }

  // Global histograms
  h_scan_tsent->Reset();
  h_scan_tsent->SetBins(scan_max_points, scan_lo-half_step, scan_hi+half_step);

  // MuSTARD histograms
  for(n=0;n<MUS_NUMBER;n++){
    h_scan_evcnt[n]->Reset();
    h_scan_evcnt[n]->SetBins(scan_max_points, scan_lo-half_step, scan_hi+half_step);
    h_scan_evcnt[n]->SetXTitle(scan_name);

    h_scan_ercnt[n]->Reset();
    h_scan_ercnt[n]->SetBins(scan_max_points, scan_lo-half_step, scan_hi+half_step);
    h_scan_ercnt[n]->SetXTitle(scan_name);

    h_scan_tocnt[n]->Reset();
    h_scan_tocnt[n]->SetBins(scan_max_points, scan_lo-half_step, scan_hi+half_step);
    h_scan_tocnt[n]->SetXTitle(scan_name);

    h_scan_xecnt[n]->Reset();
    h_scan_xecnt[n]->SetBins(scan_max_points, scan_lo-half_step, scan_hi+half_step);
    h_scan_xecnt[n]->SetXTitle(scan_name);
  }

  // Module histograms
  for(n=0;n<SCT_MAX_MODULES;n++){
    m[n]->InitScan(scan_max_points,scan_start,scan_stop,scan_step,scan_name);
  }


  // Finally, open ascii ntuple if requested
  if(ascii_nt){
    stout_open(runnum);
    if(scannum==1) stout_sor();
  }
}
  
  
void TST::ScanPoint(){

  // Execute burst(s) for current scan_point;  calling routine must repeat this call until finished
  // This version has PWP additions to allow increasing or decreasing scans - could be tidied up...

  short cal;

  if (!scan_running) {
    printf("ScanPoint: stopping because scan_running == 0 at point %d %3.0f\n", burst_count, scan_point); 
    return;
  } else if (burst_count >= ST_MAX_SCANPOINTS) {  // PWP bugfix 28.09.00
    printf("ScanPoint: stopping because burst_count %d > max_scanpoints %d\n", 
       burst_count, ST_MAX_SCANPOINTS); 
    return;
  } else if ((scan_point < scan_stop) && (scan_stop < scan_start)) {
    printf("ScanPoint: stopping because scan_point %3.0f < scan_stop %3.0f < scan_start %3.0f at point %d\n", 
       scan_point, scan_stop, scan_start, burst_count); 
    return;
  } else if ((scan_current < scan_stop) && (scan_stop < scan_start)){
    printf("ScanPoint: stopping because scan_current %3.0f < scan_stop %3.0f < scan_start %3.0f at point %d\n", 
       scan_current, scan_stop, scan_start, burst_count); 
    return;
  } else if ((scan_point > scan_stop) && (scan_stop > scan_start)) {
    printf("ScanPoint: stopping because scan_point %3.0f > scan_stop %3.0f > scan_start %3.0f at point %d\n", 
       scan_point, scan_stop, scan_start, burst_count); 
    return;
  } else if ((scan_current > scan_stop) && (scan_stop > scan_start)){
    printf("ScanPoint: stopping because scan_current %3.0f > scan_stop %3.0f > scan_start %3.0f at point %d\n", 
       scan_current, scan_stop, scan_start, burst_count); 
    return;
  }


  if (scan_type == ST_CLOCK_DUTY){ // scans where chip may hang up
    st_hard_reset();
  }
  
  st_coerce_variable(scan_type,scan_point,&scan_current,0);
  st_configure_variable(-1,scan_type,scan_current);

  if(scan_type==ST_VDET){
    HVRamp(0);
    cout << "Now wait for current to settle " << endl;
    Sleep(5000);
  }

  sprintf(scan_status_string,"%s at cnt %d  set %3.1f req %3.1f", 
    scan_description_string, burst_count, scan_current, scan_point);
  printf("%s\n", scan_status_string);

  // For scans involving the Liverpool forward SC
  if((scan_type==ST_FWD_SC_MODE)||(scan_type==ST_BPM_DRIVE)
     ||(scan_type==ST_VDC_DAC)||(scan_type==ST_VDC_DAC0) 
     ||(scan_type==ST_VDC_DAC1) ){
    st_execute_sc_configs(1,1);
  }

  if ((do_cal_loop)&&(! ((burst_type==ST_RAWBURST)||(burst_type==ST_SENDIDBURST)||
                         (burst_type==ST_EBURST)||(burst_type==ST_DEBURST)||
                         (burst_type==ST_SEUBURST)||(burst_type==ST_VBURST2)||(burst_type==ST_DVBURST)))
                   &&((burst_trtype==3)||(burst_trtype==33)||(burst_trtype==34)||
                      (burst_trtype==37)||(burst_trtype==43)||(burst_trtype==51))){
    for (cal=0; cal<4; cal++){
      st_configure_variable(-1,ST_CAL_MODE,cal);
      st_execute_configs(1,1);
      Sleep(2);
      ExecuteBurst(cal);
    }
  } else {
    st_execute_configs(1,1);
    Sleep(2);
    ExecuteBurst(-1);
  }

  // DCS functionality moved to ExecuteBurst
  // PWP 28.11.00 

  if (scan_stop > scan_start) {  // increasing scan
    scan_point = scan_point + scan_step;
    if ((scan_point > scan_stop) || (scan_current > scan_stop)||(burst_count > ST_MAX_SCANPOINTS)){
      scan_running = 0;
    }
  }
  else {                         // decreasing scan
    scan_point = scan_point - scan_step;
    if ((scan_point < scan_stop) || (scan_current < scan_stop)||(burst_count > ST_MAX_SCANPOINTS)){
      scan_running = 0;
    }
  }
  if(burst_type==112) scan_running=0;
}

void TST::EndScan(){
  short i,j;

  if (ascii_nt) stout_close();

  if (burst_type == ST_NSEBURST){
    for (i=0; i<SCT_MAX_MODULES; i++){ // fill hists only for present modules
      if(m[i]->present) m[i]->FillNormalisedScanHist();
    }
  } else if( ((burst_type!=ST_VBURST2) && (burst_type!=ST_DVBURST)) &&
      (((burst_type==ST_DBURST)&&(do_vary_ntrigs)&&(do_cal_loop==0))==0) ){
    for (i=0; i<SCT_MAX_MODULES; i++){
      if(m[i]->present) m[i]->FillScanHist();
    }
  }
  // If one has looped over the four cal lines, the event count histogram
  // will have been filled four times.  This histogram is used to normalise
  // the scan data in stan_fitscan, hence the factor of 4 is removed here.
  // PWP 24.01.99
  if ((do_cal_loop)&&(! ((burst_type==ST_RAWBURST)||(burst_type==ST_SENDIDBURST)||
                         (burst_type==ST_EBURST)||(burst_type==ST_DEBURST)||
                         (burst_type==ST_SEUBURST)||(burst_type==ST_VBURST2)||(burst_type==ST_DVBURST)))
                   &&((burst_trtype==3)||(burst_trtype==33)||(burst_trtype==34)||
                     (burst_trtype==37)||(burst_trtype==43)||(burst_trtype==51))){
    for(j=0;j<MUS_NUMBER;j++){
      h_scan_evcnt[j]->Scale(0.25);
    }
  }
  if((burst_type==ST_DBURST)&&(L1A_per_loop>1)&&(throw_away>0)){
    for(j=0;j<MUS_NUMBER;j++){
      h_scan_evcnt[j]->Scale((float)((float)L1A_per_loop-(float)throw_away)/(float)L1A_per_loop);
    }
  }
  burst_count = 0;
}


void TST::ShowScan(){
  printf("Current scan type is %d  from %f to %f stepping %f", scan_type, scan_start, scan_stop, scan_step);
  printf("Most recent scan point was %f at burst number %d\n", scan_current, burst_count);
}

void TST::RunFree(){
  st_run(1,20);
}

void TST::RunStop(){
  st_stop();
}

void TST::Single(short ttype){
  st_single(ttype);
}

void TST::SingleTrigger(){
  st_single(0);
}

void TST::SingleCalTrig(){
  st_single(3);
}

void TST::SingleSoftReset(){
  st_soft_reset();
}

void TST::SingleBCReset(){
  st_bc_reset();
}

void TST::HardReset(){
  st_hard_reset();
}

void TST::OptoTXOn(){
  st_optotx_on();
}

void TST::OptoTXOff(){
  st_optotx_off();
}

void TST::OptifStatus(int oid){
  optif_status(oid);
}

void TST::HVOff(){
  int i;
  st_hv_off();
  for (i=0; i<SCT_MAX_MODULES; i++) {
    if(m[i]->present){
      m[i]->DCS->HVStopwatch->Stop();
      m[i]->DCS->HVStopwatchRunning=kFALSE;
    }
  }
}

void TST::HVOn(){
  int i;
  st_hv_on();
  for (i=0; i<SCT_MAX_MODULES; i++) {
    if(m[i]->present){
      m[i]->DCS->HVStopwatch->Continue();
      m[i]->DCS->HVStopwatchRunning=kTRUE;
    }
  }
}

void TST::HVRamp(short to_zero){
  // if(to_zero==0) - ramp UP
  // if(to_zero==1) - ramp DOWN
  st_hv_ramp(to_zero);
}

void TST::HVRecoverTrips(){
  st_hv_trips(1);
}

void TST::HVStatus(){
  st_hv_status();
}

void TST::LVOff(){
  int i;
  st_lv_off();
  for (i=0; i<SCT_MAX_MODULES; i++) {
    if(m[i]->present){
      m[i]->DCS->LVStopwatch->Stop();
      m[i]->DCS->LVStopwatchRunning=kFALSE;
    }
  }
}

void TST::LVOn(){
  int i;
  st_lv_on();
  for (i=0; i<SCT_MAX_MODULES; i++) {
    if(m[i]->present){
      m[i]->DCS->LVStopwatch->Continue();
      m[i]->DCS->LVStopwatchRunning=kTRUE;
    }
  }
}

void TST::LVStatus(){
  st_lv_status();
}

float TST::LVGetIdd(short id) {
  if(m[id]->present) {
    m[id]->DCS->Query();
    return m[id]->DCS->idd;
  }
  return 0;
}

float TST::LVGetIcc(short id) {
  if(m[id]->present) {
    m[id]->DCS->Query();
    return m[id]->DCS->icc;
  }
  return 0;
}

TGraph* TST::LVGetIddGraph(short id) {
  TGraph* gr = NULL;

  if(m[id]->present) {
    gr = m[id]->GetIddGraph();
  }
  return gr;
}

TGraph* TST::LVGetIccGraph(short id) {
  TGraph* gr = NULL;

  if(m[id]->present) {
    gr = m[id]->GetIccGraph();
  }
  return gr;
}

void TST::FindTrips(){
  st_lv_trips(0);
}

void TST::RecoverTrips(){
  
  int ntrips = 0;
  int ntries = 0;

  while(ntries<10){
    printf("RecoverTrips - pass %d\n", ntries);
    if (ntrips=st_lv_trips(1)){
      st_execute_configs(0,1);
      ntries++;
    }
    else ntries = 10;
  }
  if(ntrips) printf("RecoverTrips: failed to recover, %d trips remaining\n",ntrips);
}

void TST::SysmapMenu(){
  st_menu_sysmap();
}

void TST::ShowSysmap(){
  st_show_sysmap();
}

void TST::ShowModule(int imod){

  if((imod<0)||(imod>=SCT_MAX_MODULES)){
    cout << "Selected module " << imod;
    cout << " is out of range 0 to " << (SCT_MAX_MODULES-1) << endl;
    return;
  }

  if(m[imod]->present){
    st_show_module(imod);
  } else {
    cout << "Selected module " << imod;
    cout << " is not present" << endl;
  }
}

void TST::ShowTrims(int imod){

  if((imod<0)||(imod>=SCT_MAX_MODULES)){
    cout << "Selected module " << imod;
    cout << " is out of range 0 to " << (SCT_MAX_MODULES-1) << endl;
    return;
  }

  if(m[imod]->present){
    st_show_trims(imod);
  } else {
    cout << "Selected module " << imod;
    cout << " is not present" << endl;
  }
}

void TST::ShowRC(int imod){

  if((imod<0)||(imod>=SCT_MAX_MODULES)){
    cout << "Selected module " << imod;
    cout << " is out of range 0 to " << (SCT_MAX_MODULES-1) << endl;
    return;
  }

  if(m[imod]->present){
    st_show_rc(imod);
  } else {
    cout << "Selected module " << imod;
    cout << " is not present" << endl;
  }
}

int TST::UpdateRunInfo(){

  // Read last run settings from an external file

  int run, rcode;
  FILE *f;
  char fname[128];

  sprintf(fname,"%sLAST_RUN.txt", sct_etc);

  printf("\nTST::UpdateRunInfo()\n\n");
  printf(" Attempting to open last info file %s for reading\n", fname);

  if ((f = fopen(fname,"r")) == NULL){   // Can't find runnum file
    printf(" Cannot open %s for reading\n", fname);
    printf(" Setting runnum to 1\n");
    run = 1;
    rcode = -1;
  } else {
    printf(" Last info file open for reading\n");
    fscanf(f,"%d",&run);
    printf(" Last runnum was   %d\n", run);
    run++;
    fclose(f);
    rcode = 1;
  }
    
  runnum = run;
  printf(" Current runnum is now %d\n", runnum);

  printf(" Attempting to reopen last info file %s for updating\n", fname);

  if ((f = fopen(fname,"w")) == NULL){
    printf(" ERROR: Cannot open %s for writing\n", fname);
    printf(" ERROR: Run number will not increment !!!");
    runnum = 1;
    rcode = -2;
  } else {
    printf(" Last info file open for writing");
    fprintf(f,"%d",runnum);
    fclose(f);
    printf(" ...done\n");
    rcode = 0;
  }

  printf("\n");
  return(rcode);
}

void TST::StOutScanReset(){
  st_scan_reset();
}

int TST::GetMaxModules(){
  return SCT_MAX_MODULES;
}

void TST::GetModuleName(int imod, char *nam){
  st_give_modname(imod, nam);
}

void TST::GetDeviceType(int imod, char *nam){
  st_give_devtype(imod, nam);
}

void TST::ReadMaskFile(int imod){
  st_read_mask(imod);
}

void TST::Sleep(int milliseconds){
  st_sleep(milliseconds);
}

void TST::AddHistograms(short cal_mode){

  // Add the MuSTARD Histograms to histograms in memory
  // PWP 27.04.01

  short hdata[1024];     // We read only 768 channels at a time
  short* hpointer;
  short hlength;

  short soft_link;
  short stream;
  
  short start,step;
  short mid;
  int n,i;

  if ((cal_mode<0)||(cal_mode>3)) {
    start=0;
    step=1;
  } else {
    start=cal_mode; // if looping over all 4 cal lines,
    step=4;         // fill ROOT histogram only for strobed line
  }

  for (n=0; n<SCT_MAX_MODULES; n++){
    if(m[n]->present){

      // Module is present - read the Mustard histograms.

      mid = st_find_mustard(n);

      for(soft_link=0;soft_link<2;soft_link++){
        stream = st_find_stream(n,soft_link);

        hpointer = hdata;
        mus_read_histogram(mid,soft_link,stream,&hlength,&hpointer);

        if (hlength>768) hlength = 768;

        for (i=start; i<hlength; i+=step){
          // scan is type short.
          // will overflow at 2^15, then reset at 2^16.
          // allow it to overflow, but do not then fill further.
          // PWP (temporary?) bugfix 18.03.02
          if(m[n]->scan[burst_count][soft_link][i]>=0){
            m[n]->scan[burst_count][soft_link][i] += hdata[i];
          }
        } 
      }
    } 
  }
}

void TST::ReadHistograms(short cal_mode){

  // Read the MuSTARD Histograms for each module
  // PWP 27.04.01

  short hdata[1024];     // We read only 768 channels at a time
  short* hpointer;
  short hlength;

  short soft_link;
  short stream;
  
  short start,step;
  short mid;
  int n,i;

  if ((cal_mode<0)||(cal_mode>3)) {
    start=0;
    step=1;
  } else {
    start=cal_mode; // if looping over all 4 cal lines,
    step=4;         // fill ROOT histogram only for strobed line
  }

  for (n=0; n<SCT_MAX_MODULES; n++){
    if(m[n]->present){

      // Module is present - read the Mustard histograms.

      mid = st_find_mustard(n);

      for(soft_link=0;soft_link<2;soft_link++){
        stream = st_find_stream(n,soft_link);

        hpointer = hdata;
        mus_read_histogram(mid,soft_link,stream,&hlength,&hpointer);

        if (hlength>768) hlength = 768;

        for (i=start; i<hlength; i+=step){
          m[n]->scan[burst_count][soft_link][i] = hdata[i];
        } 
      }
    } 
  }
}

int TST::ReadEvent(){

  // Read MUS_NUMBER MuSTARDs
  //   check for TIMEOUT and CONTROL errors
  //   store data in evdata[MUS_NUMBER][0x8000]
  // PWP 15.03.00


  short rcode,status;
  short *evpointer;
  short evlength;

  short mid;
  int   n_errs = 0;

  short stream;
  short module;
  short hard_link;
  short error_count;
  int   event_count;

  for(mid=0; mid<MUS_NUMBER; mid++) {
    if(st_used_streams(mid)>0){
      evpointer = evdata[mid];
      rcode = mus_read_event(mid,&evlength,&evpointer);

      error_count =  evdata[mid][1];
      event_count = (evdata[mid][3]<<16) + (unsigned short) evdata[mid][2] +1; // PWP 19.03.02

      if (debug_level>8) {
        printf("ReadEvent(mid %d): burst %d tsent %d evcnt %d got evlength %d\n", 
                mid,burst_count,tsent,event_count,evlength);
      }

      if (rcode != 0) {
        mus_read_status(mid,&status);
        if (debug_level>5){
          printf("ReadEvent: ERROR %d from Mustard %d\n", rcode,mid);
          mus_print_status(status);
        }
        if (rcode == -4) {

          // TIMEOUT: stream can be determined from MuSTARD status register
          tocnt_sum++;
          tocnt[mid]++;

          evlength = 0;
          evdata[mid][0]=0;  // such that DecodeEvent does not try....

          stream = status & 0x00f;
          if (stream >= 12) break;
          module = st_find_module(mid,stream);
          hard_link = st_find_link(mid,stream); 

          if(module>-1)
            m[module]->FillTocnt(hard_link);
          else if (debug_level>5){
            printf("ReadEvent: st_find_module returned invalid module for MuSTARD %d stream %d\n",mid,stream);
          }
        } else {
          // MuSTARD control error
          xecnt_sum++;
          xecnt[mid]++;
        }
      }

      if((rcode != -4)&&(evlength==0)&&(debug_level>5)){
          printf("ReadEvent: zero length event for MuSTARD %d\n",mid);
      }

      // Any stream errors this event? 

      n_errs+= (error_count-ercnt[mid]);  // PWP bugfix 12.04.00


      // Update counters and fill histograms

      ercnt[mid]=error_count;
      evcnt[mid]=event_count;

      h_evlen[mid]->Fill((float)evlength);   // Fill even if zero
    }
  } // end for(mid)

  return(n_errs);
}


int TST::TestEvent(){

  // Read MUS_NUMBER MuSTARDs
  //   check for TIMEOUT and CONTROL errors
  //   if there are any, store data in evdata[MUS_NUMBER][0x8000]
  // PWP 16.03.00


  short rcode,status;
  short *evpointer;
  short evlength;

  short mid;
  int   n_errs = 0;

  short stream;
  short module;
  short hard_link;
  short error_count;
  int   event_count;

  for(mid=0; mid<MUS_NUMBER; mid++) {
    if(st_used_streams(mid)>0){
      evpointer = evdata[mid];
      error_count = ercnt[mid]; // PWP bugfix 16.08.00
      rcode = mus_test_event(mid,&evlength,&evpointer,&event_count,&error_count,0);

      event_count++;

      if (debug_level > 8) {
        printf("TestEvent: burst %d tsent %d MuSTARD %d gives evcnt %d evlength %d\n", 
                burst_count,tsent,mid,event_count,evlength);
      }

      if (rcode != 0) {
        mus_read_status(mid,&status);
        if (debug_level>5){
           // revised PWP 19.07.00 to show burst and trigger counters
          printf("TestEvent: ERROR %d from Mustard %d burst %d tsent %d\n",
                  rcode,mid,burst_count,tsent);
          mus_print_status(status);
        }
          if (rcode == -5) {
            if (debug_level>5){
              printf("TestEvent: STREAM ERROR detected by Mustard %d\n", mid);
            }
            n_errs++;

          } else if (rcode == -4) {

          // TIMEOUT: stream can be determined from MuSTARD status register
          tocnt_sum++;
          tocnt[mid]++;

          evlength = 0;
          evdata[mid][0]=0;  // such that DecodeEvent does not try....

          stream = status & 0x00f;
          if (stream >= 12) break;
          module = st_find_module(mid,stream);
          hard_link = st_find_link(mid,stream); 

          if(module>-1)
            m[module]->FillTocnt(hard_link);
          else if (debug_level>5) {
            printf("TestEvent: st_find_module returned invalid module for MuSTARD %d stream %d\n",mid,stream);
          }

        } else {
          // MuSTARD control error
          xecnt_sum++;
          xecnt[mid]++;
        }
      } else {
        // No error this MuSTARD - nothing for DecodeEvent to do!!
        evdata[mid][0]=0;
      }

      if((rcode != -4)&&(evlength==0)&&(debug_level>5)){
          printf("TestEvent: zero length event for MuSTARD %d\n",mid);
      }

      // Update counters and fill histograms

      ercnt[mid]=error_count;
      evcnt[mid]=event_count;

      h_evlen[mid]->Fill((float)evlength);   // Fill even if zero
    }
  } // end for(mid)

  return(n_errs);
}


void TST::DecodeEvent(short cal_mode){

  // Assume MUS_NUMBER MuSTARDS have been read, their data is stored in
  // data[MUS_NUMBER].  This routine will decode events, associating a 
  // number of error conditions with the causal module/link.
  // PWP 15.03.00

  short mid,n;
  short* data;

  short word_count;

  short chipid;
  short module;
  short stream;
  short hard_link;
  short soft_link;
  short L1_count;
  short BC_count;
  short strip_address;
  short cluster_size;
  short bc[SCT_MAX_MODULES][2];
  short l1[SCT_MAX_MODULES][2];
  short multiplicity[SCT_MAX_MODULES][2];
  short allow_continued_hits;

  short most_entries;
  short median_bc=0;
  short median_l1=0;

  unsigned char bc_hist[256];
  unsigned char l1_hist[16];

  for(n=0;n<SCT_MAX_MODULES;n++){
    bc[n][0]=0;
    l1[n][0]=0;
    multiplicity[n][0]=0;
    multiplicity[n][1]=0;
  }

  for(n=0;n<256;n++){
    bc_hist[n]=0;
  }

  for(n=0;n<16;n++){
    l1_hist[n]=0;
  }

  if(do_fill_evtree){  // reset tree arrays
    for(module=0;module<SCT_MAX_MODULES;module++){
      if(m[module]->present){ // fill ntps only for present modules
        m[module]->ResetEvent();
      }
    }
  }

  for(mid=0; mid<MUS_NUMBER; mid++) {
    if(st_used_streams(mid)>0){

      data = evdata[mid];
      word_count  =  data[0];

      if (word_count > 0){   // Decode the event data
                           // ### ignore timebins for the moment
        stream = 0;
        module = 0;
        hard_link = 0;
        soft_link = 0;
        strip_address=0;
        cluster_size=0;
        allow_continued_hits=0;

        for (n=4; n<word_count; n++){

          // First test must be for presence of the error bit.
          // The error bit can be set together with other bits - 
          // such words should not be interpreted other than as errors!!!

          if (data[n]&0x4000) {  // ERROR BIT SET
 
            allow_continued_hits = 0; 

            if (debug_level>6){
              printf("DecodeEvent: trig %d mid %d sid %d err at short %d val %4hx\n",tsent,mid,stream,n,data[n]);
            }

            // Error bit will remain set until MuSTARD "finds its feet"
            // => log only each transition of error bit from 0 to 1

            if((n==4)||((data[n-1]&0x4000)==0)){  
              if (stream >= 12) break;
              module = st_find_module(mid,stream);
              hard_link = st_find_link(mid,stream);

              if(module>-1)  // valid module
                m[module]->FillErcnt(hard_link);
              else if (debug_level>6){
                printf("DecodeEvent: trig %d mid %d sid %d st_find_module returns invalid module\n",tsent,mid,stream);
              }
            }

          } else if (data[n]&0x1000) { // HIT
            // HIT must follow either HIT or NEW CLUSTER
            // anything else is OUT OF SEQUENCE

            if(allow_continued_hits){  // in sequence hit

              if(module>-1){  // valid module
                // Fill strip data histograms with respect to cal_mode
                if ((cal_mode==-1)||(strip_address%4)==cal_mode) {

                  if (strip_address<768) {
                    // Prevent overflowed counters from "looping the loop"
                    // PWP bugfix 18.03.02
                    if(m[module]->scan[burst_count][soft_link][strip_address]>=0){
                      m[module]->scan[burst_count][soft_link][strip_address]++;
                    }
                    if(do_fill_evtree & 1){
                      if(soft_link){
                        m[module]->evtr_strm1[strip_address]=kTRUE;
                      }else{
                        m[module]->evtr_strm0[strip_address]=kTRUE;
                      }
                    }
                    if(do_fill_evtree & 2){
                      chipid = strip_address/128;
                      if(chipid<0 || chipid>5) chipid=6;
                      if(soft_link){
                        (m[module]->hits_strm1[chipid])++;
                      }else{
                        (m[module]->hits_strm0[chipid])++;
                      }
                    }
                  } else {
                    if (debug_level>6){
                      printf("DecodeEvent: trig %d mid %d sid %d non-existent strip %d (cluster %d)\n",tsent,mid,stream,strip_address,cluster_size); 
                    }
                    if (stream >= 12) break;
                    module = st_find_module(mid,stream);
                    hard_link = st_find_link(mid,stream);

                    if(module>-1)
                      m[module]->FillNescnt(hard_link);
                  }
                }
                // Always increment cluster_size and multiplicity
                cluster_size++;
                multiplicity[module][soft_link]++;
                strip_address++;
              }
            } else { // out of sequence hit
              if (debug_level>6){
                printf("DecodeEvent: trig %d mid %d sid %d OSH at short %d val %4hx\n",tsent,mid,stream,n,data[n]);
              }  
              if (stream >= 12) break;
              module = st_find_module(mid,stream);
              hard_link = st_find_link(mid,stream);
              if(module>-1)
                m[module]->FillOshcnt(hard_link);
            } 

          } else if (data[n]&0x2000) {  // NEW CLUSTER

            // Fill histogram with previous cluster and reset
            if ((cluster_size>0)&&(module>-1)){
              m[module]->FillClusterHist(soft_link,cluster_size);
              cluster_size=0;
            }

            // For the time being, we have two scan histograms, one per "link"
            // in practice, "link 0" handles strips    0 to  768
            //              "link 1" handles strips 1024 to 1792
            // We may come to regret having taken this approach....

            strip_address =  data[n] & 0x03ff;
            soft_link     = (data[n] & 0x0400)>>10;

            allow_continued_hits = 1; 

          } else if (data[n]&0x8000) {  // NEW STREAM
//          } else if (data[n]&0x8000 ||    // NEW STREAM  --  temporary fix whilst d15 is not set in
//                    ((data[n]&0x000f)==(data[n+1]&0x000f))) { // stream event header - JGK 11.07.01

            allow_continued_hits = 0; 

            // Fill histogram with previous cluster and reset
            if ((cluster_size>0)&&(module>-1)){
              m[module]->FillClusterHist(soft_link,cluster_size);
              cluster_size=0;
            }

            if (data[n+1]&0x4000) {  // ERROR IN STREAM ID
              // we no longer know which stream has caused the error
              // How do we tally this condition??
              if (debug_level>6){
                printf("DecodeEvent: trig %d mid %d Stream Assignment error \n",tsent,mid);
                printf("              short n=%d data(n)=%4x data(n+1)=%4x\n",n, data[n],data[n+1]);
              }
            }else{

              // TODO:  Histogram, or validity check, of BC and L1 data
              BC_count =  data[n] & 0x00ff;
              L1_count = (data[n] & 0x0f00) >> 8;
              stream   =  data[n+1]&0x000f; 

              if (stream >= 12) break;
              module = st_find_module(mid,stream);
              hard_link = st_find_link(mid,stream);
              if(module>-1){
                bc_hist[BC_count]++;
                l1_hist[L1_count]++;
                bc[module][hard_link]=BC_count;
                l1[module][hard_link]=L1_count;
                m[module]->FillEvcnt(hard_link);
              }
              else if (debug_level>6){
                printf("DecodeEvent: trig %d mid %d sid %d st_find_module returns invalid module\n",tsent,mid,stream);
                printf("              short n=%d data(n)=%4x data(n+1)=%4x\n",n, data[n],data[n+1]);
              }
            }
          } // end new stream
        } // end for(n)
      } // end if(word_count>0)
    }
  } // end for (mid)


  // Find most common BC and L1 values
  most_entries=0;
  for(n=0;n<256;n++){
    if(bc_hist[n]>most_entries){
      most_entries=(short)bc_hist[n];
      median_bc=n;
    }
  }

  most_entries=0;
  for(n=0;n<16;n++){
    if(l1_hist[n]>most_entries){
      most_entries=(short)l1_hist[n];
      median_l1=n;
    }
  }

  for(n=0;n<SCT_MAX_MODULES;n++){
    // End of event:
    // Fill any multiplicity / correlation histograms here
    if(m[n]->present){
       m[n]->FillCorrHist(multiplicity[n][0],multiplicity[n][1]);

       // BC and L1 counter checking
       if(bc[n][0]!=median_bc){
         m[n]->FillBcErcnt(0);
       }
       if(bc[n][1]!=median_bc){
         m[n]->FillBcErcnt(1);
       }
       if(l1[n][0]!=median_l1){
         m[n]->FillL1Ercnt(0);
       }
       if(l1[n][1]!=median_l1){
         m[n]->FillL1Ercnt(1);
       }
    }
  }

  if(do_fill_evtree){  // Fill Evtree
    for(module=0;module<SCT_MAX_MODULES;module++){
      if(m[module]->present){ // fill ntps only for present modules
        m[module]->FillTree();
      }
    }
  }

}

void TST::DumpEvent(FILE* dumpfile){

  short mid,n;
  short* data;
  short word_count;
  short error_count;
  int event_count;

  short strip = 0;
  short L1_count;
  short BC_count;

  if(dumpfile==NULL){

    printf("DumpEvent: ERROR: dumpfile is not open\n");
    return;
  }

  for(mid=0; mid<MUS_NUMBER; mid++) {
    if(st_used_streams(mid)>0){

      data = evdata[mid];

      word_count  = data[0];
      error_count = data[1];
      event_count =(data[3]<<16) + (unsigned short) data[2]; // PWP 19.03.02

      // Write data to file always - even if zero length
  
      fprintf(dumpfile,"%d  %4x  WDCNT %d\n", mid,(unsigned short)data[0],word_count);
      fprintf(dumpfile,"%d  %4x  ERCNT %d\n", mid,(unsigned short)data[1],error_count);
      fprintf(dumpfile,"%d  %4x\n",           mid,(unsigned short)data[2]);
      fprintf(dumpfile,"%d  %4x  EVCNT %d\n", mid,(unsigned short)data[3],event_count);

      if(debug_level==998){
        // dump counter info only
        for (n=4; n<word_count; n++){
          if (data[n]&0x8000) {
            L1_count = (data[n] & 0x0f00) >> 8;
            BC_count = data[n] & 0x00ff;
            fprintf(dumpfile,"%d  %4x  ", mid,(unsigned short)data[n]);
            fprintf(dumpfile,"  STR HDR  L1 %1x  BC %2x\n", L1_count, BC_count);
            n++;
            fprintf(dumpfile,"%d  %4x    STR NO.\n", mid, (unsigned short)data[n]);
          }
        }
      }else{
        for (n=4; n<word_count; n++){
        // dump everything
          fprintf(dumpfile,"%d  %4x  ", mid,(unsigned short)data[n]);

          if (data[n]&0x4000) {
            fprintf(dumpfile,"    ERR\n");
          }
          if (data[n]&0x2000) { 
            strip = data[n]&0x07ff;
            fprintf(dumpfile,"\n");
          }
          if((data[n]&0xf000)==0x1000) { // PWP bugfix 26.04.01
            fprintf(dumpfile,"    %3d\n", strip);
            strip++;
          }
          if (data[n]&0x8000) {
            L1_count = (data[n] & 0x0f00) >> 8;
            BC_count = data[n] & 0x00ff;
            fprintf(dumpfile,"  STR HDR  L1 %1x  BC %2x\n", L1_count, BC_count);
            n++;
            fprintf(dumpfile,"%d  %4x    STR NO.\n", mid, (unsigned short)data[n]);
          }
        }
      }
    }
  }
}


void TST::DecodeEvent3bins(short cal_mode){

  // Assume MUS_NUMBER MuSTARDS have been read, their data is stored in
  // data[MUS_NUMBER].  This routine will decode events, associating a 
  // number of error conditions with the causal module/link.
  // This version, DecodeEvent3bins, fills
  //    burst_count    with timebin t-1
  //    burst_count+1  with timebin t
  //    burst_count+2  with timebin t+1
  // As such, this is a routine for diagnostic purposes only.
  // PWP 16.03.00

  short mid,n,t;
  short* data;

  short word_count;

  short module;
  short stream;
  short hard_link;
  short soft_link;
  short L1_count;
  short BC_count;
  short strip_address;
  short cluster_size;
  short multiplicity[SCT_MAX_MODULES][2][3];
  short allow_continued_hits;

  for(n=0;n<SCT_MAX_MODULES;n++){
    for(t=0;t<3;t++){
      multiplicity[n][0][t]=0;
      multiplicity[n][1][t]=0;
    }
  }

  for(mid=0; mid<MUS_NUMBER; mid++) {
    if(st_used_streams(mid)>0){

      data = evdata[mid];
      word_count  =  data[0];

      if (word_count > 0){   // Decode the event data
                           // ### ignore timebins for the moment
        stream = 0;
        module = 0;
        hard_link = 0;
        soft_link = 0;
        strip_address=0;
        cluster_size=0;
        allow_continued_hits=0;

        for (n=4; n<word_count; n++){

          // First test must be for presence of the error bit.
          // The error bit can be set together with other bits - 
          // such words should not be interpreted other than as errors!!!

          if (data[n]&0x4000) {  // ERROR BIT SET
 
            allow_continued_hits = 0; 

            printf("DecodeEvent: trig %d mid %d sid %d err at short %d val %4hx\n",tsent,mid,stream,n,data[n]);

            // Error bit will remain set until MuSTARD "finds its feet"
            // => log only each transition of error bit from 0 to 1

            if((n==4)||((data[n-1]&0x4000)==0)){  
              if (stream >= 12) break;
              module = st_find_module(mid,stream);
              hard_link = st_find_link(mid,stream);

              if(module>-1)  // valid module
                m[module]->FillErcnt(hard_link);
              else 
                printf("DecodeEvent: trig %d mid %d sid %d st_find_module returns invalid module\n",tsent,mid,stream);
            }

          } else if (data[n]&0x1000) { // HIT
            // HIT must follow either HIT or NEW CLUSTER
            // anything else is OUT OF SEQUENCE

            if(allow_continued_hits){  // in sequence hit

              if(module>-1){  // valid module
                // Fill strip data histograms with respect to cal_mode
                if ((cal_mode==-1)||(strip_address%4)==cal_mode) {

                  if (strip_address<768) {
                    if (data[n]&0x0004) {
                      m[module]->scan[burst_count][soft_link][strip_address]++;
                      multiplicity[module][soft_link][0]++;
                    }
                    if (data[n]&0x0002) {
                      m[module]->scan[burst_count+1][soft_link][strip_address]++;
                      multiplicity[module][soft_link][1]++;
                    }
                    if (data[n]&0x0001) {
                      m[module]->scan[burst_count+2][soft_link][strip_address]++;
                      multiplicity[module][soft_link][2]++;
                    }
                  } else {
                    printf("DecodeEvent: trig %d mid %d sid %d non-existent strip %d (cluster %d)\n",tsent,mid,stream,strip_address,cluster_size); 
                    if (stream >= 12) break;
                    module = st_find_module(mid,stream);
                    hard_link = st_find_link(mid,stream);

                    if(module>-1)
                      m[module]->FillNescnt(hard_link);
                  }
                }
                // Always increment cluster_size
                cluster_size++;
                strip_address++;
              }
            } else { // out of sequence hit
              printf("DecodeEvent: trig %d mid %d sid %d OSH at short %d val %4hx\n",tsent,mid,stream,n,data[n]);
            
              if (stream >= 12) break;
              module = st_find_module(mid,stream);
              hard_link = st_find_link(mid,stream);
              if(module>-1)
                m[module]->FillOshcnt(hard_link);
            } 

          } else if (data[n]&0x2000) {  // NEW CLUSTER

            // Fill histogram with previous cluster and reset
            if ((cluster_size>0)&&(module>-1)){
              m[module]->FillClusterHist(soft_link,cluster_size);
              cluster_size=0;
            }

            // For the time being, we have two scan histograms, one per "link"
            // in practice, "link 0" handles strips    0 to  768
            //              "link 1" handles strips 1024 to 1792
            // We may come to regret having taken this approach....

            strip_address =  data[n] & 0x03ff;
            soft_link     = (data[n] & 0x0400)>>10;

            allow_continued_hits = 1; 

          } else if (data[n]&0x8000) {  // NEW STREAM

            allow_continued_hits = 0; 

            // Fill histogram with previous cluster and reset
            if ((cluster_size>0)&&(module>-1)){
              m[module]->FillClusterHist(soft_link,cluster_size);
              cluster_size=0;
            }

            if (data[n+1]&0x4000) {  // ERROR IN STREAM ID
              // we no longer know which stream has caused the error
              // How do we tally this condition??
              printf("DecodeEvent: trig %d mid %d Stream Assignment error \n",tsent,mid);
              printf("              short n=%d data(n)=%4x data(n+1)=%4x\n",n, data[n],data[n+1]);

            }else{

              // TODO:  Histogram, or validity check, of BC and L1 data
              BC_count =  data[n] & 0x00ff;
              L1_count = (data[n] & 0x0f00) >> 8;
              stream   =  data[n+1]&0x000f; 

              if (stream >= 12) break;
              module = st_find_module(mid,stream);
              hard_link = st_find_link(mid,stream);
              if(module>-1)
                m[module]->FillEvcnt(hard_link);
              else{ 
                printf("DecodeEvent: trig %d mid %d sid %d st_find_module returns invalid module\n",tsent,mid,stream);
                printf("              short n=%d data(n)=%4x data(n+1)=%4x\n",n, data[n],data[n+1]);
              }
            }
          } // end new stream
        } // end for(n)
      } // end if(word_count>0)
    }
  } // end for (mid)

  for(n=0;n<SCT_MAX_MODULES;n++){
    // End of event:
    // Fill any multiplicity / correlation histograms here
    if(m[n]->present){
      m[n]->FillCorrHist(multiplicity[n][0][0],multiplicity[n][1][0]);
      m[n]->FillCorrHist(multiplicity[n][0][1],multiplicity[n][1][1]);
      m[n]->FillCorrHist(multiplicity[n][0][2],multiplicity[n][1][2]);
    }
  }
}


// Following routines for writing binary event data to files,
// using the TB convention (int data, 4-word headers).  
// Data is accumulated into a buffer before writing under the
// assumption of a burst model of data flow.


int TST::AppendEvent(){  
    
  // Append current event data to output buffer using TB conventions
  // return total ints appended (positive) or error (negative)

  short mid;
  int n,nwords,nints;
  int size;
  int firstpnt;
  int startpnt;  
  int sizepnt;  
  char sid[8];
  int *pnt;

  if (bufpnt >= (ST_BUF_LENGTH - 0x1000)) {
    printf("AppendEvent: ERROR: Buffer pointer %d > buffer length %d\n",
      bufpnt, ST_BUF_LENGTH);
    return -1;
  }

  firstpnt = bufpnt;

  for (mid=0; mid<MUS_NUMBER; mid++){

    startpnt = bufpnt;

    buf[bufpnt++] = 0xAAAAAAAA;     

    sprintf(sid,"MUS%1x", mid);
    pnt = (int *)sid;
    buf[bufpnt++] = *pnt;

    sizepnt = bufpnt;
    buf[bufpnt++] = sizepnt - startpnt;

    buf[bufpnt++] = bufevt;

    /* Mustard has a short buffer */
    nwords = evdata[mid][0];
    nints = (nwords + 1)/2;

    pnt = (int*) &evdata[mid][0];
    for (n=0; n<nints; n++) buf[bufpnt++] = *(pnt+n);
 
    size = bufpnt - startpnt;
    buf[sizepnt] = size;
  }

  if (bufpnt > (ST_BUF_LENGTH - 0x1000)){
    printf("AppendEvent: WARNING: Buffer almost full %d\n", bufpnt);
  }

  bufevt++;  // ### This is in place of some other system-wide unique event counter

  return (bufpnt - firstpnt);
}

int TST::AppendArray(char *id, int length, int count, int *data){

  // Append arbitrary int array to output buffer
  // return total ints appended (positive) or error (negative)

  int n;
  int size;
  int startpnt;  
  int sizepnt;
  int *pnt;

  if ((bufpnt+length) > ST_BUF_LENGTH) {
    printf("AppendArray: ERROR: Buffer overflow\n");
    return -1;
  }

  startpnt = bufpnt;

  buf[bufpnt++] = 0xAAAAAAAA;   // Universal fragment marker

  pnt = (int *)id;              // Fragment identifier
  buf[bufpnt++] = *pnt;

  sizepnt = bufpnt;
  buf[bufpnt++] = sizepnt - startpnt;

  buf[bufpnt++] = count;        // An appropriate counter, unique for this fragment type

  for (n=0; n<length; n++) buf[bufpnt++] = data[n];
    
  size = bufpnt - startpnt;
  buf[sizepnt] = size;

  if (bufpnt > (ST_BUF_LENGTH - 0x1000)){
    printf("AppendEvent: WARNING: Buffer almost full %d\n", bufpnt);
  }

  return (bufpnt - startpnt);
}


int TST::WriteBuffer(){

  // Append recording buffer to binary file "stdat[run]_[scan].dat,
  // creating it if necessary
  // GFM 20000329

  FILE *f;
  char fname[128];
  int nints;
  
  sprintf(fname,"%sstdat%d_%d.dat",sct_data,runnum,scannum);

//  printf("WriteBuffer: Attempting to open datafile %s\n", fname);
  fflush(stdout);

  if ((f = fopen(fname,"ab")) == NULL) {
    printf("WriteBuffer: ERROR: Unable to open data file %s for appending\n", fname);
    fflush(stdout);
    return(-1);
  }
// printf("WriteBuffer: Opened data file %s for appending\n",fname);
  fflush(stdout);

  nints = fwrite(buf, sizeof(buf[0]), bufpnt, f);
  fclose(f);

  printf("WriteBuffer: Written %d ints of %d\n",nints,bufpnt);

  bufpnt = 0;  // Reset buffer pointer;
  buf[0] = 0;

  return nints;
}


// Single Event Upset Tests
//   - This function uses a spare "busy" input of CLOAC
//     to monitor the state of the PS spill cycle.  
//     Make sure that an ECL active-HIGH signal corresponding 
//     to SPILL is present at CLOAC fanout-D BUSY-IN (pins 1,2);


int TST::SEUTest(){

  return SEUTest(1);
}


int TST::SEUTest(int hard_time){    // Single-Event Upset study with 
                                    // SOFTWARE or HARDWARE timing
   
  int anomaly = 0;
  int seu_header[10];
  char seu_id[4];
  int seu_event, seu_index;
  int timeout_count;
  short mid, rcode;
  unsigned short spill_state;
  unsigned short cloac_busy_state;
  int date, time;
  short event,link,chip;
  long ldelay, t_zero;
  long t_spill =0;
  int mflip2_nerrs = 0;

  if(hard_time) printf("\n\n\nSEUTest: HARDWARE TIMER");  // PS spill cycle
  else          printf("\n\n\nSEUTest: SOFTWARE TIMER");  // for continuous beams or control measurements

  if(seu_invert)printf(" polarity 1->0\n");
  else          printf(" polarity 0->1\n");

  for (link=0; link<SEU_NLINKS; link++){
    for (event=0; event<SEU_NEVENTS; event++){   // Clear SEU dedicated event arrays
      seu_wc[event][link] = 0;
      seu_l1[event][link] = 0; 
      seu_bc[event][link] = 0;

      for(chip=0;chip<6;chip++) seu_flip[event][link][chip] = 0;
      seu_l1err[event][link] = 0;
      seu_wcerr[event][link] = 0;

      seu_data[event][link][0] = 0;
      seu_data[event][link][1] = 0;
      seu_data[event][link][2] = 0;
      seu_data[event][link][3] = 0;
      seu_data[event][link][4] = 0;
    }
    for(chip=0;chip<6;chip++){
      seu_conf[link][chip] = 0;
      seu_mask[link][chip] = 0;
      seu_pipe[link][chip] = 0;
      seu_mystery[link][chip] = 0;
    }
  }

  TDatime dt;
  date = dt.GetDate();
  time = dt.GetTime();

  printf("Run %d  Scan %d  Burst %d  SEUTestCount %d Date %d  Time %d\n", 
          runnum, scannum, burstnum, seu_test_count, date,time);


  // Create special Start of SEU-test data fragment for binary data recording

  sprintf(seu_id,"SEU ");
  seu_header[0] = runnum;
  seu_header[1] = scannum;
  seu_header[2] = burstnum;
  seu_header[3] = seu_test_count;
  seu_header[4] = date;
  seu_header[5] = time;
  seu_header[6] = hard_time;  // => SEU or CONTROL in case of PS
  seu_header[7] = 0;  // stage x
  seu_header[8] = 0;  // stage y
  AppendArray(seu_id,9,seu_test_count,seu_header);


  // Enable CLOAC External Busy 4 which we use to monitor SPILL;  
  cc_menu_enable(ENEXTBUSY4);

  // Put Mustards into appropriate modes for DECODED data:
  //              DECODED, SAMPLE-ALL
  for(mid=0; mid<MUS_NUMBER; mid++) {
    if ((rcode = mus_set_readout_modes(mid,0,1)) != 0) {
      printf("SEUTest: ERROR for Mustard %d: mus_set_readout_modes returns %d\n", mid, rcode);
      return -5;
    }
    if ((rcode = mus_set_readout_cntl(mid,1,1)) != 0) {
      printf("SEUTest: ERROR for Mustard %d: mus_set_readout_cntl returns %d\n", mid, rcode);
      return -6;
    }
  }

  timeout_count = 0;

  cc_read_state(&cloac_busy_state);
  spill_state = cloac_busy_state & 0x0008;   // using fanout-D, register 6 bit 3

  if (spill_state) printf("SEUTest: Already in SPILL at start ... waiting until end of spill\n");

  while (spill_state) {
    cc_read_state(&cloac_busy_state);
    spill_state = cloac_busy_state & 0x0008;
    Sleep(1);
    if (timeout_count > 10000) {
      printf("SEUTest: ERROR: initial state timeout\n");
      return -1;
    }
  }



  // Execute full configuration sequence before spill starts

  ConfigureVariable(ST_ACCUMULATE,0); // Set Accumulator OFF
  ConfigureVariable(ST_DTM, 0);       // Disable auto sending of EnableDatataking command
  
  if(seu_invert){  // look for SEU 1->0
    ConfigureVariable(ST_MASK,3);       // Set all Mask bits to 1  
  }else{           // look for SEU 0->1
    ConfigureVariable(ST_MASK,0);       // Set all Mask bits to 0  
  }
  ConfigureVariable(ST_SEND_MASK,1);  // Disable Input Register and route Mask to Pipeline      
  st_execute_configs(0,1);            // Send full config sequence (dump_level 0)

  // Load SLOG with sequence ready for execution at start of spill:
  //   Write Mask Register (all chips)
  //   Soft Reset, DTM, ~144 BCOs, Soft Reset, L1A and then
  //   configuration register commands (all streams in parallel)
  //     -> any changes must be made to the module configuration 
  //        structure now
  ConfigureVariable(ST_ACCUMULATE,1);  // Set Accumulator ON
  
  st_seu_slog_load(seu_invert);

  if(hard_time){
    printf("SEUTest: NOT SPILL ... now waiting until next spill\n");
    do {
      cc_read_state(&cloac_busy_state);
      spill_state = cloac_busy_state & 0x0008;
    } while (!spill_state);
  }
  

  // Event SEU_DREF
  //  - SendData reference event

  t_zero = (long) clock();
  st_seu_slog_run();
  tsent++;
  ReadEvent();

  seu_index = SEU_DREF;
  seu_event = seu_index+1;
  printf("****SEUTest %6d  Event %d\n", seu_test_count, seu_event);

  AppendEvent();     // Append event to binary recording buffer to be written at end of spill
  SEUDecodeData(seu_index);  // Process Decoded Mode Event, storing vitals in SEU-specific arrays


  // Put Mustards into RAW, SAMPLE-ALL ready for SendID events
  for(mid=0; mid<MUS_NUMBER; mid++) {
    if ((rcode = mus_set_readout_modes(mid,1,1)) != 0) {
      printf("SEUTest: ERROR for Mustard %d: mus_set_readout_modes returns %d\n", mid, rcode);
      return -5;
    }
    if ((rcode = mus_set_readout_cntl(mid,1,1)) != 0) {
      printf("SEUTest: ERROR for Mustard %d: mus_set_readout_cntl returns %d\n", mid, rcode);
      return -6;
    }
  }

  // Event SEU_CREF
  //  - SendID reference event
  
  st_set_slog_loadtrig(1);   // ABCD already in SendID mode - send L1A 
  Single(30);
  tsent++;

  seu_index = SEU_CREF;
  seu_event = seu_index+1;
  printf("****SEUTest %6d  Event %d\n", seu_test_count, seu_event);

  ReadEvent();
  AppendEvent(); 
  SEUDecodeEvent(seu_index);


  printf("\n==================\n");
  if(hard_time)
    printf("= HARDWARE TIMER =\n");
  else
    printf("= SOFTWARE TIMER =\n");
  printf(  "==================\n\n");

  // Accumulator is ON so Pipeline is being continuously sampled
  // Send Mask is ON so Mask is being sent into pipeline continuously 
  // All Mask bits are 0, so output of Pipeline into Accumulator should be all 0

  if(hard_time){
    // Hardware timer
    do {
      cc_read_state(&cloac_busy_state);
      spill_state = cloac_busy_state & 0x0008;
    } while (spill_state);
  } else {
    // Software timer
    ldelay = (long)(seu_duration*(CLOCKS_PER_SEC/1000));
    if(ldelay<=0) ldelay=1;
    while(t_spill<ldelay) t_spill = (long)clock()-t_zero;
  }
  t_spill = (long)clock()-t_zero;


  // Event SEU_CFLIP1
  //  - to detect flips in the configuration register 
  Single(0);
  tsent++;

  seu_index = SEU_CFLIP1;
  seu_event = seu_index+1;
  printf("SEUTest %6d  Event %d\n", seu_test_count, seu_event);

  ReadEvent();
  AppendEvent(); 
  SEUDecodeEvent(seu_index);


  // Event SEU_CFLIP2
  //  - to confirm flips in the configuration register 

  Single(0);
  tsent++;

  seu_index = SEU_CFLIP2;
  seu_event = seu_index+1;
  printf("SEUTest %6d  Event %d\n", seu_test_count, seu_event);

  ReadEvent();
  AppendEvent(); 
  SEUDecodeEvent(seu_index);


  // Put Mustards into appropriate modes for DECODED data:
  //              DECODED, SAMPLE-ALL

  for(mid=0; mid<MUS_NUMBER; mid++) {
    if ((rcode = mus_set_readout_modes(mid,0,1)) != 0) {
      printf("SEUTest: ERROR for Mustard %d: mus_set_readout_modes returns %d\n", mid, rcode);
      return -5;
    }
    if ((rcode = mus_set_readout_cntl(mid,1,1)) != 0) {
      printf("SEUTest: ERROR for Mustard %d: mus_set_readout_cntl returns %d\n", mid, rcode);
      return -6;
    }
  }

  // Now we do send enable_datataking to switch to SendData mode; 
  // Accumulator remains ON
  st_enable_datataking();


  // Event SEU_PFLIP1
  //  - to detect flips in the accumulator/pipeline/mask

  Single(0);
  tsent++;

  seu_index = SEU_PFLIP1;
  seu_event = seu_index+1;
  printf("SEUTest %6d  Event %d\n", seu_test_count, seu_event);

  ReadEvent();
  AppendEvent(); 
  SEUDecodeData(seu_index);


  // Event SEU_PFLIP2
  //  - to confirm flips in the accumulator/pipeline/mask

  Single(0);
  tsent++;

  seu_index = SEU_PFLIP2;
  seu_event = seu_index+1;
  printf("SEUTest %6d  Event %d\n", seu_test_count, seu_event);

  ReadEvent();
  AppendEvent(); 
  SEUDecodeData(seu_index);


  // Now switch off accumulator _without_ overwriting the mask:
  // next sample will be live from the mask register - to identify mask flips

  ConfigureVariable(ST_ACCUMULATE,0);  // Switch Accumulator OFF so next sample is live from Mask
  ConfigureVariable(ST_DTM, 1);        // Re-enable auto sending of EnableDatataking command
  st_execute_configs(0,0);             // Send only configuration register commands
  st_enable_datataking();              // extra DTM command for good measure!!

  // Event SEU_MFLIP1
  //  - to detect flips in the mask register

  Single(0);
  tsent++;

  seu_index = SEU_MFLIP1;
  seu_event = seu_index+1;
  printf("SEUTest %6d  Event %d\n", seu_test_count, seu_event);

  ReadEvent();
  AppendEvent(); 
  SEUDecodeData(seu_index);


  // Event SEU_MFLIP2
  //  - to confirm flips in the mask register

  Single(0);
  tsent++;

  seu_index = SEU_MFLIP2;
  seu_event = seu_index+1;
  printf("SEUTest %6d  Event %d\n", seu_test_count, seu_event);

  mflip2_nerrs = ReadEvent();
  AppendEvent(); 
  SEUDecodeData(seu_index);

  // if there are any protocol violations / errors in the last event
  // take one more event in RAW MODE so we can see what the error is!!

  if(mflip2_nerrs){

    // Put Mustards into RAW, SAMPLE-ALL
    for(mid=0; mid<MUS_NUMBER; mid++) {
      if ((rcode = mus_set_readout_modes(mid,1,1)) != 0) {
        printf("SEUTest: ERROR for Mustard %d: mus_set_readout_modes returns %d\n", mid, rcode);
        return -5;
      }
      if ((rcode = mus_set_readout_cntl(mid,1,1)) != 0) {
        printf("SEUTest: ERROR for Mustard %d: mus_set_readout_cntl returns %d\n", mid, rcode);
        return -6;
      }
    }

    Single(0);
    tsent++;
    
    seu_index = SEU_RAWONERROR;
    seu_event = seu_index+1;
    printf("SEUTest %6d  Event %d - EXTRA RAW EVENT DUE TO PROTOCOL VIOLATION\n", seu_test_count, seu_event);

    ReadEvent();
    AppendEvent();
    SEUDecodeEvent(seu_index);

    st_hard_reset();
    Sleep(1000);
    st_execute_configs(0,1);            // Send full config sequence (dump_level 0)
  }


  // Compare SendID event with reference event - tally actual bitflips
  // TODO - replace this with proper decoding and verification of config register

  anomaly+=SEUCompareEvent(SEU_CFLIP1,SEU_CREF);
  anomaly+=SEUCompareEvent(SEU_CFLIP2,SEU_CREF);

  ///////////////////////////////
  //     Further Analysis      //
  // (sounds like a math text) //
  ///////////////////////////////

  if(seu_invert){
    for (link=0; link<SEU_NLINKS; link++){
      for (chip=0; chip<6; chip++){
        seu_flip[SEU_PFLIP1][link][chip]=0;
        seu_flip[SEU_PFLIP2][link][chip]=0;
        //seu_flip[SEU_PFLIP1][link][chip]=128-seu_flip[SEU_PFLIP1][link][chip];
        //seu_flip[SEU_PFLIP2][link][chip]=128-seu_flip[SEU_PFLIP2][link][chip];
        seu_flip[SEU_MFLIP1][link][chip]=128-seu_flip[SEU_MFLIP1][link][chip];
        seu_flip[SEU_MFLIP2][link][chip]=128-seu_flip[SEU_MFLIP2][link][chip];
      }
    }
  }
  
  for (link=0; link<SEU_NLINKS; link++){

    // Check validity of l1 count
    for(seu_index=0;seu_index<seu_event;seu_index++){
      if (seu_wc[seu_index][link]>0) {
        if (seu_l1[seu_index][link] != (seu_index+1)) { 
          anomaly++; 
          seu_l1err[seu_index][link]++; 
        }
      }
    }

    // Check validity of word count
    for(seu_index=SEU_CFLIP1;seu_index<SEU_PFLIP1;seu_index++){
      if (seu_wc[seu_index][link] != seu_wc[SEU_CREF][link]) {
        anomaly++; 
        seu_wcerr[seu_index][link]++; 
      }
    }
    for(seu_index=SEU_PFLIP1;seu_index<seu_event;seu_index++){
      if (seu_wc[seu_index][link] != seu_wc[SEU_DREF][link]) {
        anomaly++; 
        seu_wcerr[seu_index][link]++; 
      }
    }

    // Chip by chip, determine total number of resolved SEUs
    for (chip=0; chip<6; chip++){

      if(seu_flip[0][link][chip]) { // Flips in any event
        if (seu_flip[SEU_CFLIP1][link][chip]||seu_flip[SEU_CFLIP2][link][chip]) {

          if (seu_flip[SEU_CFLIP1][link][chip]==seu_flip[SEU_CFLIP2][link][chip])
            //seu_conf[link][chip] = seu_flip[SEU_CFLIP1][link][chip];
            seu_conf[link][chip] = 1;

          else seu_mystery[link][chip] += 1;
        }

        if (seu_flip[SEU_PFLIP1][link][chip]||seu_flip[SEU_PFLIP2][link][chip]
          ||seu_flip[SEU_MFLIP1][link][chip]||seu_flip[SEU_MFLIP2][link][chip]) {

          if (seu_flip[SEU_MFLIP1][link][chip]==seu_flip[SEU_MFLIP2][link][chip]) {
            seu_mask[link][chip] = seu_flip[SEU_MFLIP1][link][chip];

            if (seu_flip[SEU_PFLIP1][link][chip]==seu_flip[SEU_PFLIP2][link][chip])
              seu_pipe[link][chip] = seu_flip[SEU_PFLIP1][link][chip] - seu_flip[SEU_MFLIP1][link][chip];

            else seu_mystery[link][chip] += 2;
          } else seu_mystery[link][chip] += 4;

        } // end if any pipeline/accumulator/mask flips
      } // end if any flips
    } // end for chip
  } // end for link

  // Disable BUSYIN and reset MuSTARDs

  cc_menu_disable(ENEXTBUSY4);

  for(mid=0; mid<MUS_NUMBER; mid++) {
    mus_set_readout_cntl(mid,0,0);  
  }

  // ### AppendArray("SEUX") an End-of-SEU summary record ?

  WriteBuffer();   // Write binary record data to disk

  if (anomaly) {
    printf("\n");
    printf("????????????????????????????????????????????????????\n");
    printf("?? SEUTest %d Date %d Time %d Anomaly count %d\n", seu_test_count, date, time, anomaly);
    printf("????????????????????????????????????????????????????\n");
    SEUAppendLog(seu_event);   // Write SEU text file record to disk
  }


  seu_header[9] = (int) t_spill;  // spill duration

  SEUAppendLog2(10,seu_header);
  seu_test_count++;

  return mflip2_nerrs;
}


int TST::SEUDecodeEvent(short seu_event){

  // Decode MUSTARD RAW-mode data in evdata[0][] as much as necessary for SEU study
  // Results into seu_xxx arrays
  short l,n;
  short word_count;
  short error_count;
  short event_count;
  short event_count_high;

  short endstrm;
  short word;
  short link;
  short word_in_event;
  short word_in_link;
  short wc;

  if (seu_event > SEU_NEVENTS) {
    printf("SEUDecodeEvent: ERROR: SEU Event Number out of bounds ... exit\n");
    return -1;
  }

  word_count = evdata[0][0];
  if (word_count & 0x8000) {
    printf("SEUDecodeEvent: ERROR: MUSTARD non-event ... exit\n");
    return -2;
  }

  word_count = word_count & 0x3fff;
  if (word_count <= 4) {
    printf("SEUDecodeEvent: ERROR: EMPTY event  ... exit\n");
    return 0;
  }

  error_count = evdata[0][1];   //  ### Should log error counts & check for increments

  event_count = evdata[0][2];
  if (event_count != seu_event) {
    printf("SEUDecodeEvent: WARNING: event_count %d != seu_count %d\n", event_count, seu_event);
  }

  event_count_high = evdata[0][3];    // ### Should check this is 0

  endstrm = 0;
  word_in_link = 0;
  for (word_in_event=4; word_in_event<word_count; word_in_event++){
    if (endstrm) word_in_link = 0;
    word = evdata[0][word_in_event];
    endstrm = (word & 0x0100);
    link = (evdata[0][word_in_event] & 0x1e00) >> 9;
    if (link<0) link = 0;
    if (link>=SEU_NLINKS) link = SEU_NLINKS-1;
    if (word_in_link < SEU_BUFSIZE){  // only copy data within array bounds
      seu_data[seu_event][link][word_in_link] = word & 0x00ff;
      seu_wc[seu_event][link] = word_in_link;
    }
    word_in_link++;
  }
  
  printf("SEUDecodeEvent:  SEU Event %d  Mus_Event %d  Word_count %d\n", seu_event, event_count, word_count);
  for (l = 0; l<SEU_NLINKS; l++){
    if (seu_wc[seu_event][l] > 0) {
      printf("LINK %2d  WC %d  DATA ", l, seu_wc[seu_event][l]);

      if(seu_wc[seu_event][l]<12) wc = seu_wc[seu_event][l];
      else                     wc = 12;

      for (n=0; n<wc; n++) {
        printf(" %2x", seu_data[seu_event][l][n]);
      }

      seu_l1[seu_event][l] = ((seu_data[seu_event][l][0] & 0x03) << 2) | ((seu_data[seu_event][l][1] & 0xc0) >> 6);
      seu_bc[seu_event][l] = ((seu_data[seu_event][l][1] & 0x3f) << 2) | ((seu_data[seu_event][l][2] & 0xc0) >> 6);
      seu_data[seu_event][l][2] = seu_data[seu_event][l][2] & 0x3f;  // ### suppress least sig 2 bits of BC

      printf("  L1 %d   BC %d\n", seu_l1[seu_event][l], seu_bc[seu_event][l]);
    }
  }

  return 0; 
}


int TST::SEUCompareEvent(short event, short reference){

  int nflips = 0;
  short l, wc, word, bit;

  for (l=0; l<SEU_NLINKS; l++){

    wc = seu_wc[event][l];
    if (wc > 0) {

      if (wc!=seu_wc[reference][l]){

        printf("SEUCompareEvent:  different lengths  link%d  event%d len%d  ref%d len%d\n",
                l,event,wc, reference, seu_wc[reference][l]);

        if (wc > seu_wc[reference][l]) wc = seu_wc[reference][l];
      }

      for (word=2;word<wc;word++){

        if (seu_data[event][l][word]!=seu_data[reference][l][word]){
          printf("SEUCompareEvent: different words link%d event%d word%d d%02x ref%d d%2x\n",
                 l,event,word,seu_data[event][l][word],
                 reference,seu_data[reference][l][word]);
          seu_flip[0][l][0]++; // tally number of words with errors
		                       // for now, just associate config flips with chip 0

          for (bit=0x01;bit<0xff;bit<<=1) {
            if ((seu_data[event][l][word]&bit)!=(seu_data[reference][l][word]&bit))
              nflips++;
              seu_flip[event][l][0]++;
          }
        }
      }

    } // end if (wc > 0)

  } // end for l

  return nflips;
}


void TST::SEUAppendLog(short n_events) {

  FILE *f;
  char fname[128];
  int event,link,n;
 
  int date, time;
  TDatime dt;

  date = dt.GetDate();
  time = dt.GetTime();

  if (n_events > SEU_NEVENTS) n_events = SEU_NEVENTS;

  sprintf(fname,"%sSEUlog%d_%d.dat",sct_data,runnum,scannum);

  printf("SEUAppendLog: Attempting to open datafile %s\n", fname);
  fflush(stdout);

  if ((f = fopen(fname,"a")) == NULL) {
    printf("SEUAppendLog: ERROR: Unable to open data file %s for appending\n", fname);
    fflush(stdout);
    return;
  }
  printf("SEUAppendLog: Opened data file %s for appending\n",fname);
  fflush(stdout);

  fprintf(f, "#SEU run %d scan %d burst %d seu_cnt %d date %d time %d\n", runnum, scannum, burstnum, seu_test_count, date, time);
  
  for (event=0; event<n_events; event++){
    for (link=0; link<SEU_NLINKS; link++){
      if (seu_wc[event][link] > 0){
        fprintf(f,"%d %d %d %d %d", event, link, seu_wc[event][link], seu_l1[event][link], seu_bc[event][link]);
        for (n=0; n<seu_wc[event][link]; n++) fprintf(f," %4x", seu_data[event][link][n]);
        fprintf(f,"\n");
      }
    }
  }
  
  fclose(f);
}

void TST::SEUAppendLog2(int length, int* header) {

  FILE *f;
  char fname[128];
  int link,chip,n;

  sprintf(fname,"%sSEUlog2_%d_%d.dat",sct_data,runnum,scannum);

  //printf("SEUAppendLog2: Attempting to open datafile %s\n", fname);
  fflush(stdout);

  if ((f = fopen(fname,"a")) == NULL) {
    printf("SEUAppendLog2: ERROR: Unable to open data file %s for appending\n", fname);
    fflush(stdout);
    return;
  }
  //printf("SEUAppendLog2: Opened data file %s for appending\n",fname);
  fflush(stdout);

  fprintf(f,"-1 ");
  for (n=0;n<length;n++) fprintf(f, "%d ", header[n]);
  fprintf(f,"\n");
  
  for (link=0; link<SEU_NLINKS; link++){
    fprintf(f,"Lk %d  ",link);
    for(chip=0;chip<6;chip++) fprintf(f,"%d ",seu_conf[link][chip]);
    fprintf(f,"  ");
    for(chip=0;chip<6;chip++) fprintf(f,"%d ",seu_pipe[link][chip]);
    fprintf(f,"  ");
    for(chip=0;chip<6;chip++) fprintf(f,"%d ",seu_mask[link][chip]);
    fprintf(f,"  ");
    for(chip=0;chip<6;chip++) fprintf(f,"%d ",seu_mystery[link][chip]);
    fprintf(f,"\n");
  }
  
  fclose(f);
}

void TST::SEUAppendSummary(int length, int* header, short grand_total) {

  FILE *f;
  char fname[128];
  int link,chip,n;
  short start, stop;

  sprintf(fname,"%sSEUSummary_%d_%d.dat",sct_data,runnum,scannum);

  printf("SEUAppendSummary: Attempting to open datafile %s\n", fname);
  fflush(stdout);

  if ((f = fopen(fname,"a")) == NULL) {
    printf("SEUAppendSummary: ERROR: Unable to open data file %s for appending\n", fname);
    fflush(stdout);
    return;
  }
  printf("SEUAppendSummary: Opened data file %s for appending\n",fname);
  fflush(stdout);

  if(grand_total){
    fprintf(f,"-2 ");
    start = 2;
    stop = 4;
  } else{
    fprintf(f,"-1 ");
    start = 0;
    stop = 2;
  }

  for (n=0;n<length;n++) fprintf(f, "%d ", header[n]);
  fprintf(f,"\n");
  
  for(n=start;n<stop;n++){
    for (link=0; link<SEU_NLINKS; link++){
      fprintf(f,"Lk %d type %d  ",link, n);
      for(chip=0;chip<6;chip++) fprintf(f,"%d ",seu_tot_conf[n][link][chip]);
      fprintf(f,"  ");
      for(chip=0;chip<6;chip++) fprintf(f,"%d ",seu_tot_pipe[n][link][chip]);
      fprintf(f,"  ");
      for(chip=0;chip<6;chip++) fprintf(f,"%d ",seu_tot_mask[n][link][chip]);
      fprintf(f,"  ");
      for(chip=0;chip<6;chip++) fprintf(f,"%d ",seu_tot_mystery[n][link][chip]);
      fprintf(f,"\n");
    }
  }
  
  fclose(f);
}


int TST::FindSurvivors(short mid){

  return FindSurvivors(mid,1,1);
}


int TST::FindSurvivors(short mid, short use_link0, short use_link1){

  short i,j;
  short good_chips = 0;
  short good_links = 0;
  short max_chips = 0;
  short last_working;
  short chips_per_link[2]={0,0};
  short chip_status[3][12];

  short old_burst_type = burst_type;
  short old_burst_trtype = burst_trtype;
  int   old_burst_maxtrigs = burst_maxtrigs;

  // PWP 06.04.00 - routine is working

  if((use_link0==0)&&(use_link1==0)){
    printf("FindSurvivors: No lifelines - exiting\n");
    return 0;
  }

  burst_type = ST_TBURST2;
  burst_trtype = 0;    // single L1A
  burst_maxtrigs = 10; // How many do you need for this?

  ConfigureVariable(ST_COMPRESSION,1);  // ANY HIT compression 
  ConfigureVariable(ST_EDGE_DETECT,0);  // Level Mode 
  ConfigureVariable(ST_MASK,3);         // Set all Mask bits to 1  
  ConfigureVariable(ST_SEND_MASK,1);    // Disable Input Register and route Mask to Pipeline      

  for(i=0;i<12;i++){
    chip_status[0][i]=0;
    chip_status[1][i]=0;
    chip_status[2][i]=0;
  }

  // How many chips can be read through each link?

  if(use_link0){
    st_configure_enable(mid,1,1,2,0);
    st_execute_enables();
    if( (chips_per_link[0]=ChipHunt(mid,0,chip_status[0])) >0) good_links++;
  }

  if(use_link1){
    st_configure_enable(mid,1,1,0,2);
    st_execute_enables();
    if( (chips_per_link[1]=ChipHunt(mid,1,chip_status[1])) >0) good_links++;
  }

  // Now derive the optimum configuration
  // Two links are better than one (??)

  if(good_links==2){

    last_working=0;
    for(i=0;i<6;i++){
      if(chip_status[0][i]>1) last_working = i;
      chip_status[2][i]=chip_status[0][i];
    }
    if(last_working>0) chip_status[2][last_working] = END;
    else               chip_status[2][last_working] = LONELY;

    last_working=6;
    for(i=6;i<12;i++){
      if(chip_status[1][i]>1) last_working = i;
      chip_status[2][i]=chip_status[1][i];
    }
    if(last_working>6) chip_status[2][last_working] = END;
    else               chip_status[2][last_working] = LONELY;

    st_configure_enable(mid,1,1,1,1);

  } else if(chips_per_link[0]>0){
    for(i=0;i<12;i++) chip_status[2][i]=chip_status[0][i];
    st_configure_enable(mid,1,1,2,0);


  } else if(chips_per_link[1]>0){
    for(i=0;i<12;i++) chip_status[2][i]=chip_status[1][i];
    st_configure_enable(mid,1,1,0,2);

  } else {
    st_configure_enable(mid,1,1,0,0);
  }

  // The optimum configuration is set here

  for(i=0;i<12;i++){

    st_set_role(mid,i,chip_status[2][i]);
  
    if(chip_status[2][i]>1) max_chips++; 
    if( (chip_status[0][i]>1)||(chip_status[1][i]>1) ) good_chips++; 
  }
  st_execute_enables();
  ExecuteConfigs(0);

  //Diagnostic printout: content of chip_status array

  for(i=0;i<3;i++){
    printf("chip status[%d]:",i);
    for(j=0;j<12;j++){
      printf("  %d",chip_status[i][j]);
    }
    printf("\n");
  }

  printf("FindSurvivors: Module %d has %d working links\n",mid,good_links);
  if(use_link0)
    printf("               %d chips can be read through link 0\n",chips_per_link[0]);
  if(use_link1)
    printf("               %d chips can be read through link 1\n",chips_per_link[1]);
  printf("               %d survivors of which %d fit in the lifeboat!!\n",good_chips,max_chips);
  printf("               *** SEARCH CONCLUDED ***\n"); 

  burst_type = old_burst_type; 
  burst_trtype = old_burst_trtype; 
  burst_maxtrigs = old_burst_maxtrigs; 

  return(max_chips);
}


int TST::ChipHunt(short mid, short link, short* chip_status){

  short i;
  short chip, master;
  short last_working_chip;
  short npass = 0;
  short n_consecutive_failures=0;
  short running = 1;

  int sum =0;

  // PWP 06.04.00 - routine is working

  // MISSING    0   or UNKNOWN
  // DEAD       1
  // END        2
  // MASTER     3
  // SLAVE      4
  // LONELY     5

  for(chip=0;chip<12;chip++){
    chip_status[chip]=0;
    st_set_role(mid,chip,MISSING);
  }

  // Try Master chip

  if(link==1) master = 6;
  else        master = 0;

  st_set_role(mid,master,LONELY);

  SingleSoftReset();
  ExecuteConfigs(0);
  ExecuteBurst();

  sum = 0;
  for (i=0; i<128; i++){
    sum += m[mid]->scan[(burst_count-1)][master/6][i];
  } 

  if((m[mid]->GetTocnt(link)==0)&&(sum>(64*burst_ntrigs))){
    printf("ChipHunt: Module %d link %d Master chip is alive\n",mid,link);  
    chip_status[master]=MASTER;
    st_set_role(mid,master,MASTER);
    npass++;
    n_consecutive_failures=0;
    last_working_chip = master;
  } else {
    printf("ChipHunt: Module %d link %d Master chip missing presumed dead\n",mid,link);  
    chip_status[master]=DEAD;
    st_set_role(mid,master,DEAD);
    running = 0;
  }

  chip = master+1;

  while(running){

    st_set_role(mid,chip,END);

    SingleSoftReset();
    ExecuteConfigs(0);
    ExecuteBurst();

    // This should really incorporate a test for hits...

    sum = 0;
    for (i=((chip%6)*128); i<(((chip%6)+1)*128); i++){
      sum += m[mid]->scan[(burst_count-1)][(chip/6)][i];
    } 

    if((m[mid]->GetTocnt(link)==0)&&(sum>(64*burst_ntrigs))){
      printf("ChipHunt: Module %d chip %d is alive\n",mid,chip);  
      chip_status[chip]=SLAVE;
      st_set_role(mid,chip,SLAVE);
      npass++;
      n_consecutive_failures=0;
      last_working_chip = chip;
    } else {
      printf("ChipHunt: Module %d chip %d missing presumed dead\n",mid,chip);  
      chip_status[chip]=DEAD;
      st_set_role(mid,chip,DEAD);
      n_consecutive_failures++;
    }

    if(n_consecutive_failures>1) {
      printf("ChipHunt: Module %d link %d has two consecutive dead chips\n",mid,link);  
      running = 0;
    } else {
      chip++;
      if(chip>11) chip = 0;
      if(chip==master) running = 0;
    }
  }

  printf("          *** SEARCH CONCLUDED FOR THIS LINK ***\n"); 

  // now tidy up setting of last chip

  if(last_working_chip==master){
    chip_status[last_working_chip]=LONELY;
    st_set_role(mid,last_working_chip,LONELY);
  } else {
    chip_status[last_working_chip]=END;
    st_set_role(mid,last_working_chip,END);
  }

  return npass;
}

int TST::SEUDecodeData(short seu_event){

  // Assume MUS_NUMBER MuSTARDS have been read, their data is stored in
  // data[MUS_NUMBER].  This routine will decode events, associating a 
  // number of error conditions with the causal module/link.
  // PWP 09.04.00

  short mid,n;
  short* data;

  short word_count;
  short word_in_stream;

  short module;
  short stream;
  short hard_link;
  short soft_link;
  short soft_stream;
  short strip_address;
  short allow_continued_hits;

  short nflips=0;

  if (seu_event > SEU_NEVENTS) {
    printf("SEUDecodeData: ERROR: SEU Event Number out of bounds ... exit\n");
    return -1;
  }

  word_count = evdata[0][0];
  if (word_count & 0x8000) {
    printf("SEUDecodeData: ERROR: MUSTARD non-event ... exit\n");
    return -2;
  }

  word_count = word_count & 0x3fff;
  if (word_count <= 4) {
    printf("SEUDecodeData: ERROR: EMPTY event  ... exit\n");
    return 0;
  }

  for(mid=0; mid<MUS_NUMBER; mid++) {

    data = evdata[mid];
    word_count  =  data[0];

    if (word_count > 0){   // Decode the event data
                           // ### ignore timebins for the moment
      stream = SEU_NLINKS;
      module = 0;
      hard_link = 0;
      soft_link = 0;
      soft_stream=0;
      strip_address=0;
      allow_continued_hits=0;
      word_in_stream=0;

      for (n=4; n<word_count; n++){


        // First test must be for presence of the error bit.
        // The error bit can be set together with other bits - 
        // such words should not be interpreted other than as errors!!!

        if (data[n]&0x4000) {  // ERROR BIT SET
 
          allow_continued_hits = 0; 

          printf("SEUDecodeData: trig %d mid %d sid %d err at short %d val %4hx\n",tsent,mid,stream,n,data[n]);

          // Error bit will remain set until MuSTARD "finds its feet"
          // => log only each transition of error bit from 0 to 1

          if((n==4)||((data[n-1]&0x4000)==0)){  
            if (stream >= SEU_NLINKS) break;
            module = st_find_module(mid,stream);
            hard_link = st_find_link(mid,stream);

            if(module>-1)  // valid module
              m[module]->FillErcnt(hard_link);
            else 
              printf("SEUDecodeData: trig %d mid %d sid %d st_find_module returns invalid module\n",tsent,mid,stream);
          }

        } else if (data[n]&0x1000) { // HIT
          // HIT must follow either HIT or NEW CLUSTER
          // anything else is OUT OF SEQUENCE

          if(allow_continued_hits){  // in sequence hit

            if(module>-1){  // valid module

              if (strip_address<768) {
                seu_flip[0][soft_stream][strip_address/128]++;
                seu_flip[seu_event][soft_stream][strip_address/128]++;
                nflips++;
              } else {
                printf("SEUDecodeData: trig %d mid %d sid %d non-existent strip %d\n",tsent,mid,stream,strip_address); 
                if (stream >= SEU_NLINKS) break;
                module = st_find_module(mid,stream);
                hard_link = st_find_link(mid,stream);

                if(module>-1)
                  m[module]->FillNescnt(hard_link);
              }

              strip_address++;
            }
          } else { // out of sequence hit
            printf("SEUDecodeData: trig %d mid %d sid %d OSH at short %d val %4hx\n",tsent,mid,stream,n,data[n]);
            
            if (stream >= SEU_NLINKS) break;
            module = st_find_module(mid,stream);
            hard_link = st_find_link(mid,stream);
            if(module>-1)
              m[module]->FillOshcnt(hard_link);
          } 

        } else if (data[n]&0x2000) {  // NEW CLUSTER

          // For the time being, we have two scan histograms, one per "link"
          // in practice, "link 0" handles strips    0 to  768
          //              "link 1" handles strips 1024 to 1792
          // We may come to regret having taken this approach....

          strip_address =  data[n] & 0x03ff;
          soft_link     = (data[n] & 0x0400)>>10;
          soft_stream   = (st_find_module(mid,stream)*2) + soft_link;

          allow_continued_hits = 1; 

        } else if (data[n]&0x8000) {  // NEW STREAM

          allow_continued_hits = 0; 
          word_in_stream = 0;

          if (data[n+1]&0x4000) {  // ERROR IN STREAM ID
            // we no longer know which stream has caused the error
            // How do we tally this condition??
            printf("SEUDecodeData: trig %d mid %d Stream Assignment error \n",tsent,mid);
            printf("              short n=%d data(n)=%4x data(n+1)=%4x\n",n, data[n],data[n+1]);

          }else{

            stream = data[n+1]&0x000f; 
            if (stream >= SEU_NLINKS) break;

            seu_bc[seu_event][stream] =  data[n] & 0x00ff;
            seu_l1[seu_event][stream] = (data[n] & 0x0f00) >> 8;

            module = st_find_module(mid,stream);
            hard_link = st_find_link(mid,stream);
            if(module>-1)
              m[module]->FillEvcnt(hard_link);
            else{ 
              printf("SEUDecodeData: trig %d mid %d sid %d st_find_module returns invalid module\n",tsent,mid,stream);
              printf("              short n=%d data(n)=%4x data(n+1)=%4x\n",n, data[n],data[n+1]);
            }
          }
        } // end new stream

        // Copy the data into SEU specific array and update word count
        if((stream<SEU_NLINKS)&&word_in_stream<SEU_BUFSIZE){
          seu_data[seu_event][stream][word_in_stream++] = (data[n]&0xffff);
          seu_wc[seu_event][stream] = word_in_stream;
        }

      } // end for(n)
    } // end if(word_count>0)

  } // end for (mid)

  return nflips;
}


void TST::DecodeErrors(){

  // Assume MUS_NUMBER MuSTARDS have been read, their data is stored in
  // data[MUS_NUMBER], and that MuSTARD has reported one or more errors
  // during the event. This routine will associate a number of error 
  // conditions with the causal module/link - without filling any histograms.
  // PWP 19.07.00

  short mid,n;
  short* data;

  short word_count;

  short module;
  short stream;
  short hard_link;
  short soft_link;
  short L1_count;
  short BC_count;
  short strip_address;
  short multiplicity[SCT_MAX_MODULES][2];
  short allow_continued_hits;

  for(n=0;n<SCT_MAX_MODULES;n++){
    multiplicity[n][0]=0;
    multiplicity[n][1]=0;
  }

  for(mid=0; mid<MUS_NUMBER; mid++) {
    if(st_used_streams(mid)>0){
 
      data = evdata[mid];
      word_count  =  data[0];

      if (word_count > 0){   // Parse the event data to locate errors
        stream = 0;
        module = 0;
        hard_link = 0;
        soft_link = 0;
        strip_address=0;
        allow_continued_hits=0;

        for (n=4; n<word_count; n++){

          // First test must be for presence of the error bit.
          // The error bit can be set together with other bits - 
          // such words should not be interpreted other than as errors!!!

          if (data[n]&0x4000) {  // ERROR BIT SET
 
            allow_continued_hits = 0; 

            if(debug_level>6){
              printf("DecodeErrors: trig %d mid %d sid %d err at short %d val %4hx\n",tsent,mid,stream,n,data[n]);
            }

            // Error bit will remain set until MuSTARD "finds its feet"
            // => log only each transition of error bit from 0 to 1

            if((n==4)||((data[n-1]&0x4000)==0)){  
              if (stream >= 12) break;
              module = st_find_module(mid,stream);
              hard_link = st_find_link(mid,stream);

              if(module>-1)  // valid module
                m[module]->FillErcnt(hard_link);
              else if(debug_level>6){
                printf("DecodeErrors: trig %d mid %d sid %d st_find_module returns invalid module\n",tsent,mid,stream);
              }
            }

          } else if (data[n]&0x1000) { // HIT
            // HIT must follow either HIT or NEW CLUSTER
            // anything else is OUT OF SEQUENCE

            if(allow_continued_hits){  // in sequence hit

              if(module>-1){  // valid module

                if (strip_address>=768) {
                  if(debug_level>6){
                    printf("DecodeErrors: trig %d mid %d sid %d non-existent strip %d\n",tsent,mid,stream,strip_address);
                  } 
                  if (stream >= 12) break;
                  module = st_find_module(mid,stream);
                  hard_link = st_find_link(mid,stream);

                  if(module>-1) m[module]->FillNescnt(hard_link);
                }
                strip_address++;
              }
            } else { // out of sequence hit
              if(debug_level>6){
                printf("DecodeErrors: trig %d mid %d sid %d OSH at short %d val %4hx\n",tsent,mid,stream,n,data[n]);
              }

              if (stream >= 12) break;
              module = st_find_module(mid,stream);
              hard_link = st_find_link(mid,stream);
              if(module>-1)
                m[module]->FillOshcnt(hard_link);
            } 

          } else if (data[n]&0x2000) {  // NEW CLUSTER

            // DO NOT FILL ANY HISTOGRAMS!!

            strip_address =  data[n] & 0x03ff;
            soft_link     = (data[n] & 0x0400)>>10;

            allow_continued_hits = 1; 

          } else if (data[n]&0x8000) {  // NEW STREAM

            allow_continued_hits = 0; 

            // DO NOT FILL ANY HISTOGRAMS!!

            if (data[n+1]&0x4000) {  // ERROR IN STREAM ID
              // we no longer know which stream has caused the error
              // How do we tally this condition??
              if(debug_level>6){
                printf("DecodeErrors: trig %d mid %d Stream Assignment error \n",tsent,mid);
                printf("              short n=%d data(n)=%4x data(n+1)=%4x\n",n, data[n],data[n+1]);
              }
            }else{

              // TODO:  Validity check of BC and L1 data
              BC_count =  data[n] & 0x00ff;
              L1_count = (data[n] & 0x0f00) >> 8;
              stream   =  data[n+1]&0x000f; 

              if (stream >= 12) break;
              module = st_find_module(mid,stream);
              hard_link = st_find_link(mid,stream);

              if((module<0)&&(debug_level>6)){
                printf("DecodeErrors: trig %d mid %d sid %d st_find_module returns invalid module\n",tsent,mid,stream);
                printf("              short n=%d data(n)=%4x data(n+1)=%4x\n",n, data[n],data[n+1]);
              }
            }
          } // end new stream
        } // end for(n)
      } // end if(word_count>0)
    }
  } // end for (mid)

 // End of event
}


void TST::StreamStatus(short* empty, short* not_empty, short* almost_full, short* overflow){

  // How many (used) streams are empty, not empty, almost_full or overflowing?
  // PWP 08.05.01

  short mid,n;
  short soft_link,status;
  short stream, last_stream;

  short l_empty = 0;
  short l_not_empty = 0;
  short l_almost_full = 0;
  short l_overflow = 0;

  for (n=0; n<SCT_MAX_MODULES; n++){
    if(m[n]->present){ // read Mustard stream status flag for enabled streams

      mid = st_find_mustard(n);
      last_stream = -1;

      for(soft_link=0;soft_link<2;soft_link++){
        stream = st_find_stream(n,soft_link);

        if(stream!=last_stream){  // do not read same stream twice when redundancy in use
          mus_read_stream_status(mid,stream,&status);
          if (status&0x1) l_empty++; 
          else l_not_empty++;
          if (status&0x2) l_almost_full++;
          if (status&0x8) l_overflow++;
        }
        last_stream = stream;
      } 
    }
  } // end for(n<SCT_MAX_MODULES)

  *empty       = l_empty;
  *not_empty   = l_not_empty;
  *almost_full = l_almost_full;
  *overflow    = l_overflow;
}

short TST::ResetMustards(){
  // Reset MuSTARD, clear all histograms,
  // set paf and trlim to designated values (otherwise cleared by reset)
  // PWP 16.08.00

  short mid, stream, rcode;

  for(mid=0; mid<MUS_NUMBER; mid++) {

    if ((rcode = mus_reset(mid)) != 0){
      printf("ResetMustards: ERROR for Mustard %d: mus_reset returns %d\n", mid, rcode);
      return -1;
    }

    for (stream=0; stream<12; stream++){
      if ((rcode = mus_clear_histogram(mid,stream)) != 0){
        printf("ResetMustards: ERROR for Mustard %d stream %d: mus_clear_histogram returns %d\n", 
                mid, stream, rcode);
        return -2;
      }
    }

    for(stream=0; stream<=8; stream+=4){
      if ((rcode = mus_set_stream_paf(mid,stream,paf)) != 0) { 
        printf("ResetMustards: ERROR for Mustard %d stream %d: mus_set_stream_paf returns %d\n", mid, stream, rcode);
        return -3;
      }
      if ((rcode = mus_set_stream_trlim(mid,stream,trlim)) != 0) {
        printf("ResetMustards: ERROR for Mustard %d stream %d: mus_set_stream_trlim returns %d\n", mid, stream, rcode);
        return -4;
      }
    }

  }

  return 0;
}


int TST::PreBurst(){

  // Send up to 12 single L1A triggers and check for timeout after
  // each event.  For each timeout which is detected, readout of 
  // the causal stream is suspended.
  // PWP 11.04.01

  short mid;
  short rcode, status;
  short *evpointer;
  short evlength;
  short error_count;
  int   event_count;

  short running = 1;
  short n_pre_events = 0;
  short timeout;
  short timeout_total = 0;
  short timeout_stream[MUS_NUMBER];

  while (running) {     // Enter event loop 

    // Put MuSTARDs in appropriate mode to receive data

    for(mid=0; mid<MUS_NUMBER; mid++) {
      if ((rcode = mus_set_readout_modes(mid,0,1)) != 0) { /* DECODED, SAMPLE ALL */
        printf("PreBurst: ERROR for Mustard %d: mus_set_readout_modes returns %d\n", mid, rcode);
        return -5;
      }
   
      if ((rcode = mus_set_readout_cntl(mid,1,1)) != 0) {
        printf("PreBurst: ERROR for Mustard %d: mus_set_readout_cntl returns %d\n", mid, rcode);
        return -6;
      }
    }

    SingleSoftReset();  // Single Soft Reset
    Sleep(10); // Give MuSTARD time to be ready

    if(debug_level>7) {
      printf("Sending PreBurst trigger %d\n",n_pre_events);
    }
    Single(0);          // Single L1A 
    n_pre_events++;
 
    // Now wait for each MuSTARD to return an event, or to TIMEOUT

    timeout=0;
    for(mid=0; mid<MUS_NUMBER; mid++) {

      timeout_stream[mid]=-1;
      if(st_used_streams(mid)>0){

        evpointer = evdata[mid];
        error_count = ercnt[mid];
        rcode = mus_test_event(mid,&evlength,&evpointer,&event_count,&error_count,0);

        if (debug_level > 8) {
          printf("PreBurst: burst %d n_pre_events %d MuSTARD %d gives evcnt %d evlength %d\n", 
                  burst_count,n_pre_events,mid,event_count,evlength);
        }

        if (rcode != 0) {
          mus_read_status(mid,&status);
          if (debug_level>5){
            printf("PreBurst: ERROR %d from Mustard %d burst %d n_pre_events %d\n",
                    rcode,mid,burst_count,n_pre_events);
            mus_print_status(status);
          }

          if (rcode == -4) {
            // TIMEOUT: stream can be determined from MuSTARD status register
            timeout++;
            timeout_total++;
            timeout_stream[mid] = status & 0x00f;
          }
        }
      }
    } // end for(mid)
    

    for(mid=0; mid<MUS_NUMBER; mid++) {
      mus_set_readout_cntl(mid,0,0);
    }

    // Suspend streams which have given rise to timeouts

    if(timeout){                          // if timeout
      for(mid=0; mid<MUS_NUMBER; mid++) { // loop over MuSTARDs
        if(timeout_stream[mid]>-1){       // timeout this MuSTARD

          if(timeout_stream[mid]<12){     // suspend valid stream
            st_suspend_stream(mid,timeout_stream[mid]);
          }else if(debug_level>5){        // invalid stream
            printf("PreBurst: timeout from invalid stream,  MuSTARD %d stream %d\n",
                    mid,timeout_stream[mid]);
          }
        }
      }

    }else{                                // NO TIMEOUTS
      running = 0;
    }

    if(n_pre_events >= 12){
      running = 0;
    }
  }

  // Do not read the MuSTARD histograms 
  // - but reset MuSTARDs instead

  ResetMustards();

  return 0;
}

void TST::ClearSuspense(){
  // Clear suspense for all modules
  // PWP 17.04.01

  st_clear_suspense();
}

void TST::ClearSuspense(short mid){
  // Clear suspense for a single module
  // PWP 17.04.01

  st_clear_1suspense(mid);
}

void TST::ClockOff(){
  // Stop clock to all modules
  // (disabled at SLOG)

  st_clock_off();
}

void TST::ClockOn(){
  // Stop clock to all modules
  // (disabled at SLOG)

  st_clock_on();
}

short TST::FindModule(short mid, short sid){
  // Is a valid module connected to stream sid
  // of MuSTARD mid?  If not, return -1.

  short module;

  module = st_find_module(mid,sid);

  if((module>=0)&&(module<SCT_MAX_MODULES)){
    if(m[module]->present==0){
      module = -1;
    }
  }else{
    module = -1;
  }
  return module;
}

short TST::FindLink(short mid, short sid){
  // Return the link number (0,1) of the link
  // connected to stream sid of MuSTARD mid.
  // Return -1 if stream not enabled.

  short module,link;

  module = st_find_module(mid,sid);

  if((module>=0)&&(module<SCT_MAX_MODULES)){
    link = st_find_link(mid,sid);
  }else{
    link = -1;
  }

  return link;
}

void TST::WriteConfigs(){
  // Generate configuration files to describe
  // the present state of each module

  short module;

  for(module=0;module<SCT_MAX_MODULES;module++){
    if(m[module]->present){
      st_write_module(module,runnum,scannum);
    }
  }
}

void TST::WriteRODConfigs(){
  // Generate configuration files to describe
  // the present state of each module in ROD format

  short module;

  for(module=0;module<SCT_MAX_MODULES;module++){
    if(m[module]->present){
      st_write_module2(module,runnum,scannum);
    }
  }
}

void TST::WriteXMLConfigs(char* location="unknown"){
  // Generate configuration files to describe
  // the present state of each module in ROD format

  short module;

  for(module=0;module<SCT_MAX_MODULES;module++){
    if(m[module]->present){
      st_write_module_xml(module,runnum,scannum,version,location);
    }
  }
}

void TST::SlogLoadTrig(){
  // Set the flag to force the loading of slog triggers
  st_set_slog_loadtrig(1);
}

TPaveText* TST::GetBurstPaveText(){
  char s[128];
  int n;

  if(last_BPT == burstnum){
    return BurstPaveText;
  }

  if(BurstPaveText){
    BurstPaveText->Delete();
    BurstPaveText=0;
  }

  BurstPaveText = new TPaveText(0.0,0.0,1.0,1.0);
  BurstPaveText->SetFillColor(10);
  BurstPaveText->SetTextAlign(13);
  BurstPaveText->SetTextSize((float)0.15);
  BurstPaveText->SetTextColor(4);

  TDatime *dt = new TDatime;
  sprintf(s,"ATLAS SCT Module Test  Run %d  Burst %d  Date %d  Time %d",
             runnum, burstnum, dt->GetDate(),dt->GetTime());
  BurstPaveText->AddText(s);
  sprintf(s,"Burst Trigs_sent %u", tsent);
  BurstPaveText->AddText(s);

  for(n=0;n<nMuSTARDs;n++){
    sprintf(s,"MuSTARD %d Events_seen %u  Stream_errors %u  Timeouts %u  Other_errors %u",
       n, evcnt[n], ercnt[n], tocnt[n], xecnt[n]);
    BurstPaveText->AddText(s);
  }

  sprintf(s,"Run Trigs_sent %lu  Events_seen %lu  Stream_errors %lu  Timeouts %lu  Other_errors %lu",
     run_tsent, run_evcnt, run_ercnt, run_tocnt, run_xecnt);
  BurstPaveText->AddText(s); 
  
  if (scan_running){
    BurstPaveText->AddText(scan_status_string);
  }
  BurstPaveText->Draw();

  delete dt;
  last_BPT=burstnum;
  return BurstPaveText;
}

TPaveText* TST::GetScanPaveText(){
  char s[128];

  if(last_SPT == scannum){
    return ScanPaveText;
  }

  if(ScanPaveText){
    ScanPaveText->Delete();
    ScanPaveText=0;
  }

  ScanPaveText = new TPaveText(0.0,0.0,1.0,1.0);
  ScanPaveText->SetFillColor(10);
  ScanPaveText->Draw();
  ScanPaveText->SetTextAlign(13);
  ScanPaveText->SetTextSize((float)0.15);
  ScanPaveText->SetTextColor(4);
  ScanPaveText->AddText("ATLAS SCT Module Test");
  sprintf(s,"Run %d  Scan %d  Burst %d  Run_tsent %lu  Run_ercnt %lu", 
     runnum, scannum, burstnum, run_tsent, run_ercnt);
  ScanPaveText->AddText(s);
  sprintf(s,"%s", scan_description_string);
  ScanPaveText->AddText(s);

  last_SPT=scannum;
  return ScanPaveText;
}

